/*
Module : OSMCtrl.h
Purpose: Defines the interface for an MFC GUI control which implements display of OpenStreetMaps tiles
Created: PJN / 28-11-2009

Copyright (c) 2009 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Macros / Includes ////////////////////////////////////////

#pragma once

#ifndef __OSMCTRL_H__
#define __OSMCTRL_H__

#include "OSMCtrlIcon.h"
#include "OSMCtrlMarker.h"
#include "OSMCtrlPolyline.h"
#include "OSMCtrlPolygon.h"
#include "OSMCtrlCircle.h"
#include "OSMCtrlMapOperations.h"
#include "OSMCtrlTileProviders.h"
#include "OSMCtrlHandler.h"
#include "OSMCtrlCachedTile.h"

#ifdef COSMCTRL_NOD2D
#ifndef _GDIPLUS_H
#pragma message("To avoid this message, please put gdiplus.h in your pre compiled header (normally stdafx.h)")
#include <gdiplus.h>
#endif
#else
#include <afxrendertarget.h>
#endif //#ifdef COSMCTRL_NOD2D

#ifndef _VECTOR_
#pragma message("To avoid this message, please put vector in your pre compiled header (normally stdafx.h)")
#include <vector>
#endif //#ifndef _VECTOR_

#include "WinHTTPWrappers.h" //If you get a compilation error about this missing header file, then you need to download my WinHTTPWrappers classes from http://www.naughter.com/winhttpwrappers.html


//////////////////// Classes //////////////////////////////////////////////////

//A simple UDT which contains the tiles to download
class COSMCtrlDownloadTileItem
{
public:
//Constructors / Destructors
#pragma warning(suppress: 26455)
  COSMCtrlDownloadTileItem() = default;
#ifdef COSMCTRL_NOD2D
  COSMCtrlDownloadTileItem(_In_ int nTileX, _In_ int nTileY, _In_ const Gdiplus::RectF& rTile, _In_ double fDistanceFromCenterPosition, _In_ const CString& sFile) : m_nTileX{nTileX},
                                                                                                                                                                     m_nTileY{nTileY},
                                                                                                                                                                     m_rTile{rTile},
                                                                                                                                                                     m_fDistanceFromCenterPosition{fDistanceFromCenterPosition},
                                                                                                                                                                     m_sFile{sFile}
  {
  }
#else
  COSMCtrlDownloadTileItem(_In_ int nTileX, _In_ int nTileY, _In_ const CD2DRectF& rTile, _In_ double fDistanceFromCenterPosition, _In_ const CString& sFile) : m_nTileX{nTileX},
                                                                                                                                                                m_nTileY{nTileY},
                                                                                                                                                                m_rTile{rTile},
                                                                                                                                                                m_fDistanceFromCenterPosition{fDistanceFromCenterPosition},
                                                                                                                                                                m_sFile{sFile}
  {
  }
#endif
  COSMCtrlDownloadTileItem(const COSMCtrlDownloadTileItem&) = default;
#pragma warning(suppress: 26439)
  COSMCtrlDownloadTileItem(COSMCtrlDownloadTileItem&&) = default;
  ~COSMCtrlDownloadTileItem() = default;

//Methods
  COSMCtrlDownloadTileItem& operator=(const COSMCtrlDownloadTileItem&) = default;
#pragma warning(suppress: 26439)
  COSMCtrlDownloadTileItem& operator=(COSMCtrlDownloadTileItem&&) = default;

//Member variables
  int m_nTileX{-1}; //The X tile value
  int m_nTileY{-1}; //The Y tile value
#ifdef COSMCTRL_NOD2D
  Gdiplus::RectF m_rTile; //The client coordinates of the tile
#else
  CD2DRectF m_rTile; //The client coordinates of the tile
#endif //#ifdef COSMCTRL_NOD2D
  double m_fDistanceFromCenterPosition{0}; //The distance of the center of this tile from the center position of the map
  CString m_sFile; //The local cached filename
};


//The helper class to do async tile downloads via WinHTTP
class COSMCtrlWinHTTPRequest : public WinHTTPWrappers::CAsyncDownloader
{
public:
//Constructors / Destructors
  COSMCtrlWinHTTPRequest() = default;
  COSMCtrlWinHTTPRequest(const COSMCtrlWinHTTPRequest&) = delete;
  COSMCtrlWinHTTPRequest(COSMCtrlWinHTTPRequest&&) = delete;
  ~COSMCtrlWinHTTPRequest() = default; //NOLINT(modernize-use-override)

//Methods
  COSMCtrlWinHTTPRequest& operator=(const COSMCtrlWinHTTPRequest&) = delete;
  COSMCtrlWinHTTPRequest& operator=(COSMCtrlWinHTTPRequest&&) = delete;
  void SetOSMCtrl(_In_ COSMCtrl* pOSMCtrl) noexcept;

protected:
//Methods
  HRESULT OnCallbackComplete(_In_ HRESULT hr, _In_ HINTERNET hInternet, _In_ DWORD dwInternetStatus, _In_opt_ LPVOID lpvStatusInformation, _In_ DWORD dwStatusInformationLength) override;

//Member variables
  COSMCtrl* m_pOSMCtrl{nullptr};
};


//The main control class
class COSMCtrl : public CStatic
{
public:
//Enums
  enum class DownloadOrder
  {
    ClosestToCenterFirst,
    YOuterXInnerLeftToRight
  };
  enum class ControlAnchorPosition
  {
    Undefined = -1,
    TopLeft,
    TopRight,
    BottomLeft,
    BottomRight
  };
  enum Misc
  {
    OSMTileWidth = 256,
    OSMTileHeight = 256,
    OSMHalfTileWidth = 128,
    OSMHalfTileHeight = 128,
    OSMMaxZoom = 18,
    OSMMinZoom = 0,
    SLIDER_ID = 1,
    COPYRIGHT_ID = 2
  };
  enum class MapItem
  {
    None,
    Map,
    ScrollerUp,
    ScrollerDown,
    ScrollerRight,
    ScrollerLeft,
    ZoomIn,
    ZoomOut,
    ZoomBar,
    Marker,
    Polyline,
    Polygon,
    PolylineNode,
    PolygonNode,
    Circle,
    CircleCircumference,
  };
  enum class MapMode
  {
    Normal,
    Selection,
    PolylineCreation,
    PolygonCreation,
    MarkerCreation,
    CircleCreation,
    Rotate,
  };
  enum class ScaleBarUnits
  {
    UseOSDefault,
    Metric,
    Imperial
  };

//Constructors / Destructors
  COSMCtrl();
  COSMCtrl(const COSMCtrl&) = delete;
  COSMCtrl(COSMCtrl&&) = delete;
  ~COSMCtrl() = default; //NOLINT(modernize-use-override)

//static methods
  static CString GetTileCachePath(_In_ const CString& sCacheDirectory, _In_ int nZoom, _In_ int nTileX, _In_ int nTileY);
#ifdef COSMCTRL_NOD2D
  _Return_type_success_(return != 0) static bool PositionToClient(_In_ const COSMCtrlPosition& position, _In_ const COSMCtrlPosition& centerPosition, _In_ double fZoom, _In_ const CRect& rClient, _Out_ Gdiplus::PointF& point) noexcept;
  _Return_type_success_(return != 0) static bool ClientToPosition(_In_ const Gdiplus::PointF& point, _In_ const COSMCtrlPosition& centerPosition, _In_ double fZoom, _In_ const CRect& rClient, _Out_ COSMCtrlPosition& position) noexcept;
#else
  _Return_type_success_(return != 0) static bool PositionToClient(_In_ const COSMCtrlPosition& position, _In_ const COSMCtrlPosition& centerPosition, _In_ double fZoom, _In_ double fBearingOfTopOfMap, _In_ const CRect& rClient, _Out_ CD2DPointF& point) noexcept;
  _Return_type_success_(return != 0) static bool ClientToPosition(_In_ const CD2DPointF& point, _In_ const COSMCtrlPosition& centerPosition, _In_ double fZoom, _In_ double fBearingOfTopOfMap, _In_ const CRect& rClient, _Out_ COSMCtrlPosition& position) noexcept;
#endif //#ifdef COSMCTRL_NOD2D

//Methods
  COSMCtrl& operator=(const COSMCtrl&) = delete;
  COSMCtrl& operator=(COSMCtrl&&) = delete;
  BOOL Create(LPCTSTR lpszText, DWORD dwStyle, const RECT& rect, CWnd* pParentWnd, UINT nID = 0xFFFF) override;
#ifdef COSMCTRL_NOD2D
  virtual bool Draw(_In_ CDC* pDC, _In_ const CRect& rSource, _In_opt_ const CRect* prDest = nullptr, _In_ bool bDrawScrollRose = true, _In_ bool bDrawZoomBar = true, _In_ bool bDrawScaleBar = true, _In_ bool bDrawMarkers = true, _In_ bool bDrawPolylines = true, _In_ bool bDrawPolygons = true, _In_ bool bDrawCircles = true, _In_ bool bDrawGPSTracks = true);
  virtual bool Draw(_In_ Gdiplus::Graphics& graphics, _In_ const CRect& rClient, _In_opt_ const CRect* prClip = nullptr, _In_ bool bDrawScrollRose = true, _In_ bool bDrawZoomBar = true, _In_ bool bDrawScaleBar = true, _In_ bool bDrawMarkers = true, _In_ bool bDrawPolylines = true, _In_ bool bDrawPolygons = true, _In_ bool bDrawCircles = true, _In_ bool bDrawGPSTracks = true);
#else
  virtual HRESULT CreateImageFromWICBitmap(_In_ IWICBitmapSource* WICBitmapSource, _Inout_ ATL::CImage& image);
  virtual HRESULT Draw(_In_ CDC* pDC, _In_ const CRect& rSource, _In_opt_ const CRect* prDest = nullptr, _In_ bool bDrawScrollRose = true, _In_ bool bDrawZoomBar = true, _In_ bool bDrawScaleBar = true, _In_ bool bDrawMarkers = true, _In_ bool bDrawPolylines = true, _In_ bool bDrawPolygons = true, _In_ bool bDrawCircles = true, _In_ bool bDrawGPSTracks = true);
  virtual HRESULT Draw(_In_ CRenderTarget* pRenderTarget, _In_ const CRect& rClient, _In_opt_ const CRect* prClip = nullptr, _In_ bool bDrawScrollRose = true, _In_ bool bDrawZoomBar = true, _In_ bool bDrawScaleBar = true, _In_ bool bDrawMarkers = true, _In_ bool bDrawPolylines = true, _In_ bool bDrawPolygons = true, _In_ bool bDrawCircles = true, _In_ bool bDrawGPSTracks = true, _In_ bool bResetDPI = true);
  _NODISCARD double GetBearingOfTopOfMap() const noexcept { return m_fBearingOfTopOfMap; };
  virtual bool SetBearingOfTopOfMap(_In_ double fBearingOfTopOfMap, _In_ double fAnimationDuration, _In_ bool bRestartDownloading = true);
  virtual void DestroyD2DResources();
#endif //#ifdef COSMCTRL_NOD2D
  _NODISCARD double GetZoom() const noexcept { return m_fZoom; };
  virtual bool SetZoom(_In_ double fZoom, _In_ double fAnimationDuration, _In_ bool bRestartDownloading = true);
  virtual bool SetCenter(_In_ const COSMCtrlPosition& position, _In_ double fAnimationDuration, _In_ bool bRestartDownloading = true);
  virtual bool SetCenterAndZoom(_In_ const COSMCtrlPosition& position, _In_ double fZoom, _In_ double fAnimationDuration, _In_ bool bRestartDownloading = true);
  virtual bool SetCenterAndZoomAndBearing(_In_ const COSMCtrlPosition& position, _In_ double fZoom, _In_ double fBearingOfTopOfMap, _In_ double fAnimationDuration, _In_ bool bRestartDownloading = true);
  _NODISCARD COSMCtrlPosition GetCenter() const noexcept { return m_CenterPosition; };
  virtual bool ZoomToBounds(_In_ const COSMCtrlPosition& position1, _In_ const COSMCtrlPosition& position2, _In_ double fAnimationDuration, _In_ bool bRestartDownloading = true);
  _NODISCARD CString GetCacheDirectory() const { return m_sCacheDirectory; };
  virtual void SetCacheDirectory(_In_ const CString& sCacheDirectory, _In_ bool bInvalidateCache = false);
  _NODISCARD bool GetDownloadTiles() const noexcept { return m_bDownloadTiles; };
  virtual void SetDownloadTiles(_In_ bool bDownloadTiles);
  virtual void Refresh();
  virtual void SetDrawCenterCrossHairs(_In_ bool bDrawCenterCrossHairs);
  _NODISCARD bool GetDrawCenterCrossHairs() const noexcept { return m_bDrawCenterCrossHairs; };
  virtual void SetMapMode(_In_ MapMode mode);
  _NODISCARD MapMode GetMapMode() const noexcept { return m_MapMode; };
  _NODISCARD bool GetDrawScrollRose() const noexcept { return m_bDrawScrollRose; };
  virtual bool SetDrawScrollRose(_In_ bool bDrawScrollRose);
  _NODISCARD ControlAnchorPosition GetScrollRoseAnchorPosition() const noexcept { return m_ScrollRoseAnchorPosition; };
  virtual bool SetScrollRoseAnchorPosition(_In_ ControlAnchorPosition controlAnchorPosition);
  _NODISCARD CPoint GetScrollRoseAnchorPositionOffset() const noexcept { return m_ptOffsetScrollRose; };
  virtual bool SetScrollRoseAnchorPositionOffset(_In_ const CPoint& ptOffset);
  _NODISCARD bool GetDrawZoomBar() const noexcept { return m_bDrawZoomBar; };
  virtual bool SetDrawZoomBar(_In_ bool bDrawZoomBar);
  _NODISCARD ControlAnchorPosition GetZoomBarAnchorPosition() const noexcept { return m_ZoomBarAnchorPosition; };
  virtual bool SetZoomBarAnchorPosition(_In_ ControlAnchorPosition controlAnchorPosition);
  _NODISCARD CPoint GetZoomBarAnchorPositionOffset() const noexcept { return m_ptOffsetZoomBar; };
  virtual bool SetZoomBarAnchorPositionOffset(_In_ const CPoint& ptOffset);
  _NODISCARD bool GetUseTransparencyForZoomBar() const noexcept { return m_bTransparencyForZoomBar; };
  virtual void SetUseTransparencyForZoomBar(_In_ bool bUseTransparencyForZoomBar);
  _NODISCARD bool GetDrawZoomBarAsSlider() const noexcept { return m_bDrawZoomBarAsSlider; };
  virtual bool SetDrawZoomBarAsSlider(_In_ bool bDrawZoomBarAsSlider);
  _NODISCARD bool GetDrawScaleBar() const noexcept { return m_bDrawScaleBar; };
  virtual void SetDrawScaleBar(_In_ bool bDrawScaleBar);
  _NODISCARD ControlAnchorPosition GetScaleBarAnchorPosition() const noexcept { return m_ScaleBarAnchorPosition; };
  void SetScaleBarAnchorPosition(_In_ ControlAnchorPosition controlAnchorPosition);
  _NODISCARD CPoint GetScaleBarAnchorPositionOffset() const noexcept { return m_ptOffsetScaleBar; };
  virtual void SetScaleBarAnchorPositionOffset(_In_ const CPoint& ptOffset);
  virtual void SetScaleBarUnits(_In_ ScaleBarUnits units) noexcept { m_ScaleBarUnits = units; };
  _NODISCARD ScaleBarUnits GetScaleBarUnits() const noexcept { return m_ScaleBarUnits; };
  _NODISCARD bool GetDrawCopyright() const noexcept { return m_bDrawCopyright; };
  virtual bool SetDrawCopyright(_In_ bool bDrawCopyright);
  _NODISCARD ControlAnchorPosition GetCopyrightAnchorPosition() const noexcept { return m_CopyrightAnchorPosition; };
  virtual bool SetCopyrightAnchorPosition(_In_ ControlAnchorPosition controlAnchorPosition);
  _NODISCARD CPoint GetCopyrightAnchorPositionOffset() const noexcept { return m_ptOffsetCopyright; };
  virtual bool SetCopyrightAnchorPositionOffset(_In_ const CPoint& ptOffset);
  _NODISCARD bool GetAllowDrag() const noexcept { return m_bAllowDrag; };
  void SetAllowDrag(_In_ bool bAllowDrag) noexcept { m_bAllowDrag = bAllowDrag; };
  _NODISCARD bool GetAllowKeyboard() const noexcept { return m_bAllowKeyboard; };
  void SetAllowKeyboard(_In_ bool bAllowKeyboard) noexcept { m_bAllowKeyboard = bAllowKeyboard; };
  _NODISCARD bool GetAllowMouseZoom() const noexcept { return m_bAllowMouseZoom; };
  void SetAllowMouseZoom(_In_ bool bAllowMouseZoom) noexcept { m_bAllowMouseZoom = bAllowMouseZoom; };
  _NODISCARD bool GetAllowDoubleClickZoom() const noexcept { return m_bAllowDoubleClickZoom; };
  void SetAllowDoubleClickZoom(_In_ bool bAllowDoubleClickZoom) noexcept { m_bAllowDoubleClickZoom = bAllowDoubleClickZoom; };
  _NODISCARD bool GetAllowUsePreviousZoomStretch() const noexcept { return m_bAllowPreviousZoomStretch; };
  void SetAllowUsePreviousZoomStretch(_In_ bool bAllowPreviousZoomStretch) noexcept { m_bAllowPreviousZoomStretch = bAllowPreviousZoomStretch; };
  _NODISCARD bool GetAllowUseNextZoomSqueeze() const noexcept { return m_bAllowNextZoomSqueeze; };
  void SetAllowUseNextZoomSqueeze(_In_ bool bAllowNextZoomSqueeze) noexcept { m_bAllowNextZoomSqueeze = bAllowNextZoomSqueeze; };
  _NODISCARD CString GetUserAgent() const { return m_sUserAgent; };
  void SetUserAgent(_In_ const CString& sUserAgent) { m_sUserAgent = sUserAgent; };
  _NODISCARD CString GetProxyServer() const { return m_sProxyServer; };
  void SetProxyServer(_In_ const CString& sProxyServer) { m_sProxyServer = sProxyServer; };
  _NODISCARD CString GetProxyBypassList() const { return m_sProxyBypassList; };
  void SetProxyBypassList(_In_ const CString& sProxyBypassList) { m_sProxyBypassList = sProxyBypassList; };
  _NODISCARD bool GetDoAnimations() const noexcept { return m_bAnimations; };
  void SetDoAnimations(_In_ bool bAnimations) noexcept { m_bAnimations = bAnimations; };
  _NODISCARD DownloadOrder GetDownloadOrder() const noexcept { return m_DownloadOrder; };
  void SetDownloadOrder(_In_ DownloadOrder downloadOrder) noexcept { m_DownloadOrder = downloadOrder; };
  _NODISCARD bool GetDrawTileOutlines() const noexcept { return m_bDrawTileOutlines; };
  virtual void SetDrawTileOutlines(_In_ bool bDrawTileOutlines);
  virtual MapItem HitTest(_In_ const CPoint& point, _Out_ INT_PTR& nItem, _Out_ INT_PTR& nSubItem, _In_ const CRect& rClient) const;
  _Return_type_success_(return != 0) virtual bool GetBoundingRect(_In_ const COSMCtrlMarker& marker, _Out_ CRect& rBounds, _In_ const CRect& rClient) const;
  _Return_type_success_(return != 0) virtual bool GetBoundingNodeRect(_In_ const COSMCtrlPolyline& polyline, _In_ INT_PTR nIndex, _Out_ CRect& rBounds, _In_ const CRect& rClient) const;
  _Return_type_success_(return != 0) virtual bool GetBoundingNodeRect(_In_ const COSMCtrlPolygon& polygon, _In_ INT_PTR nIndex, _Out_ CRect& rBounds, _In_ const CRect& rClient) const;
#ifdef COSMCTRL_NOD2D
  _Return_type_success_(return != 0) virtual bool ClientToPosition(_In_ const Gdiplus::PointF& point, _In_ const CRect& rClient, _Out_ COSMCtrlPosition& position) const noexcept;
  _Return_type_success_(return != 0) virtual bool PositionToClient(_In_ const COSMCtrlPosition& position, _In_ const CRect& rClient, _Out_ Gdiplus::PointF& point) const noexcept;
#else
  _Return_type_success_(return != 0) virtual bool ClientToPosition(_In_ const CD2DPointF& point, _In_ const CRect& rClient, _Out_ COSMCtrlPosition& position) const noexcept;
  _Return_type_success_(return != 0) virtual bool PositionToClient(_In_ const COSMCtrlPosition& position, _In_ const CRect& rClient, _Out_ CD2DPointF& point) const noexcept;
#endif //#ifdef COSMCTRL_NOD2D
  _NODISCARD int GetDownloadTilesEdgeCount() const noexcept { return m_nDownloadTilesEdgeCount; };
  void SetDownloadTilesEdgeCount(_In_ int nDownloadTilesEdgeCount) noexcept { m_nDownloadTilesEdgeCount = nDownloadTilesEdgeCount; };
  virtual void ScrollTileUp();
  virtual void ScrollTileDown();
  virtual void ScrollTileLeft();
  virtual void ScrollTileRight();
  virtual bool ScrollUp(_In_ int nPixels);
  virtual bool ScrollDown(_In_ int nPixels);
  virtual bool ScrollLeft(_In_ int nPixels);
  virtual bool ScrollRight(_In_ int nPixels);
  virtual void DeselectAllItems();
  virtual void SelectAllItems();
  virtual void SelectMarker(_In_ INT_PTR nIndex);
  virtual void SelectPolyline(_In_ INT_PTR nIndex);
  virtual void SelectPolylineNode(_In_ INT_PTR nIndex, _In_ INT_PTR nSubItemIndex);
  virtual void SelectPolygon(_In_ INT_PTR nIndex);
  virtual void SelectPolygonNode(_In_ INT_PTR nIndex, _In_ INT_PTR nSubItemIndex);
  virtual void SelectCircle(_In_ INT_PTR nIndex);
  virtual size_t DeleteSelectedItems();
  _NODISCARD virtual int GetFirstSelectedPolyline() const;
  _NODISCARD virtual int GetFirstSelectedPolygon() const;
  _NODISCARD virtual int GetFirstSelectedMarker() const;
  _NODISCARD virtual int GetFirstSelectedCircle() const;
  virtual INT_PTR ShowOperationsDialog(_In_ COSMCtrlMapOperationsDlg& dialog);
  virtual bool DecimatePolyline(_Inout_ COSMCtrlPolyline& polyline, _In_ int nNewNodesPerWay, _In_ const CRect& rClient);
  virtual bool DecimatePolygon(_Inout_ COSMCtrlPolygon& polyline, _In_ int nNewNodesPerWay, _In_ const CRect& rClient);
  virtual void DeleteCachedTiles();
  virtual void SetMaxCachedTiles(_In_ size_t nMaxTiles);
  _NODISCARD INT_PTR GetMaxCachedTiles() const noexcept { return m_nMaxCachedTiles; };
  void SetScrollPixels(_In_ int nPixels) noexcept { m_nScrollPixels = nPixels; };
  _NODISCARD int GetScrollPixels() const noexcept { return m_nScrollPixels; };
  virtual void SetMaxGPSTracks(_In_ int nGPSTracks);
  _NODISCARD std::vector<COSMCtrlNode>::size_type GetMaxGPSTracks() const noexcept { return m_nMaxGPSTracks; };
  virtual void AddGPSTrack(_In_ bool bGPSFix, _In_ double fLongitude, _In_ double fLatitude, _In_ double fBearing, _In_ double fSpeed, _In_ bool bBearingValid, _In_ bool bSpeedValid, _In_ double fHDOP);
  _NODISCARD IOSMCtrlHandler* GetEventHandler() const noexcept { return m_pEventHandler; };
  void SetEventHandler(_In_ IOSMCtrlHandler* pHandler) noexcept { m_pEventHandler = pHandler; };
  _NODISCARD IOSMCtrlTileProvider* GetTileProvider() const noexcept { return m_pTileProvider; };
  virtual void SetTileProvider(_In_ IOSMCtrlTileProvider* pTileProvider);
  void SetDefaultAnimationDuration(_In_ double fDuration) noexcept { m_fDefaultAnimationDuration = fDuration; }
  _NODISCARD double GetDefaultAnimationDuration() const noexcept { return m_fDefaultAnimationDuration; }
  _NODISCARD ULONGLONG GetLastCenterTickCount() const noexcept { return m_nLastSetCenterTickCount; }
  void SetAnimationInterval(_In_ int nAnimationInterval) noexcept { m_nAnimationInterval = nAnimationInterval; }
  _NODISCARD int GetAnimationInterval() const noexcept {return m_nAnimationInterval; }

//Member variables
  std::vector<std::shared_ptr<COSMCtrlMarker>> m_Markers; //The marker array associated with this map
  std::vector<std::shared_ptr<COSMCtrlPolygon>> m_Polygons; //The polygons array associated with this map
  std::vector<std::shared_ptr<COSMCtrlPolyline>> m_Polylines; //The polylines array associated with this map
  std::vector<std::shared_ptr<COSMCtrlCircle>> m_Circles; //The circles array associated with this map
  std::vector<COSMCtrlIcon*> m_Icons; //The icon array associated with this map
  std::unique_ptr<COSMCtrlPolygon> m_SelectionPolygon; //The selection polygon
  std::unique_ptr<COSMCtrlPolyline> m_GPSTrack; //The GPS track on the map
  bool m_bGPSFix; //Do we currently have a GPS fix
#ifdef COSMCTRL_NOD2D
  Gdiplus::Color m_colorGPSTrackPointer; //The color to use for the GPS track pointer
  Gdiplus::Color m_colorGPSTrackNoBearingPointer; //The color to use for the GPS track pointer when we do not have a bearing
#endif //#ifdef COSMCTRL_NOD2D

protected:
//Message handlers
  afx_msg void OnPaint();
  afx_msg BOOL OnEraseBkgnd(CDC* pDC);
  afx_msg void OnDestroy();
  afx_msg void OnLButtonDown(UINT nFlags, CPoint point);
  afx_msg void OnLButtonUp(UINT nFlags, CPoint point);
  afx_msg void OnLButtonDblClk(UINT nFlags, CPoint point);
  afx_msg void OnMouseMove(UINT nFlags, CPoint point);
  afx_msg void OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags);
  afx_msg void OnChar(UINT nChar, UINT nRepCnt, UINT nFlags);
  afx_msg BOOL OnMouseWheel(UINT nFlags, short zDelta, CPoint pt);
  afx_msg void OnVScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar);
  afx_msg BOOL OnToolTipText(UINT nID, NMHDR* pNMHDR, LRESULT* pResult);
  afx_msg void OnNMClickCopyright(NMHDR* pNMHDR, LRESULT* pResult);
  afx_msg LRESULT OnCreateChildControls(WPARAM wParam, LPARAM lParam);
  afx_msg void OnSize(UINT nType, int cx, int cy);
  afx_msg LRESULT OnDownloadComplete(WPARAM wParam, LPARAM lParam);
#ifndef COSMCTRL_NOD2D
  afx_msg LRESULT OnDraw2D(WPARAM wParam, LPARAM lParam);
  afx_msg LRESULT OnRecreatedResources(WPARAM wParam, LPARAM lParam);
#endif //#ifndef COSMCTRL_NOD2D
  afx_msg void OnTimer(UINT_PTR nIDEvent);

  DECLARE_MESSAGE_MAP()

//Methods
  INT_PTR OnToolHitTest(CPoint point, TOOLINFO* pTI) const override;
  virtual CPoint GetControlPosition(_In_ ControlAnchorPosition anchorPosition, _In_ const CPoint& ptOffset, _In_ const CRect& rClient);
  virtual HRESULT CreateSession(_Inout_ WinHTTPWrappers::CSession& session, _In_ DWORD dwFlags);
  virtual HRESULT CreateConnection(_In_ const CString& sServer, _In_ INTERNET_PORT nPort);
  virtual HRESULT CreateRequest(_In_ const CString& sObject, _In_ const CString& sFile, _In_ bool bForcedRefresh);
  virtual bool CreateZoomBarSlider();
  virtual bool CreateCopyrightLinkCtrl();
  _NODISCARD virtual bool HitTest(_In_ const CPoint& point, _In_ const COSMCtrlPolyline& polyline, _In_ const CRect& rClient) const;
  _NODISCARD virtual bool HitTest(_In_ const CPoint& point, _In_ const COSMCtrlPolygon& polygon, _In_ const CRect& rClient) const;
  _NODISCARD virtual MapItem HitTest(_In_ const CPoint& point, _In_ const COSMCtrlCircle& circle, _In_ const CRect& rClient) const;
#ifdef COSMCTRL_NOD2D
  virtual bool DrawTileNotAvailable(_In_ Gdiplus::Graphics& graphics, _In_ const Gdiplus::RectF& rTile);
  virtual bool DrawTile(_In_ Gdiplus::Graphics& graphics, _In_ const Gdiplus::RectF& rTile, _In_ int nTileX, _In_ int nTileY);
  virtual bool DrawTileOutline(_In_ Gdiplus::Graphics& graphics, _In_ const Gdiplus::RectF& rTile);
  virtual bool DrawScrollRose(_In_ Gdiplus::Graphics& graphics, _In_ const CRect& rClip, _In_ const CRect& rClient);
  virtual bool DrawZoomBar(_In_ Gdiplus::Graphics& graphics, _In_ const CRect& rClip, _In_ const CRect& rClient);
  virtual bool DrawScaleBar(_In_ Gdiplus::Graphics& graphics, _In_ const CRect& rClip, _In_ const CRect& rClient);
  virtual bool DrawCenterCrossHairs(_In_ Gdiplus::Graphics& graphics, _In_ const CRect& rClip, _In_ const CRect& rClient);
  virtual int DrawPolylines(_In_ Gdiplus::Graphics& graphics, _In_ const CRect& rClip, _In_ const CRect& rClient);
  virtual int DrawPolygons(_In_ Gdiplus::Graphics& graphics, _In_ const CRect& rClip, _In_ const CRect& rClient);
  virtual int DrawCircles(_In_ Gdiplus::Graphics& graphics, _In_ const CRect& rClip, _In_ const CRect& rClient);
  virtual int DrawMarkers(_In_ Gdiplus::Graphics& graphics, _In_ const CRect& rClip, _In_ const CRect& rClient);
  virtual bool DrawGPSTrack(_In_ Gdiplus::Graphics& graphics, _In_ const CRect& rClient);
  virtual bool DrawGPSLocation(_In_ Gdiplus::Graphics& graphics, _In_ const CRect& rClient);
  bool Draw3dRect(_In_ Gdiplus::Graphics& graphics, _In_ const Gdiplus::RectF& r, _In_ const Gdiplus::Color& clrTopLeft, _In_ const Gdiplus::Color& clrBottomRight);
  _Return_type_success_(return != 0) virtual Gdiplus::CachedBitmap* GetCachedBitmap(_In_ Gdiplus::Graphics& graphics, _In_ const CStringW& sFile, _In_ int nTileX, _In_ int nTileY, _In_ int nZoom, _Inout_ Gdiplus::Bitmap*& pBitmap);
#else
  virtual bool DrawTileNotAvailable(_In_ CRenderTarget* pRenderTarget, _In_ const CD2DRectF& rTile);
  virtual bool DrawTile(_In_ CRenderTarget* pRenderTarget, _In_ const CD2DRectF& rTile, _In_ int nTileX, _In_ int nTileY);
  virtual bool DrawTileOutline(_In_ CRenderTarget* pRenderTarget, _In_ const CD2DRectF& rTile);
  virtual bool DrawScrollRose(_In_ CRenderTarget* pRenderTarget, _In_ const CRect& rClip, _In_ const CRect& rClient);
  virtual bool DrawZoomBar(_In_ CRenderTarget* pRenderTarget, _In_ const CRect& rClip, _In_ const CRect& rClient);
  virtual bool DrawScaleBar(_In_ CRenderTarget* pRenderTarget, _In_ const CRect& rClip, _In_ const CRect& rClient);
  virtual bool DrawCenterCrossHairs(_In_ CRenderTarget* pRenderTarget, _In_ const CRect& rClip, _In_ const CRect& rClient);
  virtual int DrawPolylines(_In_ CRenderTarget* pRenderTarget, _In_ const CRect& rClip, _In_ const CRect& rClient);
  virtual int DrawPolygons(_In_ CRenderTarget* pRenderTarget, _In_ const CRect& rClip, _In_ const CRect& rClient);
  virtual int DrawCircles(_In_ CRenderTarget* pRenderTarget, _In_ const CRect& rClip, _In_ const CRect& rClient);
  virtual int DrawMarkers(_In_ CRenderTarget* pRenderTarget, _In_ const CRect& rClip, _In_ const CRect& rClient);
  virtual bool DrawGPSTrack(_In_ CRenderTarget* pRenderTarget, _In_ const CRect& rClient);
  virtual bool DrawGPSHDOP(_In_ CRenderTarget* pRenderTarget, _In_ const CRect& rClient);
  virtual bool DrawGPSLocation(_In_ CRenderTarget* pRenderTarget, _In_ const CRect& rClient);
  bool Draw3dRect(_In_ CRenderTarget* pRenderTarget, _In_ const CD2DRectF& r, _In_ const D2D1::ColorF& clrTopLeft, _In_ const D2D1::ColorF& clrBottomRight);
  _Return_type_success_(return != 0) virtual CD2DBitmap* GetCachedBitmap(_In_ CRenderTarget* pRebderTarget, _In_ const CString& sFile, _In_ int nTileX, _In_ int nTileY, _In_ int nZoom);
#endif //#ifdef COSMCTRL_NOD2D
  virtual CString FormScaleBarText(_In_ double fScaleDistance, _In_ bool bMetric);
  virtual CString FormCopyrightText();
  virtual void PerformInMemoryGPSTrackMaintenance(_In_ std::vector<COSMCtrlNode>::size_type nMaxTracks);
  virtual void EnsureBoundingRectForGPSTrackTriangle(_In_ const COSMCtrlPosition& position, _In_ CRect& rBounds, _In_ const CRect& rClient) const;
  virtual bool UseMetric();
  virtual void StartDrag(_In_ const CPoint& point, _In_ MapItem item, _In_ INT_PTR nItem, _In_ INT_PTR nSubItem);
  virtual void CalculateScaleBar(_Out_ double& fDistance, _Out_ double& fScaleDistance, _Out_ bool& bMetric, _Out_ int& nScaleLength);
  virtual void CalculateStartTilePositions(_In_ const CRect& rClient, _Out_ double& fClientX, _Out_ double& fClientY, _Out_ double& fEndClientX, _Out_ double& fEndClientY, _Out_ int& nStartX, _Out_ int& nStartY, _Out_ double& fOSMTileWidth, _Out_ double& fOSMTileHeight) noexcept;
  bool ContainsPosition(_In_ const COSMCtrlPosition& topLeft, _In_ const COSMCtrlPosition& bottomRight, _In_ const COSMCtrlPosition& point) noexcept;
  virtual CStringW GetFileNameForCache(_In_ const CStringW& sFile);
  virtual void PerformTileCacheMaintenance(_In_ size_t nMaxTiles);
  virtual void InitializeNewPolyline(_Inout_ COSMCtrlPolyline& polyline);
  virtual void InitializeNewPolygon(_Inout_ COSMCtrlPolygon& polygon);
  virtual void InitializeNewSelectionPolygon(_Inout_ COSMCtrlPolygon& polygon);
  virtual void InitializeNewMarker(_Inout_ COSMCtrlMarker& marker);
  virtual void InitializeNewCircle(_Inout_ COSMCtrlCircle& circle);
  virtual void HandleLButtonDownStandard(_In_ UINT nFlags, _In_ const CPoint& point, _In_ MapItem item, _In_ INT_PTR nItem, _In_ INT_PTR nSubItem);
  virtual void HandleLButtonDownPolylineCreation(_In_ UINT nFlags, _In_ const CPoint& point, _In_ MapItem item, _In_ INT_PTR nItem, _In_ INT_PTR nSubItem);
  virtual void HandleLButtonDownPolygonCreation(_In_ UINT nFlags, _In_ const CPoint& point, _In_ MapItem item, _In_ INT_PTR nItem, _In_ INT_PTR nSubItem);
  virtual void HandleLButtonDownPolylineCreationWithNode(_In_ UINT nFlags, _In_ const CPoint& point, _In_ MapItem item, _In_ INT_PTR nItem, _In_ INT_PTR nSubItem);
  virtual void HandleLButtonDownPolygonCreationWithNode(_In_ UINT nFlags, _In_ const CPoint& point, _In_ MapItem item, _In_ INT_PTR nItem, _In_ INT_PTR nSubItem);
  virtual void HandleLButtonDownMarkerCreation(_In_ UINT nFlags, _In_ const CPoint& point, _In_ MapItem item, _In_ INT_PTR nItem, _In_ INT_PTR nSubItem);
  virtual void HandleLButtonDownCircleCreation(_In_ UINT nFlags, _In_ const CPoint& point, _In_ MapItem item, _In_ INT_PTR nItem, _In_ INT_PTR nSubItem);
  virtual void HandleLButtonDownCircleCircumference(_In_ UINT nFlags, _In_ const CPoint& point, _In_ MapItem item, _In_ INT_PTR nItem, _In_ INT_PTR nSubItem);
  virtual void HandleLButtonDownSelection(_In_ UINT nFlags, _In_ const CPoint& point, _In_ MapItem item, _In_ INT_PTR nItem, _In_ INT_PTR nSubItem);
  virtual void HandleLButtonDownPolygonNode(_In_ UINT nFlags, _In_ const CPoint& point, _In_ MapItem item, _In_ INT_PTR nItem, _In_ INT_PTR nSubItem);
  virtual void HandleLButtonDownPolygon(_In_ UINT nFlags, _In_ const CPoint& point, _In_ MapItem item, _In_ INT_PTR nItem, _In_ INT_PTR nSubItem);
  virtual void HandleLButtonDownPolylineNode(_In_ UINT nFlags, _In_ const CPoint& point, _In_ MapItem item, _In_ INT_PTR nItem, _In_ INT_PTR nSubItem);
  virtual void HandleLButtonDownPolyline(_In_ UINT nFlags, _In_ const CPoint& point, _In_ MapItem item, _In_ INT_PTR nItem, _In_ INT_PTR nSubItem);
  virtual void HandleLButtonDownMarker(_In_ UINT nFlags, _In_ const CPoint& point, _In_ MapItem item, _In_ INT_PTR nItem, _In_ INT_PTR nSubItem);
  virtual void HandleLButtonDownCircle(_In_ UINT nFlags, _In_ const CPoint& point, _In_ MapItem item, _In_ INT_PTR nItem, _In_ INT_PTR nSubItem);
  virtual void HandleLButtonDownMap(_In_ UINT nFlags, _In_ const CPoint& point, _In_ MapItem item, _In_ INT_PTR nItem, _In_ INT_PTR nSubItem);
#ifndef COSMCTRL_NOD2D
  virtual void HandleLButtonDownRotate(_In_ UINT nFlags, _In_ const CPoint& point, _In_ MapItem item, _In_ INT_PTR nItem, _In_ INT_PTR nSubItem);
#endif //#ifndef COSMCTRL_NOD2D
  virtual void HandleMouseMoveSelection(_In_ UINT nFlags, _In_ const CPoint& point);
  virtual void HandleMouseMoveCircleCreation(_In_ UINT nFlags, _In_ const CPoint& point);
  virtual void HandleMouseMovePolylineCreation(_In_ UINT nFlags, _In_ const CPoint& point);
  virtual void HandleMouseMovePolygonCreation(_In_ UINT nFlags, _In_ const CPoint& point);
  virtual void HandleMouseMovePolygonCreationWithNode(_In_ UINT nFlags, _In_ const CPoint& point);
  virtual void HandleMouseMoveMap(_In_ UINT nFlags, _In_ const CPoint& point);
  virtual void HandleMouseMoveMarker(_In_ UINT nFlags, _In_ const CPoint& point);
  virtual void HandleMouseMoveCircle(_In_ UINT nFlags, _In_ const CPoint& point);
  virtual void HandleMouseMovePolylineNode(_In_ UINT nFlags, _In_ const CPoint& point);
  virtual void HandleMouseMovePolyline(_In_ UINT nFlags, _In_ const CPoint& point);
  virtual void HandleMouseMovePolygonNode(_In_ UINT nFlags, _In_ const CPoint& point);
  virtual void HandleMouseMovePolygon(_In_ UINT nFlags, _In_ const CPoint& point);
  virtual void HandleLButtonDblClickPolyline(_In_ UINT nFlags, _In_ const CPoint& point, _In_ MapItem item, _In_ INT_PTR nItem, _In_ INT_PTR nSubItem);
  virtual void HandleLButtonDblClickPolygon(_In_ UINT nFlags, _In_ const CPoint& point, _In_ MapItem item, _In_ INT_PTR nItem, _In_ INT_PTR nSubItem);
  virtual void HandleLButtonDblClickMap(_In_ UINT nFlags, _In_ const CPoint& point, _In_ MapItem item, _In_ INT_PTR nItem, _In_ INT_PTR nSubItem);
  virtual void HandleLButtonDblClickMarker(_In_ UINT nFlags, _In_ const CPoint& point, _In_ MapItem item, _In_ INT_PTR nItem, _In_ INT_PTR nSubItem);
  virtual void HandleLButtonDblClickCircle(_In_ UINT nFlags, _In_ const CPoint& point, _In_ MapItem item, _In_ INT_PTR nItem, _In_ INT_PTR nSubItem);
#ifndef COSMCTRL_NOD2D
  virtual void HandleMouseMoveRotate(_In_ UINT nFlags, _In_ const CPoint& point);
#endif //#ifndef COSMCTRL_NOD2D
  virtual HRESULT StartAsyncDownload(_In_ bool bBuildArray, _In_ bool bForcedRefresh);
  virtual HRESULT StopAsyncDownload(bool bCloseConnection) noexcept;
  static ID2D1Factory* GetD2DFactory();
  virtual void BuildDownloadArray(_Inout_ std::vector<COSMCtrlDownloadTileItem>& tilesToDownload, _In_ bool bForceRefresh);
  virtual bool CreateSelectionPolygon();
  virtual bool CreateGPSTrack();

//Member variables
  double m_fZoom; //The current zoom level
  double m_fStartZoom; //The start zoom level (used for animations)
  double m_fFinalZoom; //The final zoom level (used for animations)
  double m_fDefaultAnimationDuration; //The default animation duration in seconds
  COSMCtrlPosition m_CenterPosition; //The center point of the map
  COSMCtrlPosition m_StartCenterPosition; //The start center position (used for animations)
  COSMCtrlPosition m_FinalCenterPosition; //The final center position (used for animations)
  CString m_sCacheDirectory; //The directory where downloaded OSM tiles will be cached
  IOSMCtrlTileProvider* m_pTileProvider; //The tile provider for this map
  COSMCtrlMapnikTileProvider m_DefaultTileProvider; //The default tile provider (Mapnik)
  bool m_bDownloadTiles; //Should we download tiles
  bool m_bMouseCapturedForDrag; //Is the mouse currently captured for dragging the map
  bool m_bMouseCapturedForZoom; //Is the mouse currently captured for zooming the map
  CPoint m_pointMouseCapture; //the client coordinates where we started dragging from
  bool m_bDrawTileOutlines; //Should we draw outlines around all the tiles
  bool m_bDrawScrollRose; //Should we draw the scroll rose
  ControlAnchorPosition m_ScrollRoseAnchorPosition; //The anchor position for the scroll rose
  CPoint m_ptOffsetScrollRose; //the offset position for the scroll rose
  bool m_bDrawZoomBar; //Should we draw the zoom bar
  ControlAnchorPosition m_ZoomBarAnchorPosition; //The anchor position for the Zoom bar
  CPoint m_ptOffsetZoomBar; //the offset position for the Zoom bar
  bool m_bTransparencyForZoomBar; //Should we use transparency to draw the zoom bar
  bool m_bDrawScaleBar; //Should we draw the scale bar
  ControlAnchorPosition m_ScaleBarAnchorPosition; //The anchor position for the Scale bar
  CPoint m_ptOffsetScaleBar; //the offset position for the Scale bar
  bool m_bAllowDrag; //Is the map allowed to be dragged
  bool m_bAllowKeyboard; //Is the map allowed to be scrolled/zoomed via the keyboard
  bool m_bDeltaMode; //Is "Delta" mode enabled
  bool m_bAllowMouseZoom; //Is the map allowed to be zoomed via the mouse wheel
  bool m_bAllowDoubleClickZoom; //Is the map allowed to be zoomed via a double click
  bool m_bAllowPreviousZoomStretch; //Do we allow the previous zoom level image to be stretched
  bool m_bAllowNextZoomSqueeze; //Do we allow the next zoom level image to be squeezed
  CString m_sUserAgent; //The HTTP_USER_AGENT header to use
  DownloadOrder m_DownloadOrder; //The order in which we will download tiles
  int m_nDownloadTilesEdgeCount; //The number of rows or columns of tiles to cache which are outside of the visible client area
  CRect m_rNorthScrollRose; //Used for hittesting
  CRect m_rSouthScrollRose; //Used for hittesting
  CRect m_rEastScrollRose; //Used for hittesting
  CRect m_rWestScrollRose; //Used for hittesting
  CRect m_rZoomInZoomBar; //Used for hittesting
  CRect m_rZoomOutZoomBar; //Used for hittesting
  CRect m_rZoomBar; //Used for hittesting
  CSliderCtrl m_ctrlZoomBar; //The Zoom bar if we are using a native slider control
  bool m_bDrawZoomBarAsSlider; //Should we use a slider control for the zoom bar
  CString m_sCurrentToolTipText; //The current tooltip text
  MapItem m_MouseDragItem; //What we are dragging
  MapItem m_ToolHitItem; //What we hit for tooltips
  INT_PTR m_nMouseDragItemIndex; //The index value associated with what we are dragging
  INT_PTR m_nMouseDragSubItemIndex; //The node index value associated with what we are dragging
  CSize m_DragOffset; //Used to implement smooth / non jumpy dragging of markers, polylines and polygons
  std::vector<COSMCtrlNode> m_OriginalDragNodes; //The nodes we are dragging from (used for dragging of polylines and polygons)
  bool m_bModifyingCircleRadius; //Used to implement circle editing
  MapMode m_MapMode; //The mapping mode currently active
  bool m_bDrawCenterCrossHairs; //Should cross hairs be shown on the map
  bool m_bDrawCopyright; //Should we draw the copyright on the map
  ControlAnchorPosition m_CopyrightAnchorPosition; //The anchor position for the copyright message
  CPoint m_ptOffsetCopyright; //the offset position for the copyright message
  ScaleBarUnits m_ScaleBarUnits; //What units should we use for drawing in the scale bar
  CLinkCtrl m_ctrlCopyright; //The copyright hyperlink control
  std::vector<COSMCtrlCachedTile> m_CachedTiles; //The cache of tiles we have
  INT_PTR m_nMaxCachedTiles; //The maximum number of tiles which are allowed in "m_CachedTiles"
  int m_nInMemoryCachedTilesInsertionCounter; //The insertion counter value for the next item to be added to "m_CachedTiles"
#ifndef COSMCTRL_NOD2D
  CD2DBitmap* m_pEmptyTile; //Empty tile bitmap
#endif //#ifdef COSMCTRL_NOD2D
  int m_nScrollPixels; //The amount to scroll in pixels 
  std::vector<COSMCtrlNode>::size_type m_nMaxGPSTracks; //The maximum number of GPS tracks to display on the map
  IOSMCtrlHandler* m_pEventHandler; //The event handler instance
  double m_fBearingOfTopOfMap; //The compass bearing of the top of the map (i.e. at 12 o clock)
  double m_fBearingAtTopOfMapMouseCapture; //The bearing at the top of the map when the mouse was captured
  double m_fStartBearingAtTopOfMap; //The initial bearing (used for animation)
  double m_fFinalBearingAtTopOfMap; //The final bearing (used for animations)
  bool m_bAnimations; //Should we do animations
  WinHTTPWrappers::CSession m_WinHTTPSession; //The WinHTTP session object
  WinHTTPWrappers::CConnection m_WinHTTPConnection; //The WinHTTP connection object
  std::vector<COSMCtrlDownloadTileItem> m_tilesToDownload; //The tiles to download
  std::vector<COSMCtrlDownloadTileItem> m_tilesPendingToDownload; //The pending tiles to download
  COSMCtrlDownloadTileItem m_CurrentlyDownloadingTile; //The tile currently being downloaded
  bool m_bDownloading; //Are we currently downloading
  COSMCtrlWinHTTPRequest m_WinHTTPRequest; //The WinHTTP request object
  bool m_bInitializedWinHTTPRequest;
  CString m_sProxyServer; //The proxy server to use
  CString m_sProxyBypassList; //The proxy bypass list to use
  ULONGLONG m_nLastSetCenterTickCount; //The Win32 TickCount64 for the last time the map was centered
  ULONGLONG m_nStartAnimationTickCount; //The Win32 TickCount64 for the start time for animations
  UINT_PTR m_nAnimationTimer; //The timer id used for animations
  int m_nAnimationInterval; //The interval in ms for the animation timer
  double m_fCurrentAnimationDuration; //The current animation duration
  bool m_bDoingBearingAnimation;
  bool m_bDoingZoomAnimation;
  bool m_bDoingPositionAnimation;

  friend class COSMCtrlMapOperationsDlg;
  friend class COSMCtrlTimerEventHandler;
  friend class COSMCtrlPolyline;
  friend class COSMCtrlPolygon;
  friend class COSMCtrlCircle;
  friend class COSMCtrlMarker;
};


#endif //#ifndef __OSMCTRL_H__
