/*
Module : OSMCtrlCircle.cpp
Purpose: Implementation for the COSMCtrlCircle class
Created: PJN / 10-04-2011
History: PJN / None

Copyright (c) 2011 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Includes /////////////////////////////////////////////////

#include "stdafx.h"
#include "OSMCtrlCircle.h"
#include "OSMCtrl.h"


//////////////////// Implementation ///////////////////////////////////////////

_Return_type_success_(return != 0) bool COSMCtrlCircle::GetBoundingRect(_Out_ COSMCtrlPosition& topLeft, _Out_ COSMCtrlPosition& bottomRight) const noexcept
{
  double fTopLatitude{-91};
  double fBottomLatitude{91};
  double fLeftLongitude{181};
  double fRightLongitude{-181};

  //Look through the four cardinal positions from the center
  double fBearing{0};
  for (INT_PTR i{0}; i<4; i++)
  {
    //Move the specific distance from the center of the circle
    const COSMCtrlPosition position{COSMCtrlHelper::GetPosition(m_Position, fBearing, m_fRadius, nullptr)};

    if (position.m_fLatitude > fTopLatitude)
      fTopLatitude = position.m_fLatitude;
    if (position.m_fLatitude < fBottomLatitude)
      fBottomLatitude = position.m_fLatitude;
    if (position.m_fLongitude < fLeftLongitude)
      fLeftLongitude = position.m_fLongitude;
    if (position.m_fLongitude > fRightLongitude)
      fRightLongitude = position.m_fLongitude;

    //prepare for the next time around
    fBearing += 90;
  }

  topLeft.m_fLatitude = fTopLatitude;
  topLeft.m_fLongitude = fLeftLongitude;
  bottomRight.m_fLatitude = fBottomLatitude;
  bottomRight.m_fLongitude = fRightLongitude;

  return true;
}

#ifdef COSMCTRL_NOD2D
bool COSMCtrlCircle::Draw(_In_ const COSMCtrl& ctrl, _In_ Gdiplus::GraphicsPath& path, _In_ bool bCircumference, _In_ const CRect& rClient) const
{
  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

  if (m_bVisible && (ctrl.m_fZoom >= m_nMinZoomLevel) && (ctrl.m_fZoom <= m_nMaxZoomLevel))
  {
    //Work out the radius of the circle to draw in pixels
    const COSMCtrlPosition position90{COSMCtrlHelper::GetPosition(m_Position, 90, m_fRadius, nullptr)};
    const COSMCtrlPosition position270{COSMCtrlHelper::GetPosition(m_Position, 270, m_fRadius, nullptr)};
    Gdiplus::PointF pt90;
    Gdiplus::PointF pt270;
    Gdiplus::PointF ptCenter;
    if (ctrl.PositionToClient(position90, rClient, pt90) && ctrl.PositionToClient(position270, rClient, pt270) && ctrl.PositionToClient(m_Position, rClient, ptCenter))
    {
      //Finally draw the circle using GDI+
      Gdiplus::REAL fDiameter{pt90.X - pt270.X};
      if (bCircumference)
        fDiameter += max(m_fLinePenWidth, 2);
      else
        fDiameter -= max(m_fLinePenWidth, 2);
      const Gdiplus::REAL fRadius{fDiameter / 2};
      bSuccess = (path.AddEllipse(ptCenter.X - fRadius, ptCenter.Y - fRadius, fDiameter, fDiameter) == Gdiplus::Ok);
    }
  }

  return bSuccess;
}

bool COSMCtrlCircle::Draw(_In_ COSMCtrl& ctrl, _In_ Gdiplus::Graphics& graphics, _In_ const CRect& rClient)
{
  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

  if (m_bVisible && (ctrl.m_fZoom >= m_nMinZoomLevel) && (ctrl.m_fZoom <= m_nMaxZoomLevel))
  {
    //Create the pens we need
    Gdiplus::Pen penStandard{m_colorPen, m_fLinePenWidth};
    if (penStandard.GetLastStatus() != Gdiplus::Ok)
      return false;
    Gdiplus::Pen penSelected{m_colorSelection, m_fLinePenWidth};
    if (penSelected.GetLastStatus() != Gdiplus::Ok)
      return false;

    //Create the brush we need
    Gdiplus::SolidBrush brush{m_colorBrush};
    if (brush.GetLastStatus() != Gdiplus::Ok)
      return false;

    //Set up the pen attributes
    penStandard.SetDashStyle(m_DashStyle);

    //Work out the radius of the circle to draw in pixels
    const COSMCtrlPosition position90{COSMCtrlHelper::GetPosition(m_Position, 90, m_fRadius, nullptr)};
    const COSMCtrlPosition position270{COSMCtrlHelper::GetPosition(m_Position, 270, m_fRadius, nullptr)};
    Gdiplus::PointF pt90;
    Gdiplus::PointF pt270;
    Gdiplus::PointF ptCenter;
    if (ctrl.PositionToClient(position90, rClient, pt90) && ctrl.PositionToClient(position270, rClient, pt270) && ctrl.PositionToClient(m_Position, rClient, ptCenter))
    {
      //Finally draw the circle using GDI+
      const Gdiplus::REAL fDiameter{pt90.X - pt270.X};
      if (m_DrawingStyle == COSMCtrlCircle::DrawingStyle::LineAndInside)
        bSuccess = (graphics.FillEllipse(&brush, pt270.X, ptCenter.Y - fDiameter / 2, fDiameter, fDiameter) == Gdiplus::Ok);
      if ((m_DrawingStyle == COSMCtrlCircle::DrawingStyle::LineOnly) || (m_DrawingStyle == COSMCtrlCircle::DrawingStyle::LineAndInside))
        bSuccess = (graphics.DrawEllipse(m_bSelected ? &penSelected : &penStandard, pt270.X, ptCenter.Y - fDiameter / 2, fDiameter, fDiameter) == Gdiplus::Ok);
    }
  }

  return bSuccess;
}
#else
#pragma warning(suppress: 26429)
bool COSMCtrlCircle::Draw(_In_ COSMCtrl& ctrl, _In_ CRenderTarget* pRenderTarget, _In_ const CRect& rClient)
{
  //Validate our parameters
  ASSERT(pRenderTarget != nullptr);

  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

  if (m_bVisible && (ctrl.m_fZoom >= m_nMinZoomLevel) && (ctrl.m_fZoom <= m_nMaxZoomLevel))
  {
    //Create the brushes we need
    CD2DBrushProperties brushProperties{m_colorPen.a};
    CD2DSolidColorBrush brushStandard{pRenderTarget, m_colorPen, m_colorPen.a == 1 ? nullptr : &brushProperties};
    CD2DBrushProperties brushProperties2{m_colorSelection.a};
    CD2DSolidColorBrush brushSelected{pRenderTarget, m_colorSelection, m_colorSelection.a == 1 ? nullptr : &brushProperties2};
    CD2DBrushProperties brushProperties3{m_colorBrush.a};
    CD2DSolidColorBrush brush{pRenderTarget, m_colorBrush, m_colorBrush.a == 1 ? nullptr : &brushProperties3};

    //Create the stroke style we need
    ATL::CComPtr<ID2D1StrokeStyle> strokeStyle;
    ID2D1Factory* pFactory{ctrl.GetD2DFactory()};
    if (pFactory == nullptr)
      return false;
    //Note we hardcode most of the circle stroke properties because it does not make sense to expose
    //these values in the case of a circle since there never will be any corners to the object
    //where most of the stroke properties would be actually used
    D2D1_STROKE_STYLE_PROPERTIES strokeStyleProperties;
    strokeStyleProperties.startCap = D2D1_CAP_STYLE_FLAT;
    strokeStyleProperties.endCap = D2D1_CAP_STYLE_FLAT;
    strokeStyleProperties.dashCap = D2D1_CAP_STYLE_FLAT;
    strokeStyleProperties.lineJoin = D2D1_LINE_JOIN_MITER;
    strokeStyleProperties.miterLimit = 1;
    strokeStyleProperties.dashStyle = m_DashStyle;
    strokeStyleProperties.dashOffset = 0;
    const HRESULT hr{pFactory->CreateStrokeStyle(&strokeStyleProperties, nullptr, 0, &strokeStyle)};
    if (FAILED(hr))
      return false;

    //Work out the radius of the circle to draw in pixels
    const COSMCtrlPosition position90{COSMCtrlHelper::GetPosition(m_Position, 90 + ctrl.m_fBearingOfTopOfMap, m_fRadius, nullptr)};
    const COSMCtrlPosition position270{COSMCtrlHelper::GetPosition(m_Position, 270 + ctrl.m_fBearingOfTopOfMap, m_fRadius, nullptr)};
    CD2DPointF pt90;
    CD2DPointF pt270;
    CD2DPointF ptCenter;
    if (ctrl.PositionToClient(position90, rClient, pt90) && ctrl.PositionToClient(position270, rClient, pt270) && ctrl.PositionToClient(m_Position, rClient, ptCenter))
    {
      //Finally draw the circle using D2D
      const FLOAT fRadius{(pt90.x - pt270.x) / 2};
      if (m_DrawingStyle == COSMCtrlCircle::DrawingStyle::LineAndInside)
        pRenderTarget->FillEllipse(CD2DEllipse{CD2DPointF{ptCenter.x, ptCenter.y}, CD2DSizeF{fRadius, fRadius}}, &brush);
      if (m_DrawingStyle == COSMCtrlCircle::DrawingStyle::LineOnly || m_DrawingStyle == COSMCtrlCircle::DrawingStyle::LineAndInside)
        pRenderTarget->DrawEllipse(CD2DEllipse{CD2DPointF{ptCenter.x, ptCenter.y}, CD2DSizeF{fRadius, fRadius}}, m_bSelected ? &brushSelected : &brushStandard, m_fLinePenWidth, m_bSelected ? nullptr : strokeStyle);
      bSuccess = true;
    }
  }

  return bSuccess;
}
#endif //#ifdef COSMCTRL_NOD2D
