/*
Module : OSMCtrlGPX.h
Purpose: Defines the interface and implementation for a set of MFC class to handle GPX v1.1 files.
Created: PJN / 22-08-2010
History: PJN / 06-03-2017 1. Updated copyright details.
                          2. Fix up compile problems when importing "msxml6.dll".
         PJN / 01-10-2017 1. Added support for Elevation variable to CGPXWaypoint class.
                          2. Added support for "Calculated Speed" variable to CGPXWaypoint class.
         PJN / 19-06-2020 1. Extended the CGPX::Save method to allow the GPX "creator" value to be customized
                          2. Added support to the demo application to exporting to a GPX file. Thanks to Ralf
                          Wirtz for providing this nice addition.
                          3. Fixed a bug in CGPX::Save which was creating invalid GPX files. Thanks to Ralf
                          Wirtz for reporting this issue.
         PJN / 30-12-2021 1. Reworked the GPX loading and saving code to use ATL::CComPtr instead of #import

Copyright (c) 2010 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Macros / Includes ////////////////////////////////////////

#pragma once

#ifndef __OSMXTRLGPX_H__
#define __OSMXTRLGPX_H__

#include <msxml2.h>

#ifndef _VECTOR_
#pragma message("To avoid this message, please put vector in your pre compiled header (normally stdafx.h)")
#include <vector>
#endif //#ifndef _VECTOR_


//////////////////// Classes //////////////////////////////////////////////////

//GPX trackpoint and waypoint class
class CGPXTrackOrWayPoint
{
public:
#pragma warning(suppress: 26460)
  void Save(_In_ bool bIsWaypoint, _Inout_ ATL::CComPtr<IXMLDOMDocument2>& document, _Inout_ ATL::CComPtr<IXMLDOMNode>& waypointNode)
  {
    ATL::CComPtr<IXMLDOMNode> pointNode;
#pragma warning(suppress: 26485)
    HRESULT hr{document->createNode(ATL::CComVariant{NODE_ELEMENT}, ATL::CComBSTR{bIsWaypoint ? L"wpt" : L"trkpt"}, ATL::CComBSTR{L"http://www.topografix.com/GPX/1/1"}, &pointNode)};
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    if (bIsWaypoint)
      waypointNode = pointNode;

    ATL::CComPtr<IXMLDOMAttribute> latitudeAttribute;
    hr = document->createAttribute(ATL::CComBSTR{L"lat"}, &latitudeAttribute);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    CStringW sLatitude;
    sLatitude.Format(L"%f", m_fLatitude);
    hr = latitudeAttribute->put_text(ATL::CComBSTR{sLatitude});
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    ATL::CComPtr<IXMLDOMNamedNodeMap> attributes;
    hr = pointNode->get_attributes(&attributes);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    ATL::CComPtr<IXMLDOMNode> unusedNode1;
    hr = attributes->setNamedItem(latitudeAttribute, &unusedNode1);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));

    ATL::CComPtr<IXMLDOMAttribute> longitudeAttribute;
    hr = document->createAttribute(ATL::CComBSTR{L"lon"}, &longitudeAttribute);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    CStringW sLongitude;
    sLongitude.Format(L"%f", m_fLongitude);
    hr = longitudeAttribute->put_text(ATL::CComBSTR{sLongitude});
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    ATL::CComPtr<IXMLDOMNode> unusedNode2;
    hr = attributes->setNamedItem(longitudeAttribute, &unusedNode2);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));

    if (m_Time.wYear != 0)
    {
      ATL::CComPtr<IXMLDOMNode> timeNode;
      hr = document->createNode(ATL::CComVariant{NODE_ELEMENT}, ATL::CComBSTR{"time"}, ATL::CComBSTR{L"http://www.topografix.com/GPX/1/1"}, &timeNode);
      if (FAILED(hr))
        _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
      CStringW sTime;
      sTime.Format(L"%04d-%02d-%02dT%02d:%02d:%02dZ", static_cast<int>(m_Time.wYear), static_cast<int>(m_Time.wMonth), static_cast<int>(m_Time.wDay), 
                   static_cast<int>(m_Time.wHour), static_cast<int>(m_Time.wMinute), static_cast<int>(m_Time.wSecond));
      hr = timeNode->put_text(ATL::CComBSTR{sTime});
      if (FAILED(hr))
        _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
      ATL::CComPtr<IXMLDOMNode> unusedNode3;
      hr = pointNode->appendChild(timeNode, &unusedNode3);
      if (FAILED(hr))
        _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    }

    if (m_sName.GetLength())
    {
      ATL::CComPtr<IXMLDOMNode> nameNode;
      hr = document->createNode(ATL::CComVariant{NODE_ELEMENT}, ATL::CComBSTR{L"name"}, ATL::CComBSTR{L"http://www.topografix.com/GPX/1/1"}, &nameNode);
      if (FAILED(hr))
        _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
      hr = nameNode->put_text(ATL::CComBSTR{m_sName});
      if (FAILED(hr))
        _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
      ATL::CComPtr<IXMLDOMNode> unusedNode4;
      hr = pointNode->appendChild(nameNode, &unusedNode4);
      if (FAILED(hr))
        _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    }

    if (m_sComment.GetLength())
    {
      ATL::CComPtr<IXMLDOMNode> commentNode;
      hr = document->createNode(ATL::CComVariant{NODE_ELEMENT}, ATL::CComBSTR{L"cmt"}, ATL::CComBSTR{L"http://www.topografix.com/GPX/1/1"}, &commentNode);
      if (FAILED(hr))
        _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
      hr = commentNode->put_text(ATL::CComBSTR{m_sComment});
      if (FAILED(hr))
        _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
      ATL::CComPtr<IXMLDOMNode> unusedNode5;
      hr = pointNode->appendChild(commentNode, &unusedNode5);
      if (FAILED(hr))
        _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    }

    if (m_sLink.GetLength())
    {
      ATL::CComPtr<IXMLDOMNode> linkNode;
      hr = document->createNode(ATL::CComVariant{NODE_ELEMENT}, ATL::CComBSTR{L"link"}, ATL::CComBSTR{L"http://www.topografix.com/GPX/1/1"}, &linkNode);
      if (FAILED(hr))
        _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
      hr = linkNode->put_text(ATL::CComBSTR{m_sLink});
      if (FAILED(hr))
        _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
      ATL::CComPtr<IXMLDOMNode> unusedNode6;
      hr = pointNode->appendChild(linkNode, &unusedNode6);
      if (FAILED(hr))
        _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    }

    ATL::CComPtr<IXMLDOMNode> elevationNode;
    hr = document->createNode(ATL::CComVariant{NODE_ELEMENT}, ATL::CComBSTR{L"ele"}, ATL::CComBSTR{L"http://www.topografix.com/GPX/1/1"}, &elevationNode);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    CStringW sElevation;
    sElevation.Format(L"%f", m_fElevation);
    hr = elevationNode->put_text(ATL::CComBSTR{sElevation});
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    ATL::CComPtr<IXMLDOMNode> unusedNode7;
    hr = pointNode->appendChild(elevationNode, &unusedNode7);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));

    if (!bIsWaypoint)
    {
      ATL::CComPtr<IXMLDOMNode> unusedNode8;
      hr = waypointNode->appendChild(pointNode, &unusedNode8);
      if (FAILED(hr))
        _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    }
  }

#pragma warning(suppress: 26460)
  void Load(_Inout_ ATL::CComPtr<IXMLDOMDocument2>& document, _In_ ATL::CComPtr<IXMLDOMNode>& waypointNode)
  {
    ATL::CComPtr<IXMLDOMNamedNodeMap> attributes;
    HRESULT hr = waypointNode->get_attributes(&attributes);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    ATL::CComPtr<IXMLDOMNode> latitudeAttribute;
    hr = attributes->getNamedItem(ATL::CComBSTR{L"lat"}, &latitudeAttribute);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    if (latitudeAttribute == nullptr)
      _com_issue_errorex(E_FAIL, document, __uuidof(DOMDocument60));
    ATL::CComBSTR bstrLatitude;
    hr = latitudeAttribute->get_text(&bstrLatitude);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    CStringW sLatitude{bstrLatitude};
    m_fLatitude = _wtof(sLatitude);

    ATL::CComPtr<IXMLDOMNode> longitudeAttribute;
    hr = attributes->getNamedItem(ATL::CComBSTR{L"lon"}, &longitudeAttribute);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    if (longitudeAttribute == nullptr)
      _com_issue_errorex(E_FAIL, document, __uuidof(DOMDocument60));
    ATL::CComBSTR bstrLongitude;
    hr = longitudeAttribute->get_text(&bstrLongitude);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    CStringW sLongitude{bstrLongitude};
    m_fLongitude = _wtof(sLongitude);

    ATL::CComPtr<IXMLDOMNode> TimeNode;
    hr = waypointNode->selectSingleNode(ATL::CComBSTR{L"gpx11:time"}, &TimeNode);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    memset(&m_Time, 0, sizeof(m_Time));
    if (TimeNode != nullptr)
    {
      //Break the time down into its constituent parts. Note we support a very limited form of time parsing here
      ATL::CComBSTR bstrTime;
      hr = TimeNode->get_text(&bstrTime);
      if (FAILED(hr))
        _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
      int nStartToken{0};
      int nToken{0};
      CStringW sTime{bstrTime};
      CStringW sToken{sTime.Tokenize(L"-T:Z", nStartToken)};
      while (nStartToken != -1)
      {
        switch (nToken)
        {
          case 0:
          {
#pragma warning(suppress: 26472)
            m_Time.wYear = static_cast<WORD>(_wtoi(sToken));
            break;
          }
          case 1:
          {
#pragma warning(suppress: 26472)
            m_Time.wMonth = static_cast<WORD>(_wtoi(sToken));
            break;
          }
          case 2:
          {
#pragma warning(suppress: 26472)
            m_Time.wDay = static_cast<WORD>(_wtoi(sToken));
            break;
          }
          case 3:
          {
#pragma warning(suppress: 26472)
            m_Time.wHour = static_cast<WORD>(_wtoi(sToken));
            break;
          }
          case 4:
          {
#pragma warning(suppress: 26472)
            m_Time.wMinute = static_cast<WORD>(_wtoi(sToken));
            break;
          }
          case 5:
          {
#pragma warning(suppress: 26472)
            m_Time.wSecond = static_cast<WORD>(_wtoi(sToken));
            break;
          }
          case 6:
          {
#pragma warning(suppress: 26472)
            m_Time.wMilliseconds = static_cast<WORD>(_wtoi(sToken));
            break;
          }
          default:
          {
            break;
          }
        };

        //Prepare for the next time around
        ++nToken;
        sToken = sTime.Tokenize(L"-T:Z", nStartToken);
      }
    }

    ATL::CComPtr<IXMLDOMNode> NameNode;
    hr = waypointNode->selectSingleNode(ATL::CComBSTR{L"gpx11:name"}, &NameNode);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    m_sName.Empty();
    if (NameNode != nullptr)
    {
      ATL::CComBSTR bstrName;
      hr = NameNode->get_text(&bstrName);
      if (FAILED(hr))
        _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
      m_sName = bstrName;
    }

    ATL::CComPtr<IXMLDOMNode> CommentNode;
    hr = waypointNode->selectSingleNode(ATL::CComBSTR{L"gpx11:cmt"}, &CommentNode);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    m_sComment.Empty();
    if (CommentNode != nullptr)
    {
      ATL::CComBSTR bstrComment;
      hr = CommentNode->get_text(&bstrComment);
      if (FAILED(hr))
        _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
      m_sComment = bstrComment;
    }

    ATL::CComPtr<IXMLDOMNode> LinkNode;
    hr = waypointNode->selectSingleNode(ATL::CComBSTR{L"gpx11:link"}, &LinkNode);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    m_sLink.Empty();
    if (LinkNode != nullptr)
    {
      ATL::CComBSTR bstrLink;
      hr = LinkNode->get_text(&bstrLink);
      if (FAILED(hr))
        _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
      m_sLink = bstrLink;
    }

    ATL::CComPtr<IXMLDOMNode> ElevationNode;
    hr = waypointNode->selectSingleNode(ATL::CComBSTR{L"gpx11:ele"}, &ElevationNode);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    m_fElevation = 0;
    if (ElevationNode != nullptr)
    {
      ATL::CComBSTR bstrElevation;
      hr = ElevationNode->get_text(&bstrElevation);
      if (FAILED(hr))
        _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
      CStringW sElevation{bstrElevation};
      m_fElevation = _wtof(sElevation);
    }
  }

//Member variables
  double m_fLatitude{0};
  double m_fLongitude{0};
  double m_fElevation{0};
  double m_fCalculatedSpeed{0};
  SYSTEMTIME m_Time{};
  CString m_sName;
  CString m_sComment;
  CString m_sDescription;
  CString m_sLink;
};


//A GPX track segment
class CGPXTrackSegment
{
public:
  void Save(_In_ ATL::CComPtr<IXMLDOMDocument2>& document, _In_ ATL::CComPtr<IXMLDOMNode>& trackSegmentNode)
  {
    //Serialize each waypoint
    for (auto& trackpoint : m_TrackPoints)
      trackpoint.Save(false, document, trackSegmentNode);
  }

#pragma warning(suppress: 26460)
  void Load(_In_ ATL::CComPtr<IXMLDOMDocument2>& document, _In_ ATL::CComPtr<IXMLDOMNode>& trackSegmentNode)
  {
    //Get all the track segments
    ATL::CComPtr<IXMLDOMNodeList> trackPointNodes;
    HRESULT hr{trackSegmentNode->selectNodes(ATL::CComBSTR{L"gpx11:trkpt"}, &trackPointNodes)};
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    if (trackPointNodes == nullptr)
      _com_issue_errorex(E_FAIL, document, __uuidof(DOMDocument60));
    long nTrackPoints{0};
    hr = trackPointNodes->get_length(&nTrackPoints);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    m_TrackPoints.clear();
    m_TrackPoints.reserve(nTrackPoints);
    for (long i{0}; i<nTrackPoints; i++)
    {
      //Pull out the current tracksegment
      ATL::CComPtr<IXMLDOMNode> pointNode;
      hr = trackPointNodes->get_item(i, &pointNode);
      if (FAILED(hr))
        _com_issue_errorex(hr, document, __uuidof(DOMDocument60));

      //Load it up from the XML
      CGPXTrackOrWayPoint waypoint;
      waypoint.Load(document, pointNode);

      //Calculate the speed for the current waypoint if necessary
      if (i)
      {
#pragma warning(suppress: 26446)
        const auto& prevWaypoint{m_TrackPoints[static_cast<size_t>(i) - 1]};
        const COSMCtrlPosition pos1{prevWaypoint.m_fLongitude, prevWaypoint.m_fLatitude};
        const COSMCtrlPosition pos2{waypoint.m_fLongitude, waypoint.m_fLatitude};
        const double fDistance{COSMCtrlHelper::DistanceBetweenPoints(pos1, pos2, nullptr, nullptr)};
        if (waypoint.m_Time.wYear && prevWaypoint.m_Time.wYear)
        {
#pragma warning(suppress: 26486)
          const double fTimeSpan{(COleDateTime{waypoint.m_Time} - COleDateTime{prevWaypoint.m_Time}).GetTotalSeconds()};
          waypoint.m_fCalculatedSpeed = fDistance / fTimeSpan * 3.6;
        }
      }

      //And put it into the array
      m_TrackPoints.push_back(waypoint);
    }
  }

//Member variables
  std::vector<CGPXTrackOrWayPoint> m_TrackPoints;
};


//A GPX track
class CGPXTrack
{
public:
//Methods
#pragma warning(suppress: 26460)
  void Save(_In_ ATL::CComPtr<IXMLDOMDocument2>& document, _In_ ATL::CComPtr<IXMLDOMNode>& trackNode)
  {
    //Serialize each track segment
    for (auto& trackSegment : m_TrackSegments)
    {
      ATL::CComPtr<IXMLDOMNode> trackSegmentNode;
      HRESULT hr{document->createNode(ATL::CComVariant{NODE_ELEMENT}, ATL::CComBSTR{L"trkseg"}, ATL::CComBSTR{L"http://www.topografix.com/GPX/1/1"}, &trackSegmentNode)};
      if (FAILED(hr))
        _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
      trackSegment.Save(document, trackSegmentNode);
      ATL::CComPtr<IXMLDOMNode> unusedNode1;
      hr = trackNode->appendChild(trackSegmentNode, &unusedNode1);
      if (FAILED(hr))
        _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    }
  }

#pragma warning(suppress: 26460)
  void Load(_In_ ATL::CComPtr<IXMLDOMDocument2>& document, _In_ ATL::CComPtr<IXMLDOMNode>& trackNode)
  {
    //Get all the track segments
    ATL::CComPtr<IXMLDOMNodeList> trackSegmentNodes;
    HRESULT hr{trackNode->selectNodes(ATL::CComBSTR(L"gpx11:trkseg"), &trackSegmentNodes)};
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    if (trackSegmentNodes == nullptr)
      _com_issue_errorex(E_FAIL, document, __uuidof(DOMDocument60));
    long nTrackSegments{0};
    hr = trackSegmentNodes->get_length(&nTrackSegments);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    m_TrackSegments.clear();
    m_TrackSegments.reserve(nTrackSegments);
    for (long i{0}; i<nTrackSegments; i++)
    {
      //Pull out the current tracksegment
      ATL::CComPtr<IXMLDOMNode> trackSegmentNode;
      hr = trackSegmentNodes->get_item(i, &trackSegmentNode);
      if (FAILED(hr))
        _com_issue_errorex(hr, document, __uuidof(DOMDocument60));

      //Load it up from the XML
      CGPXTrackSegment trackSegment;
      trackSegment.Load(document, trackSegmentNode);

      //And put it into the array
#pragma warning(suppress: 26486)
      m_TrackSegments.push_back(trackSegment);
    }
  }

//Member variables
  std::vector<CGPXTrackSegment> m_TrackSegments;
};


//The main GPX class
class CGPX
{
public:
//Methods
  void Save(_In_ const CString& sFileName, _In_ const CString& sCreator)
  {
    //First create a DOM document
    ATL::CComPtr<IXMLDOMDocument2> document;
    HRESULT hr{document.CoCreateInstance(__uuidof(DOMDocument60), nullptr, CLSCTX_INPROC_SERVER)};
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));

    //Create the one and only root node in our XML
    ATL::CComPtr< IXMLDOMNode> gpxNode;
    hr = document->createNode(ATL::CComVariant{NODE_ELEMENT}, ATL::CComBSTR{L"gpx"}, ATL::CComBSTR{L"http://www.topografix.com/GPX/1/1"}, &gpxNode);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    ATL::CComPtr<IXMLDOMAttribute> xmlnssiAttribute;
    hr = document->createAttribute(ATL::CComBSTR{L"xmlns:xsi"}, &xmlnssiAttribute);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    hr = xmlnssiAttribute->put_text(ATL::CComBSTR{L"http://www.w3.org/2001/XMLSchema-instance"});
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    ATL::CComPtr<IXMLDOMNamedNodeMap> attributes;
    hr = gpxNode->get_attributes(&attributes);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    ATL::CComPtr<IXMLDOMNode> unusedNode1;
    hr = attributes->setNamedItem(xmlnssiAttribute, &unusedNode1);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    ATL::CComPtr<IXMLDOMAttribute> xsiAttribute;
    hr = document->createAttribute(ATL::CComBSTR{L"xsi:schemaLocation"}, &xsiAttribute);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    hr = xsiAttribute->put_text(ATL::CComBSTR{L"http://www.topografix.com/GPX/1/1 http://www.topografix.com/GPX/1/1/gpx.xsd"});
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    ATL::CComPtr<IXMLDOMNode> unusedNode2;
    hr = attributes->setNamedItem(xsiAttribute, &unusedNode2);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    ATL::CComPtr<IXMLDOMNode> unusedNode3;
    hr = document->appendChild(gpxNode, &unusedNode3);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    ATL::CComPtr<IXMLDOMProcessingInstruction> processingInstructions;
    hr = document->createProcessingInstruction(ATL::CComBSTR{L"xml"}, ATL::CComBSTR{L"version=\"1.0\""}, &processingInstructions);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    ATL::CComVariant firstChild{unusedNode3.p};
    ATL::CComPtr<IXMLDOMNode> unusedNode4;
    hr = document->insertBefore(processingInstructions, firstChild, &unusedNode4);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));

    //Create the version number
    ATL::CComPtr<IXMLDOMAttribute> versionAttribute;
    hr = document->createAttribute(ATL::CComBSTR{L"version"}, &versionAttribute);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    hr = versionAttribute->put_text(ATL::CComBSTR{L"1.1"});
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    ATL::CComPtr<IXMLDOMNode> unusedNode5;
    hr = attributes->setNamedItem(versionAttribute, &unusedNode5);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));

    //Create the creator value
    ATL::CComPtr<IXMLDOMAttribute> creatorAttribute;
    hr = document->createAttribute(ATL::CComBSTR{L"creator"}, &creatorAttribute);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    hr = creatorAttribute->put_text(ATL::CComBSTR{sCreator});
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    ATL::CComPtr<IXMLDOMNode> unusedNode6;
    hr = attributes->setNamedItem(creatorAttribute, &unusedNode6);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));

    //Serialize each waypoint
    for (auto& waypoint : m_Waypoints)
    {
      ATL::CComPtr<IXMLDOMNode> waypointNode;
      waypoint.Save(true, document, waypointNode);
      ATL::CComPtr<IXMLDOMNode> unusedNode7;
      hr = gpxNode->appendChild(waypointNode, &unusedNode7);
      if (FAILED(hr))
        _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    }

    //Serialize each track
    for (auto& track : m_Tracks)
    {
      ATL::CComPtr<IXMLDOMNode> trackNode;
      hr = document->createNode(ATL::CComVariant{NODE_ELEMENT}, ATL::CComBSTR{L"trk"}, ATL::CComBSTR{L"http://www.topografix.com/GPX/1/1"}, &trackNode);
      if (FAILED(hr))
        _com_issue_errorex(hr, document, __uuidof(DOMDocument60));

      if (sFileName.GetLength())
      {
        ATL::CComPtr<IXMLDOMNode> nameNode;
        hr = document->createNode(ATL::CComVariant{NODE_ELEMENT}, ATL::CComBSTR{L"name"}, ATL::CComBSTR{L"http://www.topografix.com/GPX/1/1"}, &nameNode);
        if (FAILED(hr))
          _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
        hr = nameNode->put_text(ATL::CComBSTR{sFileName});
        if (FAILED(hr))
          _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
        ATL::CComPtr<IXMLDOMNode> unusedNode8;
        hr = trackNode->appendChild(nameNode, &unusedNode8);
        if (FAILED(hr))
          _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
      }

      track.Save(document, trackNode);
      ATL::CComPtr<IXMLDOMNode> unusedNode9;
      hr = gpxNode->appendChild(trackNode, &unusedNode9);
      if (FAILED(hr))
        _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    }

    //And save the document
#pragma warning(suppress: 26486)
    hr = document->save(ATL::CComVariant{sFileName});
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
  }

  void Load(_In_ ATL::CComPtr<IXMLDOMDocument2>& document)
  {
    //Set the default namespace
    HRESULT hr{document->setProperty(ATL::CComBSTR{L"SelectionNamespaces"}, ATL::CComVariant{L"xmlns:gpx11='http://www.topografix.com/GPX/1/1'"})};
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));

    //Get all the tracks
    ATL::CComPtr<IXMLDOMNodeList> trackNodes;
    hr = document->selectNodes(ATL::CComBSTR{L"/gpx11:gpx/gpx11:trk"}, &trackNodes);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    if (trackNodes == nullptr)
      _com_issue_errorex(E_FAIL, document, __uuidof(DOMDocument60));
    long nTracks{0};
    hr = trackNodes->get_length(&nTracks);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    m_Tracks.clear();
    m_Tracks.reserve(nTracks);
    for (int i{0}; i<nTracks; i++)
    {
      //Pull out the current track
      ATL::CComPtr<IXMLDOMNode> trackNode;
      hr = trackNodes->get_item(i, &trackNode);
      if (FAILED(hr))
        _com_issue_errorex(hr, document, __uuidof(DOMDocument60));

      //Load it up from the XML
      CGPXTrack track;
      track.Load(document, trackNode);

      //And put it into the array
#pragma warning(suppress: 26486)
      m_Tracks.push_back(track);
    }

    //Get all the waypoints
    ATL::CComPtr<IXMLDOMNodeList> waypointNodes;
    hr = document->selectNodes(ATL::CComBSTR{L"/gpx11:gpx/gpx11:wpt"}, &waypointNodes);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    if (waypointNodes == nullptr)
      _com_issue_errorex(E_FAIL, document, __uuidof(DOMDocument60));
    long nWaypoints{0};
    hr = waypointNodes->get_length(&nWaypoints);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    m_Waypoints.clear();
    m_Waypoints.reserve(nWaypoints);
    for (int i{0}; i<nWaypoints; i++)
    {
      //Pull out the current waypoint
      ATL::CComPtr<IXMLDOMNode> waypointNode;
      hr = waypointNodes->get_item(i, &waypointNode);
      if (FAILED(hr))
        _com_issue_errorex(hr, document, __uuidof(DOMDocument60));

      //Load it up from the XML
      CGPXTrackOrWayPoint waypoint;
      waypoint.Load(document, waypointNode);

      //And put it into the array
      m_Waypoints.push_back(waypoint);
    }
  }

  void Load(_In_ const CString& sFileName)
  {
    //First create a DOM document
    ATL::CComPtr<IXMLDOMDocument2> document;
    HRESULT hr = document.CoCreateInstance(__uuidof(DOMDocument60), nullptr, CLSCTX_INPROC_SERVER);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));

    //Now load up the XML from disk
    ATL::CComVariant varSource{sFileName};
#pragma warning(suppress: 26486)
    hr = document->put_async(VARIANT_FALSE);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    VARIANT_BOOL bSuccess{VARIANT_FALSE};
    hr = document->load(varSource, &bSuccess);
    if (FAILED(hr))
      _com_issue_errorex(hr, document, __uuidof(DOMDocument60));
    if (bSuccess == VARIANT_FALSE)
      _com_raise_error(E_FAIL);

    //Delegate to the overload
    Load(document);
  }

//Member variables
  std::vector<CGPXTrack> m_Tracks; //The array of tracks in the GPX file
  std::vector<CGPXTrackOrWayPoint> m_Waypoints; //The array of waypoints in the GPX file
};


#endif //#define __OSMXTRLGPX_H__
