/*
Module : OSMCtrlHelper.h
Purpose: Defines the interface for a helper class for COSMCtrl
Created: PJN / 10-04-2011

Copyright (c) 2011 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Macros / Includes ////////////////////////////////////////

#pragma once

#ifndef __OSMCTRLHELPERS_H__
#define __OSMCTRLHELPERS_H__

#include "OSMCtrlPosition.h"


//////////////////// Classes //////////////////////////////////////////////////

class COSMCtrlHelper
{
public:
//Enums
  enum Misc
  {
    OSMMaxZoom = 18,
    OSMMinZoom = 0,
  };

//static methods
  static double Longitude2TileX(_In_ double Longitude, _In_ int nZoom) noexcept
  {
    return (Longitude + 180.0) / 360.0 * pow(2.0, nZoom); 
  }

  static double Latitude2TileY(_In_ double Latitude, _In_ int nZoom) noexcept
  {
    return (1.0 - log(tan(Latitude * M_PI/180.0) + 1.0 / cos(Latitude * M_PI/180.0)) / M_PI) / 2.0 * pow(2.0, nZoom); 
  }

  static double TileX2Longitude(_In_ double X, _In_ int nZoom) noexcept
  {
    return X / pow(2.0, nZoom) * 360.0 - 180;
  }

  static double TileY2Latitude(_In_ double Y, _In_ int nZoom) noexcept
  {
    const double n = M_PI - 2.0 * M_PI * Y / pow(2.0, nZoom);
    return 180.0 / M_PI * atan(0.5 * (exp(n) - exp(-n)));
  }

  //This is an implementation of Vincenty's Inverse algorithm (http://en.wikipedia.org/wiki/Vincenty's_formulae). It is based upon the C# 
  //implementation provided by Mike Gavaghan at http://www.gavaghan.org/blog/2007/08/06/c-gps-receivers-and-geocaching-vincentys-formula/
  static double DistanceBetweenPoints(_In_ const COSMCtrlPosition& position1, _In_ const COSMCtrlPosition& position2, _Out_opt_ double* pdStartBearing, _Out_opt_ double* pdReverseBearing) noexcept
  {
    //First thing to do is setup the constants for the WGS 84 ellipsoid
    constexpr const double a{6378137.0};
    constexpr const double b{6356752.314245};
    constexpr const double f{(a - b)/a};
    constexpr const double asquared{a * a};
    constexpr const double bsquared{b * b};
    constexpr const double asquaredminusbsquaredoverbsquared{(asquared - bsquared) / bsquared};

    //Convert all our parameters to radians
    const double phi1{position1.m_fLatitude / 180 * M_PI};
    const double phi2{position2.m_fLatitude / 180 * M_PI};
    const double lambda1{position1.m_fLongitude / 180 * M_PI};
    const double lambda2{position2.m_fLongitude / 180 * M_PI};

    const double omega{lambda2 - lambda1};
    const double tanphi1{tan(phi1)};
    const double tanU1{(1 - f) * tanphi1};
    const double U1{atan(tanU1)};
    const double sinU1{sin(U1)};
    const double cosU1{cos(U1)};

    const double tanphi2{tan(phi2)};
    const double tanU2{(1 - f) * tanphi2};
    const double U2{atan(tanU2)};
    const double sinU2{sin(U2)};
    const double cosU2{cos(U2)};

    const double sinU1sinU2{sinU1 * sinU2};
    const double cosU1sinU2{cosU1 * sinU2};
    const double sinU1cosU2{sinU1 * cosU2};
    const double cosU1cosU2{cosU1 * cosU2};

    double lambda{omega}; //Equation 13

    bool bConverged{false};
    bool bContinueIteration{true};
    double A{0};
    double sigma{0};
    double deltasigma{0};
    while (bContinueIteration)
    {
      //Remember the previous value
      const double lambda0{lambda};

      const double sinlambda{sin(lambda)};
      const double coslambda{cos(lambda)};

      //Equation 14
      double temp{cosU1sinU2 - sinU1cosU2 * coslambda};
      temp *= temp;
      const double sinsquaredsigma{(cosU2 * sinlambda * cosU2 * sinlambda) + temp};
      const double sinsigma{sqrt(sinsquaredsigma)};

      const double cossigma{sinU1sinU2 + (cosU1cosU2 * coslambda)}; //Equation 15

      sigma = atan2(sinsigma, cossigma); //Equation 16

      const double sinalpha{(sinsquaredsigma == 0) ? 0 : cosU1cosU2 * sinlambda / sinsigma}; //Equation 17

      const double alpha{asin(sinalpha)};
      const double cosalpha{cos(alpha)};
      const double cossquaredalpha{cosalpha * cosalpha};

      const double cos2sigmam{(cossquaredalpha == 0) ? 0 : cossigma - 2 * sinU1sinU2 / cossquaredalpha}; //Equation 18
      const double cossquared2sigmam{cos2sigmam * cos2sigmam};

      const double usquared{cossquaredalpha * asquaredminusbsquaredoverbsquared};

      A = 1 + (usquared / 16384) * (4096 + usquared * (-768 + usquared * (320 - 175 * usquared))); //Equation 3

      const double B{usquared / 1024 * (256 + usquared * (-128 + usquared * (74 - 47 *usquared)))}; //Equation 4

      deltasigma = B * sinsigma * (cos2sigmam + (B / 4) * (cossigma * (-1 + 2 * cossquared2sigmam) - (B / 6 * cos2sigmam * (-3 + 4 * sinsquaredsigma) * (-3 + 4 * cossquared2sigmam)))); //Equation 6

      const double C{f / 16 * cossquaredalpha * (4 + f * (4 - 3 * cossquaredalpha))}; //Equation 10

      lambda = omega + (1 - C) * f * sinalpha * (sigma + C * sinsigma * (cos2sigmam + C * cossigma * (-1 + 2 * cossquared2sigmam))); //Equation 11 (modified)

      if (abs(lambda) > M_PI)
        bContinueIteration = false;
      else
      {
        bConverged = abs(lambda - lambda0) < 0.0000000000001;
        if (bConverged)
          bContinueIteration = false;
      }
    }

    const double s{b * A * (sigma - deltasigma)}; //Equation 19 (return value will be in meters)

    if (!bConverged)
    {
      //The points must be anti-podal, in this case we actually have an infinite number of great circles
      if (phi1 > phi2)
      {
        //The great circle returned will by definition be a southerly bearing from position1 to position2
        if (pdStartBearing != nullptr)
          *pdStartBearing = 180;
        if (pdReverseBearing != nullptr)
          *pdReverseBearing = 0;
      }
      else if (phi1 < phi2)
      {
        //The great circle returned will by definition be a northerly bearing from position1 to position2
        if (pdStartBearing != nullptr)
          *pdStartBearing = 0;
        if (pdReverseBearing != nullptr)
          *pdReverseBearing = 180;
      }
      else if (lambda1 > lambda2)
      {
        //The great circle returned will by definition be a easterly bearing from position1 to position2
        if (pdStartBearing != nullptr)
          *pdStartBearing = 90;
        if (pdReverseBearing != nullptr)
          *pdReverseBearing = 270;
      }
      else
      {
        //The great circle returned will by defintion be a westerly bearing from position1 to position2
        if (pdStartBearing != nullptr)
          *pdStartBearing = 270;
        if (pdReverseBearing != nullptr)
          *pdReverseBearing = 90;
      }
      return s;
    }

    const double sinlambda{sin(lambda)};
    const double coslambda{cos(lambda)};

    if (pdStartBearing != nullptr)
    {
      *pdStartBearing = atan2(cosU2 * sinlambda, cosU1sinU2 - sinU1cosU2 * coslambda) * 180 / M_PI; //Equation 20

      //Ensure bearing is in the conventional range
      if (*pdStartBearing < 0)
        *pdStartBearing += 360;
      if (*pdStartBearing >= 360)
        *pdStartBearing -= 360;
    }

    if (pdReverseBearing != nullptr)
    {
      *pdReverseBearing = (atan2(cosU1 * sinlambda, -sinU1cosU2 + cosU1sinU2 * coslambda) + M_PI) * 180 / M_PI; //Equation 21

      if (*pdReverseBearing < 0)
        *pdReverseBearing += 360;
      if (*pdReverseBearing >= 360)
        *pdReverseBearing -= 360;
    }

    return s;
  }

  //This is an implementation of Vincenty's Direct algorithm (http://en.wikipedia.org/wiki/Vincenty's_formulae). It is based upon the C# 
  //implementation provided by Mike Gavaghan at http://www.gavaghan.org/blog/2007/08/06/c-gps-receivers-and-geocaching-vincentys-formula/
  static COSMCtrlPosition GetPosition(_In_ const COSMCtrlPosition& position, _In_ double dBearing, _In_ double dDistance, _Out_opt_ double* pdEndBearing) noexcept
  {
    //First thing to do is setup the constants for the WGS 84 ellipsoid
    constexpr const double a{6378137.0};
    constexpr const double b{6356752.314245};
    constexpr const double f{(a - b)/a};
    constexpr const double asquared{a * a};
    constexpr const double bsquared{b * b};

    //Convert all our parameters to radians
    const double alpha1{dBearing / 180 * M_PI};
    const double cosalpha1{cos(alpha1)};
    const double sinalpha1{sin(alpha1)};
    const double phi1{position.m_fLatitude / 180 * M_PI};

    //Start out with the approximation
    const double s{dDistance};

    const double tanU1{(1 - f) * tan(phi1)};
    const double cosU1{1 / sqrt(1 + tanU1 * tanU1)};
    const double sinU1{tanU1 * cosU1};

    const double sigma1{atan2(tanU1, cosalpha1)}; //Equation 1
    const double sinalpha{cosU1 * sinalpha1}; //Equation 2

    const double sinsquaredalpha{sinalpha * sinalpha};
    const double cossquaredalpha{1 - sinsquaredalpha};
    const double usquared{cossquaredalpha * (asquared - bsquared) / bsquared};

    const double A{1 + (usquared / 16384) * (4096 + usquared * (-768 + usquared * (320 - 175 * usquared)))}; //Equation 3
    const double B{usquared / 1024 * (256 + usquared * (-128 + usquared * (74 - 47 *usquared)))}; //Equation 4

    //Now do the iteration until there is a negligible change in sigma
    bool bContinueIteration{true};
    const double soverbA{s / (b * A)};
    double sigma{soverbA};
    double previoussigma{sigma};
    double cossquaredsigma2M{0};
    double cossigma2M{0};
    while (bContinueIteration)
    {
      const double sigma2M{2 * sigma1 + sigma}; //Equation 5
      cossigma2M = cos(sigma2M);
      cossquaredsigma2M = cossigma2M * cossigma2M;
      const double sinsigma{sin(sigma)};
      const double sinsquaredsigma{sinsigma * sinsigma};
      const double cossigma{cos(sigma)};

      const double deltasigma{B * sinsigma * (cossigma2M + (B / 4) * (cossigma * (-1 + 2 * cossquaredsigma2M) - B / 6 * cossigma2M * (-3 + 4 * sinsquaredsigma) * (-3 + 4 * cossquaredsigma2M)))}; //Equation 6

      sigma = soverbA + deltasigma; //Equation 7

      //Have we converged yet
      if (abs(sigma - previoussigma) < 0.0000000000001)
        bContinueIteration = false;
      else
      {
        //Remember the previous value for the next time in the iteration
        previoussigma = sigma;
      }
    }

    const double sinsigma{sin(sigma)};
    const double cossigma{cos(sigma)};
    double temp{(sinU1 * sinsigma) - (cosU1 * cossigma * cosalpha1)};
    temp *= temp;
    const double phi2{atan2((sinU1 * cossigma) + (cosU1 * sinsigma * cosalpha1), (1 - f) * sqrt(sinsquaredalpha + temp))}; //Equation 8

    const double lambda{atan2(sinsigma * sinalpha1, (cosU1 * cossigma) - (sinU1 * sinsigma * cosalpha1))}; //Equation 9

    const double C{f / 16 * cossquaredalpha * (4 + f * (4 - 3 * cossquaredalpha))}; //Equation 10

    const double L{lambda - (1 - C) * f * sinalpha * (sigma + C * sinsigma * (cossigma2M + C * cossigma * (-1 + 2 * cossquaredsigma2M)))}; //Equation 11

    double alpha2{atan2(sinalpha, (-sinU1 * sinsigma) + (cosU1 * cossigma * cosalpha1))}; //Equation 12

    //Setup the return values from this function
    COSMCtrlPosition finalPosition;
    finalPosition.m_fLatitude = phi2 * 180 / M_PI;
    finalPosition.m_fLongitude = position.m_fLongitude + (L * 180 / M_PI);

    //Ensure longitude is in the conventional range
    finalPosition.NormalizeLongitude();

    if (pdEndBearing != nullptr)
      *pdEndBearing = alpha2 * 180 / M_PI;

    return finalPosition;
  }
};


#endif //#ifndef __OSMCTRLHELPERS_H__
