/*
Module : OSMCtrlMapOperations.cpp
Purpose: Implementation for the various COSMCtrl map operations classes
Created: PJN / 10-04-2011
History: PJN / None

Copyright (c) 2011 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Includes /////////////////////////////////////////////////

#include "stdafx.h"
#include "OSMCtrlMapOperations.h"
#include "OSMCtrl.h"


//////////////////// Implementation ///////////////////////////////////////////

#pragma warning(suppress: 26426)
UINT COSMCtrlMapOperationsDlg::sm_nMessageAddEvent = RegisterWindowMessage(_T("COSMCtrlMapOperations_AddEvent"));


#pragma warning(suppress: 26433 26440)
BEGIN_MESSAGE_MAP(COSMCtrlMapOperationsDlg, CDialog) //NOLINT(modernize-avoid-c-arrays)
  ON_BN_CLICKED(ID_START_OPERATION, &COSMCtrlMapOperationsDlg::OnBnClickedStartOperation)
  ON_BN_CLICKED(ID_STOP_OPERATION, &COSMCtrlMapOperationsDlg::OnBnClickedStopOperation)
  ON_WM_HSCROLL()
  ON_WM_DESTROY()
  ON_REGISTERED_MESSAGE(sm_nMessageAddEvent, &COSMCtrlMapOperationsDlg::OnAddEvent)
END_MESSAGE_MAP()


#pragma warning(suppress: 26455)
COSMCtrlMapOperationsDlg::COSMCtrlMapOperationsDlg(_In_opt_ CWnd* pParent): CDialog{COSMCtrlMapOperationsDlg::IDD, pParent}, 
                                                                            m_nOperation{static_cast<int>(Operation::CacheTilesOnlyThoseNotDownloaded)},
                                                                            m_pOSMCtrl{nullptr},
                                                                            m_pWorkerThread{nullptr},
                                                                            m_hSession{nullptr}
{
}

void COSMCtrlMapOperationsDlg::DoDataExchange(CDataExchange* pDX)
{
  __super::DoDataExchange(pDX);
  DDX_Control(pDX, ID_START_OPERATION, m_ctrlStartOperation);
  DDX_Control(pDX, ID_STOP_OPERATION, m_ctrlStopOperation);
  DDX_Control(pDX, IDC_OPERATION, m_ctrlOperation);
  DDX_Control(pDX, IDC_ZOOM_LEVEL, m_ctrlZoomLevel);
  DDX_CBIndex(pDX, IDC_OPERATION, m_nOperation);
  DDX_Control(pDX, IDC_STATUS_TEXT, m_ctrlStatusText);
  DDX_Control(pDX, IDC_RESOLUTION_TEXT, m_ctrlResolutionText);
  DDX_Control(pDX, IDC_PROGRESS, m_ctrlProgress);
}

void COSMCtrlMapOperationsDlg::OnBnClickedStartOperation()
{
  if (UpdateData(TRUE))
  {
    //First thing to do is disable this button and enable the stop operation button
    m_ctrlStartOperation.EnableWindow(FALSE);
    m_ctrlStopOperation.EnableWindow(TRUE);

    //Set focus to the Stop button
    m_ctrlStopOperation.SetFocus();

    //Reset the progress control
    m_ctrlProgress.SetPos(0);

    //Disable the zoom level control
    m_ctrlZoomLevel.EnableWindow(FALSE);

    //Empty out the array
    m_PendingEvents.clear();

    //Create the worker thread
    CreateWorkerThread();
  }
}

void COSMCtrlMapOperationsDlg::OnBnClickedStopOperation()
{
  //First thing to do is disable this button and enable the start operation button
  m_ctrlStopOperation.EnableWindow(FALSE);
  m_ctrlStartOperation.EnableWindow(TRUE);

  //Set focus back to the Start button
  m_ctrlStartOperation.SetFocus();

  //Shutdown the worker thread
  DestroyWorkerThread();
}

BOOL COSMCtrlMapOperationsDlg::OnInitDialog()
{
  //Let the base class do its thing
  __super::OnInitDialog();

  //Validate our parameters
  ASSERT(m_pOSMCtrl != nullptr);

  //Setup the various controls
  const int nZoom{static_cast<int>(m_pOSMCtrl->GetZoom())};
  m_ctrlZoomLevel.SetRange(nZoom, min(COSMCtrl::OSMMaxZoom, nZoom+5), TRUE); //Only allow iteration up to 5 levels of zoom
  m_ctrlZoomLevel.EnableWindow(nZoom != COSMCtrl::OSMMaxZoom);
  m_ctrlStopOperation.EnableWindow(FALSE);
  UpdateZoom(nZoom);

  return TRUE;
}

void COSMCtrlMapOperationsDlg::UpdateZoom(_In_ int nZoomLevel)
{
  //Validate our parameters
#pragma warning(suppress: 26496)
  AFXASSUME(m_pOSMCtrl != nullptr);

  //Display a wait cursor while we do the maths to work out how many tiles we will be working on
  CWaitCursor waitCursor;

  //Reset the tiles array
  m_Tiles.clear();
  m_Tiles.reserve(100000);

  //Work out how many tiles this operation will affect
  COSMCtrlPosition topLeft;
  COSMCtrlPosition bottomRight;
  const int nZoom{static_cast<int>(m_pOSMCtrl->GetZoom())};
  if (!m_pOSMCtrl->m_SelectionPolygon->GetBoundingRect(topLeft, bottomRight))
    return;
  //Iterate across all the zoom levels we will be operating on
  for (int i{nZoom}; i<=nZoomLevel; i++)
  {
    //Work our the start and end X and Y values for the tiles at this zoom level
    const double fStartX{COSMCtrlHelper::Longitude2TileX(topLeft.m_fLongitude, i)};
    const int nStartX{static_cast<int>(fStartX)};
    const double fStartY{COSMCtrlHelper::Latitude2TileY(topLeft.m_fLatitude, i)};
    const int nStartY{static_cast<int>(fStartY)};

    const double fEndX{COSMCtrlHelper::Longitude2TileX(bottomRight.m_fLongitude, i)};
    const int nEndX{static_cast<int>(fEndX)};
    const double fEndY{COSMCtrlHelper::Latitude2TileY(bottomRight.m_fLatitude, i)};
    const int nEndY{static_cast<int>(fEndY)};

    //Add all the tiles for this zoom level to the array
    for (int j{nStartX}; j<=nEndX; j++)
    {
      for (int k{nStartY}; k<=nEndY; k++)
      {
        COSMCtrlMapOperationsDlgTile tile;
        tile.m_nTileX = j;
        tile.m_nTileY = k;
        tile.m_nZoom = i;
        m_Tiles.push_back(tile);
      }
    }
  }
  
  //Free up any unused memory in the array
  m_Tiles.shrink_to_fit();

  //Set the range on the progress control
#pragma warning(suppress: 26472)
  const int nTiles{static_cast<int>(m_Tiles.size())};
  m_ctrlProgress.SetPos(0);
  m_ctrlProgress.SetRange32(0, nTiles);

  //Finally update the text on the UI
  CString sStatus;
  CString sZoom1;
  sZoom1.Format(_T("%d"), nZoom);
  CString sTiles;
  sTiles.Format(_T("%d"), nTiles);
  if (nZoom == nZoomLevel)
    AfxFormatString2(sStatus, IDS_OSMCTRL_MAP_OPERATIONS_UPDATE_ZOOM2, sZoom1, sTiles);
  else
  {
    CString sZoom2;
    sZoom2.Format(_T("%d"), nZoomLevel);

    std::array<LPCTSTR, 3> szStrings{ sZoom1, sZoom2, sTiles };
#pragma warning(suppress: 26472)
    AfxFormatStrings(sStatus, IDS_OSMCTRL_MAP_OPERATIONS_UPDATE_ZOOM3, szStrings.data(), static_cast<int>(szStrings.size()));
  }
  m_ctrlStatusText.SetWindowText(sStatus);

  //Calculate the size of a pixel for the current zoom level (using the center of the bounding rect)
  const double fLatitude{0.5 * (topLeft.m_fLatitude + bottomRight.m_fLatitude)};
  const double fLongitude{0.5 * (topLeft.m_fLongitude + bottomRight.m_fLongitude)};
  double fX{COSMCtrlHelper::Longitude2TileX(fLongitude, nZoomLevel)};
  fX += (nZoomLevel == 0 ? 0.5 : 1);
  const double fLongitude2{COSMCtrlHelper::TileX2Longitude(fX, nZoomLevel)};
  double fDistanceWide{COSMCtrlHelper::DistanceBetweenPoints(COSMCtrlPosition(fLongitude, fLatitude), COSMCtrlPosition(fLongitude2, fLatitude), nullptr, nullptr) / COSMCtrl::OSMTileWidth};
  if (nZoomLevel == 0)
    fDistanceWide *= 2;
  double fY{COSMCtrlHelper::Latitude2TileY(fLatitude, nZoomLevel)};
  fY += (nZoomLevel == 0 ? 0.5 : 1);
  const double fLatitude2{COSMCtrlHelper::TileY2Latitude(fY, nZoomLevel)};
  const double fDistanceHigh{COSMCtrlHelper::DistanceBetweenPoints(COSMCtrlPosition(fLongitude, fLatitude), COSMCtrlPosition(fLongitude, fLatitude2), nullptr, nullptr) / COSMCtrl::OSMTileHeight};
  if (nZoomLevel == 0)
    fDistanceWide *= 2;
  CString sResolutionWide;
  sResolutionWide.Format(_T("%.2f"), fDistanceWide);
  CString sResolutionHigh;
  sResolutionHigh.Format(_T("%.2f"), fDistanceHigh);
  AfxFormatString2(sStatus, IDS_OSMCTRL_MAP_OPERATIONS_UPDATE_ZOOM1, sResolutionWide, sResolutionHigh);
  m_ctrlResolutionText.SetWindowText(sStatus);
}

#pragma warning(suppress: 26434)
void COSMCtrlMapOperationsDlg::OnHScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar)
{
  //Validate our parameters
  ASSERT(m_pOSMCtrl != nullptr);

  //Let the base class do its thing
  __super::OnHScroll(nSBCode, nPos, pScrollBar);

  //Handle the notification if it is from our control
#pragma warning(suppress: 26496)
  AFXASSUME(pScrollBar != nullptr);
  if (pScrollBar->GetSafeHwnd() == m_ctrlZoomLevel.GetSafeHwnd())
  {
    switch (nSBCode)
    {
      case TB_BOTTOM:
      {
        UpdateZoom(min(COSMCtrl::OSMMaxZoom, static_cast<int>(m_pOSMCtrl->GetZoom()) + 5));
        break;
      }
      case TB_TOP:
      {
        UpdateZoom(static_cast<int>(m_pOSMCtrl->GetZoom()));
        break;
      }
      case TB_LINEUP: //Deliberate fallthrough
      case TB_PAGEUP:
      {
        UpdateZoom(m_ctrlZoomLevel.GetPos());
        break;
      }
      case TB_LINEDOWN: //Deliberate fallthrough
      case TB_PAGEDOWN:
      {
        UpdateZoom(m_ctrlZoomLevel.GetPos());
        break;
      }
      case TB_THUMBTRACK: //deliberate fallthrough
      case TB_THUMBPOSITION:
      {
        UpdateZoom(nPos);
        break;
      }
      default:
      {
        //Nothing to do
        break;
      }
    }
  }
}

bool COSMCtrlMapOperationsDlg::CreateWorkerThread()
{
  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

  //Destroy the worker thread if currently running
  DestroyWorkerThread();

  //Spin off the background thread
  ASSERT(m_pWorkerThread == nullptr);
  m_WorkerTerminateEvent.ResetEvent();
  m_pWorkerThread = AfxBeginThread(_WorkerThread, this, THREAD_PRIORITY_IDLE, 0, CREATE_SUSPENDED);
  if (m_pWorkerThread)
  {
    //We're in charge of deletion of the thread
    m_pWorkerThread->m_bAutoDelete = FALSE;

    //Resume the thread now that everything is ready to go
    m_pWorkerThread->ResumeThread();

    bSuccess = true;
  }

  return bSuccess;
}

void COSMCtrlMapOperationsDlg::DestroyWorkerThread()
{
  //Ensure the download thread is shutdown
  if (m_pWorkerThread != nullptr)
  {
  #ifdef _DEBUG
    //If we are running a debug version of the code, display a wait cursor to visually indicate
    //how long the destruction of the worker thread takes
    CWaitCursor wait;
  #endif //#ifdef _DEBUG

    //Signal the worker thread to exit and wait for it to return
    m_WorkerTerminateEvent.SetEvent();
  
    //Close the session handle which is being used by the worker thread.
    WinHttpCloseHandle(m_hSession);
 
    //Wait for the thread to return
    WaitForSingleObject(m_pWorkerThread->m_hThread, INFINITE);
    delete m_pWorkerThread;
    m_pWorkerThread = nullptr;
  }
}

#pragma warning(suppress: 26434)
void COSMCtrlMapOperationsDlg::OnDestroy()
{
  //kill the worker thread if it is currently running
  DestroyWorkerThread();

  //Let the base class do its thing
  __super::OnDestroy();
}

UINT COSMCtrlMapOperationsDlg::_WorkerThread(LPVOID pParam)
{
  //Validate our parameters
  ASSERT(pParam != nullptr);
  auto pThis{static_cast<COSMCtrlMapOperationsDlg*>(pParam)};
#pragma warning(suppress: 26496)
  AFXASSUME(pThis != nullptr);

  //Convert from the SDK world to the C++ world
  return pThis->WorkerThread();
}

bool COSMCtrlMapOperationsDlg::DeleteCachedTilesHelper()
{
  //What will be the return value from this function (assume the best)
  bool bSuccess{true};

  //Accumulate how many tiles we have deleted and not deleted
  int nTilesDeleted{0};
  int nTilesNotDeleted{0};

  //Work thro all the tiles
  CString sCacheDirectory(m_pOSMCtrl->GetCacheDirectory());
  for (std::vector<COSMCtrlMapOperationsDlgTile>::size_type i{0}; (i<m_Tiles.size()) && bSuccess; i++)
  {
    //Pull out the tile we are working on
#pragma warning(suppress: 26446)
    const COSMCtrlMapOperationsDlgTile& tile{m_Tiles[i]};

    //Do the deletion of the tile
    CString sTile{m_pOSMCtrl->GetTileCachePath(sCacheDirectory, tile.m_nZoom, tile.m_nTileX, tile.m_nTileY)};
    COSMCtrlMapOperationsDlgEvent dlgEvent;
    if (GetFileAttributes(sTile) != INVALID_FILE_ATTRIBUTES) //Don't bother doing anything if the tile does not already exist
    {
      if (DeleteFile(sTile))
      {
        //Update the stats
        ++nTilesDeleted;

        dlgEvent.m_bSuccess = true;
      }
      else
      {
        //Update the stats
        ++nTilesNotDeleted;

        dlgEvent.m_bSuccess = false;
      }
    }

    //Update the UI
    dlgEvent.m_Event = COSMCtrlMapOperationsDlgEvent::Event::SimpleStringStatus;
    dlgEvent.m_sString = sTile;
    dlgEvent.m_nItemData = i + 1;
    AddEvent(dlgEvent);

    //Check if we have been cancelled before we loop around
    bSuccess = (WaitForSingleObject(m_WorkerTerminateEvent, 0) == WAIT_TIMEOUT);
  }

  //Finally add a event about how many items have been deleted
  COSMCtrlMapOperationsDlgEvent dlgEvent;
  dlgEvent.m_Event = COSMCtrlMapOperationsDlgEvent::Event::SimpleStringStatus;
  CString sTilesDeleted;
  sTilesDeleted.Format(_T("%d"), nTilesDeleted);
  CString sTilesNotDeleted;
  sTilesNotDeleted.Format(_T("%d"), nTilesNotDeleted);
  AfxFormatString2(dlgEvent.m_sString, IDS_OSMCTRL_DELETE_FILES_STATS, sTilesDeleted, sTilesNotDeleted);
  AddEvent(dlgEvent);

  return bSuccess;
}

bool COSMCtrlMapOperationsDlg::DownloadTiles(_In_ bool bSkipIfTileAlreadyExists)
{
  //Validate our parameters
#pragma warning(suppress: 26496)
  AFXASSUME(m_pOSMCtrl != nullptr);

  //Pull out the tile provider we will be using
  IOSMCtrlTileProvider* pTileProvider{m_pOSMCtrl->GetTileProvider()};
#pragma warning(suppress: 26496)
  AFXASSUME(pTileProvider != nullptr);

  //Next get the server to connect to
  CStringW sServer{pTileProvider->GetDownloadServer()};

  //Accumulate how many tiles we have deleted and not deleted
  int nTilesDownloaded{0};
  int nTilesNotDownloaded{0};

  //Next create the WinHTTP session object
  WinHTTPWrappers::CSession session;
  HRESULT hr = m_pOSMCtrl->CreateSession(session, 0);
  if (SUCCEEDED(hr))
  {
    //Now create the connection object from the session object
    WinHTTPWrappers::CConnection connection;
    hr = connection.Initialize(session, sServer, pTileProvider->GetDownloadPort());
    if (SUCCEEDED(hr))
    {
      //Iterate across the array of tiles to download
      bool bSuccess{true};
      for (std::vector<COSMCtrlMapOperationsDlgTile>::size_type i{0}; (i<m_Tiles.size()) && bSuccess; i++)
      {
        //Pull out the next tile to download
#pragma warning(suppress: 26446)
        const COSMCtrlMapOperationsDlgTile& tile{m_Tiles[i]};

        //Create the sub directories if we can
        CString sCacheDirectory{m_pOSMCtrl->GetCacheDirectory()};
        CString sSubDirectory;
        sSubDirectory.Format(_T("%s\\%d"), sCacheDirectory.GetString(), tile.m_nZoom);
        CreateDirectory(sSubDirectory, nullptr);
        sSubDirectory.Format(_T("%s\\%d\\%d"), sCacheDirectory.GetString(), tile.m_nZoom, tile.m_nTileX);
        CreateDirectory(sSubDirectory, nullptr);

        //Form the path to the tile we will be downloading to and determine if we should do the download
        CString sFile{COSMCtrl::GetTileCachePath(sCacheDirectory, tile.m_nZoom, tile.m_nTileX, tile.m_nTileY)};
        bool bDownload{true};
        if (bSkipIfTileAlreadyExists)
          bDownload = (GetFileAttributes(sFile) == INVALID_FILE_ATTRIBUTES);

        //Now download the specific tile to the cache if required
        COSMCtrlMapOperationsDlgEvent dlgEvent;
        dlgEvent.m_bSuccess = false;
        if (bDownload)
        {
          //We will accept any mime type
          std::array<LPCWSTR, 2> pwszAcceptTypes{ L"*/*", nullptr };
          WinHTTPWrappers::CSyncDownloader winHttpRequest;
#pragma warning(suppress: 26489)
          winHttpRequest.m_sFileToDownloadInto = CStringW{sFile};
          CString sObject{pTileProvider->GetDownloadObject(tile.m_nZoom, tile.m_nTileX, tile.m_nTileY)};
          hr = winHttpRequest.Initialize(connection, CStringW{sObject}, nullptr, nullptr, nullptr, pwszAcceptTypes.data(), WINHTTP_FLAG_REFRESH);
          if (FAILED(hr))
          {
            //report the error
            TRACE(_T("COSMCtrlMapOperationsDlg::DownloadTiles, Failed to create request for tile \"%s\", Error:%08X\n"), sFile.GetString(), hr);

            //Ensure any remnants of a bad download file are nuked
            DeleteFile(sFile);

            //Update the stats
            ++nTilesNotDownloaded;
          }
          else
          {
            hr = winHttpRequest.SendRequestSync();
            if (FAILED(hr))
            {
              //report the error
              TRACE(_T("COSMCtrlMapOperationsDlg::DownloadTiles, Failed to send request for tile \"%s\", Error:%08X\n"), sFile.GetString(), hr);
            
              //Ensure any remnants of a bad download file are nuked
              DeleteFile(sFile);
            
              //Update the stats
              ++nTilesNotDownloaded;
            }
            else
            {
              //Update the stats
              ++nTilesDownloaded;
              dlgEvent.m_bSuccess = true;
            }
          }
        }
        else
        {
          //Update the stats
          ++nTilesNotDownloaded;
        }

        //Update the UI
        dlgEvent.m_Event = COSMCtrlMapOperationsDlgEvent::Event::SimpleStringStatus;
        dlgEvent.m_sString = sFile;
        dlgEvent.m_nItemData = i + 1;
        AddEvent(dlgEvent);

        //Check if we have been cancelled before we loop around
        bSuccess = (WaitForSingleObject(m_WorkerTerminateEvent, 0) == WAIT_TIMEOUT);
      }
    }
  }

  //Finally add a event about how many items have been downloaded
  COSMCtrlMapOperationsDlgEvent dlgEvent;
  dlgEvent.m_Event = COSMCtrlMapOperationsDlgEvent::Event::SimpleStringStatus;
  CString sTilesDownloaded;
  sTilesDownloaded.Format(_T("%d"), nTilesDownloaded);
  CString sTilesNotDownloaded;
  sTilesNotDownloaded.Format(_T("%d"), nTilesNotDownloaded);
  AfxFormatString2(dlgEvent.m_sString, IDS_OSMCTRL_DOWNLOAD_TILES_STATS, sTilesDownloaded, sTilesNotDownloaded);
  AddEvent(dlgEvent);

  return true;
}

bool COSMCtrlMapOperationsDlg::ForceMapnikRerenderHelper()
{
  //Validate our parameters
#pragma warning(suppress: 26496)
  AFXASSUME(m_pOSMCtrl != nullptr);

  //Next get the server to connect to
  COSMCtrlMapnikTileProvider MapnikTileProvider;
  CStringW sServer{MapnikTileProvider.GetDownloadServer()};

  //Accumulate how many tiles we have request to rerender and which ones indicated a failure to rerender
  int nTilesRerendered{0};
  int nTilesNotRerendered{0};

  //Next create the WinHTTP session object
  WinHTTPWrappers::CSession session;
  HRESULT hr = m_pOSMCtrl->CreateSession(session, 0);
  if (SUCCEEDED(hr))
  {
    //Now create the connection object from the session object
    WinHTTPWrappers::CConnection connection;
    hr = connection.Initialize(session, sServer, MapnikTileProvider.GetDownloadPort());
    if (SUCCEEDED(hr))
    {
      //Iterate across the array of tiles to download
      bool bSuccess{true};
      for (std::vector<COSMCtrlMapOperationsDlgTile>::size_type i{0}; (i<m_Tiles.size()) && bSuccess; i++)
      {
        //Pull out the next tile to download
#pragma warning(suppress: 26446)
        const COSMCtrlMapOperationsDlgTile& tile{m_Tiles[i]};

        //Now issue the request to rerender
        COSMCtrlMapOperationsDlgEvent dlgEvent;
        dlgEvent.m_bSuccess = false;
        std::array<LPCWSTR, 2> pwszAcceptTypes{ L"*/*", nullptr };
        WinHTTPWrappers::CSyncDownloader winHttpRequest;
        CString sObject{MapnikTileProvider.GetDownloadObject(tile.m_nZoom, tile.m_nTileX, tile.m_nTileY) + _T("/dirty")};
        hr = winHttpRequest.Initialize(connection, CStringW(sObject), nullptr, nullptr, nullptr, pwszAcceptTypes.data(), WINHTTP_FLAG_REFRESH);
        if (FAILED(hr))
        {
          //report the error
          TRACE(_T("COSMCtrlMapOperationsDlg::ForceMapnikRerenderHelper, Failed to create request for object \"%s\", Error:%08X\n"), sObject.GetString(), hr);

          //Update the stats
          ++nTilesNotRerendered;
        }
        else
        {
          hr = winHttpRequest.SendRequestSync();
          if (FAILED(hr))
          {
            //report the error
            TRACE(_T("COSMCtrlMapOperationsDlg::ForceMapnikRerenderHelper, Failed to send request for object \"%s\", Error:%08X\n"), sObject.GetString(), hr);

            //Update the stats
            ++nTilesNotRerendered;
          }
          else
          {
            CStringA sResponse;
#pragma warning(suppress: 26472 26486 26489 26490)
            sResponse.Append(reinterpret_cast<LPCSTR>(winHttpRequest.m_Response.data()), static_cast<int>(winHttpRequest.m_Response.size()));
            if (sResponse.Find("Tile submitted for rendering") != -1)
            {
              //Update the stats
              ++nTilesRerendered;
              dlgEvent.m_bSuccess = true;
            }
            else
            {
              //Update the stats
              ++nTilesNotRerendered;
            }
          }
        }

        //Update the UI
        dlgEvent.m_Event = COSMCtrlMapOperationsDlgEvent::Event::SimpleStringStatus;
        dlgEvent.m_sString = sObject;
        dlgEvent.m_nItemData = i + 1;
        AddEvent(dlgEvent);

        //Check if we have been cancelled before we loop around
        bSuccess = (WaitForSingleObject(m_WorkerTerminateEvent, 0) == WAIT_TIMEOUT);
      }
    }
  }

  //Finally add a event about how many items have been downloaded
  COSMCtrlMapOperationsDlgEvent dlgEvent;
  dlgEvent.m_Event = COSMCtrlMapOperationsDlgEvent::Event::SimpleStringStatus;
  CString sTilesRerendered;
  sTilesRerendered.Format(_T("%d"), nTilesRerendered);
  CString sTilesNotRerendered;
  sTilesNotRerendered.Format(_T("%d"), nTilesNotRerendered);
  AfxFormatString2(dlgEvent.m_sString, IDS_OSMCTRL_RERENDER_TILES_STATS, sTilesRerendered, sTilesNotRerendered);
  AddEvent(dlgEvent);

  return true;
}

UINT COSMCtrlMapOperationsDlg::WorkerThread()
{
  //Validate our parameters
#pragma warning(suppress: 26496)
  AFXASSUME(m_pOSMCtrl != nullptr);

  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

  //Switch on the operation type we have been asked to work on
  switch (static_cast<Operation>(m_nOperation))
  {
    case Operation::DeleteCachedTiles:
    {
      bSuccess = DeleteCachedTilesHelper();
      break;
    }
    case Operation::CacheTilesIncludingThoseDownloaded:
    {
      bSuccess = DownloadTiles(false);
      break;
    }
    case Operation::CacheTilesOnlyThoseNotDownloaded:
    {
      bSuccess = DownloadTiles(true);
      break;
    }
    case Operation::ForceMapnikRerender:
    {
      bSuccess = ForceMapnikRerenderHelper();
      break;
    }
    default:
    {
      ASSERT(false);
      break;
    }
  }

  //Inform the UI that the work is completed
  COSMCtrlMapOperationsDlgEvent dlgEvent;
  dlgEvent.m_Event = COSMCtrlMapOperationsDlgEvent::Event::ThreadCompleted;
  dlgEvent.m_bSuccess = bSuccess ? true : false;
  AddEvent(dlgEvent);

  return 0;
}

void COSMCtrlMapOperationsDlg::AddEvent(_In_ const COSMCtrlMapOperationsDlgEvent& dlgEvent)
{
  {
    //Serialize access to the array
    CSingleLock sl{&m_csPendingEvents, TRUE};
    m_PendingEvents.push_back(dlgEvent);
  }

  //Inform the GUI thread that there has been a new event added to the array
  PostMessage(sm_nMessageAddEvent);
}

LRESULT COSMCtrlMapOperationsDlg::OnAddEvent(WPARAM /*wParam*/, LPARAM /*lParam*/)
{
  //Validate our parameters
#pragma warning(suppress: 26496)
  AFXASSUME(m_pOSMCtrl != nullptr);

  //Serialize access to the array
  CSingleLock sl{&m_csPendingEvents, TRUE};

  //Pull as many items of the queue as we can
  while (m_PendingEvents.size())
  {
#pragma warning(suppress: 26446 26486)
    COSMCtrlMapOperationsDlgEvent dlgEvent{m_PendingEvents[0]};
    m_PendingEvents.erase(m_PendingEvents.cbegin());
    switch (dlgEvent.m_Event)
    {
      case COSMCtrlMapOperationsDlgEvent::Event::ThreadCompleted:
      {
        //Disable the stop button and reenable the start button
        m_ctrlStopOperation.EnableWindow(FALSE);
        m_ctrlStartOperation.EnableWindow(TRUE);

        //Reset the progress control if necessary
        if (!dlgEvent.m_bSuccess)
          m_ctrlProgress.SetPos(0);

        //Also reset the zoom level control
        m_ctrlZoomLevel.EnableWindow(TRUE);

        //Set focus back to the Start button
        m_ctrlStartOperation.SetFocus();

        //Force a refresh of map as some of the map operations may have invalidate the tiles
        m_pOSMCtrl->Refresh();

        break;
      }
      case COSMCtrlMapOperationsDlgEvent::Event::SimpleStringStatus:
      {
        //Simple update the status static
        m_ctrlStatusText.SetWindowText(dlgEvent.m_sString);

        //Update the progress control if necessary
        if (dlgEvent.m_nItemData)
#pragma warning(suppress: 26472)
          m_ctrlProgress.SetPos(static_cast<int>(dlgEvent.m_nItemData));
        break;
      }
      default:
      {
        ASSERT(false);
        break;
      }
    }
  }

  return 0L;
}
