/*
Module : OSMCtrlMarker.cpp
Purpose: Implementation for the COSMCtrlMarker class
Created: PJN / 30-05-2020
History: PJN / None

Copyright (c) 2011 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Includes /////////////////////////////////////////////////

#include "stdafx.h"
#include "OSMCtrlPosition.h"
#include "OSMCtrl.h"


//////////////////// Implementation ///////////////////////////////////////////

#ifdef COSMCTRL_NOD2D
bool COSMCtrlMarker::Draw(_In_ COSMCtrl& ctrl, _In_ Gdiplus::Graphics& graphics, _In_ const CRect& rClient)
{
  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

  //Draw the marker if necessary
  Gdiplus::PointF ptMarker;
  if (m_bVisible && ctrl.PositionToClient(m_Position, rClient, ptMarker) && (ctrl.m_fZoom >= m_nMinZoomLevel) && (ctrl.m_fZoom <= m_nMaxZoomLevel))
  {
    //Extract the icon associated with the marker
#pragma warning(suppress: 26446)
    const COSMCtrlIcon* pIcon{ctrl.m_Icons[m_nIconIndex]};
#pragma warning(suppress: 26496)
    AFXASSUME(pIcon != nullptr);

    //Draw the outline of the marker in the selection color if we are selected
#pragma warning(suppress: 26489)
    if (m_bSelected && (pIcon->m_Image != nullptr))
    {
      //Create the pen we need
      Gdiplus::Pen pen{m_colorSelection, m_fSelectionPenWidth};
      if (pen.GetLastStatus() != Gdiplus::Ok)
        return false;
      if (pen.SetAlignment(Gdiplus::PenAlignmentInset) != Gdiplus::Ok)
        return false;

      //Draw the selection rectangle
#pragma warning(suppress: 26489)
      if (graphics.DrawRectangle(&pen, ptMarker.X - pIcon->m_ptAnchor.x + 1, ptMarker.Y - pIcon->m_ptAnchor.y + 1, static_cast<Gdiplus::REAL>(pIcon->m_Image->GetWidth() - 2), static_cast<Gdiplus::REAL>(pIcon->m_Image->GetHeight() - 2)) != Gdiplus::Ok)
        return false;
    }

    //And draw it if valid
#pragma warning(suppress: 26489)
    if (pIcon->m_Image != nullptr)
#pragma warning(suppress: 26486 26489)
      bSuccess = (graphics.DrawImage(pIcon->m_Image.get(), ptMarker.X - pIcon->m_ptAnchor.x, ptMarker.Y - pIcon->m_ptAnchor.y) == Gdiplus::Ok);
  }

  return bSuccess;
}
#else
#pragma warning(suppress: 26429)
bool COSMCtrlMarker::DrawInternal(_In_ COSMCtrl& ctrl, _In_ CRenderTarget* pRenderTarget, _In_ CD2DBitmap* pBitmap, _In_ CD2DPointF& ptMarker)
{
  //Extract the icon associated with the marker
#pragma warning(suppress: 26446)
  const COSMCtrlIcon* pIcon{ctrl.m_Icons[m_nIconIndex]};
#pragma warning(suppress: 26496)
  AFXASSUME(pIcon != nullptr);

  //Draw the outline of the marker in the selection color if we are selected
#pragma warning(suppress: 26496)
  AFXASSUME(pBitmap != nullptr);
  const CD2DSizeU sizeIcon{pBitmap->GetPixelSize()};
  if (m_bSelected)
  {
    //Create the brush we need
    CD2DBrushProperties brushProperties{m_colorSelection.a};
    CD2DSolidColorBrush brush{pRenderTarget, m_colorSelection, m_colorSelection.a == 1 ? nullptr : &brushProperties};

    //Draw the selection rectangle
    pRenderTarget->DrawRectangle(CD2DRectF{ptMarker.x - pIcon->m_ptAnchor.x + 1, ptMarker.y - pIcon->m_ptAnchor.y + 1, ptMarker.x - pIcon->m_ptAnchor.x + sizeIcon.width - 2, ptMarker.y - pIcon->m_ptAnchor.y + 1 + sizeIcon.height - 2}, &brush, m_fSelectionPenWidth);
  }

  //And draw the image
  pRenderTarget->DrawBitmap(pBitmap, CD2DRectF{ptMarker.x - pIcon->m_ptAnchor.x, ptMarker.y - pIcon->m_ptAnchor.y, ptMarker.x - pIcon->m_ptAnchor.x + sizeIcon.width, ptMarker.y - pIcon->m_ptAnchor.y + sizeIcon.height});

  return true;
}

bool COSMCtrlMarker::Draw(_In_ COSMCtrl& ctrl, _In_ CRenderTarget* pRenderTarget, _In_ const CRect& rClient)
{
  //Validate our parameters
  ASSERT(pRenderTarget != nullptr);

  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

  //Draw the marker if necessary
  CD2DPointF ptMarker;
  if (m_bVisible && ctrl.PositionToClient(m_Position, rClient, ptMarker) && (ctrl.m_fZoom >= m_nMinZoomLevel) && (ctrl.m_fZoom <= m_nMaxZoomLevel))
  {
    //Extract the icon associated with the marker
#pragma warning(suppress: 26446)
    COSMCtrlIcon* pIcon{ctrl.m_Icons[m_nIconIndex]};
#pragma warning(suppress: 26496)
    AFXASSUME(pIcon != nullptr);

    //Cache the bitmap if necessary
#pragma warning(suppress: 26489)
    if (pIcon->m_pImage == nullptr)
    {
      if (pIcon->m_nResourceID != 0)
#pragma warning(suppress: 26409)
        pIcon->m_pImage = new CD2DBitmap{pRenderTarget, pIcon->m_nResourceID, pIcon->m_sResourceType};
      else
#pragma warning(suppress: 26409)
        pIcon->m_pImage = new CD2DBitmap{pRenderTarget, pIcon->m_sFilename};
      if (FAILED(pIcon->m_pImage->Create(pRenderTarget)))
      {
        delete pIcon->m_pImage;
        pIcon->m_pImage = nullptr;
        return false;
      }
    }

    //Call the helper function
#pragma warning(suppress: 26486 26489)
    bSuccess = DrawInternal(ctrl, pRenderTarget, pIcon->m_pImage, ptMarker);
  }

  return bSuccess;
}
#endif //#ifdef COSMCTRL_NOD2D
