/*
Module : OSMCtrlPolygon.cpp
Purpose: Implementation for the COSMCtrlPolygon class
Created: PJN / 10-04-2011
History: PJN / None

Copyright (c) 2011 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Includes /////////////////////////////////////////////////

#include "stdafx.h"
#include "OSMCtrlPolygon.h"
#include "OSMCtrl.h"


//////////////////// Implementation ///////////////////////////////////////////

bool COSMCtrlPolygon::FullySelected() const noexcept
{
  //Check all the nodes
  for (const auto& node : m_Nodes)
  {
    if (!node.m_bSelected)
      return false;
  }

  return true;
}

bool COSMCtrlPolygon::AnySelected() const noexcept
{
  //Check all the nodes
  for (const auto& node : m_Nodes)
  {
    if (node.m_bSelected)
      return true;
  }

  return false;
}

void COSMCtrlPolygon::Select() noexcept
{
  //Mark all nodes as selected
  for (auto& node : m_Nodes)
    node.m_bSelected = true;
}

void COSMCtrlPolygon::Deselect() noexcept
{
  //Mark all nodes as selected
  for (auto& node : m_Nodes)
    node.m_bSelected = false;
}

#pragma warning(suppress: 26440)
double COSMCtrlPolygon::Length() const
{
  //What will be the return value from this function
  double dDistance{0};

  //Accumulate the distance between all of the nodes in this polyline
  for (std::vector<COSMCtrlNode>::size_type i{1}; i<m_Nodes.size(); i++)
#pragma warning(suppress: 26446)
    dDistance += COSMCtrlHelper::DistanceBetweenPoints(m_Nodes[i - 1].m_Position, m_Nodes[i].m_Position, nullptr, nullptr);

  return dDistance;
}

_Return_type_success_(return != 0) bool COSMCtrlPolygon::GetBoundingRect(_Out_ COSMCtrlPosition& topLeft, _Out_ COSMCtrlPosition& bottomRight) const noexcept
{
  double fTopLatitude{-91};
  double fBottomLatitude{91};
  double fLeftLongitude{181};
  double fRightLongitude{-181};

  //Look through all the nodes to find the extreme values in latitude and longitude
  for (const auto& node : m_Nodes)
  {
    if (node.m_Position.m_fLatitude > fTopLatitude)
      fTopLatitude = node.m_Position.m_fLatitude;
    if (node.m_Position.m_fLatitude < fBottomLatitude)
      fBottomLatitude = node.m_Position.m_fLatitude;
    if (node.m_Position.m_fLongitude < fLeftLongitude)
      fLeftLongitude = node.m_Position.m_fLongitude;
    if (node.m_Position.m_fLongitude > fRightLongitude)
      fRightLongitude = node.m_Position.m_fLongitude;
  }

  topLeft.m_fLatitude = fTopLatitude;
  topLeft.m_fLongitude = fLeftLongitude;
  bottomRight.m_fLatitude = fBottomLatitude;
  bottomRight.m_fLongitude = fRightLongitude;

  return (m_Nodes.size() != 0); //A polygon without no nodes fails this function
}

#ifdef COSMCTRL_NOD2D
bool COSMCtrlPolygon::DrawInternal(_In_ const COSMCtrl& ctrl, _In_ Gdiplus::GraphicsPath& path, _In_ const CRect& rClient) const
{
  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

  //Draw the polygon if necessary
  if (m_bVisible && (m_Nodes.size() > 1) && (ctrl.m_fZoom >= m_nMinZoomLevel) && (ctrl.m_fZoom <= m_nMaxZoomLevel))
  {
    //Create the array of points in GDI+ format
    std::vector<Gdiplus::PointF> points;
#pragma warning(suppress: 26472)
    const size_t nPoints{m_Nodes.size()};
    points.reserve(nPoints);
    bSuccess = true;
    for (size_t j{0}; (j<nPoints) && bSuccess; j++)
    {
#pragma warning(suppress: 26446)
      const COSMCtrlNode& node{m_Nodes[j]};
      Gdiplus::PointF point;
      bSuccess = ctrl.PositionToClient(node.m_Position, rClient, point);
      if (bSuccess)
        points.push_back(point);
    }

    if (bSuccess)
    {
      //Finally draw the polygon using GDI+
#pragma warning(suppress: 26472)
      bSuccess = (path.AddPolygon(points.data(), static_cast<INT>(nPoints)) == Gdiplus::Ok);
    }
  }

  return bSuccess;
}

bool COSMCtrlPolygon::DrawOutline(_In_ const COSMCtrl& ctrl, _In_ Gdiplus::GraphicsPath& path, _In_ const CRect& rClient) const
{
  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

  //Draw the polygon if necessary
  if (m_bVisible && (m_Nodes.size() > 1) && (ctrl.m_fZoom >= m_nMinZoomLevel) && (ctrl.m_fZoom <= m_nMaxZoomLevel))
  {
    //Create the pen we need
    Gdiplus::Pen pen{m_colorPen, m_fLinePenWidth};
    if (pen.GetLastStatus() != Gdiplus::Ok)
      return false;

    //Set up the brush attributes
    pen.SetEndCap(m_EndCap);
    pen.SetStartCap(m_StartCap);
    pen.SetLineJoin(m_LineJoin);

    //iterate thro all the positions in the polygon
    bSuccess = true;
    for (std::vector<COSMCtrlNode>::size_type j{1}; (j<m_Nodes.size()) && bSuccess; j++)
    {
      Gdiplus::PointF ptPosition1;
      Gdiplus::PointF ptPosition2;
#pragma warning(suppress: 26446)
      if (ctrl.PositionToClient(m_Nodes[j - 1].m_Position, rClient, ptPosition1) && ctrl.PositionToClient(m_Nodes[j].m_Position, rClient, ptPosition2))
      {
        //Draw a line to the next node in the polyline
        bSuccess = (path.AddLine(ptPosition1.X, ptPosition1.Y, ptPosition2.X, ptPosition2.Y) == Gdiplus::Ok);
      }
      else
        bSuccess = false; //break out of the loop
    }

    //Finally close the figure
    if (bSuccess)
      bSuccess = (path.CloseAllFigures() == Gdiplus::Ok);

    //Widen the path by the pen
    if (bSuccess)
      bSuccess = (path.Widen(&pen) == Gdiplus::Ok);
  }

  return bSuccess;
}

bool COSMCtrlPolygon::Draw(_In_ COSMCtrl& ctrl, _In_ Gdiplus::Graphics& graphics, _In_ const CRect& rClient)
{
  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

  if (m_bVisible && (m_Nodes.size() > 1) && (ctrl.m_fZoom >= m_nMinZoomLevel) && (ctrl.m_fZoom <= m_nMaxZoomLevel))
  {
    //Create the pen we need
    Gdiplus::Pen pen{m_colorPen, m_fLinePenWidth};
    if (pen.GetLastStatus() != Gdiplus::Ok)
      return false;

    //Create the brush we need
    Gdiplus::SolidBrush brush{m_colorBrush};
    if (brush.GetLastStatus() != Gdiplus::Ok)
      return false;

    //Set up the pen attributes
    pen.SetDashCap(m_DashCap);
    pen.SetDashOffset(m_fDashOffset);
    pen.SetDashStyle(m_DashStyle);
    pen.SetEndCap(m_EndCap);
    pen.SetStartCap(m_StartCap);
    pen.SetLineJoin(m_LineJoin);

    //Create the array of points in GDI+ format
    std::vector<Gdiplus::PointF> points;
#pragma warning(suppress: 26472)
    const size_t nPoints{m_Nodes.size()};
    points.reserve(nPoints);
    bSuccess = true;
    for (size_t j{0}; (j<nPoints) && bSuccess; j++)
    {
#pragma warning(suppress: 26446)
      const COSMCtrlNode& node{m_Nodes[j]};
      Gdiplus::PointF point;
      bSuccess = ctrl.PositionToClient(node.m_Position, rClient, point);
      if (bSuccess)
      {
        const Gdiplus::PointF gdiPoint{point.X, point.Y};
        points.push_back(gdiPoint);
      }
    }

    if (bSuccess)
    {
      //Finally draw the polygon using GDI+
#pragma warning(suppress: 4267 26472)
      bSuccess = (graphics.FillPolygon(&brush, points.data(), nPoints) == Gdiplus::Ok) && (graphics.DrawPolygon(&pen, points.data(), static_cast<INT>(nPoints)) == Gdiplus::Ok);
    }

    //Draw each node (selected or not) of the polygon as necessary
    Gdiplus::SolidBrush brushSelectedNode{m_colorSelectionNode};
    if (brushSelectedNode.GetLastStatus() != Gdiplus::Ok)
      return false;
    Gdiplus::SolidBrush brushNode{m_colorNode};
    if (brushNode.GetLastStatus() != Gdiplus::Ok)
      return false;
    for (const auto& node : m_Nodes)
    {
      Gdiplus::PointF ptPosition;
      if (ctrl.PositionToClient(node.m_Position, rClient, ptPosition))
      {
        if (node.m_bSelected)
          bSuccess = (graphics.FillEllipse(&brushSelectedNode, ptPosition.X - m_fSelectionNodeWidth / 2, ptPosition.Y - m_fSelectionNodeWidth / 2, m_fSelectionNodeWidth, m_fSelectionNodeWidth) == Gdiplus::Ok);
        else if (m_DrawingStyle == COSMCtrlPolygon::DrawingStyle::LinesAndNodes)
          bSuccess = (graphics.FillEllipse(&brushNode, ptPosition.X - m_fNodeWidth / 2, ptPosition.Y - m_fNodeWidth / 2, m_fNodeWidth, m_fNodeWidth) == Gdiplus::Ok);
      }
      else
        bSuccess = false;
    }
  }

  return bSuccess;
}
#else
#pragma warning(suppress: 26429)
bool COSMCtrlPolygon::Draw(_In_ COSMCtrl& ctrl, _In_ CRenderTarget* pRenderTarget, _In_ const CRect& rClient)
{
  //Validate our parameters
  ASSERT(pRenderTarget != nullptr);

  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

  if (m_bVisible && (m_Nodes.size() > 1) && (ctrl.m_fZoom >= m_nMinZoomLevel) && (ctrl.m_fZoom <= m_nMaxZoomLevel))
  {
    //Create the brushes we need
    CD2DBrushProperties brushProperties{m_colorPen.a};
    CD2DSolidColorBrush brushPen{pRenderTarget, m_colorPen, m_colorPen.a == 1 ? nullptr : &brushProperties};
    CD2DBrushProperties brushProperties2{m_colorBrush.a};
    CD2DSolidColorBrush brush{pRenderTarget, m_colorBrush, m_colorBrush.a == 1 ? nullptr : &brushProperties2};

    //Create the stroke style we need
    ATL::CComPtr<ID2D1StrokeStyle> strokeStyle;
    ID2D1Factory* pFactory{ctrl.GetD2DFactory()};
    if (pFactory == nullptr)
      return false;
    D2D1_STROKE_STYLE_PROPERTIES strokeStyleProperties;
    strokeStyleProperties.startCap = m_StartCap;
    strokeStyleProperties.endCap = m_EndCap;
    strokeStyleProperties.dashCap = m_DashCap;
    strokeStyleProperties.lineJoin = m_LineJoin;
    strokeStyleProperties.miterLimit = m_fMiterLimit;
    strokeStyleProperties.dashStyle = m_DashStyle;
    strokeStyleProperties.dashOffset = m_fDashOffset;
    const HRESULT hr = pFactory->CreateStrokeStyle(&strokeStyleProperties, nullptr, 0, &strokeStyle);
    if (FAILED(hr))
      return false;

    //Create the array of points as a D2D geometry
    CD2DPathGeometry pointsGeometry{pRenderTarget};
    if (FAILED(pointsGeometry.Create(pRenderTarget)))
      return false;
    CD2DGeometrySink geometrySink{pointsGeometry};
    CD2DPointF point;
#pragma warning(suppress: 26446)
    if (ctrl.PositionToClient(m_Nodes[0].m_Position, rClient, point))
    {
      geometrySink.BeginFigure(point, D2D1_FIGURE_BEGIN_FILLED);
      bSuccess = true;
      for (std::vector<COSMCtrlNode>::size_type j{1}; (j<m_Nodes.size()) && bSuccess; j++)
      {
#pragma warning(suppress: 26446)
        const COSMCtrlNode& node{m_Nodes[j]};
        bSuccess = ctrl.PositionToClient(node.m_Position, rClient, point);
        if (bSuccess)
          geometrySink.AddLine(CD2DPointF(point.x, point.y));
      }
      geometrySink.EndFigure(D2D1_FIGURE_END_CLOSED);
      if (!geometrySink.Close())
        return false;
      pRenderTarget->FillGeometry(&pointsGeometry, &brush);
      pRenderTarget->DrawGeometry(&pointsGeometry, &brushPen, m_fLinePenWidth, strokeStyle);
    }

    //Draw each node (selected or not) of the polyline as necessary
    CD2DBrushProperties brushProperties3{m_colorSelectionNode.a};
    CD2DSolidColorBrush brushSelectedNode{pRenderTarget, m_colorSelectionNode, m_colorSelectionNode.a == 1 ? nullptr : &brushProperties3};
    CD2DBrushProperties brushProperties4{m_colorNode.a};
    CD2DSolidColorBrush brushNode{pRenderTarget, m_colorNode, m_colorNode.a == 1 ? nullptr : &brushProperties4};
    for (const auto& node : m_Nodes)
    {
      CD2DPointF ptPosition;
      if (ctrl.PositionToClient(node.m_Position, rClient, ptPosition))
      {
        if (node.m_bSelected)
          pRenderTarget->FillEllipse(CD2DEllipse{ptPosition, CD2DSizeF{m_fSelectionNodeWidth / 2, m_fSelectionNodeWidth / 2}}, node.m_bSelected ? &brushSelectedNode : &brushNode);
        else if (m_DrawingStyle == COSMCtrlPolygon::DrawingStyle::LinesAndNodes)
          pRenderTarget->FillEllipse(CD2DEllipse{ptPosition, CD2DSizeF{m_fNodeWidth / 2, m_fNodeWidth / 2}}, &brushNode);
      }
      else
        bSuccess = false;
    }
  }

  return bSuccess;
}
#endif //#ifdef COSMCTRL_NOD2D
