/*
Module : OSMCtrlPolygon.h
Purpose: Defines the interface for COSMCtrlPolygon class
Created: PJN / 10-04-2011

Copyright (c) 2011 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Macros / Includes ////////////////////////////////////////

#pragma once

#ifndef __OSMCTRLPOLYGON_H__
#define __OSMCTRLPOLYGON_H__

#include "OSMCtrlPosition.h"
#include "OSMCtrlNode.h"
#include "OSMCtrlHelper.h"

#ifndef _VECTOR_
#pragma message("To avoid this message, please put vector in your pre compiled header (normally stdafx.h)")
#include <vector>
#endif //#ifndef _VECTOR_


//////////////////// Classes //////////////////////////////////////////////////

class COSMCtrl;

//A class which represents a polygon (i.e. a closed sequence of nodes) on the map
class COSMCtrlPolygon
{
public:
//Constructors / Destructors
  COSMCtrlPolygon() = default;
  COSMCtrlPolygon(const COSMCtrlPolygon&) = default;
#pragma warning(suppress: 26439)
  COSMCtrlPolygon(COSMCtrlPolygon&&) = default;
  virtual ~COSMCtrlPolygon() = default;

//Enums
  enum class DrawingStyle
  {
    LinesOnly,
    LinesAndNodes,
  };

//Methods
  COSMCtrlPolygon& operator=(const COSMCtrlPolygon&) = default;
#pragma warning(suppress: 26439)
  COSMCtrlPolygon& operator=(COSMCtrlPolygon&&) = default;
  _NODISCARD bool FullySelected() const noexcept;
  _NODISCARD bool AnySelected() const noexcept;
  void Select() noexcept;
  void Deselect() noexcept;
  _NODISCARD double Length() const;
  _Return_type_success_(return != 0) bool GetBoundingRect(_Out_ COSMCtrlPosition& topLeft, _Out_ COSMCtrlPosition& bottomRight) const noexcept;
#ifdef COSMCTRL_NOD2D
  virtual bool DrawInternal(_In_ const COSMCtrl& ctrl, _In_ Gdiplus::GraphicsPath& path, _In_ const CRect& rClient) const;
  virtual bool DrawOutline(_In_ const COSMCtrl& ctrl, _In_ Gdiplus::GraphicsPath& path, _In_ const CRect& rClient) const;
  virtual bool Draw(_In_ COSMCtrl& ctrl, _In_ Gdiplus::Graphics& graphics, _In_ const CRect& rClient);
#else
  virtual bool Draw(_In_ COSMCtrl& ctrl, _In_ CRenderTarget* pRenderTarget, _In_ const CRect& rClient);
#endif //#ifdef COSMCTRL_NOD2D

//Member variables
  std::vector<COSMCtrlNode> m_Nodes; //The array of position for this polyline
  float m_fLinePenWidth{2}; //The width of the line to draw
#ifdef COSMCTRL_NOD2D
  Gdiplus::Color m_colorPen{128, 0, 0, 255}; //The color of the pen to use
  Gdiplus::Color m_colorNode{0, 0, 0}; //The color of the node
  Gdiplus::Color m_colorSelectionNode{255, 0, 0}; //The color of the brush to use for selection nodes
  Gdiplus::Color m_colorBrush{128, 0, 0, 255}; //The color of the brush to use
  Gdiplus::DashCap m_DashCap{Gdiplus::DashCapFlat}; //The dash cap to use
  Gdiplus::DashStyle m_DashStyle{Gdiplus::DashStyleSolid}; //The dash style to use
  Gdiplus::LineCap m_EndCap{Gdiplus::LineCapRound}; //The end cap to use
  Gdiplus::LineCap m_StartCap{Gdiplus::LineCapRound}; //The start cap to use
  Gdiplus::LineJoin m_LineJoin{Gdiplus::LineJoinRound}; //The line join style to use
  float m_fDashOffset{0}; //The dash offset to use
#else
  D2D1::ColorF m_colorPen{0, 0, 255, 0.5F}; //The color of the pen to use
  D2D1::ColorF m_colorNode{0, 0, 0}; //The color of the node
  D2D1::ColorF m_colorSelectionNode{255, 0, 0}; //The color of the brush to use for selection nodes
  D2D1::ColorF m_colorBrush{0, 0, 255, 0.5F}; //The color of the brush to use
  D2D1_CAP_STYLE m_DashCap{D2D1_CAP_STYLE_FLAT}; //The dash cap to use
  D2D1_DASH_STYLE m_DashStyle{D2D1_DASH_STYLE_SOLID}; //The dash style to use
  D2D1_CAP_STYLE m_EndCap{D2D1_CAP_STYLE_ROUND}; //The end cap to use
  D2D1_CAP_STYLE m_StartCap{D2D1_CAP_STYLE_ROUND}; //The start cap to use
  D2D1_LINE_JOIN m_LineJoin{D2D1_LINE_JOIN_ROUND}; //The line join style to use
  FLOAT m_fDashOffset{0}; //The dash offset to use
  FLOAT m_fMiterLimit{1}; //The limit of the thickness of the join on a mitered corner
#endif //#ifdef COSMCTRL_NOD2D
  float m_fNodeWidth{4}; //The width of the node to draw
  float m_fSelectionNodeWidth{4}; //The width of the node to draw if selected
  DrawingStyle m_DrawingStyle{DrawingStyle::LinesOnly}; //The drawing style to use for this polyline
  CString m_sToolTipText; //The tooltip text for this polygon
  bool m_bDraggable{false}; //Is this polygon draggable
  bool m_bVisible{true}; //Is this polygon currently visible
  bool m_bEditable{false}; //Is this polygon editable
  bool m_bHitTest{true}; //Is this polygon to be hit-tested
  int m_nMinZoomLevel{COSMCtrlHelper::OSMMinZoom}; //The minimum zoom level at which this polyline appears
  int m_nMaxZoomLevel{COSMCtrlHelper::OSMMaxZoom}; //The max zoom level at which this polyline appears
};


#endif //#ifndef __OSMCTRLPOLYGON_H__
