/*
Module : OSMCtrlPolyline.cpp
Purpose: Implementation for the COSMCtrlPolyline class
Created: PJN / 10-04-2011
History: PJN / None

Copyright (c) 2011 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Includes /////////////////////////////////////////////////

#include "stdafx.h"
#include "OSMCtrlPolyline.h"
#include "OSMCtrl.h"


//////////////////// Implementation ///////////////////////////////////////////

bool COSMCtrlPolyline::FullySelected() const noexcept
{
  //Check all the nodes
  for (const auto& node : m_Nodes)
  {
    if (!node.m_bSelected)
      return false;
  }

  return true;
}

bool COSMCtrlPolyline::AnySelected() const noexcept
{
  //Check all the nodes
  for (const auto& node : m_Nodes)
  {
    if (node.m_bSelected)
      return true;
  }

  return false;
}

void COSMCtrlPolyline::Select() noexcept
{
  //Mark all nodes as selected
  for (auto& node : m_Nodes)
    node.m_bSelected = true;
}

void COSMCtrlPolyline::Deselect() noexcept
{
  //Mark all nodes as selected
  for (auto& node : m_Nodes)
    node.m_bSelected = false;
}

#pragma warning(suppress: 26440)
double COSMCtrlPolyline::Length() const
{
  //What will be the return value from this function
  double dDistance{0};

  //Accumulate the distance between all of the nodes in this polyline
  for (std::vector<COSMCtrlNode>::size_type i{1}; i<m_Nodes.size(); i++)
#pragma warning(suppress: 26446)
    dDistance += COSMCtrlHelper::DistanceBetweenPoints(m_Nodes[i-1].m_Position, m_Nodes[i].m_Position, nullptr, nullptr);

  return dDistance;
}

#pragma warning(suppress: 26440)
double COSMCtrlPolyline::LastBearing() const
{
  double dBearing{0};
  const auto nNodes{m_Nodes.size()};
  ASSERT(nNodes > 1);
#pragma warning(suppress: 26446)
  COSMCtrlHelper::DistanceBetweenPoints(m_Nodes[nNodes-2].m_Position, m_Nodes[nNodes-1].m_Position, &dBearing, nullptr);

  return dBearing;
}

_Return_type_success_(return != 0) bool COSMCtrlPolyline::GetBoundingRect(_Out_ COSMCtrlPosition& topLeft, _Out_ COSMCtrlPosition& bottomRight) const noexcept
{
  double fTopLatitude{-91};
  double fBottomLatitude{91};
  double fLeftLongitude{181};
  double fRightLongitude{-181};

  //Look through all the nodes to find the extreme values in latitude and longitude  
  for (const auto& node : m_Nodes)
  {
    if (node.m_Position.m_fLatitude > fTopLatitude)
      fTopLatitude = node.m_Position.m_fLatitude;
    if (node.m_Position.m_fLatitude < fBottomLatitude)
      fBottomLatitude = node.m_Position.m_fLatitude;
    if (node.m_Position.m_fLongitude < fLeftLongitude)
      fLeftLongitude = node.m_Position.m_fLongitude;
    if (node.m_Position.m_fLongitude > fRightLongitude)
      fRightLongitude = node.m_Position.m_fLongitude;
  }

  topLeft.m_fLatitude = fTopLatitude;
  topLeft.m_fLongitude = fLeftLongitude;
  bottomRight.m_fLatitude = fBottomLatitude;
  bottomRight.m_fLongitude = fRightLongitude;

  return (m_Nodes.size() != 0); //A polygon without no nodes fails this function
}

#ifdef COSMCTRL_NOD2D
bool COSMCtrlPolyline::Draw(_In_ COSMCtrl& ctrl, _In_ Gdiplus::Graphics& graphics, _In_ const CRect& rClient)
{
  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

  //Draw the polyline if necessary
  if (m_bVisible && (m_Nodes.size() > 1) && (ctrl.m_fZoom >= m_nMinZoomLevel) && (ctrl.m_fZoom <= m_nMaxZoomLevel))
  {
    //Draw the lines if necessary
    if ((m_DrawingStyle == COSMCtrlPolyline::DrawingStyle::LinesOnly) || (m_DrawingStyle == COSMCtrlPolyline::DrawingStyle::LinesAndNodes))
    {
      //Create the pen we need
      Gdiplus::Pen pen{m_colorPen, m_fLinePenWidth};
      if (pen.GetLastStatus() != Gdiplus::Ok)
        return false;

      //Set up the pen attributes
      pen.SetDashCap(m_DashCap);
      pen.SetDashOffset(m_fDashOffset);
      pen.SetDashStyle(m_DashStyle);
      pen.SetEndCap(m_EndCap);
      pen.SetStartCap(m_StartCap);
      pen.SetLineJoin(m_LineJoin);

      //iterate thro all the positions in the polyline
      bSuccess = true;
      for (std::vector<COSMCtrlNode>::size_type j{1}; (j<m_Nodes.size()) && bSuccess; j++)
      {
        Gdiplus::PointF ptPosition1;
        Gdiplus::PointF ptPosition2;
#pragma warning(suppress: 26446)
        if (ctrl.PositionToClient(m_Nodes[j - 1].m_Position, rClient, ptPosition1) && ctrl.PositionToClient(m_Nodes[j].m_Position, rClient, ptPosition2))
        {
          //Draw a line to the next node in the polyline
          bSuccess = (graphics.DrawLine(&pen, ptPosition1.X, ptPosition1.Y, ptPosition2.X, ptPosition2.Y) == Gdiplus::Ok);
        }
        else
          bSuccess = false; //break out of the loop
      }
    }

    //Draw each node (selected or not) of the polyline as necessary
    Gdiplus::SolidBrush brushSelectedNode{m_colorSelectionNode};
    if (brushSelectedNode.GetLastStatus() != Gdiplus::Ok)
      return false;
    Gdiplus::SolidBrush brushNode{m_colorNode};
    if (brushNode.GetLastStatus() != Gdiplus::Ok)
      return false;
    for (const auto& node : m_Nodes)
    {
      Gdiplus::PointF ptPosition;
      if (ctrl.PositionToClient(node.m_Position, rClient, ptPosition))
      {
        if (node.m_bSelected)
          bSuccess = (graphics.FillEllipse(&brushSelectedNode, ptPosition.X - m_fSelectionNodeWidth / 2, ptPosition.Y - m_fSelectionNodeWidth / 2, m_fSelectionNodeWidth, m_fSelectionNodeWidth) == Gdiplus::Ok);
        else if ((m_DrawingStyle == COSMCtrlPolyline::DrawingStyle::LinesAndNodes) || (m_DrawingStyle == COSMCtrlPolyline::DrawingStyle::NodesOnly))
          bSuccess = (graphics.FillEllipse(&brushNode, ptPosition.X - m_fNodeWidth / 2, ptPosition.Y - m_fNodeWidth / 2, m_fNodeWidth, m_fNodeWidth) == Gdiplus::Ok);
      }
      else
        bSuccess = false;
    }
  }

  return bSuccess;
}
#else
#pragma warning(suppress: 26429)
bool COSMCtrlPolyline::Draw(_In_ COSMCtrl& ctrl, _In_ CRenderTarget* pRenderTarget, _In_ const CRect& rClient)
{
  //Validate our parameters
  ASSERT(pRenderTarget != nullptr);

  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

  //Draw the polyline if necessary
  if (m_bVisible && (m_Nodes.size() > 1) && (ctrl.m_fZoom >= m_nMinZoomLevel) && (ctrl.m_fZoom <= m_nMaxZoomLevel))
  {
    //Draw the lines if necessary
    if ((m_DrawingStyle == COSMCtrlPolyline::DrawingStyle::LinesOnly) || (m_DrawingStyle == COSMCtrlPolyline::DrawingStyle::LinesAndNodes))
    {
      //Create the stroke style we need
      ATL::CComPtr<ID2D1StrokeStyle> strokeStyle;
      ID2D1Factory* pFactory{ctrl.GetD2DFactory()};
      if (pFactory == nullptr)
        return false;
      D2D1_STROKE_STYLE_PROPERTIES strokeStyleProperties;
      strokeStyleProperties.startCap = m_StartCap;
      strokeStyleProperties.endCap = m_EndCap;
      strokeStyleProperties.dashCap = m_DashCap;
      strokeStyleProperties.lineJoin = m_LineJoin;
      strokeStyleProperties.miterLimit = m_fMiterLimit;
      strokeStyleProperties.dashStyle = m_DashStyle;
      strokeStyleProperties.dashOffset = m_fDashOffset;
      const HRESULT hr{pFactory->CreateStrokeStyle(&strokeStyleProperties, nullptr, 0, &strokeStyle)};
      if (FAILED(hr))
        return false;

      if (m_colorPenStrokes.size())
      {
        //iterate thro all the positions in the polyline
        bSuccess = true;
        for (std::vector<COSMCtrlNode>::size_type j{1}; (j<m_Nodes.size()) && bSuccess; j++)
        {
          CD2DPointF ptPosition1;
          CD2DPointF ptPosition2;
#pragma warning(suppress: 26446)
          if (ctrl.PositionToClient(m_Nodes[j - 1].m_Position, rClient, ptPosition1) && ctrl.PositionToClient(m_Nodes[j].m_Position, rClient, ptPosition2))
          {
            //Create the brush we need
#pragma warning(suppress: 26446)
            const D2D1::ColorF& colorPen{m_colorPenStrokes[j]};
            CD2DBrushProperties brushProperties{colorPen.a};
            CD2DSolidColorBrush brushLine{pRenderTarget, colorPen, colorPen.a == 1 ? nullptr : &brushProperties};

            //Draw a line to the next node in the polyline
            pRenderTarget->DrawLine(ptPosition1, ptPosition2, &brushLine, m_fLinePenWidth, strokeStyle);
          }
          else
            bSuccess = false; //break out of the loop
        }
      }
      else
      {
        //Create the brush we need
        CD2DBrushProperties brushProperties{m_colorPen.a};
        CD2DSolidColorBrush brushLine{pRenderTarget, m_colorPen, m_colorPen.a == 1 ? nullptr : &brushProperties};

        //iterate thro all the positions in the polyline
        bSuccess = true;
        for (std::vector<COSMCtrlNode>::size_type j{1}; (j<m_Nodes.size()) && bSuccess; j++)
        {
          CD2DPointF ptPosition1;
          CD2DPointF ptPosition2;
#pragma warning(suppress: 26446)
          if (ctrl.PositionToClient(m_Nodes[j - 1].m_Position, rClient, ptPosition1) && ctrl.PositionToClient(m_Nodes[j].m_Position, rClient, ptPosition2))
          {
            //Draw a line to the next node in the polyline
            pRenderTarget->DrawLine(ptPosition1, ptPosition2, &brushLine, m_fLinePenWidth, strokeStyle);
          }
          else
            bSuccess = false; //break out of the loop
        }
      }
    }

    //Draw each node (selected or not) of the polyline as necessary
    CD2DBrushProperties brushProperties2{m_colorSelectionNode.a};
    CD2DSolidColorBrush brushSelectedNode{pRenderTarget, m_colorSelectionNode, m_colorSelectionNode.a == 1 ? nullptr : &brushProperties2};
    CD2DBrushProperties brushProperties3{m_colorNode.a};
    CD2DSolidColorBrush brushNode{pRenderTarget, m_colorNode, m_colorNode.a == 1 ? nullptr : &brushProperties3};
    for (const auto& node : m_Nodes)
    {
      CD2DPointF ptPosition;
      if (ctrl.PositionToClient(node.m_Position, rClient, ptPosition))
      {
        if (node.m_bSelected)
          pRenderTarget->FillEllipse(CD2DEllipse(ptPosition, CD2DSizeF(m_fSelectionNodeWidth / 2, m_fSelectionNodeWidth / 2)), node.m_bSelected ? &brushSelectedNode : &brushNode);
        else if ((m_DrawingStyle == COSMCtrlPolyline::DrawingStyle::LinesAndNodes) || (m_DrawingStyle == COSMCtrlPolyline::DrawingStyle::NodesOnly))
          pRenderTarget->FillEllipse(CD2DEllipse(ptPosition, CD2DSizeF(m_fNodeWidth / 2, m_fNodeWidth / 2)), &brushNode);
        bSuccess = true;
      }
      else
        bSuccess = false;
    }
  }

  return bSuccess;
}
#endif //#ifdef COSMCTRL_NOD2D
