/*
Module : OSMCtrlPolyline.h
Purpose: Defines the interface for COSMCtrlPolyline class
Created: PJN / 10-04-2011

Copyright (c) 2011 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Macros / Includes ////////////////////////////////////////

#pragma once

#ifndef __OSMCTRLPOLYLINE_H__
#define __OSMCTRLPOLYLINE_H__

#include "OSMCtrlNode.h"
#include "OSMCtrlHelper.h"

#ifndef _VECTOR_
#pragma message("To avoid this message, please put vector in your pre compiled header (normally stdafx.h)")
#include <vector>
#endif //#ifndef _VECTOR_


//////////////////// Classes //////////////////////////////////////////////////

class COSMCtrl;

//A class which represents a polyline (i.e. a sequence of nodes) on the map
class COSMCtrlPolyline
{
public:
//Constructors / Destructors
  COSMCtrlPolyline() = default;
  COSMCtrlPolyline(const COSMCtrlPolyline&) = default;
#pragma warning(suppress: 26439)
  COSMCtrlPolyline(COSMCtrlPolyline&&) = default;
  virtual ~COSMCtrlPolyline() = default;

//Enums
  enum class DrawingStyle
  {
    LinesOnly,
    LinesAndNodes,
    NodesOnly
  };

//Methods
  COSMCtrlPolyline& operator=(const COSMCtrlPolyline&) = default;
#pragma warning(suppress: 26439)
  COSMCtrlPolyline& operator=(COSMCtrlPolyline&&) = default;
  _NODISCARD bool FullySelected() const noexcept;
  _NODISCARD bool AnySelected() const noexcept;
  void Select() noexcept;
  void Deselect() noexcept;
  _NODISCARD double Length() const; //returns the length of this polyline in kilometers
  _NODISCARD double LastBearing() const; //Returns the start bearing between the second last and last node in this polyline
  _Return_type_success_(return != 0) bool GetBoundingRect(_Out_ COSMCtrlPosition& topLeft, _Out_ COSMCtrlPosition& bottomRight) const noexcept;
#ifdef COSMCTRL_NOD2D
  virtual bool Draw(_In_ COSMCtrl& ctrl, _In_ Gdiplus::Graphics& graphics, _In_ const CRect& rClient);
#else
  virtual bool Draw(_In_ COSMCtrl& ctrl, _In_ CRenderTarget* pRenderTarget, _In_ const CRect& rClient);
#endif //#ifdef COSMCTRL_NOD2D

//Member variables
  std::vector<COSMCtrlNode> m_Nodes; //The array of position for this polyline
  float m_fLinePenWidth{2}; //The width of the line to draw
#ifdef COSMCTRL_NOD2D
  Gdiplus::Color m_colorPen{255, 0, 0, 255}; //The color of the pen to use
  Gdiplus::Color m_colorNode{0, 0, 0}; //The color of the node
  Gdiplus::Color m_colorSelectionNode{255, 0, 0}; //The color of the brush to use for selection nodes
  Gdiplus::DashCap m_DashCap{Gdiplus::DashCapFlat}; //The dash cap to use
  Gdiplus::DashStyle m_DashStyle{Gdiplus::DashStyleSolid}; //The dash style to use
  Gdiplus::LineCap m_EndCap{Gdiplus::LineCapRound}; //The end cap to use
  Gdiplus::LineCap m_StartCap{Gdiplus::LineCapRound}; //The start cap to use
  Gdiplus::LineJoin m_LineJoin{Gdiplus::LineJoinRound}; //The line join style to use
  float m_fDashOffset{0}; //The dash offset to use
#else
  std::vector<D2D1::ColorF> m_colorPenStrokes; //The array of colors for this polyline (if set will be used instead of m_colorPen)
  D2D1::ColorF m_colorPen{0, 0, 255}; //The color of the pen to use
  D2D1::ColorF m_colorNode{0, 0, 0}; //The color of the node
  D2D1::ColorF m_colorSelectionNode{255, 0, 0}; //The color of the brush to use for selection nodes
  D2D1_CAP_STYLE m_DashCap{D2D1_CAP_STYLE_FLAT}; //The dash cap to use
  D2D1_DASH_STYLE m_DashStyle{D2D1_DASH_STYLE_SOLID}; //The dash style to use
  D2D1_CAP_STYLE m_EndCap{D2D1_CAP_STYLE_ROUND}; //The end cap to use
  D2D1_CAP_STYLE m_StartCap{D2D1_CAP_STYLE_ROUND}; //The start cap to use
  D2D1_LINE_JOIN m_LineJoin{D2D1_LINE_JOIN_ROUND}; //The line join style to use
  FLOAT m_fDashOffset{0}; //The dash offset to use
  FLOAT m_fMiterLimit{1}; //The limit of the thickness of the join on a mitered corner
#endif //#ifdef COSMCTRL_NOD2D
  float m_fNodeWidth{4}; //The width of the node to draw
  float m_fSelectionNodeWidth{4}; //The width of the node to draw if selected
  DrawingStyle m_DrawingStyle{DrawingStyle::LinesOnly}; //The drawing style to use for this polyline
  CString m_sToolTipText; //The tooltip text for this polyline
  bool m_bDraggable{false}; //Is this polyline draggable
  bool m_bVisible{true}; //Is this polyline currently visible
  bool m_bEditable{false}; //Is this polyline editable
  bool m_bHitTest{true}; //Is this polyline to be hit-tested
  int m_nMinZoomLevel{COSMCtrlHelper::OSMMinZoom}; //The minimum zoom level at which this polyline appears
  int m_nMaxZoomLevel{COSMCtrlHelper::OSMMaxZoom}; //The max zoom level at which this polyline appears
};


#endif //#ifndef __OSMCTRLPOLYLINE_H__
