/*
Module : OSMCtrl.cpp
Purpose: Implementation for an MFC GUI control which implements display of OpenStreetMap tiles
Created: PJN / 28-11-2009
History: PJN / 17-12-2009 1. Addition of a SetCenterPosition override method which allows the zoom level to be changed in one go along
                          with the center position of the map.
                          2. The control now handles a left mouse double click as a request to zoom in the map at the cursor position. 
                          This is consistent with how the main OpenStreetMap web site map operates
                          3. Addition of a GetPosition method which converts a client coordinates point into a latitude and longitude
                          position
                          4. Code now allows previous zoom level images to be used in a stretched fashion if the tile does not exist
                          at the current zoom level.
                          5. Code now allows next zoom level images to be used in a squeezed fashion if the tile does not exist
                          at the current zoom level.
                          6. The DrawScaleBar method now takes user preferences as to whether they are using the Metric (Kilometers) or
                          Imperial (Miles) system of measurement into account when drawing the scale
                          7. The DrawScaleBar method now displays a rounded scale value rather than the exact width of one tile
                          8. You can now customize the UserAgent string used by the class via the SetUserAgent method
                          9. You can now customize the ordering of which tiles are downloaded. By default the tiles which are closest
                          to the center of the map are downloaded first and the algorithm works outwards from there. If you prefer you
                          can download using a simple download Y outer loop and a left to right inner loop. This behaviour can be changed
                          via the SetDownloadOrder method.
                          10. The sample app now explicitly does not draw the scroll rose and zoom bar when the map is being printed.
                          11. By default the control now caches two additional rows/columns of tiles around the edge of the map in 
                          anticipation that the end user will scroll around the map. The number of additional row/columns which are 
                          cached in this mechanism can be configured via the new SetDownloadTilesEdgeCount method. If you were to use a 
                          value of 0, then only the tiles necessary to fully cover the visible client area will be downloaded.
                          12. The sample app now includes static MFC build configurations and the exe included in the download now
                          pulls in MFC this way.
         PJN / 10-01-2010 1. Fixed a bug in the display of the polar regions at low zoom settings
                          2. You can now scroll horizontally continuously around all longitudes
                          3. Fixed an issue where the scale would report a distance of 0 meters at zoom level 0.
                          4. Improved the overall responsiveness of the control by ensuring that the download thread returns quickly from any
                          blocking calls when the code needs to terminate the download thread. The trick which the code uses is to share
                          the Wininet session handle between the download thread and the main thread and when it comes time to kill the
                          download thread the session handle is closed via InternetCloseHandle in the main thread. This causes any derived 
                          Wininet handles which the download thread has created from this session handle to become invalid and causes any
                          blocking call on these handles to return with an error. This change also means that I do not need to completely
                          re-architect the code to use asynchronous Wininet calls. Having taken a look at how asynchronous Wininet works 
                          I'm real glad about this. 
                          5. Fixed an issue in DownloadThread() where the code would attempt to download a tile with an invalid Y value. 
                          This resulted in spurious TRACE messages in debug builds.
                          6. Fixed a bug where the OnPaint method was not using the values returned from GetDrawScrollRose(), 
                          GetDrawZoomBar() & GetDrawScaleBar().
                          7. As a really nice to have, the Zoom Bar is now drawn using transparency, just like on the main OpenStreetMap 
                          web site. As you may know GDI does not really handle transparency well, so I found it hard to find a good example 
                          on how to achieve this. Most of the examples I found were based on using a pre-existing image with an alpha channel
                          already provided. My COSMCtrl code does not require any pre-rendered bitmaps and I wanted to keep the code this way. 
                          I did not also want to throw away all the GDI code and replace it with GDI+ code just for this one effect. For 
                          those interested in how the code works, it uses ATL::CImage to generate a 32 bit ARGB DIB section bitmap and selects
                          that into a memory device context. Then I use standard GDI calls as before to render most of the parts of the zoom 
                          bar, then I directly access the bits of the DIB section via CImage::GetBits and CImage::GetPitch and modify the 
                          transparency of specific pixels to achieve the effect I want. At this point you also need to do the calculations to 
                          ensure the RGB components of the pixel are premultiplied by the alpha value as this is what is required. Then I use 
                          CDC::AlphaBlend to blend the just created zoom bar pixels onto the map data which has already been rendered to the 
                          device context. I was worried about the performance of this code as I need to iterate across all the pixels of the 
                          DIB section to achieve this effect, but some profiling on my main dev machine showed that it added very little 
                          overhead to the total rendering time. All of this code is in COSMCtrl::DrawZoomBar and hopefully you may find this
                          code of interest for your own projects which want to do runtime alpha blending. Finally if you do not want a
                          transparent zoom bar, you can disable this effect using SetUseTransparencyForZoomBar(FALSE).
                          8. You can now optionally draw rectangles around each rendered tile via SetDrawTileOutlines(TRUE).
                          9. The sample app now ships with a VC 2008 solution and project instead of VC 2005.
                          10. The sample app now uses separate cache directories for the different tile providers. In addition the cache 
                          directory is now located at " "CSIDL_LOCAL_APPDATA"\OSMCtrlApp\"Provider" "
                          11. The client area is now invalidated when you change tile provider.
                          12. The sample app now can display a tile properties dialog when you right mouse click on the map. This dialog uses
                          SysLink controls and as such will only show up if you build a Unicode version of the app. As such the binary 
                          included in the download is now the Static Unicode Release build. This tile properties dialog shows the following
                          tile attributes: Provider, URL to download from, local cache path, the physical dimensions of the tile vertically
                          and horizontally, The center position of the tile, the tile Coordinates and the Rerender and status URLs if any. 
                          Please note that the vertical distance is the actual great arc distance from the top left of the tile to the bottom 
                          left, while the horizontal distance is the great arc distance from the bottom left of the tile to the bottom right of 
                          the tile. It is important to emphasize this as at lower zoom levels the scale of the map becomes different depending 
                          on where you are located in the tile.
                          13. Updated the sample app icon to use the standard OpenStreetMap logo
         PJN / 16-03-2010 1. Updated the sample app to include copyright details in line with information at 
                          http://wiki.openstreetmap.org/wiki/Legal_FAQ#I_would_like_to_use_OpenStreetMap_maps._How_should_I_credit_you.3F. The
                          copyright details are displayed in the about dialog and on the main map itself. For client applications which use COSMCtrl, 
                          you should review these details to make sure you are compliant with the OpenStreetMap licensing requirements.
                          2. The zoom bar can now be shown using a standard Win32 slider control instead of custom draw code and this setting is
                          now the default. To change this behaviour you can use SetDrawZoomBarAsSlider(FALSE)
                          3. The setting to allow a zoom via double click has now been separated from the setting to allow a zoom via the wheel
                          mouse.
                          4. Sample app now shows the modified date of the tile in the Tile Properties dialog
                          5. The sample app now allows a specific tile to be refreshed via the view context menu and the main application menu
                          6. The scroll rose, zoom bar, and scale bar can all now be placed on the control in an arbitrary position. This is achieved
                          via two additional parameters to SetDrawScrollRose, SetDrawZoomBar and SetDrawScaleBar. These parameters allow an anchor
                          position on the map to be chosen for the control as well as an offset position.
                          7. Fixed an issue where the download thread would get created if the cache directory was not specified.
                          8. The class now supports markers, polylines, polygons and circles being overlaid on the map. In addition to just 
                          allowing static markers, polylines, polygons and circles to be added, the code now has comprehensive support for 
                          interactively editing, dragging, moving and deleted these items. The sample app has been extensively modified to demo 
                          these features.
                          9. Re-implemented all GDI drawing code with GDI+. This provides much better support for features such as transparency 
                          etc and will make it easier to add more features to the code base going forward.
                          10. The class now supports a full set of methods to calculate the distance between two points as well as calculate the end 
                          location from a start position given a certain distance and bearing. These methods implement C++ versions of Vincenty's 
                          Direct and Inverse algorithms. These methods are required for calculation of the scale bar as well as supporting dragging 
                          polygons and polylines. The sample app now uses these features to show the distance and bearing for the first polyline or
                          polygon which is selected in the status bar.
                          11. The code to draw the scale bar has been re-factored to allow further detailed client customization.
                          12. The various strings which are used by the class are now all stored in a string table. You need to ensure that all the 
                          string resources of ID "IDS_OSMCTRL*" and dialog resources of "IDD_OSMCTRL*" are included in your client application.
                          13. Fixed a bug where the download thread would not be created if you changed the location of the map using the cursor keys
                          14. The control now supports a Rectangular selection mode. When this mode is activated, you can select specific markers,
                          polylines and polygons on the map and in conjunction with support for the "Delete" button you can interactively edit the 
                          items on the map
                          15. Addition of a new comprehensive "Map Operations" dialog. This in conjunction with the rectangular selection mechanism
                          allows you to delete specific tiles, download specific tiles (optionally skipping files which have already been downloaded)
                          as well as support Mapnik re-render requests. This dialog uses a worker thread to remain responsive while these potentially
                          lengthy operations are taking place. In addition this dialog provides feedback via a progress control and a static text
                          notification area as the operation is taking place as well as cancelation support. This dialog on its own provides a good 
                          example on how to implement a responsive user interface while a lengthy operation is happening.
                          16. The control now supports "decimation" of polylines and polygons. This feature adds new nodes between all the existing 
                          nodes of a polyline or polygon. This can prove useful where you have a feature where the curvature of the earth can cause
                          distortion of the displayed object. By default this feature is provided for by double clicking on an editable polyline or
                          polygon. Thanks to Dermot McNally for prompting this update
                          17. The control now has support for drawing crosshairs at the center of the map.
                          18. The control now supports a ZoomToBounds method. This method takes two positions which the method will ensure will be
                          shown on the map at the highest possible zoom level. In conjunction with various new "GetBoundingRect*" methods you can 
                          now add your various items to the map and then zoom to those items. This avoids client code needing to explicitly handle 
                          zoom levels of center positions.
                          19. The arrow keys now scroll by a small amount while the likes of PageUp / PageDown keys will scroll by a tile at a time
         PJN / 09-04-2010 1. Reworked the OnPaint code to use classic double buffering via a GDI memory device context rather than using the GDI+
                          equivalent code. This restructuring of the code gives up a typical improvement of 125 ms down to 52 ms for a redraw over 
                          terminal services on my main dev machine (a 240% increase in performance) and a speed up of 140 ms down to 100 ms for a 
                          redraw on my primary graphics card on my main dev machine (a 40% increase in performance). My theory on why this new code 
                          is faster is probably due to the fact that we the code is now probably avoiding a code path in GDI+ which needs to convert 
                          from the internal GDI+ ARGB bitmap format to a format compatible with the display device context. Thanks to Frits van Veen 
                          for suggesting this update.
                          2. Following on from the previous performance optimization, the control now maintains an internal cache of 
                          Gdiplus::CachedBitmap* tiles in memory. The conversion process which GDI+ itself must perform to convert from its' internal 
                          format to the format compatible with the display device context proved to be a significant percentage of the time involved 
                          in drawing the control, hence the need for this caching optimization. The default size of this in memory cache array is 100 
                          elements which based on average type tiles in Ireland corresponds to a total application memory usage for the unicode release 
                          build of demo app of about 30 MB. You can change this limit by using the new SetMaxInMemoryCachedTiles method. If you want 
                          you can turn of this caching by calling SetMaxInMemoryCachedTiles(0) (not advised as it will adversely impact performance of 
                          the control). To give you further background on how you would pick a good value for this: at a resolution of 2560 * 1600, a 
                          maximized window will display approximately 60 OSM standard sized 256 * 256 tiles. This metric is where I arrived at 100 for 
                          the default value. With this optimization now in place and taking the previous optimization example, the drawing time for a 
                          full screen window on my primary graphics card on my main dev machine has speeded up of 140 ms to 100 ms to 15 ms. This 
                          corresponds to a 930% increase in performance! when compared to v1.03 of the control. You can really now see the difference 
                          in performance in the control when you drag the control and notice how really responsive it its. Again many thanks to Frits 
                          van Veen for doing all the low level performance testing and suggestions for this update.
                          3. If you hold down the control key while using the arrow keys to navigate around the map, the map now scrolls by a tile 
                          rather than a small pixel amount. This is consistent with how most other Map controls behave. Thanks to Frits van Veen for 
                          providing this update.
                          4. The amount which the ScrollToNorth/East/South/West() methods scroll by can now be controlled via a new SetScrollPixels()
                          method. The default value for this has been increased from 4 to 20 pixels. Thanks to Frits van Veen for providing this 
                          update.
         PJN / 10-04-2010 1. The control now only attempts to keep the position under the cursor when you do a mouse zoom if the position is in the 
                          client area, otherwise the control simple zooms in or out at the current center position
         PJN / 01-05-2010 1. The control now has the concept of a GPS position and recent track. This is achieved through the new member variables:
                          m_GPSTrack and m_colorGPSTrackPointer and the new methods of SetMaxGPSTracks, GetMaxGPSTracks & AddGPSTrack. In addition 
                          the sample app has been updated to use the authors GPSCom2 library to add comprehensive support to the sample app for GPS 
                          devices.
         PJN / 02-05-2010 1. Reduced the m_nMaxGPSTracks default value to 600, which if we are receiving GPS data every second corresponds to the last
                          10 minutes of the track will be shown.
                          2. Introduced the concept of bearing valid and speed valid to COSMCtrlPosition. In addition the COSMCtrl now visually 
                          indicates the lack of a bearing value by drawing the GPS triangle using black rather than the standard red color.
                          3. Introduced the concept of loss of a GPS fix to the control. If there is no fix, the GPS triangle will now only be drawn
                          as an outline rather than drawn filled.
                          4. The GPS track polyline's attributes are now explicitly set in the COSMCtrl constructor.
                          5. Fixed a clipping problem in the code which calculates the bounding rect for the GPS track.
                          6. Fixed a clipping problem when the GPS triangle moves a significant distance from one fix to the next.
                          7. Optimized the code in AddGPSTrack when perform invalidations of the client area.
         PJN / 02-05-2010 1. The control now plays the MB_ICONEXCLAMATION system sound whenever a GPS fix is obtained or lost
                          2. Fixed a bug in CGPSSettingsDlg::OnInitDialog where the "Do not use a GPS device" string was not been shown
         PJN / 22-07-2010 1. Updated the code to work with GDI+ v1.0 which is the version supported prior to Windows Vista. Thanks to Richard Dols for
                          reporting this issue.
                          2. Fixed a bug in DrawScaleBar where the actual distance shown would be out by a factor of two when the zoom value was 
                          anything other than zero. Thanks to Richard Dols for reporting this embarrassing mistake.
                          3. Fixed a redraw glitch for polylines by modifying COSMCtrl::GetBoundingRect(const COSMCtrlPolyline& polyline... and
                          COSMCtrl::GetBoundingRect(const COSMCtrlPolygon& polygon... to not be as aggressive with its inclusion of the extra margin.
                          4. The SetTileProvider method now calls Invalidate internally. This ensures that the map cleanly updates when you switch tile 
                          providers.
         PJN / 16-08-2010 1. The control now includes support for fractional zoom levels.
                          2. The control now uses the Windows 7 Animation API's if available for zoom level and position changes. If you do not require
                          animations then this behavior can be disabled.
                          3. Updated the sample app to compile cleanly on VC 2010. Also because the sample app now takes advantage of the Windows 7
                          Animation API's, the project files shipped in the sample are now for VC 2010 as this includes the Windows 7 SDK in the box. 
                          You can still use the control in older versions of VC though (>= VC 2005). For example if you want to exclude the Windows 7 
                          Animation support, you can define the "COSMCTRL_NOANIMATION" preprocessor value.
                          4. Included some missing status bar prompts for menu items in the sample app
         PJN / 23-08-2010 1. Fixed a bug in COSMCtrl::SetZoom which caused drawing glitches in the zoom bar
                          2. Made the default animation speed a bit faster
                          3. To avoid requirements for client code to create the control using the WS_CLIPCHILDREN style, the control now forces a repaint
                          on the zoom bar control and copyright control if they are in the clipping rect.
                          4. Fixed some maths problems in the COSMCtrl::HandleLButtonDownStandard when calculating the position of mouse clicks on the 
                          zoom bar
                          5. The sample app now includes support for logging tracks to a GPX file. This setting can be enabled in the GPS Settings dialog
                          by checking the "Enable GPX Track Log" checkbox. When this is enabled, then every time a new GPRMC NMEA sentence arrives from 
                          your GPS device it will be logged. The location of the GPX file is "CSIDL_LOCAL_APPDATA"\OSMCtrlApp\GPX\YYYYMMDD.gpx". For 
                          performance reasons the GPX file will only be written to file every 60th new waypoint. For most GPS devices this should correspond 
                          to every minute. The code will also ensure that if the app is closed and restarted, that any existing data in today's GPX file
                          will be preserved and a new track will be created. The code also correctly handles when a GPS fix has been lost under which it 
                          will create a new track segment in the GPX file. To support GPX files, a very simple set of MFC classes have been developed in
                          the header file "OSMCtrlGPX.H". These classes use MSXML 6 for its XML parsing and saving requirements. Please note that these 
                          classes are not comprehensive wrappers for all the features of GPX files, but just enough to support the COSMCtrl GPX 
                          requirements. If there is demand out there I may consider extending these classes to support all the features of GPX files.
                          6. The sample app now has support for importing GPX tracks and displaying them on the map. This can be achieved using the "File ->
                          Import GPX File" menu item.
         PJN / 06-09-2010 1. Updated the zip file to only include the VC 2010 project files. Thanks to Frits van Veen for reporting this issue.
                          2. Fixed up comment in sample app about inclusion of "enumser.h". Thanks to Frits van Veen for reporting this issue.
                          3. Updated the zip file to include missing OSMCtrlGPX.h. Thanks to Frits van Veen for reporting this issue.
                          4. Fixed a bug in COSMCtrlAppView::OnFileImportGPXFile in the sample app where the filename filter was not working correctly. Thanks 
                          to Frits van Veen for reporting this issue.
                          5. Updated the error message in the sample app when a GPX file fails to import. It now mentions the need for 1.1 schema GPX files.
                          6. The sample app now saves the track log any time the filename for the log changes e.g. midnight crossovers.
                          7. Updated the GPX class to support saving and loading GPX waypoints. In addition the sample app now displays these waypoints when you
                          import a GPX file.
                          8. Sample app now includes a more specific error message if GPSCom2 cannot be used because of a registration issue. Thanks 
                          to Frits van Veen for reporting this issue.
         PJN / 01-11-2010 1. Now supports the Mapquest tile provider as documented at http://devblog.mapquest.com/2010/08/24/mapquest-opens-tiles-and-style-enables-rapid-data-updates/
                          2. CreateCopyrightLinkCtrl method now works correctly if Windows is using a High DPI setting
                          3. Fixed an ASSERT in the Draw method which could occur when m_bDrawZoomBarAsSlider is TRUE.
                          4. The ControlAnchorPosition and offset CPoint values now refers to the edge of the child control which is closest to the border instead of the top left 
                          edge of the control. This makes it easier for client applications to control layout behaviour.
                          5. Fixed a bug in the tile properties dialog in the sample app which would report the distance in meters instead of in Kilometers or Miles.
                          6. The sample app has been updated to use the author's CNominatim library to add comprehensive support for Nominatim searches and Address lookups
                          7. Fixed a bug where the download thread would not be restarted if the control uses a Windows animation during panning
                          8. DrawScaleBar has been refactored somewhat and now works correctly if Windows is using a High DPI setting
         PJN / 20-11-2010 1. Added support for a simple event handler mechanism to the class
                          2. Reworked the sample app's Refresh Tile, Tile Properties and Address Lookup functions to allow the user to click on the part of the map where they want the 
                          operation performed rather than using the current cursor position. This means that using these options from the main menu is now more useful. Thanks to
                          Frits van Veen for suggesting this update.
                          3. Implemented a Goto Coordinates Dialog in the sample app
                          4. Updated the sample app to use a built in resource for the png default marker rather than a stand alone png file. Thanks to Frits van Veen for suggesting 
                          this update.
         PJN / 30-04-2011 1. Updated copyright details.
                          2. Implemented a "Delta" mode for the control. When this mode is enabled via SetDeltaMode(TRUE), any newly updated cached local tiles will be blinked with the
                          old tile it has been replaced with, when you do a refresh/redownload of the tile
                          3. In debug mode the code no longer reports any ERROR_INTERNET_OPERATION_CANCELLED download errors via TRACE.
                          4. Fixed a bug in the download code where you would get intermittent download errors because the file is locked because of how GDI+ locks the file when you
                          open a bitmap from it.
                          5. Sample app now handles the ISensorEvents::OnLeave event. This event occurs when your GPS sensor is removed from the system (such as unplugging the device if
                          it is GPS). The app now closes down GPS mode in the control if this occurs.
                          6. Implemented comprehensive support for Direct2D drawing in COSMCtrl. This support is provided by the new MFC D2D wrapper classes provided with Visual Studio
                          2010 SP1. By default this support is enabled so you will need to have VS 2010 and SP1 installed to take advantage. If you do not want this you can define 
                          COSMCTRL_NOD2D before pulling in the OSMCtrl classes and the code will fall back to using GDI+. This new code provides a very good sample to developers who are
                          looking to migrate their large GDI/GDI+ code bases to D2D as the before and after code in COSMCtrl can be compared to each other. If you want to exclude D2D 
                          from COSMCtrl define COSMCTRL_NOD2D and the code will fall back to the original GDI+ drawing logic. Performance testing on my laptop indicates that the 
                          drawing code time has reduced from c. 3 - 4 ms for a full screen redraw down to c. 0.3 ms i.e. a speed up of 900%!.
                          7. The m_Icons member variable in COSMCtrl now maintains COSMCtrlIcon pointers rather than actual instances. This fixes a bug in the code where previously it 
                          was not clear where ownership of the resources belong to. Now since they are pointers, client code is responsible for the lifetime of the icons
                          8. Fixed a bug in the sample app where the CMySensorEvents::OnDataUpdated and CMySensorEvents::OnLeave could cause an access violation if called during
                          shutdown of the sample app.
                          9. The GPS Location triangle is now drawn using some transparency.
                          10. Refactored the ancillary "COSMCtrl*" classes which the main control class depends on. All of these classes are now contained in separate modules.
                          11. GDI+ code path which draws fractionally zoomed tiles now uses in-memory tile cache.
                          12. When drawing fractionally zoomed tiles, the next zoom level tiles are now used in preference to the previous zoom level.
                          13. Reworked tile providers to use a new interface class of IOSMCtrlTileProvider. This allows easier addition of new tile providers.
                          14. Added support for Mapquest Open Aerial tiles via the new COSMCtrlMapquestAerialTileProvider class.
                          15. Optimized ClientToPosition and PositionToClient methods by allowing them to pass in the client rect.
                          16. Markers, Polylines, Polygons and circles can now be excluded from hit testing via a new "m_bHitTest" member variable
                          17. Fixed a memory leak in SetCacheDirectory.
                          18. The classes now have support for using WinHTTP instead of Wininet as the download API. Please note that if you use WinHTTP then the sample app 
                          will not provide Nominatim search and address lookup as currently the CNominatim class only supports Wininet and not WinHTTP.
         PJN / 30-11-2013 1. Updated copyright details.
                          2. Removed If Modified Since functionality from code.
                          3. Updated the code to compile cleanly on VC 2012 and 2013.
                          4. IOSMCtrlTileProvider interface now provides a GetDownloadPort method.
                          5. The sample app now provides Nominatim search and address lookup via WinHTTP
                          6. Reworked the code to use the author's WinHTTPWrappers classes for downloading in a truly async manner
                          7. Reworked all the OSMCtrl classes to use std::vector instead of MFC's CArray
                          8. Updated text on copyright control to refer to the new ODbL license for OpenStreetMap
                          9. Removed defunct Osmarender tile provider
                          10. Updated the URL used for the Mapquest Open Aerial Tile Provider as per http://developer.mapquest.com/web/products/open/map/
                          11. Updated the URL used for the OpenCycleMap Tile Provider as per http://wiki.openstreetmap.org/wiki/Slippy_map_tilenames
         PJN / 01-12-2013 1. Updated to use new versions of the author's MFCSensor, EnumSerialPorts and WinHTTPWrappers classes.
         PJN / 10-06-2014 1. Updated copyright details.
                          2. Updated to use new version of the author's WinHTTPWrappers classes. This addresses a bug where the drawing code for the control could draw with a tile
                          which has been partially downloaded. Thanks to Simon Orde for reporting this issue.
                          3. Fixed a bug in COSMCtrlWinHTTPRequest::OnCallbackComplete where tiles which did not download successfully were not deleted. Thanks to Simon Orde for
                          reporting this issue.
                          4. The duration of the animation duration can now be customized via a new virtual "CalculateTransitionDuration" method. Also by default the minimum the
                          minimum duration of a animation has been set to 0.5 seconds. Thanks to Simon Orde for reporting this issue.
                          5. Added some additional TRACE'ing code to the COSMCtrl::OnDownloadComplete method to show failed file operations in debug builds of the code.
                          6. Fixed the text in a number of TRACE statements in COSMCtrl::StartAsyncDownload.
                          7. Fixed a bug in the COSMCtrl::Refresh method where it did not pass the correct parameters to StartAsyncDownload which would sometimes not result in a 
                          refresh happening.
                          8. Removed unnecessary code to delete an unsuccessful download tile from COSMCtrl::OnDownloadComplete as this is already done in 
                          COSMCtrlWinHttpRequest::OnCallbackComplete.
                          9. Removed support for the in memory cache in the D2D code path as it cause drawing problems and is not really needed anyway.
                          10. Removed the delta mode feature from the control as there are better ways of looking back at OSM historical data online
                          11. Fixed an ASSERT which could occur in COSMCtrlWinHttpRequest::OnCallbackComplete when this method is called when the test application is being shutdown.
         PJN / 13-06-2014 1. Fixed a bug where you would occasionally get an access violation when there was a WinHTTP callback pending and your close the demo app.
         PJN / 25-06-2014 1. The HDOP (Horizontal Dilution Of Precision) value is now used to draw a circle around the current GPS location in the Direct2D code path.
                          2. The sample app now zooms to level 18 if the current zoom level is less than 18 upon received the first position in the gpx track. This zooming is
                          done with an animation.
                          3. Reworked the Direct2D code path which draws the current GPS position to use a nicer looking triangle as well as rename some virtual functions in this
                          area.
                          4. Fixed a bug in the DrawGPSTrack method where if the track was empty, the function would fail prematurely and would fail to draw the gps position
                          triangle.
                          5. Added support for a "m_fBearingOfTopOfMap" setting when using the D2D code path. This major new feature allows you to change the map orientation so 
                          that north is not necessarily at the top of the mode. Please note that to facilitate this a major rework of the D2D code path had to be done. The demo 
                          app has also been updated to allow the bearing value to be changed by dragging on the map. In addition the current Windows 7 Animations code path has 
                          been  extended to support animations for changing the bearing. For example if you zoom into a specific location on the map, then rotate the map 
                          interactively, then close and reopen the demo app, you will see an animation which rotates, zooms and recenters the map to the saved location.
                          6. All "BOOL bAnimation" parameters to methods have been replaced with a "double fAnimationDuration" parameter. This allows client code to directly 
                          specify the animation duration without needing to override a virtual method. Using a value of 0 for fAnimationDuration will not use an animation. This 
                          means that the CalculateTransitionDuration virtual method is now defunct and has been removed.
                          7. Removed most of the now unused "GetBoundingRect*" methods.
                          8. Added a new "m_bChangeBearingOfMap" GPS setting for the demo app. When this new setting is enabled, the map will be oriented in the direction of 
                          travel upon arrival of each new GPS location event. This new feature works for both the Windows 7 Sensors and the GPSCom2 code paths.
         PJN / 15-03-2016 1. Updated copyright details.
                          2. Updated code to compile with CNominatim v1.03, GPSCom2 v1.02, CEnumerateSerial v1.28, MFCSensor v1.03, WinHTTPWrappers v1.10.
                          3. Added SAL annotations to all the code.
                          4. Reworked all the virtual methods which take a "CPoint" parameter to now take a "const CPoint&" parameter
                          5. Improved the failure code paths in the COSMCtrl::Create method
                          6. Eliminated the need for the internal class COSMCtrlWinHTTPRequestFactory
                          7. Reviewed the text in all the TRACE calls.
                          8. Updated the code to compile cleanly on VC 2015.
         PJN / 23-07-2016 1. Fixed a bug in the GDI+ version of COSMCtrl::Draw where not all the image would appear correctly when drawn to a real Windows printer device 
                          context. Thanks to Evgheni Bobrov for reporting this issue.
                          2. Fixed an issue where Print Preview and Print functionality in the control would not work in the D2D code path if a marker was present on the control. 
                          Thanks to Evgheni Bobrov for reporting this issue.
         PJN / 01-11-2016 1. Removed the MapQuest providers from the demo app as they are no longer available.
         PJN / 06-03-2017 1. Updated copyright details.
         PJN / 16-04-2017 1. Verified the code compiles cleanly on VC 2017
                          2. Updated the control and sample app to support connecting through a proxy server. Thanks to Charalambos Ioannides for requesting this update.
         PJN / 27-08-2017 1. Fixed an issue in COSMCtrlAppView::OnViewGotoCoordinates in the demo app where the seconds value are not scaled correctly for one branch of an if 
                          statement. Thanks to Bart Duijndam for reporting this issue.
                          2. Fixed an issue in COSMCtrlMapOperationsDlg::UpdateZoom in the demo app where the "m_Tiles" array is not cleared down correctly. Thanks to Bart 
                          Duijndam for reporting this issue.
                          3. Fixed an issue in COSMCtrl::HandleAnimationTimerEvent where downloading of tiles would not restart when changing location.
                          4. Updated all the code to use nullptr instead of NULL.
                          5. Updated the sample app to support connecting through a proxy server using the current user's IE settings. Thanks to Bart Duijndam for reporting 
                          this issue.
                          6. Added code to the Map Operations dialog to show the size in meters of a pixel in a tile. Thanks to Bart Duijndam for reporting this issue.
         PJN / 01-10-2017 1. Removed an unused local variable from the COSMCtrl::Draw method. Thanks to Bart Duijndam for reporting this issue.
                          2. Reimplemented the D2D version of COSMCtrl::DrawTileNotAvailable to draw a bitmap. Thanks to Bart Duijndam for reporting this issue.
                          3. Added a Reset Map Orientation menu item to the sample app for the D2D code path
                          4. Rotating the map now uses "true angles" around the map center. So you can "grab" a point and turn it where you want it to be. Thanks to Bart 
                          Duijndam for reporting this issue.
                          5. Replaced CSortedArray usage throughout the code with std::vector.
                          6. Eliminated raw new call from COSMCtrl::SetCacheDirectory.
                          7. Removed the usage of the CCompareCOSMCtrlCachedTileCleanupItem, CCompareCOSMCtrlCachedTileCleanupItem2, CCompareCOSMCtrlCachedTilePosition
                          & CCompareCOSMCtrlDownloadTileItem functors throughout the code and replaced with lambdas.
                          8. Implemented a zoom to selection rectangle in the demo app.
                          9. Fixed an issue in COSMCtrl::HandleLButtonDownStandard, COSMCtrl::OnMouseMove, COSMCtrl::OnChar & COSMCtrl::OnMouseWheel where an incorrect 
                          animation duration value was being used.
                          10. Reworked the GPX import code to apply a red to blue color gradient for the track stroke to indicate speed at that time. Please note that
                          this support was only added to the D2D code path.
         PJN / 10-10-2017 1. Updated the sample app to work with the latest version of the author's CNominatim classes.
         PJN / 13-11-2017 1. Fixed a problem in the demo app's COSMCtrlAppView::OnRMC method where it was not using the correct declaration from the GPSCom2::CClient class
                          2. "Change Bearing Of Map" GPS setting in the demo app is now persisted.
                          3. Updated the demo app to implement a 10 second delay to recentering when the "Center the map on the current GPS location" setting is turned on
         PJN / 22-06-2019 1. Updated code to compile with CNominatim v1.06, GPSCom2 v1.08, CEnumerateSerial v1.37, MFCSensor v1.05, WinHTTPWrappers v1.18.
                          2. Updated copyright details
                          3. Fixed a number of C++ core guidelines compiler warnings. These changes mean that the code will now only compile on VC 2017 or later.
                          4. Replaced BOOL with bool throughout the codebase
                          5. Replaced NULL with nullptr throughout the codebase
                          6. Updated the code to clean compile on VC 2017 & 2019
                          7. Now uses Eraseremove idiom as necessary throughout codebase
         PJN / 13-10-2019 1. Fixed a number of compiler warnings when the code is compiled with VS 2019 Preview
         PJN / 30-05-2020 1. Updated copyright details.
                          2. Fixed Clang-Tidy static code analysis warnings in the code.
                          3. Made more COSMCtrl methods virtual.
                          4. Removed call to virtual SetTileProvider method in COSMCtrl constructor.
                          5. Improved the D2D codepath to recreate D2D resources.
                          6. Fixed a bug in COSMCtrl::DeleteSelectedItems when deleting a circle.
                          7. Replaced COSMCtrlCachedTileEquivalent functor with a lambda in COSMCtrl::GetCachedBitmap.
                          8. COSMCtrl::PerformTileCacheMaintenance now uses erase-remove idiom to improve
                          performance.
                          9. D2D codepath now implements a tile cache on the GPU just like the GDI+ code path.
                          10. The default size of the tile cache has now been increased to 1000.
                          11. Move the drawing code for circles, markers, polygons & polylines into virtual
                          methods in the respective classes instead of the COSMCtrl class.
                          12. The "m_SelectionPolygon" member variable has been made a std::unique_ptr<COSMCtrlPolygon> 
                          to allow easier end-user customization. Also added a new CreateSelectionPolygon method to 
                          customize the creation of the selection rectangle.
                          13. The "m_Markers" array is now an array of std::shared_ptr<COSMCtrlMarker> to allow easier
                          end-user customization.
                          14. The "m_Polygons" array is now an array of std::shared_ptr<COSMCtrlPolygon> to allow easier
                          end-user customization.
                          15. The "m_Polylines" array is now an array of std::shared_ptr<COSMCtrlPolygon> to allow easier
                          end-user customization.
                          16. The "m_Circles" array is now an array of std::shared_ptr<COSMCtrlCircle> to allow easier
                          end-user customization.
                          17. The "m_GPSTrack" member variable has been made a std::unique_ptr<COSMCtrlPolyline> 
                          to allow easier end-user customization. Also added a new CreateGPSTrack method to customize 
                          the creation of the selection rectangle.
                          18. Changed the default animation time to 500 ms.
         PJN / 01-06-2020 1. Fixed a problem where zooming would not work in COSMCtrl::OnMouseWheel when the current zoom level was a non-integer.
                          2. Fixed a problem where zooming would not work in COSMCtrl::HandleLButtonDownStandard when the zoom level was a non-integer.
                          3. Replaced the Windows 7 Animation API's with good old WM_TIMER logic. This improves the reliability of the animation
                          logic in the control.
         PJN / 19-06-2020 1. Optimized the code to not attempt to restart downloading tiles while dragging the map or dragging on the zoombar.
         PJN / 25-08-2022 1. Updated copyright details.
                          2. Updated the code to use C++ uniform initialization for all variable declarations.

Copyright (c) 2009 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Includes /////////////////////////////////////////////////

#include "stdafx.h"
#include "OSMCtrl.h"
#ifndef __ATLFILE_H__
#pragma message("To avoid this message, please put atlfile.h in your pre compiled header (normally stdafx.h)")
#include <atlfile.h>
#endif //#ifndef __ATLFILE_H__
#ifndef __ATLIMAGE_H__
#pragma message("To avoid this message, please put atlimage.h in your pre compiled header (normally stdafx.h)")
#include <atlimage.h>
#endif //#ifndef __ATLIMAGE_H__
#include <afxglobals.h>
#ifndef _VECTOR_
#pragma message("To avoid this message, please put vector in your pre compiled header (normally stdafx.h)")
#include <vector>
#endif //#ifndef _VECTOR_
#ifndef _ALGORITHM_
#pragma message("To avoid this message, please put algorithm in your pre compiled header (normally stdafx.h)")
#include <algorithm>
#endif //#ifndef _ALGORITHM_


//////////////////// Macros / Defines /////////////////////////////////////////

constexpr UINT WM_OSMCTRL_CREATE_CHILD_CONTROLS = WM_APP + 1;
constexpr UINT WM_OSMCTRL_WINHTTP_DOWNLOAD_COMPLETE = WM_APP + 2;
constexpr UINT_PTR OSMCTRL_ANIMATION_TIMER_ID = 1;


//////////////////// Implementation ///////////////////////////////////////////

void COSMCtrlWinHTTPRequest::SetOSMCtrl(_In_ COSMCtrl* pOSMCtrl) noexcept
{
  m_pOSMCtrl = pOSMCtrl;
}

HRESULT COSMCtrlWinHTTPRequest::OnCallbackComplete(_In_ HRESULT hr, _In_ HINTERNET hInternet, _In_ DWORD dwInternetStatus, _In_opt_ LPVOID lpvStatusInformation, _In_ DWORD dwStatusInformationLength)
{
  UNREFERENCED_PARAMETER(hInternet);
  UNREFERENCED_PARAMETER(dwInternetStatus);
  UNREFERENCED_PARAMETER(lpvStatusInformation);
  UNREFERENCED_PARAMETER(dwStatusInformationLength);

  //Free up the resources we have used
  ReleaseResources();

  //Delete any files which failed to download
  if (hr != S_OK)
    DeleteDownloadedFile();

  //Inform the UI thread that we have finished via a windows message
#pragma warning(suppress: 26496)
  AFXASSUME(m_pOSMCtrl != nullptr);
  if (IsWindow(m_pOSMCtrl->GetSafeHwnd()))
#pragma warning(suppress: 26472)
    m_pOSMCtrl->PostMessage(WM_OSMCTRL_WINHTTP_DOWNLOAD_COMPLETE, 0, static_cast<LPARAM>(hr));

  return S_OK;
}


#pragma warning(suppress: 26433 26440)
BEGIN_MESSAGE_MAP(COSMCtrl, CStatic) //NOLINT(modernize-avoid-c-arrays)
  ON_WM_PAINT()
  ON_WM_ERASEBKGND()
  ON_WM_DESTROY()
  ON_WM_LBUTTONDOWN()
  ON_WM_LBUTTONUP()
  ON_WM_MOUSEMOVE()
  ON_WM_KEYDOWN()
  ON_WM_CHAR()
  ON_WM_MOUSEWHEEL()
  ON_WM_LBUTTONDBLCLK()
  ON_WM_VSCROLL()
#pragma warning(suppress: 26454)
  ON_NOTIFY_EX_RANGE(TTN_NEEDTEXT, 0, 0xFFFF, &COSMCtrl::OnToolTipText)
#pragma warning(suppress: 26454)
  ON_NOTIFY(NM_CLICK, COPYRIGHT_ID, &COSMCtrl::OnNMClickCopyright)
  ON_WM_SIZE()
  ON_MESSAGE(WM_OSMCTRL_CREATE_CHILD_CONTROLS, &COSMCtrl::OnCreateChildControls)
  ON_MESSAGE(WM_OSMCTRL_WINHTTP_DOWNLOAD_COMPLETE, &COSMCtrl::OnDownloadComplete)
#ifndef COSMCTRL_NOD2D
  ON_REGISTERED_MESSAGE(AFX_WM_DRAW2D, &COSMCtrl::OnDraw2D)
  ON_REGISTERED_MESSAGE(AFX_WM_RECREATED2DRESOURCES, &COSMCtrl::OnRecreatedResources)
#endif //#ifndef COSMCTRL_NOD2D
  ON_WM_TIMER()
END_MESSAGE_MAP()

#pragma warning(suppress: 26455)
COSMCtrl::COSMCtrl() : m_bGPSFix{false},
                     #ifdef COSMCTRL_NOD2D
                       m_colorGPSTrackPointer{192, 255, 0, 0},
                       m_colorGPSTrackNoBearingPointer{192, 0, 0, 0},
                     #endif //#ifdef COSMCTRL_NOD2D
                       m_fZoom{3},
                       m_fStartZoom{0},
                       m_fFinalZoom{0},
                       m_fDefaultAnimationDuration{0.5},
                       m_pTileProvider{nullptr},
                       m_bDownloadTiles{true},
                       m_bMouseCapturedForDrag{false},
                       m_bMouseCapturedForZoom{false},
                       m_bDrawTileOutlines{false},
                       m_bDrawScrollRose{true},
                       m_ScrollRoseAnchorPosition{ControlAnchorPosition::TopLeft},
                       m_ptOffsetScrollRose{8, 8},
                       m_bDrawZoomBar{true},
                       m_ZoomBarAnchorPosition{ControlAnchorPosition::TopLeft},
                       m_ptOffsetZoomBar{15, 69},
                       m_bTransparencyForZoomBar{true},
                       m_bDrawScaleBar{true},
                       m_ScaleBarAnchorPosition{ControlAnchorPosition::BottomLeft},
                       m_ptOffsetScaleBar{15, 15},
                       m_bAllowDrag{true},
                       m_bAllowKeyboard{true},
                       m_bDeltaMode{false},
                       m_bAllowMouseZoom{true},
                       m_bAllowDoubleClickZoom{true},
                       m_bAllowPreviousZoomStretch{true},
                       m_bAllowNextZoomSqueeze{true},
                       m_DownloadOrder{DownloadOrder::ClosestToCenterFirst},
                       m_nDownloadTilesEdgeCount{2},
                       m_bDrawZoomBarAsSlider{true},
                       m_MouseDragItem{MapItem::None},
                       m_ToolHitItem{MapItem::None},
                       m_nMouseDragItemIndex{-1},
                       m_nMouseDragSubItemIndex{-1},
                       m_DragOffset{0, 0},
                       m_bModifyingCircleRadius{false},
                       m_MapMode{MapMode::Normal},
                       m_bDrawCenterCrossHairs{true},
                       m_bDrawCopyright{true},
                       m_CopyrightAnchorPosition{ControlAnchorPosition::BottomRight},
                       m_ptOffsetCopyright{15, 15},
                       m_ScaleBarUnits{ScaleBarUnits::UseOSDefault},
                       m_nMaxCachedTiles{1000},
                       m_nInMemoryCachedTilesInsertionCounter{0},
                    #ifndef COSMCTRL_NOD2D
                       m_pEmptyTile{nullptr},
                    #endif //#ifdef COSMCTRL_NOD2D
                       m_nScrollPixels{20},
                       m_nMaxGPSTracks{600},
                       m_pEventHandler{nullptr},
                       m_fBearingOfTopOfMap{0},
                       m_fBearingAtTopOfMapMouseCapture{0},
                       m_fStartBearingAtTopOfMap{0},
                       m_fFinalBearingAtTopOfMap{0},
                       m_bAnimations{true},
                       m_bDownloading{false},
                       m_bInitializedWinHTTPRequest{false},
                       m_nLastSetCenterTickCount{0},
                       m_nStartAnimationTickCount{0},
                       m_nAnimationTimer{0},
                       m_nAnimationInterval{50},
                       m_fCurrentAnimationDuration{0},
                       m_bDoingBearingAnimation{false},
                       m_bDoingZoomAnimation{false},
                       m_bDoingPositionAnimation{false}
{
  m_WinHTTPRequest.SetOSMCtrl(this);
  m_sUserAgent = AfxGetAppName();
  m_rNorthScrollRose.SetRectEmpty();
  m_rSouthScrollRose.SetRectEmpty();
  m_rEastScrollRose.SetRectEmpty();
  m_rWestScrollRose.SetRectEmpty();
  m_rZoomInZoomBar.SetRectEmpty();
  m_rZoomOutZoomBar.SetRectEmpty();
  m_rZoomBar.SetRectEmpty();

#ifndef COSMCTRL_NOD2D
  EnableD2DSupport();
#endif //#ifndef COSMCTRL_NOD2D

  m_pTileProvider = &m_DefaultTileProvider;
}

ID2D1Factory* COSMCtrl::GetD2DFactory()
{
  return AfxGetD2DState()->GetDirect2dFactory();
}

BOOL COSMCtrl::Create(LPCTSTR lpszText, DWORD dwStyle, const RECT& rect, CWnd* pParentWnd, UINT nID)
{
  //Let the base class do its thing
  if (!__super::Create(lpszText, dwStyle, rect, pParentWnd, nID))
    return FALSE;

  if (m_SelectionPolygon == nullptr)
  {
    if (!CreateSelectionPolygon())
      return FALSE;
  }

  if (m_GPSTrack == nullptr)
  {
    if (!CreateGPSTrack())
      return FALSE;
  }

  //Enable tooltips for this control
  EnableToolTips(TRUE);

  //Asynchronously create the child controls
  PostMessage(WM_OSMCTRL_CREATE_CHILD_CONTROLS);

  return TRUE;
}

LRESULT COSMCtrl::OnCreateChildControls(WPARAM /*wParam*/, LPARAM /*lParam*/)
{
  //Create the child slider control also if necessary
  if (m_bDrawZoomBar && m_bDrawZoomBarAsSlider && (m_ctrlZoomBar.GetSafeHwnd() == nullptr))
    CreateZoomBarSlider();

  //Create the copyright link control also if necessary
  if (m_bDrawCopyright && (m_ctrlCopyright.GetSafeHwnd() == nullptr))
    CreateCopyrightLinkCtrl();

  return 0L;
}

#ifndef COSMCTRL_NOD2D
bool COSMCtrl::SetBearingOfTopOfMap(_In_ double fBearingOfTopOfMap, _In_ double fAnimationDuration, _In_ bool bRestartDownloading)
{
  return SetCenterAndZoomAndBearing(m_CenterPosition, GetZoom(), fBearingOfTopOfMap, fAnimationDuration, bRestartDownloading);
}
#endif //#ifndef COSMCTRL_NOD2D

bool COSMCtrl::SetZoom(_In_ double fZoom, _In_ double fAnimationDuration, _In_ bool bRestartDownloading)
{
  return SetCenterAndZoom(m_CenterPosition, fZoom, fAnimationDuration, bRestartDownloading);
}

bool COSMCtrl::ZoomToBounds(_In_ const COSMCtrlPosition& position1, _In_ const COSMCtrlPosition& position2, _In_ double fAnimationDuration, _In_ bool bRestartDownloading)
{
  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

  //First thing we need to do is find the extreme latitude and longitude of the two positions
  double fTopLatitude{-91};
  double fBottomLatitude{91};
  double fLeftLongitude{181};
  double fRightLongitude{-181};
  if (position1.m_fLatitude > fTopLatitude)
    fTopLatitude = position1.m_fLatitude;
  if (position1.m_fLatitude < fBottomLatitude)
    fBottomLatitude = position1.m_fLatitude;
  if (position1.m_fLongitude < fLeftLongitude)
    fLeftLongitude = position1.m_fLongitude;
  if (position1.m_fLongitude > fRightLongitude)
    fRightLongitude = position1.m_fLongitude;
  if (position2.m_fLatitude > fTopLatitude)
    fTopLatitude = position2.m_fLatitude;
  if (position2.m_fLatitude < fBottomLatitude)
    fBottomLatitude = position2.m_fLatitude;
  if (position2.m_fLongitude < fLeftLongitude)
    fLeftLongitude = position2.m_fLongitude;
  if (position2.m_fLongitude > fRightLongitude)
    fRightLongitude = position2.m_fLongitude;

  //Form the extreme bounding positions
  const COSMCtrlPosition topLeft{fLeftLongitude, fTopLatitude};
  const COSMCtrlPosition bottomRight{fRightLongitude, fBottomLatitude};

  //Next find the center position of the extreme latitudes and longitudes found above
  const double fCenterLatitude{(fTopLatitude + fBottomLatitude) / 2};
  const double fCenterLongitude{(fLeftLongitude + fRightLongitude) / 2};
  const COSMCtrlPosition centerPosition{fCenterLongitude, fCenterLatitude};

  //Get the current client size
  CRect rClient;
  GetClientRect(&rClient);

  //Starting from the highest zoom level going down, determine which will contain position1 and position2 in the client area
  int nFoundZoomLevel{-1};
  for (int i=OSMMaxZoom; (i>=OSMMinZoom) && (nFoundZoomLevel == -1); i--)
  {
    //Determine if the current zoom level will contain the extreme positions
  #ifdef COSMCTRL_NOD2D
    Gdiplus::PointF ptTopLeft;
    Gdiplus::PointF ptBottomRight;
    if (PositionToClient(topLeft, centerPosition, i, rClient, ptTopLeft) && PositionToClient(bottomRight, centerPosition, i, rClient, ptBottomRight) && rClient.PtInRect(CPoint{static_cast<int>(ptTopLeft.X), static_cast<int>(ptTopLeft.Y)}) && rClient.PtInRect(CPoint{static_cast<int>(ptBottomRight.X), static_cast<int>(ptBottomRight.Y)}))
  #else
    CD2DPointF ptTopLeft;
    CD2DPointF ptBottomRight;
    if (PositionToClient(topLeft, centerPosition, i, m_fBearingOfTopOfMap, rClient, ptTopLeft) && PositionToClient(bottomRight, centerPosition, i, m_fBearingOfTopOfMap, rClient, ptBottomRight) && rClient.PtInRect(ptTopLeft) && rClient.PtInRect(ptBottomRight))
  #endif //#ifdef COSMCTRL_NOD2D
    {
      nFoundZoomLevel = i;
    }
  }

  //Now update the center and zoom level of the map
  if (nFoundZoomLevel != -1)
    bSuccess = SetCenterAndZoom(centerPosition, nFoundZoomLevel, fAnimationDuration, bRestartDownloading);

  return bSuccess;
}

bool COSMCtrl::SetDrawScrollRose(_In_ bool bDrawScrollRose)
{
  //What will be the return value from this function (assume the best)
  bool bSuccess{true};

  const bool bChange{bDrawScrollRose != m_bDrawScrollRose};
  m_bDrawScrollRose = bDrawScrollRose;

  //Force a redraw if required
  if (bChange && GetSafeHwnd())
  {
    Invalidate(FALSE);

    //Force a recreation of the zoom bar
    if (m_ctrlZoomBar.GetSafeHwnd())
    {
      m_ctrlZoomBar.SendMessage(WM_CLOSE);
      bSuccess = CreateZoomBarSlider();
    }
  }

  return bSuccess;
}

bool COSMCtrl::SetScrollRoseAnchorPosition(_In_ ControlAnchorPosition controlAnchorPosition)
{
  //What will be the return value from this function (assume the best)
  bool bSuccess{true};

  const bool bChange{controlAnchorPosition != m_ScrollRoseAnchorPosition};
  m_ScrollRoseAnchorPosition = controlAnchorPosition;

  //Force a redraw if required
  if (bChange && GetSafeHwnd())
  {
    Invalidate(FALSE);

    //Force a recreation of the zoom bar
    if (m_ctrlZoomBar.GetSafeHwnd())
    {
      m_ctrlZoomBar.SendMessage(WM_CLOSE);
      bSuccess = CreateZoomBarSlider();
    }
  }

  return bSuccess;
}

bool COSMCtrl::SetScrollRoseAnchorPositionOffset(_In_ const CPoint& ptOffset)
{
  //What will be the return value from this function (assume the best)
  bool bSuccess{true};

  const BOOL bChange{ptOffset != m_ptOffsetScrollRose};
  m_ptOffsetScrollRose = ptOffset;

  //Force a redraw if required
  if (bChange && m_bDrawScrollRose && GetSafeHwnd())
  {
    Invalidate(FALSE);

    //Force a recreation of the zoom bar
    if (m_ctrlZoomBar.GetSafeHwnd())
    {
      m_ctrlZoomBar.SendMessage(WM_CLOSE);
      bSuccess = CreateZoomBarSlider();
    }
  }

  return bSuccess;
}

bool COSMCtrl::SetDrawZoomBar(_In_ bool bDrawZoomBar)
{
  //What will be the return value from this function (assume the best)
  bool bSuccess{true};

  const bool bChange{bDrawZoomBar != m_bDrawZoomBar};
  m_bDrawZoomBar = bDrawZoomBar;

  //Force a redraw if required
  if (bChange && GetSafeHwnd())
  {
    Invalidate(FALSE);
    if (bDrawZoomBar)
    {
      if (m_bDrawZoomBarAsSlider)
      {
        if (m_ctrlZoomBar.GetSafeHwnd() == nullptr)
          bSuccess = CreateZoomBarSlider();
      }
      else
      {
        if (m_ctrlZoomBar.GetSafeHwnd() != nullptr)
          m_ctrlZoomBar.SendMessage(WM_CLOSE);
      }
    }
    else
    {
      if (m_ctrlZoomBar.GetSafeHwnd() != nullptr)
        m_ctrlZoomBar.SendMessage(WM_CLOSE);
    }
  }

  return bSuccess;
}

bool COSMCtrl::SetZoomBarAnchorPosition(_In_ ControlAnchorPosition controlAnchorPosition)
{
  //What will be the return value from this function (assume the best)
  bool bSuccess{true};

  const bool bChange{controlAnchorPosition != m_ZoomBarAnchorPosition};
  m_ZoomBarAnchorPosition = controlAnchorPosition;

  //Force a redraw if required
  if (bChange && m_bDrawZoomBar && GetSafeHwnd())
  {
    Invalidate(FALSE);
    if (m_bDrawZoomBar)
    {
      if (m_bDrawZoomBarAsSlider)
      {
        if (m_ctrlZoomBar.GetSafeHwnd() == nullptr)
          bSuccess = CreateZoomBarSlider();
      }
      else
      {
        if (m_ctrlZoomBar.GetSafeHwnd() != nullptr)
          m_ctrlZoomBar.SendMessage(WM_CLOSE);
      }
    }
    else
    {
      if (m_ctrlZoomBar.GetSafeHwnd() != nullptr)
        m_ctrlZoomBar.SendMessage(WM_CLOSE);
    }
  }

  return bSuccess;
}

bool COSMCtrl::SetZoomBarAnchorPositionOffset(_In_ const CPoint& ptOffset)
{
  //What will be the return value from this function (assume the best)
  bool bSuccess{true};

  const BOOL bChange{ptOffset != m_ptOffsetZoomBar};
  m_ptOffsetZoomBar = ptOffset;

  //Force a redraw if required
  if (bChange && m_bDrawZoomBar && GetSafeHwnd())
  {
    Invalidate(FALSE);
    if (m_bDrawZoomBar)
    {
      if (m_bDrawZoomBarAsSlider)
      {
        if (m_ctrlZoomBar.GetSafeHwnd() == nullptr)
          bSuccess = CreateZoomBarSlider();
      }
      else
      {
        if (m_ctrlZoomBar.GetSafeHwnd() != nullptr)
          m_ctrlZoomBar.SendMessage(WM_CLOSE);
      }
    }
    else
    {
      if (m_ctrlZoomBar.GetSafeHwnd() != nullptr)
        m_ctrlZoomBar.SendMessage(WM_CLOSE);
    }
  }

  return bSuccess;
}

bool COSMCtrl::SetDrawZoomBarAsSlider(_In_ bool bDrawZoomBarAsSlider)
{
  //What will be the return value from this function (assume the best)
  bool bSuccess{true};

  const bool bChange{bDrawZoomBarAsSlider != m_bDrawZoomBarAsSlider};
  m_bDrawZoomBarAsSlider = bDrawZoomBarAsSlider;

  //Create or destroy the slider as necessary
  if (bChange && GetSafeHwnd())
  {
    Invalidate(FALSE);
    if (m_bDrawZoomBar)
    {
      if (m_bDrawZoomBarAsSlider)
      {
        if (m_ctrlZoomBar.GetSafeHwnd() == nullptr)
          bSuccess = CreateZoomBarSlider();
      }
      else
      {
        if (m_ctrlZoomBar.GetSafeHwnd() != nullptr)
          m_ctrlZoomBar.SendMessage(WM_CLOSE);
      }
    }
    else
    {
      if (m_ctrlZoomBar.GetSafeHwnd() != nullptr)
        m_ctrlZoomBar.SendMessage(WM_CLOSE);
    }
  }
  
  return bSuccess;
}

void COSMCtrl::SetUseTransparencyForZoomBar(_In_ bool bUseTransparencyForZoomBar)
{
  const bool bChange{bUseTransparencyForZoomBar != m_bTransparencyForZoomBar};
  m_bTransparencyForZoomBar = bUseTransparencyForZoomBar;

  //Force a redraw if required
  if (bChange && GetSafeHwnd())
    Invalidate(FALSE);
}

void COSMCtrl::SetMapMode(_In_ MapMode mode)
{
  const bool bChange{mode != m_MapMode};
  m_MapMode = mode;

  //Force a redraw if required
  if (bChange && GetSafeHwnd())
    Invalidate(FALSE);
}

void COSMCtrl::SetDrawScaleBar(_In_ bool bDrawScaleBar)
{
  const bool bChange{bDrawScaleBar != m_bDrawScaleBar};
  m_bDrawScaleBar = bDrawScaleBar;

  //Force a redraw if required
  if (bChange && GetSafeHwnd())
    Invalidate(FALSE);
}

void COSMCtrl::SetScaleBarAnchorPosition(_In_ ControlAnchorPosition controlAnchorPosition)
{
  const bool bChange{controlAnchorPosition != m_ScaleBarAnchorPosition};
  m_ScaleBarAnchorPosition = controlAnchorPosition;

  //Force a redraw if required
  if (bChange && m_bDrawScaleBar && GetSafeHwnd())
    Invalidate(FALSE);
}

void COSMCtrl::SetScaleBarAnchorPositionOffset(_In_ const CPoint& ptOffset)
{
  const BOOL bChange{ptOffset != m_ptOffsetScaleBar};
  m_ptOffsetScaleBar = ptOffset;

  //Force a redraw if required
  if (bChange && m_bDrawScaleBar && GetSafeHwnd())
    Invalidate(FALSE);
}

bool COSMCtrl::SetDrawCopyright(_In_ bool bDrawCopyright)
{
  //What will be the return value from this function (assume the best)
  bool bSuccess{true};

  const bool bChange{bDrawCopyright != m_bDrawCopyright};
  m_bDrawCopyright = bDrawCopyright;

  //Force a redraw if required
  if (bChange && GetSafeHwnd())
  {
    Invalidate(FALSE);
    if (m_bDrawCopyright)
    {
      if (m_ctrlCopyright.GetSafeHwnd() == nullptr)
        bSuccess = CreateCopyrightLinkCtrl();
    }
    else
    {
      if (m_ctrlCopyright.GetSafeHwnd() != nullptr)
        m_ctrlCopyright.SendMessage(WM_CLOSE);
    }
  }

  return bSuccess;
}

bool COSMCtrl::SetCopyrightAnchorPosition(_In_ ControlAnchorPosition controlAnchorPosition)
{
  //What will be the return value from this function (assume the best)
  bool bSuccess{true};

  const bool bChange{controlAnchorPosition != m_CopyrightAnchorPosition};
  m_CopyrightAnchorPosition = controlAnchorPosition;

  //Force a redraw if required
  if (bChange && m_bDrawCopyright && GetSafeHwnd())
  {
    Invalidate(FALSE);
    if (m_bDrawCopyright)
    {
      if (m_ctrlCopyright.GetSafeHwnd() == nullptr)
        bSuccess = CreateCopyrightLinkCtrl();
    }
    else
    {
      if (m_ctrlCopyright.GetSafeHwnd() != nullptr)
        m_ctrlCopyright.SendMessage(WM_CLOSE);
    }
  }

  return bSuccess;
}

bool COSMCtrl::SetCopyrightAnchorPositionOffset(_In_ const CPoint& ptOffset)
{
  //What will be the return value from this function (assume the best)
  bool bSuccess{true};

  const BOOL bChange{ptOffset != m_ptOffsetCopyright};
  m_ptOffsetCopyright = ptOffset;

  //Force a redraw if required
  if (bChange && m_bDrawCopyright && GetSafeHwnd())
  {
    Invalidate(FALSE);
    if (m_bDrawCopyright)
    {
      if (m_ctrlCopyright.GetSafeHwnd() == nullptr)
        bSuccess = CreateCopyrightLinkCtrl();
    }
    else
    {
      if (m_ctrlCopyright.GetSafeHwnd() != nullptr)
        m_ctrlCopyright.SendMessage(WM_CLOSE);
    }
  }

  return bSuccess;
}

bool COSMCtrl::SetCenter(_In_ const COSMCtrlPosition& position, _In_ double fAnimationDuration, _In_ bool bRestartDownloading)
{
  //Simply delegate to the other version of the method
#ifdef COSMCTRL_NOD2D
  return SetCenterAndZoomAndBearing(position, GetZoom(), 0, fAnimationDuration, bRestartDownloading);
#else
  return SetCenterAndZoomAndBearing(position, GetZoom(), GetBearingOfTopOfMap(), fAnimationDuration, bRestartDownloading);
#endif //#ifdef COSMCTRL_NOD2D
}

bool COSMCtrl::SetCenterAndZoom(_In_ const COSMCtrlPosition& position, _In_ double fZoom, _In_ double fAnimationDuration, _In_ bool bRestartDownloading)
{
  //Simply delegate to the other version of the method
#ifdef COSMCTRL_NOD2D
  return SetCenterAndZoomAndBearing(position, fZoom, 0, fAnimationDuration, bRestartDownloading);
#else
  return SetCenterAndZoomAndBearing(position, fZoom, GetBearingOfTopOfMap(), fAnimationDuration, bRestartDownloading);
#endif //#ifdef COSMCTRL_NOD2D
}

bool COSMCtrl::SetCenterAndZoomAndBearing(_In_ const COSMCtrlPosition& position, _In_ double fZoom, _In_ double fBearingOfTopOfMap, _In_ double fAnimationDuration, _In_ bool bRestartDownloading)
{
  //Validate our parameters
  if ((fZoom < OSMMinZoom) || (fZoom > OSMMaxZoom))
    return false;

  //Normalize the bearing value to the range (0, 360)
  while (fBearingOfTopOfMap < 0)
    fBearingOfTopOfMap += 360;
  while (fBearingOfTopOfMap > 360)
    fBearingOfTopOfMap -= 360;

  //Fix up input parameter values
  COSMCtrlPosition tempPosition{position};
  tempPosition.NormalizeLongitude();

  //Fail any invalid input value
  const int nZoom{static_cast<int>(fZoom)};
  const double fMaxTile{pow(2.0, nZoom)};
  const double fX{COSMCtrlHelper::Longitude2TileX(tempPosition.m_fLongitude, nZoom)};
  if ((fX < 0) || (fX > fMaxTile))
    return false;
  const double fY{COSMCtrlHelper::Latitude2TileY(tempPosition.m_fLatitude, nZoom)};
  if ((fY < 0) || (fY > fMaxTile))
    return false;

  //Validate our parameters
  ASSERT((tempPosition.m_fLongitude >= -180) && (tempPosition.m_fLongitude <= 180));

  //Hive away the new settings
  const ULONGLONG nTickCount{GetTickCount64()};
  const bool bChangePosition{(tempPosition.m_fLatitude != m_CenterPosition.m_fLatitude) || (tempPosition.m_fLongitude != m_CenterPosition.m_fLongitude)};
  if (bChangePosition)
    m_nLastSetCenterTickCount = nTickCount;
  const bool bChangeZoom{fZoom != m_fZoom};
  const bool bChangeBearing{fBearingOfTopOfMap != m_fBearingOfTopOfMap};
  const bool bChange{bChangePosition || bChangeZoom || bChangeBearing};
  m_StartCenterPosition = m_CenterPosition;
  m_FinalCenterPosition = tempPosition;
  m_fStartZoom = m_fZoom;
  m_fFinalZoom = fZoom;
  m_fStartBearingAtTopOfMap = m_fBearingOfTopOfMap;
  m_fFinalBearingAtTopOfMap = fBearingOfTopOfMap;
  m_fCurrentAnimationDuration = fAnimationDuration;

  bool bDoAnimation{false};
  if (bChange && (GetSafeHwnd() != nullptr))
  {
    if ((fAnimationDuration != 0) && m_bAnimations && !GetSystemMetrics(SM_REMOTESESSION))
    {
      m_nAnimationTimer = SetTimer(OSMCTRL_ANIMATION_TIMER_ID, m_nAnimationInterval, nullptr);
      m_nStartAnimationTickCount = nTickCount;
      bDoAnimation = true;
      m_bDoingBearingAnimation = bChangeBearing;
      m_bDoingZoomAnimation = bChangeZoom;
      m_bDoingPositionAnimation = bChangePosition;
    }

    //If we are not using animations, just force a straight invalidation
    if (!bDoAnimation)
    {
      //Actually update the zoom level value now
      m_fZoom = fZoom;
      m_CenterPosition = tempPosition;
      m_fBearingOfTopOfMap = fBearingOfTopOfMap;

      Invalidate(FALSE);

      //Update the slider control with the new position if necessary also
      if (m_ctrlZoomBar.GetSafeHwnd() != nullptr)
        m_ctrlZoomBar.SetPos(static_cast<int>((OSMMaxZoom - GetZoom() - OSMMinZoom) * 10));

      //Restart downloading of tiles if necessary
      if (bRestartDownloading && m_bDownloadTiles && m_sCacheDirectory.GetLength())
      {
        StopAsyncDownload(false);
        StartAsyncDownload(true, false);
      }
    }
  }

  return true;
}

void COSMCtrl::SetCacheDirectory(_In_ const CString& sCacheDirectory, _In_ bool bInvalidateCache)
{
  //Hive away the new setting
  const bool bChange{sCacheDirectory != m_sCacheDirectory};
  m_sCacheDirectory = sCacheDirectory;
  if (bInvalidateCache)
  {
    //Create a Multi SZ string with the filename to delete
    int nChars{sCacheDirectory.GetLength() + 1};
    nChars++;
    SHFILEOPSTRUCT shfo{};
    shfo.hwnd = AfxGetMainWnd()->GetSafeHwnd();
    shfo.wFunc = FO_DELETE;
    shfo.fFlags = FOF_NOCONFIRMATION | FOF_NOERRORUI | FOF_SILENT;
#pragma warning(suppress: 26472)
    std::vector<TCHAR> pszFrom{static_cast<size_t>(nChars), std::allocator<TCHAR>{}};
#pragma warning(suppress: 26429)
    TCHAR* pszCur{pszFrom.data()};
    _tcscpy_s(pszCur, nChars, sCacheDirectory);
#pragma warning(suppress: 26481)
    pszCur[nChars - 1] = _T('\0');
    shfo.pFrom = pszFrom.data();

    //Let the shell perform the actual deletion
    SHFileOperation(&shfo);
  }
  //Create the cache directory if we can
  if (bChange || bInvalidateCache)
    CreateDirectory(sCacheDirectory, nullptr);
}

void COSMCtrl::SetTileProvider(_In_ IOSMCtrlTileProvider* pTileProvider)
{
  //Hive away the new setting
  const bool bChange{pTileProvider != m_pTileProvider};
  m_pTileProvider = pTileProvider;
  ASSERT(m_pTileProvider != nullptr);

  //Empty the tile cache if required
  if (bChange)
    DeleteCachedTiles();

  //Force a redraw if required
  if (bChange && (GetSafeHwnd() != nullptr))
    Invalidate(FALSE);

  //Irrespective of a change or not, restart downloading of tiles if necessary
  if (m_bDownloadTiles && m_sCacheDirectory.GetLength())
  {
    StopAsyncDownload(true);
    StartAsyncDownload(true, false);
  }
  else if (!m_bDownloadTiles)
  {
    StopAsyncDownload(true);
  }
}

void COSMCtrl::SetDownloadTiles(_In_ bool bDownloadTiles)
{
  //Hive away the new setting
  const bool bChange{m_bDownloadTiles != bDownloadTiles};
  m_bDownloadTiles = bDownloadTiles;

  //Force a redraw if required
  if (bChange && (GetSafeHwnd() != nullptr))
    Invalidate(FALSE);

  //Irrespective of a change or not, restart downloading of tiles if necessary
  if (m_bDownloadTiles && m_sCacheDirectory.GetLength())
  {
    StopAsyncDownload(false);
    StartAsyncDownload(true, false);
  }
  else if (!bDownloadTiles)
    StopAsyncDownload(false);
}

void COSMCtrl::SetDrawTileOutlines(_In_ bool bDrawTileOutlines)
{
  const bool bChange{bDrawTileOutlines != m_bDrawTileOutlines};
  m_bDrawTileOutlines = bDrawTileOutlines;

  //Force a redraw if required
  if (bChange && (GetSafeHwnd() != nullptr))
    Invalidate(FALSE);
}

#pragma warning(suppress: 26440)
void COSMCtrl::InitializeNewPolyline(_Inout_ COSMCtrlPolyline& polyline)
{
  //By default make any newly created polyline draggable and editable
  polyline.m_bDraggable = true;
  polyline.m_bEditable = true;
}

#pragma warning(suppress: 26440)
void COSMCtrl::InitializeNewPolygon(_Inout_ COSMCtrlPolygon& polygon)
{
  //By default make any newly created polygon draggable and editable
  polygon.m_bDraggable = true;
  polygon.m_bEditable = true;
}

#pragma warning(suppress: 26440)
void COSMCtrl::InitializeNewSelectionPolygon(_Inout_ COSMCtrlPolygon& polygon)
{
  //Use a style which creates a standard black selection
#ifdef COSMCTRL_NOD2D
  polygon.m_DashStyle = Gdiplus::DashStyleDot;
  polygon.m_colorPen = Gdiplus::Color{0, 0, 0};
#else
  polygon.m_DashStyle = D2D1_DASH_STYLE_DASH;
  polygon.m_colorPen = D2D1::ColorF::Black;
#endif //#ifdef COSMCTRL_NOD2D
  polygon.m_fLinePenWidth = 1.5;
}

#pragma warning(suppress: 26440)
void COSMCtrl::InitializeNewMarker(_Inout_ COSMCtrlMarker& marker)
{
  //By default make any newly created marker draggable
  marker.m_bDraggable = true;

  //Also set the marker's icon to the first icon if there is one
  if (m_Icons.size())
    marker.m_nIconIndex = 0;
}

#pragma warning(suppress: 26440)
void COSMCtrl::InitializeNewCircle(_Inout_ COSMCtrlCircle& circle)
{
  //By default make any newly created circle draggable and editable
  circle.m_bDraggable = true;
  circle.m_bEditable = true;
}


//A helper class to allow access to some protected data in MFC's CRenderTarget class (Used in COSMCtrl::DeleteCachedTiles).
class CMyCRenderTarget : public CRenderTarget
{
public:
  CObList& GetResources() noexcept { return m_lstResources; };
};

//A helper class to allow access to some protected data in MFC's CD2DResource class (Used in COSMCtrl::DeleteCachedTiles).
class CMyCD2DResource : public CD2DResource
{
public:
  _NODISCARD BOOL GetAutoDestroy() const noexcept { return m_bIsAutoDestroy; };
  void SetNullParentTarget() noexcept { m_pParentTarget = nullptr; };
};

#pragma warning(suppress: 26440)
void COSMCtrl::DeleteCachedTiles()
{
  //Empty out our cache
  m_CachedTiles.clear();

#ifndef COSMCTRL_NOD2D
  //Throw away all the resources in the render target
#pragma warning(suppress: 26490)
  auto pRenderTarget{reinterpret_cast<CMyCRenderTarget*>(GetRenderTarget())};
  if (pRenderTarget != nullptr)
  {
    CObList& resources{pRenderTarget->GetResources()};
    while (!resources.IsEmpty())
    {
#pragma warning(suppress: 26490)
      auto pRes{reinterpret_cast<CMyCD2DResource*>(DYNAMIC_DOWNCAST(CD2DResource, resources.RemoveHead()))};
      if ((pRes != nullptr) && pRes->GetAutoDestroy())
      {
        pRes->SetNullParentTarget();
        delete pRes;
      }
    }

    //Also invalidate our client side pointers into the resources in the render target which we have just destroyed
    DestroyD2DResources();
  }
#endif //#ifndef COSMCTRL_NOD2D
}

void COSMCtrl::Refresh()
{
  //Empty out our in memory cache
  TRACE(_T("COSMCtrl::Refresh, clearing tile cache\n"));
  DeleteCachedTiles();

  //Restart downloading of tiles if necessary
  if (m_bDownloadTiles && m_sCacheDirectory.GetLength())
  {
    StopAsyncDownload(true);
    StartAsyncDownload(true, true);
  }
  else
  {
    //Simply do an invalidation if download tiles is turned off
    Invalidate(TRUE);
  }

  TRACE(_T("COSMCtrl::Refresh, finished\n"));
}

void COSMCtrl::SetDrawCenterCrossHairs(_In_ bool bDrawCenterCrossHairs)
{
  const bool bChange{bDrawCenterCrossHairs != m_bDrawCenterCrossHairs};
  m_bDrawCenterCrossHairs = bDrawCenterCrossHairs;

  //Force a redraw if required
  if (bChange && (GetSafeHwnd() != nullptr))
    Invalidate(FALSE);
}

CString COSMCtrl::GetTileCachePath(_In_ const CString& sCacheDirectory, _In_ int nZoom, _In_ int nTileX, _In_ int nTileY)
{
  CString sPath;
  sPath.Format(_T("%s\\%d\\%d\\%d.png"), sCacheDirectory.GetString(), nZoom, nTileX, nTileY);
  return sPath;
}

#ifdef COSMCTRL_NOD2D
bool COSMCtrl::Draw3dRect(_In_ Gdiplus::Graphics& graphics, _In_ const Gdiplus::RectF& r, _In_ const Gdiplus::Color& clrTopLeft, _In_ const Gdiplus::Color& clrBottomRight)
{
  //Create the pens we need
  Gdiplus::Pen penTopLeft{clrTopLeft};
  if (penTopLeft.GetLastStatus() != Gdiplus::Ok)
    return false;
  Gdiplus::Pen penBottomRight{clrBottomRight};
  if (penBottomRight.GetLastStatus() != Gdiplus::Ok)
    return false;

  //Fix up the height and width values for our path calculations
  Gdiplus::REAL fHeight{r.Height};
  --fHeight;
  Gdiplus::REAL fWidth{r.Width};
  --fWidth;

  //Create the top left path
  Gdiplus::GraphicsPath pathTopLeft;
  pathTopLeft.StartFigure();
  pathTopLeft.AddLine(r.X, r.Y + fHeight - 1, r.X, r.Y);
  pathTopLeft.AddLine(r.X + 1, r.Y, r.X + fWidth - 1, r.Y);

  //Create the bottom right path
  Gdiplus::GraphicsPath pathBottomRight;
  pathBottomRight.StartFigure();
  pathBottomRight.AddLine(r.X + fWidth, r.Y, r.X + fWidth, r.Y + fHeight);
  pathBottomRight.AddLine(r.X + fWidth - 1, r.Y + fHeight, r.X, r.Y + fHeight);

  //Finally draw the two paths
  graphics.DrawPath(&penTopLeft, &pathTopLeft);
  graphics.DrawPath(&penBottomRight, &pathBottomRight);

  return true;
}
#else
#pragma warning(suppress: 26429)
bool COSMCtrl::Draw3dRect(_In_ CRenderTarget* pRenderTarget, _In_ const CD2DRectF& r, _In_ const D2D1::ColorF& clrTopLeft, _In_ const D2D1::ColorF& clrBottomRight)
{
  CD2DSolidColorBrush brushTopLeft{pRenderTarget, clrTopLeft};
  pRenderTarget->DrawLine(CD2DPointF{r.left, r.bottom - 1}, CD2DPointF{r.left, r.top}, &brushTopLeft);
  pRenderTarget->DrawLine(CD2DPointF{r.left, r.top}, CD2DPointF{r.right - 1, r.top}, &brushTopLeft);
  CD2DSolidColorBrush brushBottomRight(pRenderTarget, clrBottomRight);
  pRenderTarget->DrawLine(CD2DPointF{r.right - 1, r.top}, CD2DPointF{r.right - 1, r.bottom - 1}, &brushBottomRight);
  pRenderTarget->DrawLine(CD2DPointF{r.left, r.bottom - 1}, CD2DPointF{r.right - 1, r.bottom - 1}, &brushBottomRight);

  return true;
}
#endif //#ifdef COSMCTRL_NOD2D

#ifdef COSMCTRL_NOD2D
bool COSMCtrl::DrawTileNotAvailable(_In_ Gdiplus::Graphics& graphics, _In_ const Gdiplus::RectF& rTile)
{
  //Create the brush we need
  Gdiplus::SolidBrush brush{Gdiplus::Color{0, 0, 0}};
  if (brush.GetLastStatus() != Gdiplus::Ok)
    return false;

  //Fill the whole tile with the brush to start with
  if (graphics.FillRectangle(&brush, rTile) != Gdiplus::Ok)
    return false;

  //Also draw a tile outline
  if (!DrawTileOutline(graphics, rTile))
    return false;

  //Create the brush we need
  Gdiplus::SolidBrush brushWhite{Gdiplus::Color{255, 255, 255}};
  if (brushWhite.GetLastStatus() != Gdiplus::Ok)
    return false;

  //And some explanatory text
  HDC hDC{graphics.GetHDC()};
  NONCLIENTMETRICS ncm{};
  ncm.cbSize = sizeof(ncm);
  if (!SystemParametersInfo(SPI_GETNONCLIENTMETRICS, 0, &ncm, 0))
    return false;
  Gdiplus::Font font{hDC, &ncm.lfMessageFont};
  graphics.ReleaseHDC(hDC);
  if (font.GetLastStatus() != Gdiplus::Ok)
    return false;
  Gdiplus::StringFormat stringFormat;
  if (stringFormat.SetAlignment(Gdiplus::StringAlignmentCenter) != Gdiplus::Ok)
    return false;
  if (stringFormat.SetLineAlignment(Gdiplus::StringAlignmentCenter) != Gdiplus::Ok)
    return false;
  CStringW sTileNotAvailable;
  if (!sTileNotAvailable.LoadString(IDS_OSMCTRL_TILE_NOT_AVAILABLE))
    return false;
  if (graphics.DrawString(sTileNotAvailable, -1, &font, rTile, &stringFormat, &brushWhite) != Gdiplus::Ok)
    return false;

  return true;
}
#else
bool COSMCtrl::DrawTileNotAvailable(_In_ CRenderTarget* pRenderTarget, _In_ const CD2DRectF& rTile)
{
  //Cached the tile empty bitmap if necessary
  if (m_pEmptyTile == nullptr)
  {
#pragma warning(suppress: 26409)
    m_pEmptyTile = new CD2DBitmap{pRenderTarget, IDB_OSMCTRL_EMPTY_TILE, _T("PNG")};
    if (FAILED(m_pEmptyTile->Create(pRenderTarget)))
    {
      delete m_pEmptyTile;
      m_pEmptyTile = nullptr;
      return false;
    }
  }

  //Validate our parameters
#pragma warning(suppress: 26496)
  AFXASSUME(pRenderTarget != nullptr);
  pRenderTarget->DrawBitmap(m_pEmptyTile, rTile);

  return true;
}
#endif //#ifdef COSMCTRL_NOD2D

#ifdef COSMCTRL_NOD2D
bool COSMCtrl::DrawTileOutline(_In_ Gdiplus::Graphics& graphics, _In_ const Gdiplus::RectF& rTile)
{
  //Draw the rectangle
  return Draw3dRect(graphics, rTile, Gdiplus::Color{255, 255, 255}, Gdiplus::Color{128, 128, 128});
}
#else
bool COSMCtrl::DrawTileOutline(_In_ CRenderTarget* pRenderTarget, _In_ const CD2DRectF& rTile)
{
  //Draw the rectangle
  return Draw3dRect(pRenderTarget, rTile, D2D1::ColorF::White, D2D1::ColorF::Gray);
}
#endif //#ifdef COSMCTRL_NOD2D

#pragma warning(suppress: 26440)
CPoint COSMCtrl::GetControlPosition(_In_ ControlAnchorPosition anchorPosition, _In_ const CPoint& ptOffset, _In_ const CRect& rClient)
{
  //Work out the position of the control based on the anchor position and the offset
  CPoint ptPosition;
  switch (anchorPosition)
  {
    case ControlAnchorPosition::TopLeft:
    {
      ptPosition = CPoint{rClient.left + ptOffset.x, rClient.top + ptOffset.y};
      break;
    }
    case ControlAnchorPosition::TopRight:
    {
      ptPosition = CPoint{rClient.right - ptOffset.x, rClient.top + ptOffset.y};
      break;
    }
    case ControlAnchorPosition::BottomLeft:
    {
      ptPosition = CPoint{rClient.left + ptOffset.x, rClient.bottom - ptOffset.y};
      break;
    }
    case ControlAnchorPosition::BottomRight:
    {
      ptPosition = CPoint{rClient.right - ptOffset.x, rClient.bottom - ptOffset.y};
      break;
    }
    default:
    {
      ASSERT(false);
      break;
    }
  }

  return ptPosition;
}

#ifdef COSMCTRL_NOD2D
bool COSMCtrl::DrawScrollRose(_In_ Gdiplus::Graphics& graphics, _In_ const CRect& /*rClip*/, _In_ const CRect& rClient)
{
  //Get the position of the scroll rose
  const CPoint ptControl{GetControlPosition(m_ScrollRoseAnchorPosition, m_ptOffsetScrollRose, rClient)};

  //Work out the left X and top Y value
  int nLeftX = ptControl.x;
  if ((m_ScrollRoseAnchorPosition == ControlAnchorPosition::TopRight) || (m_ScrollRoseAnchorPosition == ControlAnchorPosition::BottomRight))
    nLeftX -= 36;
  int nTopY = ptControl.y;
  if ((m_ScrollRoseAnchorPosition == ControlAnchorPosition::BottomLeft) || (m_ScrollRoseAnchorPosition == ControlAnchorPosition::BottomRight))
    nTopY -= 56;

  //Create the brushes and pens we need
  Gdiplus::SolidBrush brushBlue{Gdiplus::Color{0, 0, 139}};
  if (brushBlue.GetLastStatus() != Gdiplus::Ok)
    return false;
  Gdiplus::SolidBrush brushWhite{Gdiplus::Color{255, 255, 255}};
  if (brushWhite.GetLastStatus() != Gdiplus::Ok)
    return false;

  //Draw the North scroller
  if (graphics.FillPie(&brushBlue, 8 + nLeftX, nTopY, 18, 18, 180, 180) != Gdiplus::Ok)
    return false;
  m_rNorthScrollRose = CRect{8 + nLeftX, nTopY, 26 + nLeftX, 14 + nTopY};

  //Draw the rectangle in the middle of the scroller
  if (graphics.FillRectangle(&brushBlue, 8 + nLeftX, 8 + nTopY, 18, 39) != Gdiplus::Ok)
    return false;

  //Draw the West scroller
  if (graphics.FillPie(&brushBlue, nLeftX, 18 + nTopY, 18, 18, 90, 180) != Gdiplus::Ok)
    return false;
  m_rWestScrollRose = CRect{nLeftX, 18 + nTopY, 14 + nLeftX, 36 + nTopY};

  //Draw the East scroller
  if (graphics.FillPie(&brushBlue, 16 + nLeftX, 18 + nTopY, 18, 18, 270, 180) != Gdiplus::Ok)
    return false;
  m_rEastScrollRose = CRect{22 + nLeftX, 17 + nTopY, 35 + nLeftX, 36 + nTopY};

  //Draw the south scroller
  if (graphics.FillPie(&brushBlue, 8 + nLeftX, 37 + nTopY, 18, 18, 0, 180) != Gdiplus::Ok)
    return false;
  m_rSouthScrollRose = CRect{8 + nLeftX, 40 + nTopY, 26 + nLeftX, 54 + nTopY};

  //Draw the north scroller triangle
  std::array<Gdiplus::PointF, 3> points{ { { static_cast<Gdiplus::REAL>(12 + nLeftX), static_cast<Gdiplus::REAL>(13 + nTopY) },
                                           { static_cast<Gdiplus::REAL>(22 + nLeftX), static_cast<Gdiplus::REAL>(13 + nTopY) },
                                           { static_cast<Gdiplus::REAL>(17 + nLeftX), static_cast<Gdiplus::REAL>(4 + nTopY) } } };
  if (graphics.FillPolygon(&brushWhite, points.data(), static_cast<INT>(points.size())) != Gdiplus::Ok)
    return false;

  //Draw the West scroller triangle
#pragma warning(suppress: 26446)
  points[0].X = static_cast<Gdiplus::REAL>(13 + nLeftX);
#pragma warning(suppress: 26446)
  points[0].Y = static_cast<Gdiplus::REAL>(21 + nTopY);
#pragma warning(suppress: 26446)
  points[1].X = static_cast<Gdiplus::REAL>(4 + nLeftX);
#pragma warning(suppress: 26446)
  points[1].Y = static_cast<Gdiplus::REAL>(27 + nTopY);
#pragma warning(suppress: 26446)
  points[2].X = static_cast<Gdiplus::REAL>(13 + nLeftX);
#pragma warning(suppress: 26446)
  points[2].Y = static_cast<Gdiplus::REAL>(32 + nTopY);
  if (graphics.FillPolygon(&brushWhite, points.data(), static_cast<INT>(points.size())) != Gdiplus::Ok)
    return false;

  //Draw the East scroller triangle
#pragma warning(suppress: 26446)
  points[0].X = static_cast<Gdiplus::REAL>(22 + nLeftX);
#pragma warning(suppress: 26446)
  points[0].Y = static_cast<Gdiplus::REAL>(21 + nTopY);
#pragma warning(suppress: 26446)
  points[1].X = static_cast<Gdiplus::REAL>(31 + nLeftX);
#pragma warning(suppress: 26446)
  points[1].Y = static_cast<Gdiplus::REAL>(27 + nTopY);
#pragma warning(suppress: 26446)
  points[2].X = static_cast<Gdiplus::REAL>(22 + nLeftX);
#pragma warning(suppress: 26446)
  points[2].Y = static_cast<Gdiplus::REAL>(32 + nTopY);
  if (graphics.FillPolygon(&brushWhite, points.data(), static_cast<INT>(points.size())) != Gdiplus::Ok)
    return false;

  //Draw the south scroller triangle
#pragma warning(suppress: 26446)
  points[0].X = static_cast<Gdiplus::REAL>(12 + nLeftX);
#pragma warning(suppress: 26446)
  points[0].Y = static_cast<Gdiplus::REAL>(40 + nTopY);
#pragma warning(suppress: 26446)
  points[1].X = static_cast<Gdiplus::REAL>(17 + nLeftX);
#pragma warning(suppress: 26446)
  points[1].Y = static_cast<Gdiplus::REAL>(49 + nTopY);
#pragma warning(suppress: 26446)
  points[2].X = static_cast<Gdiplus::REAL>(22 + nLeftX);
#pragma warning(suppress: 26446)
  points[2].Y = static_cast<Gdiplus::REAL>(40 + nTopY);
  if (graphics.FillPolygon(&brushWhite, points.data(), static_cast<INT>(points.size())) != Gdiplus::Ok)
    return false;

  return true;
}
#else
#pragma warning(suppress: 26429)
bool COSMCtrl::DrawScrollRose(_In_ CRenderTarget* pRenderTarget, _In_ const CRect& /*rClip*/, _In_ const CRect& rClient)
{
  //Get the position of the scroll rose
  const CPoint ptControl{GetControlPosition(m_ScrollRoseAnchorPosition, m_ptOffsetScrollRose, rClient)};

  //Work out the left X and top Y value
  int nLeftX = ptControl.x;
  if ((m_ScrollRoseAnchorPosition == ControlAnchorPosition::TopRight) || (m_ScrollRoseAnchorPosition == ControlAnchorPosition::BottomRight))
    nLeftX -= 36;
  int nTopY = ptControl.y;
  if ((m_ScrollRoseAnchorPosition == ControlAnchorPosition::BottomLeft) || (m_ScrollRoseAnchorPosition == ControlAnchorPosition::BottomRight))
    nTopY -= 56;

  //Create the brushes we need
  CD2DSolidColorBrush brushBlue{pRenderTarget, RGB(0, 0, 139)};
  CD2DSolidColorBrush brushWhite{pRenderTarget, D2D1::ColorF::White};

  //Draw the North scroller
  pRenderTarget->FillEllipse(CD2DEllipse{CD2DPointF{static_cast<FLOAT>(17 + nLeftX), static_cast<FLOAT>(nTopY + 9)}, CD2DSizeF{9, 9}}, &brushBlue);
  m_rNorthScrollRose = CRect{8 + nLeftX, nTopY, 26 + nLeftX, 14 + nTopY};

  //Draw the rectangle in the middle of the scroller
  pRenderTarget->FillRectangle(CD2DRectF{static_cast<FLOAT>(8 + nLeftX), static_cast<FLOAT>(8 + nTopY), static_cast<FLOAT>(26 + nLeftX), static_cast<FLOAT>(47 + nTopY)}, &brushBlue);

  //Draw the West scroller
  pRenderTarget->FillEllipse(CD2DEllipse{CD2DPointF{static_cast<FLOAT>(9 + nLeftX), static_cast<FLOAT>(nTopY + 27)}, CD2DSizeF{9, 9}}, &brushBlue);
  m_rWestScrollRose = CRect{nLeftX, 18 + nTopY, 14 + nLeftX, 36 + nTopY};

  //Draw the East scroller
  pRenderTarget->FillEllipse(CD2DEllipse{CD2DPointF{static_cast<FLOAT>(26 + nLeftX), static_cast<FLOAT>(nTopY + 27)}, CD2DSizeF{9, 9}}, &brushBlue);
  m_rEastScrollRose = CRect{22 + nLeftX, 17 + nTopY, 35 + nLeftX, 36 + nTopY};

  //Draw the south scroller
  pRenderTarget->FillEllipse(CD2DEllipse{CD2DPointF{static_cast<FLOAT>(17 + nLeftX), static_cast<FLOAT>(nTopY + 46)}, CD2DSizeF{9, 9}}, &brushBlue);
  m_rSouthScrollRose = CRect{8 + nLeftX, 40 + nTopY, 26 + nLeftX, 54 + nTopY};

  //Draw the north scroller triangle
  CD2DPathGeometry northScrollerTriangle{pRenderTarget};
  if (FAILED(northScrollerTriangle.Create(pRenderTarget)))
    return false;
  CD2DGeometrySink geometrySink{northScrollerTriangle};
  CD2DPointF point{static_cast<FLOAT>(12 + nLeftX), static_cast<FLOAT>(13 + nTopY)};
  geometrySink.BeginFigure(point, D2D1_FIGURE_BEGIN_FILLED);
  std::array<D2D1_POINT_2F, 2> points{ { { static_cast<FLOAT>(22 + nLeftX), static_cast<FLOAT>(13 + nTopY) }, { static_cast<FLOAT>(17 + nLeftX), static_cast<FLOAT>(4 + nTopY) } } };
#pragma warning(suppress: 26472)
  geometrySink.Get()->AddLines(points.data(), static_cast<UINT32>(points.size()));
  geometrySink.EndFigure(D2D1_FIGURE_END_CLOSED);
  if (!geometrySink.Close())
    return false;
  pRenderTarget->FillGeometry(&northScrollerTriangle, &brushWhite);

  //Draw the West scroller triangle
  CD2DPathGeometry westScrollerTriangle{pRenderTarget};
  if (FAILED(westScrollerTriangle.Create(pRenderTarget)))
    return false;
  CD2DGeometrySink geometrySink2{westScrollerTriangle};
  point = CD2DPointF{static_cast<FLOAT>(13 + nLeftX), static_cast<FLOAT>(21 + nTopY)};
  geometrySink2.BeginFigure(point, D2D1_FIGURE_BEGIN_FILLED);
#pragma warning(suppress: 26446)
  points[0].x = static_cast<FLOAT>(4 + nLeftX);
#pragma warning(suppress: 26446)
  points[0].y = static_cast<FLOAT>(27 + nTopY);
#pragma warning(suppress: 26446)
  points[1].x = static_cast<FLOAT>(13 + nLeftX);
#pragma warning(suppress: 26446)
  points[1].y = static_cast<FLOAT>(32 + nTopY);
#pragma warning(suppress: 26472)
  geometrySink2.Get()->AddLines(points.data(), static_cast<UINT32>(points.size()));
  geometrySink2.EndFigure(D2D1_FIGURE_END_CLOSED);
  if (!geometrySink2.Close())
    return false;
  pRenderTarget->FillGeometry(&westScrollerTriangle, &brushWhite);

  //Draw the East scroller triangle
  CD2DPathGeometry eastScrollerTriangle{pRenderTarget};
  if (FAILED(eastScrollerTriangle.Create(pRenderTarget)))
    return false;
  CD2DGeometrySink geometrySink3{eastScrollerTriangle};
  point = CD2DPointF{static_cast<FLOAT>(22 + nLeftX), static_cast<FLOAT>(21 + nTopY)};
  geometrySink3.BeginFigure(point, D2D1_FIGURE_BEGIN_FILLED);
#pragma warning(suppress: 26446)
  points[0].x = static_cast<FLOAT>(31 + nLeftX);
#pragma warning(suppress: 26446)
  points[0].y = static_cast<FLOAT>(27 + nTopY);
#pragma warning(suppress: 26446)
  points[1].x = static_cast<FLOAT>(22 + nLeftX);
#pragma warning(suppress: 26446)
  points[1].y = static_cast<FLOAT>(32 + nTopY);
#pragma warning(suppress: 26472)
  geometrySink3.Get()->AddLines(points.data(), static_cast<UINT32>(points.size()));
  geometrySink3.EndFigure(D2D1_FIGURE_END_CLOSED);
  if (!geometrySink3.Close())
    return false;
  pRenderTarget->FillGeometry(&eastScrollerTriangle, &brushWhite);

  //Draw the South scroller triangle
  CD2DPathGeometry southScrollerTriangle{pRenderTarget};
  if (FAILED(southScrollerTriangle.Create(pRenderTarget)))
    return false;
  CD2DGeometrySink geometrySink4{southScrollerTriangle};
  point = CD2DPointF{static_cast<FLOAT>(12 + nLeftX), static_cast<FLOAT>(40 + nTopY)};
  geometrySink4.BeginFigure(point, D2D1_FIGURE_BEGIN_FILLED);
#pragma warning(suppress: 26446)
  points[0].x = static_cast<FLOAT>(17 + nLeftX);
#pragma warning(suppress: 26446)
  points[0].y = static_cast<FLOAT>(49 + nTopY);
#pragma warning(suppress: 26446)
  points[1].x = static_cast<FLOAT>(22 + nLeftX);
#pragma warning(suppress: 26446)
  points[1].y = static_cast<FLOAT>(40 + nTopY);
#pragma warning(suppress: 26472)
  geometrySink4.Get()->AddLines(points.data(), static_cast<UINT32>(points.size()));
  geometrySink4.EndFigure(D2D1_FIGURE_END_CLOSED);
  if (!geometrySink4.Close())
    return false;
  pRenderTarget->FillGeometry(&southScrollerTriangle, &brushWhite);

  return true;
}
#endif //#ifdef COSMCTRL_NOD2D

bool COSMCtrl::CreateZoomBarSlider()
{
  //Get the client rect
  CRect rClient;
  GetClientRect(&rClient);

  //Get the position of the scroll rose
  const CPoint ptControl{GetControlPosition(m_ZoomBarAnchorPosition, m_ptOffsetZoomBar, rClient)};

  //Work out the left X and top Y value
  int nLeftX = ptControl.x;
  if ((m_ZoomBarAnchorPosition == ControlAnchorPosition::TopRight) || (m_ZoomBarAnchorPosition == ControlAnchorPosition::BottomRight))
    nLeftX -= 20;
  int nTopY = ptControl.y;
  if ((m_ZoomBarAnchorPosition == ControlAnchorPosition::BottomLeft) || (m_ZoomBarAnchorPosition == ControlAnchorPosition::BottomRight))
    nTopY -= 246;

  //Create the control
  const BOOL bSuccess{m_ctrlZoomBar.Create(WS_BORDER | WS_CHILD | WS_VISIBLE | TBS_VERT | TBS_NOTICKS, CRect(nLeftX, nTopY, nLeftX + 20, nTopY + 246), this, SLIDER_ID)};
  if (bSuccess)
  {
    m_ctrlZoomBar.SetRange(OSMMinZoom, ((OSMMaxZoom - OSMMinZoom)*10) + OSMMinZoom);
    m_ctrlZoomBar.SetPos(static_cast<int>((OSMMaxZoom - GetZoom() - OSMMinZoom) * 10));
  }
  return bSuccess;
}

bool COSMCtrl::CreateCopyrightLinkCtrl()
{
  //Get the client rect
  CRect rClient;
  GetClientRect(&rClient);

  //Get the position of the copyright control
  const CPoint ptControl{GetControlPosition(m_CopyrightAnchorPosition, m_ptOffsetCopyright, rClient)};

  //Work out the left X and top Y value
#pragma warning(suppress: 26496)
  int nLeftX{ptControl.x};
#pragma warning(suppress: 26496)
  int nTopY{ptControl.y};

#ifndef _UNICODE
  UNREFERENCED_PARAMETER(nLeftX);
  UNREFERENCED_PARAMETER(nTopY);

  return false; //Link controls are only supported in unicode builds
#else
  //Create the control
  CString sCopyright{FormCopyrightText()};
  const BOOL bSuccess{m_ctrlCopyright.Create(sCopyright, WS_BORDER | WS_CHILD | WS_VISIBLE, CRect(nLeftX, nTopY, nLeftX, nTopY), this, COPYRIGHT_ID)};
  if (bSuccess)
  {
    //Create the control which the message box font
    NONCLIENTMETRICS ncm{};
    ncm.cbSize = sizeof(ncm);
    if (!SystemParametersInfo(SPI_GETNONCLIENTMETRICS, 0, &ncm, 0))
      return false;
    CFont font;
    if (!font.CreateFontIndirect(&ncm.lfMessageFont))
      return false;
    m_ctrlCopyright.SetFont(&font);

    //Resize the control to ensure it works correctly with High DPI
    CSize size;
    m_ctrlCopyright.GetIdealSize(rClient.Width(), &size);
    if ((size.cx != 0) && (size.cy != 0))
    {
      const int nNewWidth{size.cx + 4};
      const int nNewHeight{size.cy + 4};
      if ((m_CopyrightAnchorPosition == ControlAnchorPosition::TopRight) || (m_CopyrightAnchorPosition == ControlAnchorPosition::BottomRight))
        nLeftX -= nNewWidth;
      if ((m_CopyrightAnchorPosition == ControlAnchorPosition::BottomLeft) || (m_CopyrightAnchorPosition == ControlAnchorPosition::BottomRight))
        nTopY -= nNewHeight;

      m_ctrlCopyright.SetWindowPos(nullptr, nLeftX, nTopY, nNewWidth, nNewHeight, SWP_NOZORDER | SWP_NOREPOSITION | SWP_NOACTIVATE);
    }
  }

  return bSuccess;
#endif //#ifndef _UNICODE
}

#ifdef COSMCTRL_NOD2D
bool COSMCtrl::DrawZoomBar(_In_ Gdiplus::Graphics& graphics, _In_ const CRect& /*rClip*/, _In_ const CRect& rClient)
{
  //Get the position of the zoom bar
  const CPoint ptControl{GetControlPosition(m_ZoomBarAnchorPosition, m_ptOffsetZoomBar, rClient)};

  //Work out the left X and top Y value
  int nLeftX{ptControl.x};
  if ((m_ZoomBarAnchorPosition == ControlAnchorPosition::TopRight) || (m_ZoomBarAnchorPosition == ControlAnchorPosition::BottomRight))
    nLeftX -= 20;
  int nTopY = ptControl.y;
  if ((m_ZoomBarAnchorPosition == ControlAnchorPosition::BottomLeft) || (m_ZoomBarAnchorPosition == ControlAnchorPosition::BottomRight))
    nTopY -= 246;

  //Create the brushes and pens we need
  Gdiplus::Pen penBlack{Gdiplus::Color{0, 0, 0}, 1};
  if (penBlack.GetLastStatus() != Gdiplus::Ok)
    return false;
  Gdiplus::SolidBrush brushBlue{Gdiplus::Color{0, 0, 139}};
  if (brushBlue.GetLastStatus() != Gdiplus::Ok)
    return false;
  Gdiplus::SolidBrush brushWhite{Gdiplus::Color{255, 255, 255}};
  if (brushWhite.GetLastStatus() != Gdiplus::Ok)
    return false;
  Gdiplus::SolidBrush brushGray{Gdiplus::Color{128, 192, 192, 192}};
  if (brushGray.GetLastStatus() != Gdiplus::Ok)
    return false;
  Gdiplus::SolidBrush brushGray2{Gdiplus::Color{225, 225, 225}};
  if (brushGray2.GetLastStatus() != Gdiplus::Ok)
    return false;

  //Draw the North part of the zoom bar
  if (graphics.FillPie(&brushBlue, 1 + nLeftX, nTopY, 19, 18, 180, 180) != Gdiplus::Ok)
    return false;
  if (graphics.FillRectangle(&brushBlue, 1 + nLeftX, 8 + nTopY, 19, 10) != Gdiplus::Ok)
    return false;
  if (graphics.FillRectangle(&brushWhite, 4 + nLeftX, 8 + nTopY, 13, 3) != Gdiplus::Ok)
    return false;
  if (graphics.FillRectangle(&brushWhite, 9 + nLeftX, 4 + nTopY, 3, 12) != Gdiplus::Ok)
    return false;
  m_rZoomInZoomBar = CRect{1 + nLeftX, nTopY, 19 + nLeftX, 18 + nTopY};

  //Get the rect for the outline of the zoom bar
  m_rZoomBar = CRect{1 + nLeftX, 17 + nTopY, 19 + nLeftX, 230 + nTopY};

  //Draw the divider lines and fill in the areas in between
  for (int i=OSMMinZoom; i<=OSMMaxZoom; i++)
  {
    const auto fTopPosition{static_cast<Gdiplus::REAL>((i * 11) + 18 + nTopY)};
    if (graphics.DrawRectangle(&penBlack, Gdiplus::RectF{static_cast<Gdiplus::REAL>(1 + nLeftX), fTopPosition, 19, 11}) != Gdiplus::Ok)
      return false;
    if (m_bTransparencyForZoomBar)
    {
      if (graphics.FillRectangle(&brushGray, Gdiplus::RectF{static_cast<Gdiplus::REAL>(1 + nLeftX), fTopPosition, 19, 11}) != Gdiplus::Ok)
        return false;
    }
    else
    {
      if (graphics.FillRectangle(&brushGray2, Gdiplus::RectF{static_cast<Gdiplus::REAL>(1 + nLeftX), fTopPosition, 19, 11}) != Gdiplus::Ok)
        return false;
    }
  }

  //Draw the south part of the zoom bar
  if (graphics.FillRectangle(&brushBlue, 1 + nLeftX, 227 + nTopY, 19, 10) != Gdiplus::Ok)
    return false;
  if (graphics.FillPie(&brushBlue, 1 + nLeftX, 227 + nTopY, 19, 18, 0, 180) != Gdiplus::Ok)
    return false;
  if (graphics.FillRectangle(&brushWhite, 4 + nLeftX, 234 + nTopY, 13, 3))
    return false;
  m_rZoomOutZoomBar = CRect{1 + nLeftX, 230 + nTopY, 19 + nLeftX, 246 + nTopY};

  //Draw the current zoom position
  const auto fZoomTopPosition{static_cast<Gdiplus::REAL>(19 + nTopY + (18 - m_fZoom) * 11)};
  if (graphics.FillRectangle(&brushBlue, Gdiplus::RectF{static_cast<Gdiplus::REAL>(nLeftX), fZoomTopPosition, 21, 9}) != Gdiplus::Ok)
    return false;
  if (graphics.FillRectangle(&brushWhite, Gdiplus::RectF{static_cast<Gdiplus::REAL>(3 + nLeftX), fZoomTopPosition + 3, 15, 3}) != Gdiplus::Ok)
    return false;

  return true;
}
#else
#pragma warning(suppress: 26429)
bool COSMCtrl::DrawZoomBar(_In_ CRenderTarget* pRenderTarget, _In_ const CRect& /*rClip*/, _In_ const CRect& rClient)
{
  //Get the position of the zoom bar
  const CPoint ptControl{GetControlPosition(m_ZoomBarAnchorPosition, m_ptOffsetZoomBar, rClient)};

  //Work out the left X and top Y value
  int nLeftX{ptControl.x};
  if ((m_ZoomBarAnchorPosition == ControlAnchorPosition::TopRight) || (m_ZoomBarAnchorPosition == ControlAnchorPosition::BottomRight))
    nLeftX -= 20;
  int nTopY{ptControl.y};
  if ((m_ZoomBarAnchorPosition == ControlAnchorPosition::BottomLeft) || (m_ZoomBarAnchorPosition == ControlAnchorPosition::BottomRight))
    nTopY -= 246;

  //Create the brushes we need
  CD2DSolidColorBrush brushBlack{pRenderTarget, D2D1::ColorF::Black};
  CD2DSolidColorBrush brushBlue{pRenderTarget, RGB(0, 0, 139)};
  CD2DSolidColorBrush brushWhite{pRenderTarget, D2D1::ColorF::White};
  CD2DSolidColorBrush brushGray{pRenderTarget, RGB(192, 192, 192), 128};

  //Draw the North part of the zoom bar
  pRenderTarget->FillEllipse(CD2DEllipse{CD2DPointF{static_cast<FLOAT>(10 + nLeftX), static_cast<FLOAT>(nTopY + 9)}, CD2DSizeF{9, 9}}, &brushBlue);
  pRenderTarget->FillRectangle(CD2DRectF{static_cast<FLOAT>(1 + nLeftX), static_cast<FLOAT>(8 + nTopY), static_cast<FLOAT>(19 + nLeftX), static_cast<FLOAT>(18 + nTopY)}, &brushBlue);
  pRenderTarget->FillRectangle(CD2DRectF{static_cast<FLOAT>(4 + nLeftX), static_cast<FLOAT>(8 + nTopY), static_cast<FLOAT>(17 + nLeftX), static_cast<FLOAT>(11 + nTopY)}, &brushWhite);
  pRenderTarget->FillRectangle(CD2DRectF{static_cast<FLOAT>(9 + nLeftX), static_cast<FLOAT>(4 + nTopY), static_cast<FLOAT>(12 + nLeftX), static_cast<FLOAT>(16 + nTopY)}, &brushWhite);
  m_rZoomInZoomBar = CRect{1 + nLeftX, nTopY, 19 + nLeftX, 18 + nTopY};

  //Get the rect for the outline of the zoom bar
  m_rZoomBar = CRect{1 + nLeftX, 17 + nTopY, 19 + nLeftX, 230 + nTopY};

  //Draw the divider lines and fill in the areas in between
  for (int i=OSMMinZoom; i<=OSMMaxZoom; i++)
  {
    const auto fTopPosition{static_cast<FLOAT>((i * 11) + 18 + nTopY)};
    pRenderTarget->DrawRectangle(CD2DRectF{static_cast<FLOAT>(2 + nLeftX), fTopPosition, static_cast<FLOAT>(18 + nLeftX), fTopPosition + 11}, &brushBlack);
    if (m_bTransparencyForZoomBar)
      pRenderTarget->FillRectangle(CD2DRectF{static_cast<FLOAT>(2 + nLeftX), fTopPosition, static_cast<FLOAT>(18 + nLeftX), fTopPosition + 11}, &brushGray);
    else
      pRenderTarget->FillRectangle(CD2DRectF{static_cast<FLOAT>(2 + nLeftX), fTopPosition, static_cast<FLOAT>(18 + nLeftX), fTopPosition + 11}, &brushWhite);
  }

  //Draw the south part of the zoom bar
  pRenderTarget->FillRectangle(CD2DRectF{static_cast<FLOAT>(1 + nLeftX), static_cast<FLOAT>(227 + nTopY), static_cast<FLOAT>(19 + nLeftX), static_cast<FLOAT>(237 + nTopY)}, &brushBlue);
  pRenderTarget->FillEllipse(CD2DEllipse{CD2DPointF{static_cast<FLOAT>(10 + nLeftX), static_cast<FLOAT>(236 + nTopY)}, CD2DSizeF{9, 9}}, &brushBlue);
  pRenderTarget->FillRectangle(CD2DRectF{static_cast<FLOAT>(4 + nLeftX), static_cast<FLOAT>(234 + nTopY), static_cast<FLOAT>(17 + nLeftX), static_cast<FLOAT>(236 + nTopY)}, &brushWhite);
  m_rZoomOutZoomBar = CRect{1 + nLeftX, 230 + nTopY, 19 + nLeftX, 246 + nTopY};

  //Draw the current zoom position
  const auto fZoomTopPosition{static_cast<FLOAT>(19 + nTopY + (18 - m_fZoom) * 11)};
  pRenderTarget->FillRectangle(CD2DRectF{static_cast<FLOAT>(nLeftX), fZoomTopPosition, static_cast<FLOAT>(21 + nLeftX), static_cast<FLOAT>(9 + fZoomTopPosition)}, &brushBlue);
  pRenderTarget->FillRectangle(CD2DRectF{static_cast<FLOAT>(3 + nLeftX), fZoomTopPosition + 3, static_cast<FLOAT>(18 + nLeftX), static_cast<FLOAT>(6 + fZoomTopPosition)}, &brushWhite);

  return true;
}
#endif //#ifdef COSMCTRL_NOD2D

CString COSMCtrl::FormScaleBarText(_In_ double fScaleDistance, _In_ bool bMetric)
{
  //Create the text mode version of the distance formatted correctly
  CString sText;
  if (bMetric)
  {
    if (fScaleDistance < 1)
    {
      CString sDistance;
      sDistance.Format(_T("%.0f"), fScaleDistance*1000);
      AfxFormatString1(sText, IDS_OSMCTRL_SCALE_BAR_METERS, sDistance);
    }
    else
    {
      CString sDistance;
      sDistance.Format(_T("%.0f"), fScaleDistance);
      AfxFormatString1(sText, IDS_OSMCTRL_SCALE_BAR_KILOMETERS, sDistance);
    }
  }
  else
  {
    CString sDistance;
    if (fScaleDistance < 1)
      sDistance.Format(_T("%.3f"), fScaleDistance);
    else
      sDistance.Format(_T("%.0f"), fScaleDistance);
    AfxFormatString1(sText, IDS_OSMCTRL_SCALE_BAR_MILES, sDistance);
  }

  return sText;
}

bool COSMCtrl::UseMetric()
{
  //Determine if we should use the Metric or Imperial system of measurement
  bool bMetric{true};
  switch (m_ScaleBarUnits)
  {
    case ScaleBarUnits::UseOSDefault:
    {
      CString sMeasure;
      const int nGLI{GetLocaleInfo(LOCALE_USER_DEFAULT, LOCALE_IMEASURE, sMeasure.GetBuffer(3), 3)};
      sMeasure.ReleaseBuffer();
      if (nGLI)
      {
        if (sMeasure == _T("1"))
        {
          //Convert the measured distance to miles
          bMetric = false;
        }
      }
      break;
    }
    case ScaleBarUnits::Metric:
    {
      bMetric = true;
      break;
    }
    case ScaleBarUnits::Imperial:
    {
      bMetric = false;
      break;
    }
    default:
    {
      ASSERT(false);
      break;
    }
  }

  return bMetric;
}

void COSMCtrl::CalculateScaleBar(_Out_ double& fDistance, _Out_ double& fScaleDistance, _Out_ bool& bMetric, _Out_ int& nScaleLength)
{
  //Work out the distance between the center point and either a full tile to the left or right (or half a tile if the zoom level is 0, This is 
  //to ensure we do not end up at the point we started at because at zoom level 0 one tile completely wraps the earth)
  const int nZoom{static_cast<int>(m_fZoom)};
  double fX{COSMCtrlHelper::Longitude2TileX(m_CenterPosition.m_fLongitude, nZoom)};
  fX += (m_fZoom == 0 ? 0.5 : 1);
  const double fLongitude2{COSMCtrlHelper::TileX2Longitude(fX, nZoom)};
  fDistance = COSMCtrlHelper::DistanceBetweenPoints(COSMCtrlPosition{m_CenterPosition.m_fLongitude, m_CenterPosition.m_fLatitude}, COSMCtrlPosition{fLongitude2, m_CenterPosition.m_fLatitude}, nullptr, nullptr) / 1000; //We want KM not meters!
  if (m_fZoom == 0)
    fDistance *= 2;

  //Convert the distance to miles if necessary
  bMetric = UseMetric();
  if (!bMetric)
    fDistance *= 0.621371192;

  //Next lets scale down the measure tile width to a most significant digit unit
  double fMSDUnit{pow(10, floor(log10(fDistance)))};
  fScaleDistance = fMSDUnit;
  while (fScaleDistance < fDistance)
    fScaleDistance += fMSDUnit;

  double fInt{0};
  const double fFractionalZoom{modf(m_fZoom, &fInt)};
  const auto fOSMTileWidth{static_cast<FLOAT>(OSMTileWidth + (fFractionalZoom * OSMTileWidth))};
  nScaleLength = static_cast<int>((fScaleDistance / fDistance * static_cast<double>(m_fZoom == 0 ? fOSMTileWidth/2 : fOSMTileWidth)) + 0.5);
}

//Note that the scale drawn represents the scale at the center of the map and not the bottom left where the scale is actually shown by default
#ifdef COSMCTRL_NOD2D
bool COSMCtrl::DrawScaleBar(_In_ Gdiplus::Graphics& graphics, _In_ const CRect& /*rClip*/, _In_ const CRect& rClient)
{
  //Get the position of the scale bar
  const CPoint ptControl{GetControlPosition(m_ScaleBarAnchorPosition, m_ptOffsetScaleBar, rClient)};

  //Work out the left X and top Y value
  int nLeftX{ptControl.x};
  int nTopY{ptControl.y};

  //Call the helper function
  double fDistance{0};
  double fScaleDistance{0};
  bool bMetric{false};
  int nScaleLength{0};
  CalculateScaleBar(fDistance, fScaleDistance, bMetric, nScaleLength);

  //Create the text mode version of the distance formatted correctly
  CStringW sText{FormScaleBarText(fScaleDistance, bMetric)};

  //Get the default font to use
  HDC hDC{graphics.GetHDC()};
  NONCLIENTMETRICS ncm{};
  ncm.cbSize = sizeof(ncm);
  if (!SystemParametersInfo(SPI_GETNONCLIENTMETRICS, 0, &ncm, 0))
  {
    graphics.ReleaseHDC(hDC);
    return false;
  }
  Gdiplus::Font font{hDC, &ncm.lfMessageFont};
  graphics.ReleaseHDC(hDC);
  if (font.GetLastStatus() != Gdiplus::Ok)
    return false;

  //Setup the text alignment and get the height of the text 
  Gdiplus::StringFormat stringFormat;
  if (stringFormat.SetAlignment(Gdiplus::StringAlignmentCenter) != Gdiplus::Ok)
    return false;
  Gdiplus::RectF rectBoundingText;
  if (graphics.MeasureString(sText, -1, &font, Gdiplus::PointF{0, 0}, &stringFormat, &rectBoundingText) != Gdiplus::Ok)
    return false;

  const int nTextHeight{static_cast<int>(rectBoundingText.Height + 4)};
  if ((m_ScaleBarAnchorPosition == ControlAnchorPosition::BottomLeft) || (m_ScaleBarAnchorPosition == ControlAnchorPosition::BottomRight))
    nTopY -= nTextHeight;

  //Create the pen and brush we need
  Gdiplus::Pen penBlack{Gdiplus::Color{0, 0, 0}, 1};
  if (penBlack.GetLastStatus() != Gdiplus::Ok)
    return false;
  Gdiplus::SolidBrush blackBrush{Gdiplus::Color{0, 0, 0}};
  if (blackBrush.GetLastStatus() != Gdiplus::Ok)
    return false;

  if ((m_ScaleBarAnchorPosition == ControlAnchorPosition::TopRight) || (m_ScaleBarAnchorPosition == ControlAnchorPosition::BottomRight))
    nLeftX -= nScaleLength;

  //Draw the lines for the scale
  const int nScaleY{nTopY + nTextHeight};
  if (graphics.DrawLine(&penBlack, nLeftX, nScaleY - 3, nLeftX, nScaleY) != Gdiplus::Ok)
    return false;
  if (graphics.DrawLine(&penBlack, nLeftX, nScaleY, nLeftX + nScaleLength, nScaleY) != Gdiplus::Ok)
    return false;
  if (graphics.DrawLine(&penBlack, nLeftX + nScaleLength, nScaleY, nLeftX + nScaleLength, nScaleY - 3) != Gdiplus::Ok)
    return false;

  const Gdiplus::RectF rectText{static_cast<Gdiplus::REAL>(nLeftX), static_cast<Gdiplus::REAL>(nTopY), static_cast<Gdiplus::REAL>(nScaleLength), static_cast<Gdiplus::REAL>(static_cast<int>(rectBoundingText.Height + 4))};
  if (graphics.DrawString(sText, -1, &font, rectText, &stringFormat, &blackBrush) != Gdiplus::Ok)
    return false;

  return true;
}
#else
#pragma warning(suppress: 26429)
bool COSMCtrl::DrawScaleBar(_In_ CRenderTarget* pRenderTarget, _In_ const CRect& /*rClip*/, _In_ const CRect& rClient)
{
  //Get the position of the scale bar
  const CPoint ptControl{GetControlPosition(m_ScaleBarAnchorPosition, m_ptOffsetScaleBar, rClient)};

  //Work out the left X and top Y value
  auto fLeftX{static_cast<FLOAT>(ptControl.x)};
  auto fTopY{static_cast<FLOAT>(ptControl.y)};

  //Call the helper function
  double fDistance{0};
  double fScaleDistance{0};
  bool bMetric{false};
  int nScaleLength{0};
  CalculateScaleBar(fDistance, fScaleDistance, bMetric, nScaleLength);

  //Create the text mode version of the distance formatted correctly
  CString sText{FormScaleBarText(fScaleDistance, bMetric)};

  //Get the font size to use
  NONCLIENTMETRICS ncm{};
  ncm.cbSize = sizeof(ncm);
  if (!SystemParametersInfo(SPI_GETNONCLIENTMETRICS, 0, &ncm, 0))
    return false;
  CDC* pDC{GetDC()};
  if (pDC == nullptr)
    return false;
  const int nPixelsY{pDC->GetDeviceCaps(LOGPIXELSY)};
#pragma warning(suppress: 26486)
  ReleaseDC(pDC);
  auto fFontSize{static_cast<FLOAT>(static_cast<double>(ncm.lfMessageFont.lfHeight) * nPixelsY / 96.0)};
  if (fFontSize < 0)
    fFontSize = -fFontSize;

  //Create the brush we need
  CD2DSolidColorBrush brushBlack{pRenderTarget, D2D1::ColorF::Black};

  //Setup the text alignment and get the height of the text 
#pragma warning(suppress: 26485)
  CD2DTextFormat textFormat{pRenderTarget, ncm.lfMessageFont.lfFaceName, fFontSize};
#pragma warning(suppress: 26429)
  IDWriteTextFormat* pTextFormat{textFormat.Get()};
  ASSERT(pTextFormat != nullptr);
  pTextFormat->SetTextAlignment(DWRITE_TEXT_ALIGNMENT_CENTER);
  pTextFormat->SetParagraphAlignment(DWRITE_PARAGRAPH_ALIGNMENT_CENTER);
  CD2DTextLayout textLayout{pRenderTarget, sText, textFormat, pRenderTarget->GetSize()};
  DWRITE_TEXT_METRICS dWriteTextMetrics;
  if (FAILED(textLayout.Get()->GetMetrics(&dWriteTextMetrics)))
    return false;
  const FLOAT fTextHeight{dWriteTextMetrics.height + 4};
  if ((m_ScaleBarAnchorPosition == ControlAnchorPosition::BottomLeft) || (m_ScaleBarAnchorPosition == ControlAnchorPosition::BottomRight))
    fTopY -= fTextHeight;

  if ((m_ScaleBarAnchorPosition == ControlAnchorPosition::TopRight) || (m_ScaleBarAnchorPosition == ControlAnchorPosition::BottomRight))
    fLeftX -= nScaleLength;

  //Draw the lines for the scale
  const FLOAT fBottom{fTopY + fTextHeight};
  pRenderTarget->DrawLine(CD2DPointF{fLeftX, fBottom - 3}, CD2DPointF{fLeftX, fBottom}, &brushBlack);
  pRenderTarget->DrawLine(CD2DPointF{fLeftX, fBottom}, CD2DPointF{fLeftX + nScaleLength, fBottom}, &brushBlack);
  pRenderTarget->DrawLine(CD2DPointF{fLeftX + nScaleLength, fBottom}, CD2DPointF{fLeftX + nScaleLength, fBottom - 3}, &brushBlack);

  //Draw the text
  const CD2DRectF rectText{fLeftX, fTopY, fLeftX + nScaleLength, fTopY + dWriteTextMetrics.height + 4};
  pRenderTarget->DrawText(sText, rectText, &brushBlack, &textFormat);

  return true;
}
#endif //#ifdef COSMCTRL_NOD2D

CString COSMCtrl::FormCopyrightText()
{
  //By default load up the copyright details from a string table
  CString sText;
  VERIFY(sText.LoadString(IDS_OSMCTRL_COPYRIGHT));
  return sText;
}

#ifdef COSMCTRL_NOD2D
bool COSMCtrl::HitTest(_In_ const CPoint& point, _In_ const COSMCtrlPolyline& polyline, _In_ const CRect& rClient) const
{
  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

  //Note that hit testing for a polyline uses a standard solid line which is increased in width by 2 pixels

  //Draw the polyline if necessary
  if (polyline.m_bVisible && (polyline.m_Nodes.size() > 1) && (m_fZoom >= polyline.m_nMinZoomLevel) && (m_fZoom <= polyline.m_nMaxZoomLevel))
  {
    //Create the pen we need
    Gdiplus::Pen pen{polyline.m_colorPen, polyline.m_fLinePenWidth + 2};
    if (pen.GetLastStatus() != Gdiplus::Ok)
      return false;

    //Set up the brush attributes
    pen.SetEndCap(polyline.m_EndCap);
    pen.SetStartCap(polyline.m_StartCap);
    pen.SetLineJoin(polyline.m_LineJoin);

    //iterate thro all the positions in the polyline
    bSuccess = true;
    Gdiplus::GraphicsPath path;
    for (std::vector<COSMCtrlNode>::size_type j{1}; (j<polyline.m_Nodes.size()) && bSuccess; j++)
    {
      Gdiplus::PointF ptPosition1;
      Gdiplus::PointF ptPosition2;
#pragma warning(suppress: 26446)
      if (PositionToClient(polyline.m_Nodes[j-1].m_Position, rClient, ptPosition1) && PositionToClient(polyline.m_Nodes[j].m_Position, rClient, ptPosition2))
      {
        //Draw a line to the next node in the polyline
        bSuccess = (path.AddLine(ptPosition1.X, ptPosition1.Y, ptPosition2.X, ptPosition2.Y) == Gdiplus::Ok);
      }
      else
        bSuccess = false; //break out of the loop
    }

    if (bSuccess)
    {
      //Widen the path by the pen and do the actual hittesting
      if (path.Widen(&pen) == Gdiplus::Ok)
        bSuccess = path.IsVisible(point.x, point.y);
      else
        bSuccess = false;
    }
  }

  return bSuccess;
}
#else
bool COSMCtrl::HitTest(_In_ const CPoint& point, _In_ const COSMCtrlPolyline& polyline, _In_ const CRect& rClient) const
{
  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

  //Hittest the polyline if necessary
  if (polyline.m_bVisible && (polyline.m_Nodes.size() > 1) && (m_fZoom >= polyline.m_nMinZoomLevel) && (m_fZoom <= polyline.m_nMaxZoomLevel))
  {
    //Create the stroke style we need
    ATL::CComPtr<ID2D1StrokeStyle> strokeStyle;
    ID2D1Factory* pFactory{GetD2DFactory()};
    if (pFactory == nullptr)
      return false;
    D2D1_STROKE_STYLE_PROPERTIES strokeStyleProperties;
    strokeStyleProperties.startCap = polyline.m_StartCap;
    strokeStyleProperties.endCap = polyline.m_EndCap;
    strokeStyleProperties.dashCap = polyline.m_DashCap;
    strokeStyleProperties.lineJoin = polyline.m_LineJoin;
    strokeStyleProperties.miterLimit = polyline.m_fMiterLimit;
    strokeStyleProperties.dashStyle = polyline.m_DashStyle;
    strokeStyleProperties.dashOffset = polyline.m_fDashOffset;
    const HRESULT hr{pFactory->CreateStrokeStyle(&strokeStyleProperties, nullptr, 0, &strokeStyle)};
    if (FAILED(hr))
      return false;

    //Create the D2D path geometry which will do the hit testing for use
    CD2DPathGeometry pathGeometry{nullptr};
    if (FAILED(pathGeometry.Create(nullptr)))
      return false;
    CD2DGeometrySink geometrySink{pathGeometry};
    CD2DPointF D2DPoint;
#pragma warning(suppress: 26446)
    if (PositionToClient(polyline.m_Nodes[0].m_Position, rClient, D2DPoint))
    {
      geometrySink.BeginFigure(D2DPoint, D2D1_FIGURE_BEGIN_HOLLOW);
      bSuccess = true;
      for (std::vector<COSMCtrlNode>::size_type j{1}; (j<polyline.m_Nodes.size()) && bSuccess; j++)
      {
#pragma warning(suppress: 26446)
        const COSMCtrlNode& node{polyline.m_Nodes[j]};
        bSuccess = PositionToClient(node.m_Position, rClient, D2DPoint);
        if (bSuccess)
          geometrySink.AddLine(CD2DPointF{D2DPoint.x, D2DPoint.y});
      }
      geometrySink.EndFigure(D2D1_FIGURE_END_OPEN);
      if (!geometrySink.Close())
        return false;

      //Finally do the actual hittesting
      BOOL bSuccess2{FALSE};
      if (FAILED(pathGeometry.Get()->StrokeContainsPoint(CD2DPointF{static_cast<FLOAT>(point.x), static_cast<FLOAT>(point.y)}, polyline.m_fLinePenWidth, strokeStyle, nullptr, &bSuccess2)) || !bSuccess2)
        bSuccess = false;
    }
  }

  return bSuccess;
}
#endif //#ifdef COSMCTRL_NOD2D

#ifdef COSMCTRL_NOD2D
int COSMCtrl::DrawPolylines(_In_ Gdiplus::Graphics& graphics, _In_ const CRect& /*rClip*/, _In_ const CRect& rClient)
{
  //What will be the return value from this function
  int nPolylines{0};

  //iterate across all the polylines and draw them
  for (auto& polyline : m_Polylines)
  {
    //Increment the count for the return value if it was drawn successfully
    if (polyline->Draw(*this, graphics, rClient))
      ++nPolylines;
  }

  return nPolylines;
}
#else
int COSMCtrl::DrawPolylines(_In_ CRenderTarget* pRenderTarget, _In_ const CRect& /*rClip*/, _In_ const CRect& rClient)
{
  //What will be the return value from this function
  int nPolylines{0};

  //iterate across all the polylines and draw them
  for (auto& polyline : m_Polylines)
  {
    //Increment the count for the return value if it was drawn successfully
    if (polyline->Draw(*this, pRenderTarget, rClient))
      ++nPolylines;
  }

  return nPolylines;
}
#endif //#ifdef COSMCTRL_NOD2D

#ifdef COSMCTRL_NOD2D
bool COSMCtrl::HitTest(_In_ const CPoint& point, _In_ const COSMCtrlPolygon& polygon, _In_ const CRect& rClient) const
{
  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

  Gdiplus::GraphicsPath path1;
  Gdiplus::GraphicsPath path2;
  if (polygon.DrawInternal(*this, path1, rClient) && polygon.DrawOutline(*this, path2, rClient))
  {
    //Convert the paths into a combined region
    Gdiplus::Region region{&path1};
    if (region.GetLastStatus() == Gdiplus::Ok)
    {
      if ((region.Union(&path2) == Gdiplus::Ok) && region.IsVisible(point.x, point.y))
        bSuccess = true;
    }
  }

  return bSuccess;
}
#else
bool COSMCtrl::HitTest(_In_ const CPoint& point, _In_ const COSMCtrlPolygon& polygon, _In_ const CRect& rClient) const
{
  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

  //Hittest the polygon if necessary
  if (polygon.m_bVisible && (polygon.m_Nodes.size() > 1) && (m_fZoom >= polygon.m_nMinZoomLevel) && (m_fZoom <= polygon.m_nMaxZoomLevel))
  {
    //Create the D2D path geometry which will do the hit testing for the internal part of the polygon
    CD2DPathGeometry pathGeometry{nullptr};
    if (FAILED(pathGeometry.Create(nullptr)))
      return false;
    CD2DGeometrySink geometrySink{pathGeometry};
    CD2DPointF D2DPoint;
#pragma warning(suppress: 26446)
    if (PositionToClient(polygon.m_Nodes[0].m_Position, rClient, D2DPoint))
    {
      geometrySink.BeginFigure(D2DPoint, D2D1_FIGURE_BEGIN_FILLED);
      bSuccess = true;
      for (std::vector<COSMCtrlNode>::size_type j{1}; (j<polygon.m_Nodes.size()) && bSuccess; j++)
      {
#pragma warning(suppress: 26446)
        const COSMCtrlNode& node{polygon.m_Nodes[j]};
        bSuccess = PositionToClient(node.m_Position, rClient, D2DPoint);
        if (bSuccess)
          geometrySink.AddLine(CD2DPointF{D2DPoint.x, D2DPoint.y});
      }
      geometrySink.EndFigure(D2D1_FIGURE_END_CLOSED);
      if (!geometrySink.Close())
        return false;

      //Finally do the actual hittesting
      BOOL bSuccess2{FALSE};
      if (FAILED(pathGeometry.Get()->FillContainsPoint(CD2DPointF{static_cast<FLOAT>(point.x), static_cast<FLOAT>(point.y)}, nullptr, &bSuccess2)) || !bSuccess2)
        bSuccess = false;
    }

    //Try the outline of the polygon
    if (!bSuccess)
    {
      //Create the D2D path geometry which will do the hit testing for the external part of the polygon

      //Create the stroke style we need
      ATL::CComPtr<ID2D1StrokeStyle> strokeStyle;
      ID2D1Factory* pFactory{GetD2DFactory()};
      if (pFactory == nullptr)
        return false;
      D2D1_STROKE_STYLE_PROPERTIES strokeStyleProperties;
      strokeStyleProperties.startCap = polygon.m_StartCap;
      strokeStyleProperties.endCap = polygon.m_EndCap;
      strokeStyleProperties.dashCap = polygon.m_DashCap;
      strokeStyleProperties.lineJoin = polygon.m_LineJoin;
      strokeStyleProperties.miterLimit = polygon.m_fMiterLimit;
      strokeStyleProperties.dashStyle = polygon.m_DashStyle;
      strokeStyleProperties.dashOffset = polygon.m_fDashOffset;
      const HRESULT hr{pFactory->CreateStrokeStyle(&strokeStyleProperties, nullptr, 0, &strokeStyle)};
      if (FAILED(hr))
        return false;

      //Create the D2D path geometry which will do the hit testing for use
      CD2DPathGeometry pathGeometry2{nullptr};
      if (FAILED(pathGeometry2.Create(nullptr)))
        return false;
      CD2DGeometrySink geometrySink2{pathGeometry2};
#pragma warning(suppress: 26446)
      if (PositionToClient(polygon.m_Nodes[0].m_Position, rClient, D2DPoint))
      {
        geometrySink2.BeginFigure(D2DPoint, D2D1_FIGURE_BEGIN_FILLED);
        bSuccess = true;
        for (std::vector<COSMCtrlNode>::size_type j{1}; (j<polygon.m_Nodes.size()) && bSuccess; j++)
        {
#pragma warning(suppress: 26446)
          const COSMCtrlNode& node{polygon.m_Nodes[j]};
          bSuccess = PositionToClient(node.m_Position, rClient, D2DPoint);
          if (bSuccess)
            geometrySink2.AddLine(CD2DPointF{D2DPoint.x, D2DPoint.y});
        }
        geometrySink2.EndFigure(D2D1_FIGURE_END_CLOSED);
        if (!geometrySink2.Close())
          return false;

        //Finally do the actual hittesting
        BOOL bSuccess2 = FALSE;
        if (FAILED(pathGeometry2.Get()->StrokeContainsPoint(CD2DPointF{static_cast<FLOAT>(point.x), static_cast<FLOAT>(point.y)}, polygon.m_fLinePenWidth, strokeStyle, nullptr, &bSuccess2)) || !bSuccess2)
          bSuccess = false;
      }
    }
  }

  return bSuccess;
}
#endif //#ifdef COSMCTRL_NOD2D

#ifdef COSMCTRL_NOD2D
int COSMCtrl::DrawPolygons(_In_ Gdiplus::Graphics& graphics, _In_ const CRect& /*rClip*/, _In_ const CRect& rClient)
{
  //What will be the return value from this function
  int nPolygons{0};

  //iterate across all the polygons and draw them
  for (auto& polygon : m_Polygons)
  {
    //Attempt to draw it
    if (polygon->Draw(*this, graphics, rClient))
      ++nPolygons;
  }

  return nPolygons;
}
#else
int COSMCtrl::DrawPolygons(_In_ CRenderTarget* pRenderTarget, _In_ const CRect& /*rClip*/, _In_ const CRect& rClient)
{
  //What will be the return value from this function
  int nPolygons{0};

  //iterate across all the polygons and draw them
  for (auto& polygon : m_Polygons)
  {
    //Attempt to draw it
    if (polygon->Draw(*this, pRenderTarget, rClient))
      ++nPolygons;
  }

  return nPolygons;
}
#endif //#ifdef COSMCTRL_NOD2D

#ifdef COSMCTRL_NOD2D
COSMCtrl::MapItem COSMCtrl::HitTest(_In_ const CPoint& point, _In_ const COSMCtrlCircle& circle, _In_ const CRect& rClient) const
{
  if (circle.m_bVisible && (m_fZoom >= circle.m_nMinZoomLevel) && (m_fZoom <= circle.m_nMaxZoomLevel))
  {
    //Try the inside circle first
    Gdiplus::GraphicsPath path1;
    if (circle.Draw(*this, path1, false, rClient))
    {
      //Convert the path into a region
      Gdiplus::Region region{&path1};
      if (region.GetLastStatus() == Gdiplus::Ok)
      {
        if (region.IsVisible(point.x, point.y))
          return MapItem::Circle;
      }
    }

    //Try the outside circle
    Gdiplus::GraphicsPath path2;
    if (circle.Draw(*this, path2, true, rClient))
    {
      //Convert the path into a region
      Gdiplus::Region region{&path2};
      if (region.GetLastStatus() == Gdiplus::Ok)
      {
        if (region.IsVisible(point.x, point.y))
          return MapItem::CircleCircumference;
      }
    }
  }

  return MapItem::None;
}
#else
COSMCtrl::MapItem COSMCtrl::HitTest(_In_ const CPoint& point, _In_ const COSMCtrlCircle& circle, _In_ const CRect& rClient) const
{
  //Hittest the circle if necessary
  if (circle.m_bVisible && (m_fZoom >= circle.m_nMinZoomLevel) && (m_fZoom <= circle.m_nMaxZoomLevel))
  {
    //Create the stroke style we need
    ATL::CComPtr<ID2D1StrokeStyle> strokeStyle;
    ID2D1Factory* pFactory{GetD2DFactory()};
    if (pFactory == nullptr)
      return MapItem::None;

    const COSMCtrlPosition position90{COSMCtrlHelper::GetPosition(circle.m_Position, 90, circle.m_fRadius, nullptr)};
    const COSMCtrlPosition position270{COSMCtrlHelper::GetPosition(circle.m_Position, 270, circle.m_fRadius, nullptr)};
    CD2DPointF pt90;
    CD2DPointF pt270;
    CD2DPointF ptCenter;
    if (PositionToClient(position90, rClient, pt90) && PositionToClient(position270, rClient, pt270) && PositionToClient(circle.m_Position, rClient, ptCenter))
    {
      //Create the D2D ellipse geometry which will do the hit testing for use
      const FLOAT fRadius{(pt90.x - pt270.x) / 2};
      ATL::CComPtr<ID2D1EllipseGeometry> ellipseGeometry;
      if (FAILED(pFactory->CreateEllipseGeometry(CD2DEllipse{CD2DPointF{ptCenter.x, ptCenter.y}, CD2DSizeF{fRadius, fRadius}}, &ellipseGeometry)))
        return MapItem::None;

      //Do the hit testing on the inner circle
      BOOL bContainsPoint{FALSE};
      const CD2DPointF D2DPoint{static_cast<FLOAT>(point.x), static_cast<FLOAT>(point.y)};
      if (SUCCEEDED(ellipseGeometry->FillContainsPoint(D2DPoint, nullptr, &bContainsPoint)) && bContainsPoint)
        return MapItem::Circle;

      //Note we hardcode most of the circle stroke properties because it does not make sense to expose
      //these values in the case of a circle since there never will be any corners to the object
      //where most of the stroke properties would be actually used
      D2D1_STROKE_STYLE_PROPERTIES strokeStyleProperties;
      strokeStyleProperties.startCap = D2D1_CAP_STYLE_FLAT;
      strokeStyleProperties.endCap = D2D1_CAP_STYLE_FLAT;
      strokeStyleProperties.dashCap = D2D1_CAP_STYLE_FLAT;
      strokeStyleProperties.lineJoin = D2D1_LINE_JOIN_MITER;
      strokeStyleProperties.miterLimit = 1;
      strokeStyleProperties.dashStyle = circle.m_DashStyle;
      strokeStyleProperties.dashOffset = 0;
      const HRESULT hr{pFactory->CreateStrokeStyle(&strokeStyleProperties, nullptr, 0, &strokeStyle)};
      if (FAILED(hr))
        return MapItem::None;

      if (SUCCEEDED(ellipseGeometry->StrokeContainsPoint(D2DPoint, circle.m_fLinePenWidth, strokeStyle, nullptr, &bContainsPoint)) && bContainsPoint)
        return MapItem::CircleCircumference;
    }
  }

  return MapItem::None;
}
#endif //#ifdef COSMCTRL_NOD2D

#ifdef COSMCTRL_NOD2D
int COSMCtrl::DrawCircles(_In_ Gdiplus::Graphics& graphics, _In_ const CRect& /*rClip*/, _In_ const CRect& rClient)
{
  //What will be the return value from this function
  int nCircles{0};

  //iterate across all the circles and draw them
  for (auto& circle : m_Circles)
  {
    //Attempt to draw it
    if (circle->Draw(*this, graphics, rClient))
      ++nCircles;
  }

  return nCircles;
}
#else
int COSMCtrl::DrawCircles(_In_ CRenderTarget* pRenderTarget, _In_ const CRect& /*rClip*/, _In_ const CRect& rClient)
{
  //What will be the return value from this function
  int nCircles{0};

  //iterate across all the circles and draw them
  for (auto& circle : m_Circles)
  {
    //Attempt to draw it
    if (circle->Draw(*this, pRenderTarget, rClient))
      ++nCircles;
  }

  return nCircles;
}
#endif //#ifdef COSMCTRL_NOD2D

#ifdef COSMCTRL_NOD2D
int COSMCtrl::DrawMarkers(_In_ Gdiplus::Graphics& graphics, _In_ const CRect& /*rClip*/, _In_ const CRect& rClient)
{
  //What will be the return value from this function
  int nMarkers{0};

  //iterate across all the markers and draw them
  for (auto& marker : m_Markers)
  {
    //Increment the count for the return value if it was drawn successfully
    if (marker->Draw(*this, graphics, rClient))
      ++nMarkers;
  }

  return nMarkers;
}
#else
int COSMCtrl::DrawMarkers(_In_ CRenderTarget* pRenderTarget, _In_ const CRect& /*rClip*/, _In_ const CRect& rClient)
{
  //What will be the return value from this function
  int nMarkers{0};

  //iterate across all the markers and draw them
  for (auto& marker : m_Markers)
  {
    //Increment the count for the return value if it was drawn successfully
    if (marker->Draw(*this, pRenderTarget, rClient))
      ++nMarkers;
  }

  return nMarkers;
}
#endif //#ifdef COSMCTRL_NOD2D

#ifdef COSMCTRL_NOD2D
bool COSMCtrl::DrawCenterCrossHairs(_In_ Gdiplus::Graphics& graphics, _In_ const CRect& /*rClip*/, _In_ const CRect& rClient)
{
  //Get the center point of the map
  const Gdiplus::PointF ptCenter{static_cast<Gdiplus::REAL>(rClient.left + rClient.Width()/2.0), static_cast<Gdiplus::REAL>(rClient.top + rClient.Height()/2.0)};

  //Create the pen we need
  Gdiplus::Pen penBlack{Gdiplus::Color{0, 0, 0}, 1};
  if (penBlack.GetLastStatus() != Gdiplus::Ok)
    return false;

  //Draw the lines for the cross hairs
  if (graphics.DrawLine(&penBlack, ptCenter.X, ptCenter.Y - 10, ptCenter.X, ptCenter.Y - 2) != Gdiplus::Ok)
    return false;
  if (graphics.DrawLine(&penBlack, ptCenter.X, ptCenter.Y + 10, ptCenter.X, ptCenter.Y + 2) != Gdiplus::Ok)
    return false;
  if (graphics.DrawLine(&penBlack, ptCenter.X - 10, ptCenter.Y, ptCenter.X - 2, ptCenter.Y) != Gdiplus::Ok)
    return false;
  if (graphics.DrawLine(&penBlack, ptCenter.X + 10, ptCenter.Y, ptCenter.X + 2, ptCenter.Y) != Gdiplus::Ok)
    return false;

  return true;
}
#else
#pragma warning(suppress: 26429)
bool COSMCtrl::DrawCenterCrossHairs(_In_ CRenderTarget* pRenderTarget, _In_ const CRect& /*rClip*/, _In_ const CRect& rClient)
{
  //Get the center point of the map
  const CD2DPointF ptCenter{static_cast<FLOAT>(rClient.left + rClient.Width()/2.0), static_cast<FLOAT>(rClient.top + rClient.Height()/2.0)};

  //Create the brush we need
  CD2DSolidColorBrush brushBlack{pRenderTarget, D2D1::ColorF::Black};

  //Draw the lines for the cross hairs
  pRenderTarget->DrawLine(CD2DPointF{ptCenter.x, ptCenter.y - 10}, CD2DPointF{ptCenter.x, ptCenter.y - 2}, &brushBlack);
  pRenderTarget->DrawLine(CD2DPointF{ptCenter.x, ptCenter.y + 10}, CD2DPointF{ptCenter.x, ptCenter.y + 2}, &brushBlack);
  pRenderTarget->DrawLine(CD2DPointF{ptCenter.x - 10, ptCenter.y}, CD2DPointF{ptCenter.x - 2, ptCenter.y}, &brushBlack);
  pRenderTarget->DrawLine(CD2DPointF{ptCenter.x + 10, ptCenter.y}, CD2DPointF{ptCenter.x + 2, ptCenter.y}, &brushBlack);

  return true;
}
#endif //#ifdef COSMCTRL_NOD2D

#ifdef COSMCTRL_NOD2D
bool COSMCtrl::DrawTile(_In_ Gdiplus::Graphics& graphics, _In_ const Gdiplus::RectF& rTile, _In_ int nTileX, _In_ int nTileY)
{
  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

  //Form the path to the cache file which we want to draw
  const int nZoom{static_cast<int>(m_fZoom)};
  CStringW sFile{GetTileCachePath(m_sCacheDirectory, nZoom, nTileX, nTileY)};

  //Get the fractional value of the zoom
  double fInt{0};
  const double fFractionalZoom{modf(m_fZoom, &fInt)};

  //Try to obtain the tile from the in memory cache
  Gdiplus::Bitmap* pBitmap{nullptr};
  Gdiplus::CachedBitmap* pCachedBitmap{GetCachedBitmap(graphics, sFile, nTileX, nTileY, nZoom, pBitmap)};

  //Determine how the tile should be draw
  bool bStandardTile{false};
  if ((fFractionalZoom == 0) && (pCachedBitmap != nullptr))
    bStandardTile = true;

  //Draw the correct bitmaps
  if (bStandardTile)
  {
    //Draw the image to the screen at the specified position
#pragma warning(suppress: 26486)
    bSuccess = (graphics.DrawCachedBitmap(pCachedBitmap, static_cast<INT>(rTile.X), static_cast<INT>(rTile.Y)) == Gdiplus::Ok);
  }
  else if ((fFractionalZoom != 0) && (pBitmap != nullptr) && pBitmap->GetLastStatus() == Gdiplus::Ok)
  {
    //Draw the image to the screen at the specified position
    bSuccess = graphics.DrawImage(pBitmap, rTile, 0, 0, OSMTileWidth, OSMTileHeight, Gdiplus::UnitPixel) == Gdiplus::Ok;
  }
  else
  {
    //Try loading up the four images from the next zoom level and use them instead if configured to allow it
    if (m_bAllowNextZoomSqueeze && (nZoom < OSMMaxZoom))
    {
      //Load up the 4 images which form part of the next zoom level
      const int nNextZoom{nZoom + 1};
      const int nNextTileX{nTileX*2};
      const int nNextTileY{nTileY*2};
      sFile = GetTileCachePath(m_sCacheDirectory, nNextZoom, nNextTileX, nNextTileY);
      Gdiplus::Bitmap* pNextBitmap1{nullptr};
      const Gdiplus::CachedBitmap* pNextImage1{GetCachedBitmap(graphics, sFile, nNextTileX, nNextTileY, nNextZoom, pNextBitmap1)};
      if (pNextImage1 != nullptr)
      {
        sFile = GetTileCachePath(m_sCacheDirectory, nNextZoom, nNextTileX + 1, nNextTileY);
        Gdiplus::Bitmap* pNextBitmap2{nullptr};
        const Gdiplus::CachedBitmap* pNextImage2{GetCachedBitmap(graphics, sFile, nNextTileX + 1, nNextTileY, nNextZoom, pNextBitmap2)};
        if (pNextImage2 != nullptr)
        {
          sFile = GetTileCachePath(m_sCacheDirectory, nNextZoom, nNextTileX, nNextTileY + 1);
          Gdiplus::Bitmap* pNextBitmap3{nullptr};
          const Gdiplus::CachedBitmap* pNextImage3{GetCachedBitmap(graphics, sFile, nNextTileX, nNextTileY + 1, nNextZoom, pNextBitmap3)};
          if (pNextImage3 != nullptr)
          {
            sFile = GetTileCachePath(m_sCacheDirectory, nNextZoom, nNextTileX + 1, nNextTileY + 1);
            Gdiplus::Bitmap* pNextBitmap4{nullptr};
            const Gdiplus::CachedBitmap* pNextImage4{GetCachedBitmap(graphics, sFile, nNextTileX + 1, nNextTileY + 1, nNextZoom, pNextBitmap4)};
            if (pNextImage4 != nullptr)
            {
              //Work out the size of a quarter tile at the current zoom level
              const auto fOSMHalfTileWidth{static_cast<Gdiplus::REAL>((OSMTileWidth + (fFractionalZoom * OSMTileWidth))/2)};
              const auto fOSMHalfTileHeight{static_cast<Gdiplus::REAL>((OSMTileHeight + (fFractionalZoom * OSMTileHeight))/2)};

              //If all 4 images are rendered correctly then we do not need to draw the tile not available
              const Gdiplus::RectF tileSrc{0, 0, OSMTileWidth, OSMTileHeight};
              Gdiplus::RectF tileDest{rTile.X, rTile.Y, fOSMHalfTileWidth, fOSMHalfTileHeight};
              bSuccess = (graphics.DrawImage(pNextBitmap1, tileDest, tileSrc.X, tileSrc.Y, tileSrc.Width, tileSrc.Height, Gdiplus::UnitPixel) == Gdiplus::Ok);

              tileDest = Gdiplus::RectF{rTile.X + fOSMHalfTileWidth, rTile.Y, fOSMHalfTileWidth, fOSMHalfTileHeight};
              bSuccess = bSuccess && (graphics.DrawImage(pNextBitmap2, tileDest, tileSrc.X, tileSrc.Y, tileSrc.Width, tileSrc.Height, Gdiplus::UnitPixel) == Gdiplus::Ok);

              tileDest = Gdiplus::RectF{rTile.X, rTile.Y + fOSMHalfTileHeight, fOSMHalfTileWidth, fOSMHalfTileHeight};
              bSuccess = bSuccess && (graphics.DrawImage(pNextBitmap3, tileDest, tileSrc.X, tileSrc.Y, tileSrc.Width, tileSrc.Height, Gdiplus::UnitPixel) == Gdiplus::Ok);

              tileDest = Gdiplus::RectF{rTile.X + fOSMHalfTileWidth, rTile.Y + fOSMHalfTileHeight, fOSMHalfTileWidth, fOSMHalfTileHeight};
              bSuccess = bSuccess && (graphics.DrawImage(pNextBitmap4, tileDest, tileSrc.X, tileSrc.Y, tileSrc.Width, tileSrc.Height, Gdiplus::UnitPixel) == Gdiplus::Ok);
            }
          }
        }
      }
    }

    //Try loading the image for the previous zoom level and use it instead if configured to allow it
    if (!bSuccess && m_bAllowPreviousZoomStretch && (nZoom > OSMMinZoom))
    {
      const int nPrevZoom{nZoom - 1};
      const int nPrevTileX{nTileX/2};
      const int nPrevTileY{nTileY/2};
      sFile = GetTileCachePath(m_sCacheDirectory, nPrevZoom, nPrevTileX, nPrevTileY);
      Gdiplus::Bitmap* pPrevBitmap{nullptr};
      const Gdiplus::CachedBitmap* pPrevImage{GetCachedBitmap(graphics, sFile, nPrevTileX, nPrevTileY, nPrevZoom, pPrevBitmap)};
      if (pPrevImage != nullptr)
      {
        const Gdiplus::RectF tileSrc{static_cast<Gdiplus::REAL>((nTileX % 2) ? OSMHalfTileWidth : 0), static_cast<Gdiplus::REAL>((nTileY % 2) ? OSMHalfTileHeight : 0), OSMHalfTileWidth, OSMHalfTileHeight};
        bSuccess = graphics.DrawImage(pPrevBitmap, rTile, tileSrc.X, tileSrc.Y, tileSrc.Width, tileSrc.Height, Gdiplus::UnitPixel) == Gdiplus::Ok;
      }
    }
  }

  return bSuccess;
}
#else
#pragma warning(suppress: 26429)
bool COSMCtrl::DrawTile(_In_ CRenderTarget* pRenderTarget, _In_ const CD2DRectF& rTile, _In_ int nTileX, _In_ int nTileY)
{
  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

  //Form the path to the cache file which we want to draw
  const int nZoom{static_cast<int>(m_fZoom)};
  CString sFile{GetTileCachePath(m_sCacheDirectory, nZoom, nTileX, nTileY)};

  //Get the fractional value of the zoom
  double fInt{0};
  const double fFractionalZoom{modf(m_fZoom, &fInt)};

  //Try to obtain the tile from the in memory cache
  CD2DBitmap* pCachedBitmap{GetCachedBitmap(pRenderTarget, sFile, nTileX, nTileY, nZoom)};

  //Draw the correct bitmaps
  if (pCachedBitmap != nullptr)
  {
    //Draw the image to the screen at the specified position
    pRenderTarget->DrawBitmap(pCachedBitmap, rTile, 1.0, D2D1_BITMAP_INTERPOLATION_MODE_NEAREST_NEIGHBOR);
    bSuccess = true;
  }
  else
  {
    //Try loading up the four images from the next zoom level and use them instead if configured to allow it
    if (m_bAllowNextZoomSqueeze && (nZoom < OSMMaxZoom))
    {
      //Load up the 4 images which form part of the next zoom level
      const int nNextZoom{nZoom + 1};
      const int nNextTileX{nTileX*2};
      const int nNextTileY{nTileY*2};

      sFile = GetTileCachePath(m_sCacheDirectory, nNextZoom, nNextTileX, nNextTileY);
      CD2DBitmap* pNextImage1{GetCachedBitmap(pRenderTarget, sFile, nNextTileX, nNextTileY, nNextZoom)};
      if (pNextImage1 != nullptr)
      {
        sFile = GetTileCachePath(m_sCacheDirectory, nNextZoom, nNextTileX + 1, nNextTileY);
        CD2DBitmap* pNextImage2{GetCachedBitmap(pRenderTarget, sFile, nNextTileX + 1, nNextTileY, nNextZoom)};
        if (pNextImage2 != nullptr)
        {
          sFile = GetTileCachePath(m_sCacheDirectory, nNextZoom, nNextTileX, nNextTileY + 1);
          CD2DBitmap* pNextImage3{GetCachedBitmap(pRenderTarget, sFile, nNextTileX, nNextTileY + 1, nNextZoom)};
          if (pNextImage3 != nullptr)
          {
            sFile = GetTileCachePath(m_sCacheDirectory, nNextZoom, nNextTileX + 1, nNextTileY + 1);
            CD2DBitmap* pNextImage4{GetCachedBitmap(pRenderTarget, sFile, nNextTileX + 1, nNextTileY + 1, nNextZoom)};
            if (pNextImage4 != nullptr)
            {
              //Work out the size of a quarter tile at the current zoom level
              const auto fOSMHalfTileWidth{static_cast<FLOAT>((OSMTileWidth + (fFractionalZoom * OSMTileWidth))/2)};
              const auto fOSMHalfTileHeight{static_cast<FLOAT>((OSMTileHeight + (fFractionalZoom * OSMTileHeight))/2)};

              //If all 4 images are rendered correctly then we do not need to draw the tile not available
              const CD2DRectF tileSrc{0, 0, OSMTileWidth, OSMTileHeight};
              CD2DRectF tileDest{rTile.left, rTile.top, rTile.left + fOSMHalfTileWidth, rTile.top + fOSMHalfTileHeight};
              pRenderTarget->DrawBitmap(pNextImage1, tileDest, 1.0, D2D1_BITMAP_INTERPOLATION_MODE_LINEAR, &tileSrc);

              tileDest = CD2DRectF{rTile.left + fOSMHalfTileWidth, rTile.top, rTile.left + fOSMHalfTileWidth + fOSMHalfTileWidth, rTile.top + fOSMHalfTileHeight};
              pRenderTarget->DrawBitmap(pNextImage2, tileDest, 1.0, D2D1_BITMAP_INTERPOLATION_MODE_LINEAR, &tileSrc);

              tileDest = CD2DRectF{rTile.left, rTile.top + fOSMHalfTileHeight, rTile.left + fOSMHalfTileWidth, rTile.top + fOSMHalfTileHeight + fOSMHalfTileHeight};
              pRenderTarget->DrawBitmap(pNextImage3, tileDest, 1.0, D2D1_BITMAP_INTERPOLATION_MODE_LINEAR, &tileSrc);

              tileDest = CD2DRectF{rTile.left + fOSMHalfTileWidth, rTile.top + fOSMHalfTileHeight, rTile.left + fOSMHalfTileWidth + fOSMHalfTileWidth, rTile.top + fOSMHalfTileHeight + fOSMHalfTileHeight};
              pRenderTarget->DrawBitmap(pNextImage4, tileDest, 1.0, D2D1_BITMAP_INTERPOLATION_MODE_LINEAR, &tileSrc);

              bSuccess = true;
            }
          }
        }
      }
    }

    //Try loading the image for the previous zoom level and use it instead if configured to allow it
    if (!bSuccess && m_bAllowPreviousZoomStretch && (nZoom > OSMMinZoom))
    {
      const int nPrevZoom{nZoom - 1};
      const int nPrevTileX{nTileX/2};
      const int nPrevTileY{nTileY/2};
      sFile = GetTileCachePath(m_sCacheDirectory, nPrevZoom, nPrevTileX, nPrevTileY);
      CD2DBitmap* pPrevImage{GetCachedBitmap(pRenderTarget, sFile, nPrevTileX, nPrevTileY, nPrevZoom)};
      if (pPrevImage != nullptr)
      {
        CD2DRectF tileSrc{static_cast<FLOAT>((nTileX % 2) ? OSMHalfTileWidth : 0), static_cast<FLOAT>((nTileY % 2) ? OSMHalfTileHeight : 0), OSMHalfTileWidth, OSMHalfTileHeight};
        tileSrc.bottom += tileSrc.top;
        tileSrc.right += tileSrc.left;
        pRenderTarget->DrawBitmap(pPrevImage, rTile, 1.0, D2D1_BITMAP_INTERPOLATION_MODE_LINEAR, &tileSrc);
        bSuccess = true;
      }
    }
  }

  return bSuccess;
}
#endif //#ifdef COSMCTRL_NOD2D

void COSMCtrl::SetMaxCachedTiles(_In_ size_t nMaxTiles)
{
  //Validate our parameters
  ASSERT(nMaxTiles >= 0);

  //Hive away the new value
  m_nMaxCachedTiles = nMaxTiles;

  //Cull the cache down to the new maximum value
  PerformTileCacheMaintenance(m_nMaxCachedTiles);
}

void COSMCtrl::SetMaxGPSTracks(_In_ int nGPSTracks)
{
  m_nMaxGPSTracks = nGPSTracks;
  PerformInMemoryGPSTrackMaintenance(m_nMaxGPSTracks);
}

#pragma warning(suppress: 26440)
void COSMCtrl::PerformInMemoryGPSTrackMaintenance(_In_ std::vector<COSMCtrlNode>::size_type nMaxTracks)
{
  //Throw away any tracks which have exceeded the max size
  const auto nTracks{m_GPSTrack->m_Nodes.size()};
  if (nTracks > m_nMaxGPSTracks)
    m_GPSTrack->m_Nodes.erase(m_GPSTrack->m_Nodes.begin(), m_GPSTrack->m_Nodes.begin() + nTracks - nMaxTracks);
}

void COSMCtrl::EnsureBoundingRectForGPSTrackTriangle(_In_ const COSMCtrlPosition& position, _In_ CRect& rBounds, _In_ const CRect& rClient) const
{
#ifdef COSMCTRL_NOD2D
  Gdiplus::PointF ptPosition;
#else
  CD2DPointF ptPosition;
#endif //#ifdef COSMCTRL_NOD2D
  if (PositionToClient(position, rClient, ptPosition))
  {
  #ifdef COSMCTRL_NOD2D
    if (rBounds.left > ptPosition.X - 30)
      rBounds.left = static_cast<LONG>(ptPosition.X - 30);
    if (rBounds.right < ptPosition.X + 30)
      rBounds.right = static_cast<LONG>(ptPosition.X + 30);
    if (rBounds.top > ptPosition.Y - 30)
      rBounds.top = static_cast<LONG>(ptPosition.Y - 30);
    if (rBounds.bottom < ptPosition.Y + 30)
      rBounds.bottom = static_cast<LONG>(ptPosition.Y + 30);
  #else
    if (rBounds.left > ptPosition.x - 30)
      rBounds.left = static_cast<LONG>(ptPosition.x - 30);
    if (rBounds.right < ptPosition.x + 30)
      rBounds.right = static_cast<LONG>(ptPosition.x + 30);
    if (rBounds.top > ptPosition.y - 30)
      rBounds.top = static_cast<LONG>(ptPosition.y - 30);
    if (rBounds.bottom < ptPosition.y + 30)
      rBounds.bottom = static_cast<LONG>(ptPosition.y + 30);
  #endif //#ifdef COSMCTRL_NOD2D
  }
}

void COSMCtrl::AddGPSTrack(_In_ bool bGPSFix, _In_ double fLongitude, _In_ double fLatitude, _In_ double fBearing, _In_ double fSpeed, _In_ bool bBearingValid, _In_ bool bSpeedValid, _In_ double fHDOP)
{
  //Play a sound if we have just lost the fix or if we have just got a fix
  if ((!m_bGPSFix && bGPSFix) || (m_bGPSFix && !bGPSFix))
    MessageBeep(MB_ICONEXCLAMATION);

  //Store the current fix value
  m_bGPSFix = bGPSFix;

  //Add the track to the array
  if (bGPSFix)
  {
    const COSMCtrlNode node{fLongitude, fLatitude, fBearing, fSpeed, false, bBearingValid, bSpeedValid, fHDOP};
    m_GPSTrack->m_Nodes.push_back(node);
  }

  //Invalidate the track after we have updated it but before we have done the maintenance
  Invalidate(FALSE);

  //and perform maintenance if necessary
  if (bGPSFix)
    PerformInMemoryGPSTrackMaintenance(m_nMaxGPSTracks);
}

#ifdef COSMCTRL_NOD2D
bool COSMCtrl::DrawGPSLocation(_In_ Gdiplus::Graphics& graphics, _In_ const CRect& rClient)
{
  //What will be the return value from this function (assume the best)
  bool bSuccess{true};

  //Finally draw the current position as a triangle pointing in the direction of travel
  const auto nNodes{m_GPSTrack->m_Nodes.size()};
  if (nNodes)
  {
    //Pull out the current node in the track
#pragma warning(suppress: 26446)
    const COSMCtrlNode& node{m_GPSTrack->m_Nodes[nNodes - 1]};

    //Work out the client coordinates of the current position
    Gdiplus::PointF ptPosition;
    if (PositionToClient(node.m_Position, rClient, ptPosition))
    {
      //Draw a simple triangle in the current node pointing in the bearing position

      //First rotate the three positions by the current bearing
      constexpr const double X1{-10};
      constexpr const double Y1{20};
      constexpr const double X2{10};
      constexpr const double Y2{20};
      constexpr const double X3{0};
      constexpr const double Y3{-20};
      const double radBearing{node.m_Position.m_fBearing * 0.0174532925};
      const double cosBearing{cos(radBearing)};
      const double sinBearing{sin(radBearing)};
      const double X11{(cosBearing * X1) - (sinBearing * Y1)};
      const double Y11{(sinBearing * X1) + (cosBearing * Y1)};
      const double X21{(cosBearing * X2) - (sinBearing * Y2)};
      const double Y21{(sinBearing * X2) + (cosBearing * Y2)};
      const double X31{(cosBearing * X3) - (sinBearing * Y3)};
      const double Y31{(sinBearing * X3) + (cosBearing * Y3)};

      //then offset by the center position
      std::array<Gdiplus::PointF, 3> points{ { { static_cast<Gdiplus::REAL>(ptPosition.X + X11), static_cast<Gdiplus::REAL>(ptPosition.Y + Y11) },
                                               { static_cast<Gdiplus::REAL>(ptPosition.X + X21), static_cast<Gdiplus::REAL>(ptPosition.Y + Y21) },
                                               { static_cast<Gdiplus::REAL>(ptPosition.X + X31), static_cast<Gdiplus::REAL>(ptPosition.Y + Y31) } } };

      //And finally draw the triangle
      if (m_bGPSFix)
      {
        //Create the brush
        Gdiplus::SolidBrush brushTriangle{node.m_Position.m_bBearingValid ? m_colorGPSTrackPointer : m_colorGPSTrackNoBearingPointer};
        if (brushTriangle.GetLastStatus() != Gdiplus::Ok)
          return false;

        bSuccess = (graphics.FillPolygon(&brushTriangle, points.data(), static_cast<INT>(points.size())) == Gdiplus::Ok);
      }
      else
      {
        //Create the pens we need
        Gdiplus::Pen penTriangle{node.m_Position.m_bBearingValid ? m_colorGPSTrackPointer : m_colorGPSTrackNoBearingPointer};
        if (penTriangle.GetLastStatus() != Gdiplus::Ok)
          return false;

#pragma warning(suppress: 26472)
        bSuccess = (graphics.DrawPolygon(&penTriangle, points.data(), static_cast<INT>(points.size())) == Gdiplus::Ok);
      }
    }
    else
      bSuccess = false;
  }
  return bSuccess;
}
#else
bool COSMCtrl::DrawGPSHDOP(_In_ CRenderTarget* pRenderTarget, _In_ const CRect& rClient)
{
  const auto nNodes{m_GPSTrack->m_Nodes.size()};
  if (nNodes)
  {
    //Pull out the current node in the track
#pragma warning(suppress: 26446)
    const COSMCtrlNode& node{m_GPSTrack->m_Nodes[nNodes - 1]};

    //Draw the HDOP circle if required
    if (node.m_Position.m_fHDOP != 0)
    {
      //Calculate what the radius of the HDOP circle will be in pixels to decide if it is big enough to draw
      const COSMCtrlPosition posHDOP{COSMCtrlHelper::GetPosition(node.m_Position, 0, node.m_Position.m_fHDOP, nullptr)};
      CD2DPointF pointGPS;
      if (!PositionToClient(node.m_Position, rClient, pointGPS))
        return false;
      CD2DPointF pointHDOP;
      if (!PositionToClient(posHDOP, rClient, pointHDOP))
        return false;
      const double fX{static_cast<double>(pointGPS.x) - pointHDOP.x};
      const double fY{static_cast<double>(pointGPS.y) - pointHDOP.y};
      const double fHDOPRadiusInPixels{sqrt((fX * fX) + (fY * fY))};

      //Only bother drawing the HDOP circle if the radius is >= 10 pixels
      if (fHDOPRadiusInPixels >= 10)
      {
        COSMCtrlCircle hdopCircle;
        hdopCircle.m_colorBrush = D2D1::ColorF{0, 0, 255, 0.25F};
        hdopCircle.m_Position = node.m_Position;
        hdopCircle.m_fRadius = node.m_Position.m_fHDOP;
        hdopCircle.m_DashStyle = D2D1_DASH_STYLE_SOLID;
        if (!hdopCircle.Draw(*this, pRenderTarget, rClient))
          return false;
      }
    }
  }

  return true;
}

#pragma warning(suppress: 26429)
bool COSMCtrl::DrawGPSLocation(_In_ CRenderTarget* pRenderTarget, _In_ const CRect& rClient)
{
  //What will be the return value from this function (assume the best)
  bool bSuccess{true};

  //Finally draw the current position as a triangle pointing in the direction of travel
  const auto nNodes{m_GPSTrack->m_Nodes.size()};
  if (nNodes)
  {
    //Pull out the current node in the track
#pragma warning(suppress: 26446)
    const COSMCtrlNode& node{m_GPSTrack->m_Nodes[nNodes - 1]};

    //Work out the client coordinates of the current position
    CD2DPointF ptPosition;
    if (PositionToClient(node.m_Position, rClient, ptPosition))
    {
      //Draw the HDOP circle
      DrawGPSHDOP(pRenderTarget, rClient);

      //Create the geometries to represent the gps pointer
      CD2DPathGeometry triangle1{pRenderTarget};
      if (FAILED(triangle1.Create(pRenderTarget)))
        return false;
      CD2DGeometrySink geometrySink1{triangle1};
      CD2DPointF point{static_cast<FLOAT>(ptPosition.x - 10), static_cast<FLOAT>(ptPosition.y + 20)};
      geometrySink1.BeginFigure(point, D2D1_FIGURE_BEGIN_FILLED);
      std::array<CD2DPointF, 2> points{ { { static_cast<FLOAT>(ptPosition.x), static_cast<FLOAT>(ptPosition.y + 16) }, { static_cast<FLOAT>(ptPosition.x), static_cast<FLOAT>(ptPosition.y - 20) } } };
      geometrySink1.Get()->AddLines(points.data(), static_cast<UINT32>(points.size()));
      geometrySink1.EndFigure(D2D1_FIGURE_END_CLOSED);
      if (!geometrySink1.Close())
        return false;

      CD2DPathGeometry triangle2{pRenderTarget};
      if (FAILED(triangle2.Create(pRenderTarget)))
        return false;
      CD2DGeometrySink geometrySink2{triangle2};
      point = CD2DPointF{static_cast<FLOAT>(ptPosition.x), static_cast<FLOAT>(ptPosition.y + 16)};
      geometrySink2.BeginFigure(point, D2D1_FIGURE_BEGIN_FILLED);
#pragma warning(suppress: 26446)
      points[0].x = static_cast<FLOAT>(ptPosition.x + 10);
#pragma warning(suppress: 26446)
      points[0].y = static_cast<FLOAT>(ptPosition.y + 20);
#pragma warning(suppress: 26446)
      points[1].x = static_cast<FLOAT>(ptPosition.x);
#pragma warning(suppress: 26446)
      points[1].y = static_cast<FLOAT>(ptPosition.y - 20);
      geometrySink2.Get()->AddLines(points.data(), static_cast<UINT32>(points.size()));
      geometrySink2.EndFigure(D2D1_FIGURE_END_CLOSED);
      if (!geometrySink2.Close())
        return false;

      //Creation the rotation matrix for the GPS pointer
      D2D1_MATRIX_3X2_F oldMatrix;
      pRenderTarget->GetTransform(&oldMatrix);
      D2D1_MATRIX_3X2_F rotateMatrix;
      if (FAILED(AfxGetD2DState()->D2D1MakeRotateMatrix(static_cast<FLOAT>(node.m_Position.m_fBearing - m_fBearingOfTopOfMap), ptPosition, &rotateMatrix)))
        return false;
      pRenderTarget->SetTransform(rotateMatrix);

      //Draw the GPS pointer
      const D2D1::ColorF color1{255, 0, 0, 0.8F};
      CD2DBrushProperties brushProperties1{color1.a};
      CD2DSolidColorBrush brush1{pRenderTarget, color1, (color1.a == 1.0F) ? nullptr : &brushProperties1};
      if (m_bGPSFix)
        pRenderTarget->FillGeometry(&triangle1, &brush1);
      else
        pRenderTarget->DrawGeometry(&triangle1, &brush1);
      const D2D1::ColorF color2{235, 0, 0, 0.8F};
      const CD2DBrushProperties brushProperties2{color2.a};
      CD2DSolidColorBrush brush2{pRenderTarget, color2, nullptr};
      if (m_bGPSFix)
        pRenderTarget->FillGeometry(&triangle2, &brush2);
      else
        pRenderTarget->DrawGeometry(&triangle2, &brush2);
      bSuccess = true;

      //Restore the transform
      pRenderTarget->SetTransform(oldMatrix);
    }
    else
      bSuccess = false;
  }
  return bSuccess;
}
#endif //#ifdef COSMCTRL_NOD2D

#ifdef COSMCTRL_NOD2D
bool COSMCtrl::DrawGPSTrack(_In_ Gdiplus::Graphics& graphics, _In_ const CRect& rClient)
{
  //Draw all the points in the track as a standard polyline
  bool bSuccess{true};
  if (m_GPSTrack->m_Nodes.size() > 1)
    bSuccess = m_GPSTrack->Draw(*this, graphics, rClient);

  //Also draw the current GPS position
  if (!DrawGPSLocation(graphics, rClient))
    bSuccess = false;

  return bSuccess;
}
#else
bool COSMCtrl::DrawGPSTrack(_In_ CRenderTarget* pRenderTarget, _In_ const CRect& rClient)
{
  //Draw all the points in the track as a standard polyline
  bool bSuccess{true};
  if (m_GPSTrack->m_Nodes.size() > 1)
    bSuccess = m_GPSTrack->Draw(*this, pRenderTarget, rClient);

  //Also draw the current GPS position
  if (!DrawGPSLocation(pRenderTarget, rClient))
    bSuccess = false;

  return bSuccess;
}
#endif //#ifdef COSMCTRL_NOD2D

//A simple UDT which is used during the maintenance operation on the in memory tile cache
class COSMCtrlCachedTileCleanupItem
{
public:
//Member variables
  INT_PTR m_nArrayIndex{-1};
  int m_nInsertionCounter{-1};
};

void COSMCtrl::PerformTileCacheMaintenance(_In_ size_t nMaxTiles)
{
  //Validate our parameters
  ASSERT(nMaxTiles >= 0);

  //Get the current size of the array
  const auto nSize{m_CachedTiles.size()};

  //Nothing to do if we have not exceeded the max size yet
  if (nSize <= nMaxTiles)
    return;

  //Build up a temp sorted array of insertion values (in ascending "m_nInsertionCounter" order)
  std::vector<COSMCtrlCachedTileCleanupItem> cleanupArray;
  cleanupArray.reserve(nSize);
  for (size_t i{0}; i<nSize; i++)
  {
    COSMCtrlCachedTileCleanupItem cleanupItem;
    cleanupItem.m_nArrayIndex = i;
#pragma warning(suppress: 26446)
    cleanupItem.m_nInsertionCounter = m_CachedTiles[i].m_nInsertionCounter;
    cleanupArray.push_back(cleanupItem);
  }
  std::sort(cleanupArray.begin(), cleanupArray.end(), 
    [](_In_ const COSMCtrlCachedTileCleanupItem& item1, _In_ const COSMCtrlCachedTileCleanupItem& item2) noexcept
    {
      return item1.m_nInsertionCounter < item2.m_nInsertionCounter;
    }
  );

  //Leave only the items in "cleanupArray" which we want to delete
  const size_t nTilesToDelete{nSize - nMaxTiles};
  cleanupArray.resize(nTilesToDelete);

  //Delete all items from the in memory cache that have an index value as specified in "cleanupArray"
  m_CachedTiles.erase(std::remove_if(m_CachedTiles.begin(), m_CachedTiles.end(), [this, &cleanupArray](const COSMCtrlCachedTile& tile)
    {
      for (const auto& cleanupItem : cleanupArray)
      {
#pragma warning(suppress: 26446)
        if (cleanupItem.m_nArrayIndex == (&tile - &m_CachedTiles[0]))
          return true;
      }
      return false;
#pragma warning(suppress: 26440)
    }), m_CachedTiles.end());
}

CStringW COSMCtrl::GetFileNameForCache(_In_ const CStringW& sFile)
{
  CStringW sMemFile;
  CStringW sDrive;
  CStringW sDir;
  CStringW sFname;
  CStringW sExt;
  _wsplitpath_s(sFile, sDrive.GetBuffer(_MAX_DRIVE), _MAX_DRIVE, sDir.GetBuffer(_MAX_DIR), _MAX_DIR, sFname.GetBuffer(_MAX_FNAME), _MAX_FNAME, sExt.GetBuffer(_MAX_EXT), _MAX_EXT);
  sExt.ReleaseBuffer();
  sFname.ReleaseBuffer();
  sDir.ReleaseBuffer();
  sDrive.ReleaseBuffer();
  sFname += L"_mem";
  CStringW sPath;
  _wmakepath_s(sPath.GetBuffer(_MAX_PATH), _MAX_PATH, sDrive, sDir, sFname, sExt);
  sPath.ReleaseBuffer();
  if (CopyFileW(sFile, sPath, FALSE))
    sMemFile = sPath;

  return sMemFile;
}

void COSMCtrl::CalculateStartTilePositions(_In_ const CRect& rClient, _Out_ double& fClientX, _Out_ double& fClientY, _Out_ double& fEndClientX, _Out_ double& fEndClientY, _Out_ int& nStartX, _Out_ int& nStartY, _Out_ double& fOSMTileWidth, _Out_ double& fOSMTileHeight) noexcept
{
  //First thing we need to do is get the X and Y values for the center point of the map
  const int nZoom{static_cast<int>(m_fZoom)};
  double fStartX{COSMCtrlHelper::Longitude2TileX(m_CenterPosition.m_fLongitude, nZoom)};
  nStartX = static_cast<int>(fStartX);
  double fStartY{COSMCtrlHelper::Latitude2TileY(m_CenterPosition.m_fLatitude, nZoom)};
  nStartY = static_cast<int>(fStartY);

  //Work out the size of a tile at the current zoom level
  double fInt{0};
  double fFractionalZoom{modf(m_fZoom, &fInt)};
  fOSMTileWidth = OSMTileWidth + (fFractionalZoom * OSMTileWidth);
  fOSMTileHeight = OSMTileHeight + (fFractionalZoom * OSMTileHeight);

#ifdef COSMCTRL_NOD2D
  const double fHalfHeight{rClient.Height()/2.0};
  const double fClientRadiusY{fHalfHeight};
  const double fHalfWidth{rClient.Width()/2.0};
  const double fClientRadiusX{fHalfWidth};
  fEndClientX = rClient.right;
  fEndClientY = rClient.bottom;
#else
  double fHalfHeight{rClient.Height()/2.0};
  double fClientRadiusY{fHalfHeight};
  double fHalfWidth{rClient.Width()/2.0};
  double fClientRadiusX{fHalfWidth};
  if (m_fBearingOfTopOfMap != 0)
  {
    fClientRadiusX = max(fClientRadiusX, fClientRadiusY);
    fClientRadiusX *= sqrt(2.0);
    fClientRadiusY = fClientRadiusX;
  }
  fEndClientX = rClient.left + fHalfWidth + fClientRadiusX;
  fEndClientY = rClient.top + fHalfHeight + fClientRadiusY;
#endif //#ifdef COSMCTRL_NOD2D

  //Next we need to find the X and Y values which occur just before the top left position of the client area
  double fCenterX{rClient.left + fHalfWidth};
  double fCenterY{rClient.top + fHalfHeight};
  fClientX = fCenterX - (modf(fStartX, &fInt) * fOSMTileWidth);
  while ((fClientX + fClientRadiusX) > fCenterX)
  {
    --nStartX;
    fClientX -= fOSMTileWidth;
  }
  fClientY = fCenterY - (modf(fStartY, &fInt) * fOSMTileHeight);
  while ((fClientY + fClientRadiusY) > fCenterY)
  {
    --nStartY;
    fClientY -= fOSMTileWidth;
  }
}

template<typename T, typename Pred>
typename std::vector<T>::iterator insert_sorted(std::vector<T>& vec, T const& item, Pred pred)
{
#pragma warning(suppress: 26487 26489)
  return vec.insert(std::upper_bound(vec.begin(), vec.end(), item, pred), item);
}

#ifdef COSMCTRL_NOD2D
_Return_type_success_(return != 0) Gdiplus::CachedBitmap* COSMCtrl::GetCachedBitmap(_In_ Gdiplus::Graphics& graphics, _In_ const CStringW& sFile, _In_ int nTileX, _In_ int nTileY, _In_ int nZoom, _Inout_ Gdiplus::Bitmap*& pBitmap)
{
  //Set the output parameter to a default value
  pBitmap = nullptr;

  //Look through our cache for a match
  COSMCtrlCachedTile tile;
  tile.m_nTileX = nTileX;
  tile.m_nTileY = nTileY;
  tile.m_nZoom = nZoom;

#pragma warning(suppress: 26486)
  const auto iterFind{std::find_if(m_CachedTiles.cbegin(), m_CachedTiles.cend(), [&tile](const COSMCtrlCachedTile& element) noexcept
    {
      return (element.m_nTileX == tile.m_nTileX) && (element.m_nTileY == tile.m_nTileY) && (element.m_nZoom == tile.m_nZoom);
    })};
  if (iterFind != m_CachedTiles.cend())
  {
    const COSMCtrlCachedTile& tileFound{*iterFind};
#pragma warning(suppress: 26489)
    ASSERT(tileFound.m_pBitmap != nullptr);
#pragma warning(suppress: 26489)
    pBitmap = tileFound.m_pBitmap.get();
#pragma warning(suppress: 26489)
    ASSERT(tileFound.m_pCachedBitmap != nullptr);
#pragma warning(suppress: 26487 26489)
    return tileFound.m_pCachedBitmap.get();
  }

  //Get the name of the file we will load from
  CStringW sMemFile{GetFileNameForCache(sFile)};
  if (sMemFile.IsEmpty())
#pragma warning(suppress: 26487)
    return nullptr;

  //Ok, the tile is not in the cache, lets load it from file and add it to the cache
  tile.m_nInsertionCounter = m_nInMemoryCachedTilesInsertionCounter;
  ++m_nInMemoryCachedTilesInsertionCounter;
  const auto iterAdded{insert_sorted(m_CachedTiles, tile, [](_In_ const COSMCtrlCachedTile& tile1, _In_ const COSMCtrlCachedTile& tile2) noexcept
    {
      if (tile1.m_nTileX > tile2.m_nTileX)
        return true;
      else if (tile1.m_nTileX < tile2.m_nTileX)
        return false;
      else
      {
        if (tile1.m_nTileY > tile2.m_nTileY)
          return true;
        else if (tile1.m_nTileY < tile2.m_nTileY)
          return false;
        else
        {
          if (tile1.m_nZoom > tile2.m_nZoom)
            return true;
          else if (tile1.m_nZoom < tile2.m_nZoom)
            return false;
          else
            return false;
        }
      }
    }
  )};

  COSMCtrlCachedTile& addedTile{*iterAdded};
#pragma warning(suppress: 26489)
  addedTile.m_pBitmap = std::make_shared<Gdiplus::Bitmap>(sMemFile.GetString());
  if (addedTile.m_pBitmap->GetLastStatus() != Gdiplus::Ok)
  {
    m_CachedTiles.erase(iterAdded);
#pragma warning(suppress: 26487)
    return nullptr;
  }
#pragma warning(suppress: 26489)
  addedTile.m_pCachedBitmap = std::make_shared<Gdiplus::CachedBitmap>(addedTile.m_pBitmap.get(), &graphics);
  if (addedTile.m_pCachedBitmap->GetLastStatus() != Gdiplus::Ok)
  {
    m_CachedTiles.erase(iterAdded);
    return nullptr;
  }
  pBitmap = addedTile.m_pBitmap.get();
  return addedTile.m_pCachedBitmap.get();
#pragma warning(suppress: 26487)
}
#else
_Return_type_success_(return != 0) CD2DBitmap* COSMCtrl::GetCachedBitmap(_In_ CRenderTarget* pRenderTarget, _In_ const CString& sFile, _In_ int nTileX, _In_ int nTileY, _In_ int nZoom)
{
  //Look through our cache for a match
  COSMCtrlCachedTile tile;
  tile.m_nTileX = nTileX;
  tile.m_nTileY = nTileY;
  tile.m_nZoom = nZoom;

#pragma warning(suppress: 26486)
  const auto iterFind{std::find_if(m_CachedTiles.cbegin(), m_CachedTiles.cend(), [&tile](const COSMCtrlCachedTile& element) noexcept
    {
      return (element.m_nTileX == tile.m_nTileX) && (element.m_nTileY == tile.m_nTileY) && (element.m_nZoom == tile.m_nZoom);
    })};
  if (iterFind != m_CachedTiles.cend())
  {
    const COSMCtrlCachedTile& tileFound{*iterFind};
#pragma warning(suppress: 26489)
    ASSERT(tileFound.m_pCachedBitmap != nullptr);
#pragma warning(suppress: 26487 26489)
    return tileFound.m_pCachedBitmap;
  }

  //Get the name of the file we will load from
#ifdef _UNICODE
  CString sMemFile{GetFileNameForCache(sFile)};
#else
  CString sMemFile{GetFileNameForCache(CStringW{sFile})};
#endif //#ifdef _UNICODE
  if (sMemFile.IsEmpty())
#pragma warning(suppress: 26487)
    return nullptr;

  //Ok, the tile is not in the cache, lets load it from file and add it to the cache
  tile.m_nInsertionCounter = m_nInMemoryCachedTilesInsertionCounter;
  ++m_nInMemoryCachedTilesInsertionCounter;
  const auto iterAdded{insert_sorted(m_CachedTiles, tile, [](_In_ const COSMCtrlCachedTile& tile1, _In_ const COSMCtrlCachedTile& tile2) noexcept
    {
      if (tile1.m_nTileX > tile2.m_nTileX)
        return true;
      else if (tile1.m_nTileX < tile2.m_nTileX)
        return false;
      else
      {
        if (tile1.m_nTileY > tile2.m_nTileY)
          return true;
        else if (tile1.m_nTileY < tile2.m_nTileY)
          return false;
        else
        {
          if (tile1.m_nZoom > tile2.m_nZoom)
            return true;
          else if (tile1.m_nZoom < tile2.m_nZoom)
            return false;
          else
            return false;
        }
      }
    }
  )};

  COSMCtrlCachedTile& addedTile{*iterAdded};
#pragma warning(suppress: 26409)
  addedTile.m_pCachedBitmap = new CD2DBitmap(pRenderTarget, sMemFile);
  if (FAILED(addedTile.m_pCachedBitmap->Create(pRenderTarget)))
  {
    delete addedTile.m_pCachedBitmap;
    m_CachedTiles.erase(iterAdded);
#pragma warning(suppress: 26487)
    return nullptr;
  }
  return addedTile.m_pCachedBitmap;
}
#endif //#ifdef COSMCTRL_NOD2D

bool COSMCtrl::CreateSelectionPolygon()
{
  m_SelectionPolygon = std::make_unique<COSMCtrlPolygon>();
  return true;
}

bool COSMCtrl::CreateGPSTrack()
{
  m_GPSTrack = std::make_unique<COSMCtrlPolyline>();

  //Setup the various attributes for the GPS Track
  m_GPSTrack->m_DrawingStyle = COSMCtrlPolyline::DrawingStyle::NodesOnly;
  m_GPSTrack->m_bDraggable = false;
  m_GPSTrack->m_bEditable = false;
  m_GPSTrack->m_bHitTest = false;
  m_GPSTrack->m_fNodeWidth = 2;
#ifdef COSMCTRL_NOD2D
  m_GPSTrack->m_colorNode = Gdiplus::Color{0, 0, 255};
#else
  m_GPSTrack->m_colorNode = D2D1::ColorF{0, 0, 255};
#endif //#ifdef COSMCTRL_NOD2D

  return true;
}

#ifdef COSMCTRL_NOD2D
bool COSMCtrl::Draw(_In_ Gdiplus::Graphics& graphics, _In_ const CRect& rClient, _In_opt_ const CRect* prClip, _In_ bool bDrawScrollRose, _In_ bool bDrawZoomBar, _In_ bool bDrawScaleBar, _In_ bool bDrawMarkers, _In_ bool bDrawPolylines, _In_ bool bDrawPolygons, _In_ bool bDrawCircles, _In_ bool bDrawGPSTracks)
{
  //Work out the clipping rect
  CRect rClip{rClient};
  if (prClip != nullptr)
    rClip = *prClip;

  //Calculate the starting position for the tiles
  double fClientX{0};
  double fClientY{0};
  double fEndClientX{0};
  double fEndClientY{0};
  int nStartX{0};
  int nStartY{0};
  double fOSMTileWidth{0};
  double fOSMTileHeight{0};
  CalculateStartTilePositions(rClient, fClientX, fClientY, fEndClientX, fEndClientY, nStartX, nStartY, fOSMTileWidth, fOSMTileHeight);

  //Iterate across the checkerboard of tiles
  const int nZoom{static_cast<int>(m_fZoom)};
  const int nMaxTile{static_cast<int>(pow(2.0, nZoom))};
  auto fY{static_cast<Gdiplus::REAL>(fClientY)};
  int nTileY{nStartY};
  while (fY <= fEndClientY)
  {
    auto fX{static_cast<Gdiplus::REAL>(fClientX)};
    int nTileX{nStartX};
    while (fX <= fEndClientX)
    {
      //Perform wrapping of invalid x tile values to valid values
      while (nTileX < 0)
        nTileX += nMaxTile;
      while (nTileX >= nMaxTile)
        nTileX -= nMaxTile;

      //Form the rect to the tile
      const Gdiplus::RectF rTile{fX, fY, static_cast<Gdiplus::REAL>(fOSMTileWidth), static_cast<Gdiplus::REAL>(fOSMTileHeight)};

      //Only draw the tile if it intersects the clip rect
      Gdiplus::RectF rIntersect{static_cast<Gdiplus::REAL>(rClip.left), static_cast<Gdiplus::REAL>(rClip.top), static_cast<Gdiplus::REAL>(rClip.Width()), static_cast<Gdiplus::REAL>(rClip.Height())};
      if (rIntersect.Intersect(rTile))
      {
        if ((nTileY >= 0) && (nTileY < nMaxTile))
        {
          //Try to draw the tile
          if (!DrawTile(graphics, rTile, nTileX, nTileY))
          {
            //Draw the "Tile not available" cell
            DrawTileNotAvailable(graphics, rTile);
          }
          else if (m_bDrawTileOutlines)
          {
            //Draw the tile outline cell
            DrawTileOutline(graphics, rTile);
          }
        }
        else
        {
          //Draw the "Tile not available" cell
          DrawTileNotAvailable(graphics, rTile);
        }
      }

      //Move onto the next column
      fX += static_cast<Gdiplus::REAL>(fOSMTileWidth);
      ++nTileX;
    }

    //Move onto the next row
    fY += static_cast<Gdiplus::REAL>(fOSMTileHeight);
    ++nTileY;
  }

  //Maintain the size of the cache (Note up to this point we can temporarily have more tiles in the cache than the maximum amount)
  PerformTileCacheMaintenance(m_nMaxCachedTiles);

  //Draw the GPS positions
  if (bDrawGPSTracks)
    DrawGPSTrack(graphics, rClient);

  //Draw the cross hairs if necessary
  if (m_bDrawCenterCrossHairs)
    DrawCenterCrossHairs(graphics, rClip, rClient);

  //Draw the selection polygon if necessary
  if (m_SelectionPolygon->m_Nodes.size())
    m_SelectionPolygon->Draw(*this, graphics, rClient);

  //Next draw the polygons
  if (bDrawPolygons)
    DrawPolygons(graphics, rClip, rClient);

  //Next draw the circles
  if (bDrawCircles)
    DrawCircles(graphics, rClip, rClient);

  //Next draw the polylines
  if (bDrawPolylines)
    DrawPolylines(graphics, rClip, rClient);

  //Next draw the markers
  if (bDrawMarkers)
    DrawMarkers(graphics, rClip, rClient);

  //Next draw the scroll rose
  if (bDrawScrollRose)
    DrawScrollRose(graphics, rClip, rClient);
  else
  {
    m_rNorthScrollRose.SetRectEmpty();
    m_rSouthScrollRose.SetRectEmpty();
    m_rEastScrollRose.SetRectEmpty();
    m_rWestScrollRose.SetRectEmpty();
  }

  //Next draw the zoom bar
  if (bDrawZoomBar)
  {
    if (!m_bDrawZoomBarAsSlider)
      DrawZoomBar(graphics, rClip, rClient);
    else
    {
      //Force a redraw of the slider control if necessary
      CRect rZoom;
      if (m_ctrlZoomBar.GetSafeHwnd())
      {
        m_ctrlZoomBar.GetWindowRect(rZoom);
        ScreenToClient(&rZoom);
        CRect rIntersect;
        if (rIntersect.IntersectRect(rClip, rZoom))
        {
          m_ctrlZoomBar.Invalidate(FALSE);
          m_ctrlZoomBar.UpdateWindow();
        }
      }
    }
  }
  else
  {
    m_rZoomInZoomBar.SetRectEmpty();
    m_rZoomOutZoomBar.SetRectEmpty();
    m_rZoomBar.SetRectEmpty();
  }

  //Next draw the scale bar
  if (bDrawScaleBar)
    DrawScaleBar(graphics, rClip, rClient);

  //Force a redraw of the copyright control if necessary
  if (m_ctrlCopyright.GetSafeHwnd())
  {
    CRect rCopyright;
    m_ctrlCopyright.GetWindowRect(rCopyright);
    ScreenToClient(&rCopyright);
    CRect rIntersect;
    if (rIntersect.IntersectRect(rClip, rCopyright))
    {
      m_ctrlCopyright.Invalidate(FALSE);
      m_ctrlCopyright.UpdateWindow();
    }
  }

  return true;
}
#else
HRESULT COSMCtrl::Draw(_In_ CRenderTarget* pRenderTarget, _In_ const CRect& rClient, _In_opt_ const CRect* prClip, _In_ bool bDrawScrollRose, _In_ bool bDrawZoomBar, _In_ bool bDrawScaleBar, _In_ bool bDrawMarkers, _In_ bool bDrawPolylines, _In_ bool bDrawPolygons, _In_ bool bDrawCircles, _In_ bool bDrawGPSTracks, _In_ bool bResetDPI)
{
  //Validate our parameters
#pragma warning(suppress: 26496)
  AFXASSUME(pRenderTarget != nullptr);

  //Creation the rotation matrix for the full map if necessary
  D2D1_MATRIX_3X2_F oldMatrix;
  pRenderTarget->GetTransform(&oldMatrix);
  D2D1_MATRIX_3X2_F rotateMatrix;
  pRenderTarget->GetTransform(&rotateMatrix);
  if (m_fBearingOfTopOfMap != 0)
  {
    //Get the center point of the map
    const CD2DPointF ptCenter{static_cast<FLOAT>(rClient.left + rClient.Width()/2.0), static_cast<FLOAT>(rClient.top + rClient.Height()/2.0)};

    const HRESULT hr{AfxGetD2DState()->D2D1MakeRotateMatrix(static_cast<FLOAT>(-m_fBearingOfTopOfMap), ptCenter, &rotateMatrix)};
    if (FAILED(hr))
      return hr;
    pRenderTarget->SetTransform(rotateMatrix);
  }

  //Set the DPI settings back to default because we will be doing our own bitmap scaling
  if (bResetDPI)
    pRenderTarget->SetDpi(CD2DSizeF(96, 96));

  //Work out the clipping rect
  CRect rClip{rClient};
  if (prClip != nullptr)
    rClip = *prClip;

  //Calculate the starting position for the tiles
  double fClientX{0};
  double fClientY{0};
  double fEndClientX{0};
  double fEndClientY{0};
  int nStartX{0};
  int nStartY{0};
  double fOSMTileWidth{0};
  double fOSMTileHeight{0};
  CalculateStartTilePositions(rClient, fClientX, fClientY, fEndClientX, fEndClientY, nStartX, nStartY, fOSMTileWidth, fOSMTileHeight);

  //Iterate across the checkerboard of tiles
  const int nZoom{static_cast<int>(m_fZoom)};
  const int nMaxTile{static_cast<int>(pow(2.0, nZoom))};
  auto fY{static_cast<FLOAT>(fClientY)};
  int nTileY{nStartY};
  while (fY <= fEndClientY)
  {
    auto fX{static_cast<FLOAT>(fClientX)};
    int nTileX{nStartX};
    while (fX <= fEndClientX)
    {
      //Perform wrapping of invalid x tile values to valid values
      while (nTileX < 0)
        nTileX += nMaxTile;
      while (nTileX >= nMaxTile)
        nTileX -= nMaxTile;

      //Form the rect to the tile
      const CD2DRectF rTile{fX, fY, fX + static_cast<FLOAT>(fOSMTileWidth), fY + static_cast<FLOAT>(fOSMTileHeight)};

      if ((nTileY >= 0) && (nTileY < nMaxTile))
      {
        //Try to draw the tile
        const bool bTileDrawn{DrawTile(pRenderTarget, rTile, nTileX, nTileY)};
        if (!bTileDrawn)
        {
          //Draw the "Tile not available" cell
          DrawTileNotAvailable(pRenderTarget, rTile);
        }
        else if (m_bDrawTileOutlines)
        {
          //Draw the tile outline cell
          DrawTileOutline(pRenderTarget, rTile);
        }
      }
      else
      {
        //Draw the "Tile not available" cell
        DrawTileNotAvailable(pRenderTarget, rTile);
      }

      //Move onto the next column
      fX += static_cast<FLOAT>(fOSMTileWidth);
      ++nTileX;
    }

    //Move onto the next row
    fY += static_cast<FLOAT>(fOSMTileHeight);
    ++nTileY;
  }

  //Maintain the size of the cache (Note up to this point we can temporarily have more tiles in the cache than the maximum amount)
  PerformTileCacheMaintenance(m_nMaxCachedTiles);

  //Everything after this point is not using a transform
  pRenderTarget->SetTransform(oldMatrix);

  //Draw the GPS positions
  if (bDrawGPSTracks)
    DrawGPSTrack(pRenderTarget, rClient);

  //Draw the cross hairs if necessary
  if (m_bDrawCenterCrossHairs)
    DrawCenterCrossHairs(pRenderTarget, rClip, rClient);

  //Draw the selection polygon if necessary
  if (m_SelectionPolygon->m_Nodes.size())
    m_SelectionPolygon->Draw(*this, pRenderTarget, rClient);

  //Next draw the polygons
  if (bDrawPolygons)
    DrawPolygons(pRenderTarget, rClip, rClient);

  //Next draw the circles
  if (bDrawCircles)
    DrawCircles(pRenderTarget, rClip, rClient);

  //Next draw the polylines
  if (bDrawPolylines)
    DrawPolylines(pRenderTarget, rClip, rClient);

  //Next draw the markers
  if (bDrawMarkers)
    DrawMarkers(pRenderTarget, rClip, rClient);

  //Next draw the scroll rose
  if (bDrawScrollRose)
    DrawScrollRose(pRenderTarget, rClip, rClient);
  else
  {
    m_rNorthScrollRose.SetRectEmpty();
    m_rSouthScrollRose.SetRectEmpty();
    m_rEastScrollRose.SetRectEmpty();
    m_rWestScrollRose.SetRectEmpty();
  }

  //Next draw the zoom bar
  if (bDrawZoomBar)
  {
    if (!m_bDrawZoomBarAsSlider)
      DrawZoomBar(pRenderTarget, rClip, rClient);
    else
    {
      //Force a redraw of the slider control if necessary
      CRect rZoom;
      if (m_ctrlZoomBar.GetSafeHwnd())
      {
        m_ctrlZoomBar.GetWindowRect(rZoom);
        ScreenToClient(&rZoom);
        CRect rIntersect;
        if (rIntersect.IntersectRect(rClip, rZoom))
        {
          m_ctrlZoomBar.Invalidate(FALSE);
          m_ctrlZoomBar.UpdateWindow();
        }
      }
    }
  }
  else
  {
    m_rZoomInZoomBar.SetRectEmpty();
    m_rZoomOutZoomBar.SetRectEmpty();
    m_rZoomBar.SetRectEmpty();
  }

  //Next draw the scale bar
  if (bDrawScaleBar)
    DrawScaleBar(pRenderTarget, rClip, rClient);

  //Force a redraw of the copyright control if necessary
  if (m_ctrlCopyright.GetSafeHwnd())
  {
    CRect rCopyright;
    m_ctrlCopyright.GetWindowRect(rCopyright);
    ScreenToClient(&rCopyright);
    CRect rIntersect;
    if (rIntersect.IntersectRect(rClip, rCopyright))
    {
      m_ctrlCopyright.Invalidate(FALSE);
      m_ctrlCopyright.UpdateWindow();
    }
  }

  return S_OK;
}
#endif //#ifdef COSMCTRL_NOD2D

#ifdef COSMCTRL_NOD2D
#pragma warning(suppress: 26429)
bool COSMCtrl::Draw(_In_ CDC* pDC, _In_ const CRect& rSource, _In_opt_ const CRect* prDest, _In_ bool bDrawScrollRose, _In_ bool bDrawZoomBar, _In_ bool bDrawScaleBar, _In_ bool bDrawMarkers, _In_ bool bDrawPolylines, _In_ bool bDrawPolygons, _In_ bool bDrawCircles, _In_ bool bDrawGPSTracks)
{
  //Create a GDI+ graphics object backed by a bitmap
  Gdiplus::Bitmap bitmap{rSource.Width(), rSource.Height()};
  if (bitmap.GetLastStatus() != Gdiplus::Ok)
  {
    TRACE(_T("COSMCtrl::Draw, failed to create Gdiplus::Bitmap object\n"));
    return false;
  }

  //And from the bitmap create a GDI+ Graphics object
  Gdiplus::Graphics graphics{&bitmap};
  if (graphics.GetLastStatus() != Gdiplus::Ok)
  {
    TRACE(_T("COSMCtrl::Draw, failed to create Gdiplus::Graphics object\n"));
    return false;
  }

  //Setup appropriate values for the graphics object
  graphics.SetSmoothingMode(Gdiplus::SmoothingModeHighQuality);

  //Delegate to the other version
  const bool bSuccess{Draw(graphics, rSource, nullptr, bDrawScrollRose, bDrawZoomBar, bDrawScaleBar, bDrawMarkers, bDrawPolylines, bDrawPolygons, bDrawCircles, bDrawGPSTracks)};
  if (bSuccess)
  {
    //Create a graphic object associated with the DC
    Gdiplus::Graphics graphicsDC{pDC->GetSafeHdc()};
    if (graphicsDC.GetLastStatus() != Gdiplus::Ok)
    {
      TRACE(_T("COSMCtrl::Draw, failed to create Gdiplus::Graphics object from DC\n"));
      return false;
    }

    //Ensure the Graphics object is in Pixel units
    graphicsDC.SetPageUnit(Gdiplus::UnitPixel);

    //Transfer the bitmap which we have drawn on via GDI+ to the graphics object which wraps the DC (this in effect implements double buffering which 
    //helps avoid flicking if we were to render to the DC graphics object directly)
    CRect rDest{rSource};
    if (prDest != nullptr)
      rDest = *prDest;
    if (graphicsDC.DrawImage(&bitmap, rDest.left, rDest.top, rDest.Width(), rDest.Height()) != Gdiplus::Ok)
    {
      TRACE(_T("COSMCtrl::Draw, failed in call to DrawImage\n"));
      return false;
    }
  }
  else
  {
    TRACE(_T("COSMCtrl::Draw, failed in call to Draw\n"));
  }

  return bSuccess;
}
#else
#pragma warning(suppress: 26429)
HRESULT COSMCtrl::CreateImageFromWICBitmap(_In_ IWICBitmapSource* WICBitmapSource, _Inout_ ATL::CImage& image)
{
  //get image attributes and check for a valid image
  UINT nWidth{0};
  UINT nHeight{0};
  const HRESULT hr{WICBitmapSource->GetSize(&nWidth, &nHeight)};
  if (FAILED(hr))
    return hr;
  if ((nWidth == 0) || (nHeight == 0))
    return E_INVALIDARG;

  //Create the image
#pragma warning(suppress: 26472)
  if (!image.Create(nWidth, -(static_cast<int>(nHeight)), 32))
    return MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WIN32, GetLastError());

  //extract the WIC bits into the image
  const UINT cbStride{nWidth * 4};
  const UINT cbImage{cbStride * nHeight};
  return WICBitmapSource->CopyPixels(nullptr, cbStride, cbImage, static_cast<BYTE*>(image.GetBits()));
}

HRESULT COSMCtrl::Draw(_In_ CDC* pDC, _In_ const CRect& rSource, _In_opt_ const CRect* prDest, _In_ bool bDrawScrollRose, _In_ bool bDrawZoomBar, _In_ bool bDrawScaleBar, _In_ bool bDrawMarkers, _In_ bool bDrawPolylines, _In_ bool bDrawPolygons, _In_ bool bDrawCircles, _In_ bool bDrawGPSTracks)
{
  //Get the destination rect
  CRect rDest{rSource};
  if (prDest != nullptr)
    rDest = *prDest;

  //Create the WICBitmap which we will use as the backing for the rendering
  ATL::CComPtr<IWICBitmap> wicBitmap;
  IWICImagingFactory* pImageFactory{AfxGetD2DState()->GetWICFactory()};
  if (pImageFactory == nullptr)
    return E_POINTER;
  HRESULT hr{pImageFactory->CreateBitmap(rSource.Width(), rSource.Height(), GUID_WICPixelFormat32bppBGR, WICBitmapCacheOnLoad, &wicBitmap)};
  if (FAILED(hr))
  {
    TRACE(_T("COSMCtrl::Draw, failed in call to IWICImagingFactory::CreateBitmap, Error:%08X\n"), hr);
    return hr;
  }

  //Get the D2D factory
  ID2D1Factory* pD2DFactory{GetD2DFactory()};
  if (pD2DFactory == nullptr)
  {
    TRACE(_T("COSMCtrl::Draw, no D2D factory available\n"));
    return E_POINTER;
  }

  //Create the WIC render target;
  ID2D1RenderTarget* pRenderTarget{nullptr};
  hr = pD2DFactory->CreateWicBitmapRenderTarget(wicBitmap, D2D1::RenderTargetProperties(D2D1_RENDER_TARGET_TYPE_DEFAULT, D2D1::PixelFormat(DXGI_FORMAT_B8G8R8A8_UNORM, D2D1_ALPHA_MODE_IGNORE),
                                                96, 96, D2D1_RENDER_TARGET_USAGE_NONE, D2D1_FEATURE_LEVEL_DEFAULT), &pRenderTarget);
  if (FAILED(hr))
  {
    TRACE(_T("COSMCtrl::Draw, failed in call to IWICImagingFactory::CreateBitmap, Error:%08X\n"), hr);
    return hr;
  }
  CRenderTarget renderTarget;
  renderTarget.Attach(pRenderTarget);

  //Next do the drawing to the render target
  renderTarget.BeginDraw();
  hr = Draw(&renderTarget, rSource, nullptr, bDrawScrollRose, bDrawZoomBar, bDrawScaleBar, bDrawMarkers, bDrawPolylines, bDrawPolygons, bDrawCircles, bDrawGPSTracks, false);
  if (FAILED(hr))
  {
    TRACE(_T("COSMCtrl::Draw, failed in call to Draw, Error:%08X\n"), hr);
    return hr;
  }
  hr = renderTarget.EndDraw();
  if (FAILED(hr))
  {
    TRACE(_T("COSMCtrl::Draw, failed in call to CRenderTarget::EndDraw, Error:%08X\n"), hr);
    return hr;
  }

  //Next lets create a DIB section from the WIC Bitmap
  ATL::CImage image;
#pragma warning(suppress: 26486)
  hr = CreateImageFromWICBitmap(wicBitmap, image);
  if (FAILED(hr))
  {
    TRACE(_T("COSMCtrl::Draw, failed in call to CreateImageFromWICBitmap. Error:%08X\n"), hr);
    return E_FAIL;
  }

  //Finally bitblit the DIB section to the DC
#pragma warning(suppress: 26496)
  AFXASSUME(pDC != nullptr);
  if (!image.StretchBlt(pDC->GetSafeHdc(), rDest, rSource, SRCCOPY))
  {
    const DWORD dwLastError{GetLastError()};
    TRACE(_T("COSMCtrl::Draw, failed in call to ATL::CImage::StretchBlt, Error:%d\n"), dwLastError);
    return MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WIN32, dwLastError);
  }

  return S_OK;
}
#endif //#ifdef COSMCTRL_NOD2D

#pragma warning(suppress: 26434 26440)
void COSMCtrl::OnPaint()
{
#ifdef COSMCTRL_NOD2D
  //Get the extent of the client area
  CRect rClient;
  GetClientRect(&rClient);

  //device context for painting
  CPaintDC dc{this};

  //Get the clipbox for the DC
  const CRect rClip{dc.m_ps.rcPaint};

  //Create a memory DC compatible with "dc"
  CDC memDC;
  if (!memDC.CreateCompatibleDC(&dc))
  {
    TRACE(_T("COSMCtrl::OnPaint, failed to create memory DC\n"));
    return;
  }

  //Create a bitmap of the client size
  CBitmap bitmap;
  if (!bitmap.CreateCompatibleBitmap(&dc, rClient.Width(), rClient.Height()))
  {
    TRACE(_T("COSMCtrl::OnPaint, failed to create compatible bitmap\n"));
    return;
  }

  //Select the bitmap into the memory DC
  CBitmap* pOldBitmap{memDC.SelectObject(&bitmap)};

  //Take the current contents of the screen into the memory DC. Without this there can be some grid
  //artefacts on the control
  memDC.BitBlt(rClient.left, rClient.top, rClient.Width(), rClient.Height(), &dc, rClient.left, rClient.top, SRCCOPY);

  //Create a GDI+ graphics object from the memory DC
#pragma warning(suppress: 26486)
  Gdiplus::Graphics graphics{memDC.m_hDC};
  if (graphics.GetLastStatus() != Gdiplus::Ok)
  {
    TRACE(_T("COSMCtrl::OnPaint, failed to create GDI+ Graphics object from DC, status=%d\n"), graphics.GetLastStatus());

    //restore the memory DC before we return
    memDC.SelectObject(pOldBitmap);

    return;
  }

  //Setup appropriate values for the graphics object
  graphics.SetSmoothingMode(Gdiplus::SmoothingModeHighQuality);

  //Delegate to the other version
  if (!Draw(graphics, rClient, &rClip, GetDrawScrollRose(), GetDrawZoomBar(), GetDrawScaleBar()))
  {
    TRACE(_T("COSMCtrl::OnPaint, failed in call to Draw\n"));

    //restore the memory DC before we return
    memDC.SelectObject(pOldBitmap);

    return;
  }

  //Finally bitblt the memory DC bitmap onto the screen DC
#pragma warning(suppress: 26486)
  if (!dc.BitBlt(rClip.left, rClip.top, rClip.Width(), rClip.Height(), &memDC, rClip.left, rClip.top, SRCCOPY))
    TRACE(_T("COSMCtrl::OnPaint, failed in call to dc.BitBlt\n"));

  //Restore the memory DC before we return
  memDC.SelectObject(pOldBitmap);
#endif //#ifdef COSMCTRL_NOD2D
}

#pragma warning(suppress: 26434 26440)
BOOL COSMCtrl::OnEraseBkgnd(CDC* /*pDC*/)
{
  //Nothing to do
  return TRUE;
}

HRESULT COSMCtrl::CreateSession(_Inout_ WinHTTPWrappers::CSession& session, _In_ DWORD dwFlags)
{
  const bool bUseProxy{!m_sProxyServer.IsEmpty()};
  const bool bUseProxyBypassList{!m_sProxyBypassList.IsEmpty()};

  HRESULT hr{S_OK};
  if (session == nullptr)
#pragma warning(suppress: 26477)
    hr = session.Initialize(CStringW{m_sUserAgent}, bUseProxy ? WINHTTP_ACCESS_TYPE_NAMED_PROXY : WINHTTP_ACCESS_TYPE_NO_PROXY, bUseProxy ? CStringW{m_sProxyServer} : WINHTTP_NO_PROXY_NAME, bUseProxyBypassList ? CStringW{m_sProxyBypassList} : WINHTTP_NO_PROXY_BYPASS, dwFlags);

  return hr;
}

HRESULT COSMCtrl::CreateConnection(_In_ const CString& sServer, _In_ INTERNET_PORT nPort)
{
  //Close the handle if it is already open
  m_WinHTTPConnection.Close();

  return m_WinHTTPConnection.Initialize(m_WinHTTPSession, CStringW{sServer}, nPort);
}

HRESULT COSMCtrl::CreateRequest(_In_ const CString& sObject, _In_ const CString& sFile, _In_ bool bForcedRefresh)
{
  //Close the handle if it is already open
  m_WinHTTPRequest.Close();

  //Release any resources which may be still around
  if (m_bInitializedWinHTTPRequest)
    m_WinHTTPRequest.ReleaseResources();

  //We will accept any mime type
  std::array<LPCWSTR, 2> pwszAcceptTypes{ { L"*/*", nullptr } };
  m_WinHTTPRequest.m_sFileToDownloadInto = CStringW{sFile};
  const HRESULT hr{m_WinHTTPRequest.Initialize(m_WinHTTPConnection, CStringW{sObject}, nullptr, nullptr, nullptr, pwszAcceptTypes.data(), bForcedRefresh ? WINHTTP_FLAG_REFRESH : 0)};
  m_bInitializedWinHTTPRequest = SUCCEEDED(hr);
  return hr;
}

void COSMCtrl::BuildDownloadArray(_Inout_ std::vector<COSMCtrlDownloadTileItem>& tilesToDownload, _In_ bool bForceRefresh)
{
  //Get the client rect of the window
  CRect rClient;
  GetClientRect(&rClient);

  //Next get the server to connect to
  CString sServer{m_pTileProvider->GetDownloadServer()};

  //First thing we need to do is get the X and Y values for the center point of the map
  const int nZoom{static_cast<int>(m_fZoom)};
  const double fStartX{COSMCtrlHelper::Longitude2TileX(m_CenterPosition.m_fLongitude, nZoom)};
  int nStartX{static_cast<int>(fStartX)};
  const double fStartY{COSMCtrlHelper::Latitude2TileY(m_CenterPosition.m_fLatitude, nZoom)};
  int nStartY{static_cast<int>(fStartY)};

  //Work out the size of a tile at the current zoom level
  double fInt{0};
  const double fFractionalZoom{modf(m_fZoom, &fInt)};
#ifdef COSMCTRL_NOD2D
  const auto fOSMTileWidth{static_cast<Gdiplus::REAL>(OSMTileWidth + (fFractionalZoom * OSMTileWidth))};
  const auto fOSMTileHeight{static_cast<Gdiplus::REAL>(OSMTileHeight + (fFractionalZoom * OSMTileHeight))};
#else
  const auto fOSMTileWidth{static_cast<FLOAT>(OSMTileWidth + (fFractionalZoom * OSMTileWidth))};
  const auto fOSMTileHeight{static_cast<FLOAT>(OSMTileHeight + (fFractionalZoom * OSMTileHeight))};
#endif //#ifdef COSMCTRL_NOD2D

  //Next we need to find the X and Y values which occur just before the top left position of the client area
#ifdef COSMCTRL_NOD2D
  auto fClientX{static_cast<Gdiplus::REAL>(rClient.left + rClient.Width() / 2.0 - (modf(fStartX, &fInt) * fOSMTileWidth))};
#else
  auto fClientX{static_cast<FLOAT>(rClient.left + rClient.Width() / 2.0 - (modf(fStartX, &fInt) * fOSMTileWidth))};
#endif //#ifdef COSMCTRL_NOD2D
  while (fClientX > rClient.left)
  {
    --nStartX;
    fClientX -= fOSMTileWidth;
  }
  //Also take into account the edge count
  for (int i{0}; (i<m_nDownloadTilesEdgeCount) && (nStartX >= 0); i++)
  {
    --nStartX;
    fClientX -= fOSMTileWidth;
  }
#ifdef COSMCTR_NOD2D
  auto fClientY{static_cast<Gdiplus::REAL>(rClient.top + rClient.Height() / 2.0 - (modf(fStartY, &fInt) * fOSMTileWidth))};
#else
  auto fClientY{static_cast<FLOAT>(rClient.top + rClient.Height() / 2.0 - (modf(fStartY, &fInt) * fOSMTileWidth))};
#endif //#ifdef COSMCTR_NOD2D
  while (fClientY > rClient.top)
  {
    --nStartY;
    fClientY -= fOSMTileWidth;
  }
  //Also take into account the edge count
  for (int i{0}; (i<m_nDownloadTilesEdgeCount) && (nStartY >= 0); i++)
  {
    --nStartY;
    fClientY -= fOSMTileHeight;
  }

  //Also work out the end X coordinate
  int nEndX{rClient.right};
  //Also take into account the edge count
  for (int i{0}; i<m_nDownloadTilesEdgeCount; i++)
    nEndX += OSMTileWidth;

  //Also work out the end Y coordinate
  int nEndY{rClient.bottom};
  //Also take into account the edge count
  for (int i{0}; i<m_nDownloadTilesEdgeCount; i++)
    nEndY += OSMTileWidth;

  //Work out the client coordinates of the center position
#ifdef COSMCTRL_NOD2D
  const auto fClientCenterX{static_cast<Gdiplus::REAL>(rClient.left + rClient.Width() / 2.0)};
  const auto fClientCenterY{static_cast<Gdiplus::REAL>(rClient.top + rClient.Height() / 2.0)};
#else
  const auto fClientCenterX{static_cast<FLOAT>(rClient.left + rClient.Width() / 2.0)};
  const auto fClientCenterY{static_cast<FLOAT>(rClient.top + rClient.Height() / 2.0)};
#endif //#ifdef COSMCTRL_NOD2D

  //Build of the array of tiles we will be downloading
  tilesToDownload.clear();
  tilesToDownload.reserve(100);
  const int nMaxTile{static_cast<int>(pow(2.0, nZoom))};
#ifdef COSMCTRL_NOD2D
  Gdiplus::REAL fY{fClientY};
#else
  FLOAT fY{fClientY};
#endif //#ifdef COSMCTRL_NOD2D
  int nTileY{nStartY};
  while (fY <= nEndY)
  {
#ifdef COSMCTRL_NOD2D
    Gdiplus::REAL fX{fClientX};
#else
    FLOAT fX{fClientX};
#endif //#ifdef COSMCTRL_NOD2D
    int nTileX{nStartX};
    while (fX <= nEndX)
    {
      //Perform wrapping of invalid x tile values to valid values
      while (nTileX < 0)
        nTileX += nMaxTile;
      while (nTileX >= nMaxTile)
        nTileX -= nMaxTile;

      //Form the path to the file to cache the tile into
      if ((nTileY >= 0) && (nTileY < nMaxTile)) //Ensure we only look to download valid Y tiles
      {
        //Only bother trying to download the tile if the cached version does not exist or if we have been asked to do a refresh
        CString sFile{GetTileCachePath(m_sCacheDirectory, nZoom, nTileX, nTileY)};
        const bool bDownload{bForceRefresh || (GetFileAttributes(sFile) == INVALID_FILE_ATTRIBUTES)};
        if (bDownload)
        {
          //Calculate the distance of the tile to the center of the client area. Note we use the pixel distance rather than the
          //true ellipsoid distance as a speed optimization
          const double fTileXCenter{static_cast<double>(fX) + (fOSMTileWidth / 2) - fClientCenterX};
          const double fTileYCenter{static_cast<double>(fY) + (fOSMTileHeight / 2) - fClientCenterY};
          const double fDistance{sqrt((fTileXCenter * fTileXCenter) + (fTileYCenter * fTileYCenter))};

          //Add the tile to the array of items to download
#ifdef COSMCTRL_NOD2D
          COSMCtrlDownloadTileItem tileToDownload{nTileX, nTileY, Gdiplus::RectF(fX, fY, fOSMTileWidth, fOSMTileHeight), fDistance, sFile};
#else
          COSMCtrlDownloadTileItem tileToDownload{nTileX, nTileY, CD2DRectF(fX, fY, fOSMTileWidth, fOSMTileHeight), fDistance, sFile};
#endif //#ifdef COSMCTRL_NOD2D
#pragma warning(suppress: 26489)
          tilesToDownload.push_back(tileToDownload);
        }
      }

      //Move onto the next column
      fX += fOSMTileWidth;
      ++nTileX;
    }

    //Move onto rhe next row
    fY += fOSMTileHeight;
    ++nTileY;
  }

  //Sort the array if necessary
  if (m_DownloadOrder == DownloadOrder::ClosestToCenterFirst)
  {
#pragma warning(suppress: 26489)
    std::sort(tilesToDownload.begin(), tilesToDownload.end(),
      [](_In_ const COSMCtrlDownloadTileItem& item1, _In_ const COSMCtrlDownloadTileItem& item2) noexcept
      {
        return (item1.m_fDistanceFromCenterPosition < item2.m_fDistanceFromCenterPosition);
      }
    );
  }
}

HRESULT COSMCtrl::StopAsyncDownload(bool bCloseConnection) noexcept
{
  m_WinHTTPRequest.Close();
  if (bCloseConnection)
    m_WinHTTPConnection.Close();
  return S_OK;
}

HRESULT COSMCtrl::StartAsyncDownload(_In_ bool bBuildArray, _In_ bool bForceRefresh)
{
  //Create a new download array if we are already downloading
  if (m_bDownloading)
  {
    BuildDownloadArray(m_tilesPendingToDownload, bForceRefresh);
    return E_UNEXPECTED;
  }
  else if (bBuildArray)
  {
    //Else just download to the standard array
    BuildDownloadArray(m_tilesToDownload, bForceRefresh);
  }

  //Download the first tile if necessary
  const auto nTiles{m_tilesToDownload.size()};
  if (nTiles)
  {
    //Pull out the first tile to download
#pragma warning(suppress: 26446)
    const auto& tile{m_tilesToDownload[0]};

    //Create the session object
    if (m_WinHTTPSession == nullptr)
    {
      const HRESULT hr{CreateSession(m_WinHTTPSession, WINHTTP_FLAG_ASYNC)};
      if (FAILED(hr))
      {
        TRACE(_T("COSMCtrl::StartAsyncDownload, failed to create WinHTTP session, Error:%08X\n"), hr);
        return hr;
      }
    }

    //Create the connection object
    if (m_WinHTTPConnection == nullptr)
    {
      const HRESULT hr{CreateConnection(m_pTileProvider->GetDownloadServer(), m_pTileProvider->GetDownloadPort())};
      if (FAILED(hr))
      {
        TRACE(_T("COSMCtrl::StartAsyncDownload, failed to create WinHTTP connection, Error:%08X\n"), hr);
        return hr;
      }
    }

    //Create the sub directories of the tile if we can
    CString sSubDirectory;
    const int nZoom{static_cast<int>(m_fZoom)};
    sSubDirectory.Format(_T("%s\\%d"), m_sCacheDirectory.GetString(), nZoom);
    CreateDirectory(sSubDirectory, nullptr);
    sSubDirectory.Format(_T("%s\\%d\\%d"), m_sCacheDirectory.GetString(), nZoom, tile.m_nTileX);
    CreateDirectory(sSubDirectory, nullptr);

    //Create the request
    HRESULT hr{CreateRequest(m_pTileProvider->GetDownloadObject(nZoom, tile.m_nTileX, tile.m_nTileY), tile.m_sFile, bForceRefresh)};
    if (FAILED(hr))
    {
      TRACE(_T("COSMCtrl::StartAsyncDownload, failed to create WinHTTP request, Error:%08X\n"), hr);
      return hr;
    }

    //Start the download
    hr = m_WinHTTPRequest.SendRequest();
    if (SUCCEEDED(hr))
    {
#pragma warning(suppress: 26486)
      m_CurrentlyDownloadingTile = tile;
#pragma warning(suppress: 26486)
      m_tilesToDownload.erase(m_tilesToDownload.begin());
      m_bDownloading = true;
    }
    return hr;
  }
  else
    return E_FAIL;
}

#pragma warning(suppress: 26434)
void COSMCtrl::OnDestroy()
{
  //Stop downloading of tiles
  StopAsyncDownload(true);

  //Let the base class do its thing
  __super::OnDestroy();
}

#ifdef COSMCTRL_NOD2D
_Return_type_success_(return != 0) bool COSMCtrl::PositionToClient(_In_ const COSMCtrlPosition& position, _In_ const CRect& rClient, _Out_ Gdiplus::PointF& point) const noexcept
{
  //Delegate to the other version of the function
  return PositionToClient(position, m_CenterPosition, m_fZoom, rClient, point);
}
#else
_Return_type_success_(return != 0) bool COSMCtrl::PositionToClient(_In_ const COSMCtrlPosition& position, _In_ const CRect& rClient, _Out_ CD2DPointF& point) const noexcept
{
  //Delegate to the other version of the function
  return PositionToClient(position, m_CenterPosition, m_fZoom, m_fBearingOfTopOfMap, rClient, point);
}
#endif //#ifdef COSMCTRL_NOD2D

#ifdef COSMCTRL_NOD2D
_Return_type_success_(return != 0) bool COSMCtrl::PositionToClient(_In_ const COSMCtrlPosition& position, _In_ const COSMCtrlPosition& centerPosition, _In_ double fZoom, _In_ const CRect& rClient, _Out_ Gdiplus::PointF& point) noexcept
{
  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

  //Do the calculations
  const int nZoom{static_cast<int>(fZoom)};
  const double fXCenter{COSMCtrlHelper::Longitude2TileX(centerPosition.m_fLongitude, nZoom)};
  const double fYCenter{COSMCtrlHelper::Latitude2TileY(centerPosition.m_fLatitude, nZoom)};
  const double fX{COSMCtrlHelper::Longitude2TileX(position.m_fLongitude, nZoom)};
  const double fY{COSMCtrlHelper::Latitude2TileY(position.m_fLatitude, nZoom)};
  const double fMaxTile{pow(2.0, nZoom)};

  if ((fX >= 0) && (fX <= fMaxTile) && (fY >= 0) && (fY <= fMaxTile))
  {
    bSuccess = true;

    double fInt{0};
    const double fFractionalZoom{modf(fZoom, &fInt)};
    const auto fOSMTileWidth{static_cast<Gdiplus::REAL>(OSMTileWidth + (fFractionalZoom * OSMTileWidth))};
    const auto fOSMTileHeight{static_cast<Gdiplus::REAL>(OSMTileHeight + (fFractionalZoom * OSMTileHeight))};

    point.X = static_cast<Gdiplus::REAL>(rClient.Width() / 2.0 + rClient.left + ((fX - fXCenter) * fOSMTileWidth));
    point.Y = static_cast<Gdiplus::REAL>(rClient.Height() / 2.0 + rClient.top + ((fY - fYCenter) * fOSMTileHeight));
  }

  return bSuccess;
}
#else
_Return_type_success_(return != 0) bool COSMCtrl::PositionToClient(_In_ const COSMCtrlPosition& position, _In_ const COSMCtrlPosition& centerPosition, _In_ double fZoom, _In_ double fBearingOfTopOfMap, _In_ const CRect& rClient, _Out_ CD2DPointF& point) noexcept
{
  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

  //Do the calculations
  const int nZoom{static_cast<int>(fZoom)};
  const double fXCenter{COSMCtrlHelper::Longitude2TileX(centerPosition.m_fLongitude, nZoom)};
  const double fYCenter{COSMCtrlHelper::Latitude2TileY(centerPosition.m_fLatitude, nZoom)};
  const double fX{COSMCtrlHelper::Longitude2TileX(position.m_fLongitude, nZoom)};
  const double fY{COSMCtrlHelper::Latitude2TileY(position.m_fLatitude, nZoom)};
  const double fMaxTile{pow(2.0, nZoom)};

  if ((fX >= 0) && (fX <= fMaxTile) && (fY >= 0) && (fY <= fMaxTile))
  {
    bSuccess = true;

    double fInt{0};
    const double fFractionalZoom{modf(fZoom, &fInt)};
    const double fOSMTileWidth{OSMTileWidth + (fFractionalZoom * OSMTileWidth)};
    const double fOSMTileHeight{OSMTileHeight + (fFractionalZoom * OSMTileHeight)};
    const double fDeltaX{(fX - fXCenter) * fOSMTileWidth};
    const double fDeltaY{(fY - fYCenter) * fOSMTileHeight};

    //Rotate the position around the center position
    const double fRadBearing{-fBearingOfTopOfMap * 0.017453292519943295769236907684886};
    const double fCosBearing{cos(fRadBearing)};
    const double fSinBearing{sin(fRadBearing)};
    const double fRotateX{fDeltaX*fCosBearing - fDeltaY*fSinBearing};
    const double fRotateY{fDeltaX*fSinBearing + fDeltaY*fCosBearing};

    //Finally calculate the output parameter values
    point.x = static_cast<FLOAT>(rClient.Width() / 2.0 + rClient.left + fRotateX);
    point.y = static_cast<FLOAT>(rClient.Height() / 2.0 + rClient.top + fRotateY);
  }

  return bSuccess;
}
#endif //#ifdef COSMCTRL_NOD2D

#ifdef COSMCTRL_NOD2D
_Return_type_success_(return != 0) bool COSMCtrl::ClientToPosition(_In_ const Gdiplus::PointF& point, _In_ const CRect& rClient, _Out_ COSMCtrlPosition& position) const noexcept
{
  //Delegate to the other version of the function
  return ClientToPosition(point, m_CenterPosition, m_fZoom, rClient, position);
}
#else
_Return_type_success_(return != 0) bool COSMCtrl::ClientToPosition(_In_ const CD2DPointF& point, _In_ const CRect& rClient, _Out_ COSMCtrlPosition& position) const noexcept
{
  //Delegate to the other version of the function
  return ClientToPosition(point, m_CenterPosition, m_fZoom, m_fBearingOfTopOfMap, rClient, position);
}
#endif //#ifdef COSMCTRL_NOD2D

#ifdef COSMCTRL_NOD2D
_Return_type_success_(return != 0) bool COSMCtrl::ClientToPosition(_In_ const Gdiplus::PointF& point, _In_ const COSMCtrlPosition& centerPosition, _In_ double fZoom, _In_ const CRect& rClient, _Out_ COSMCtrlPosition& position) noexcept
{
  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

  //Do the calculations
  const int nZoom{static_cast<int>(fZoom)};
  double fX{COSMCtrlHelper::Longitude2TileX(centerPosition.m_fLongitude, nZoom)};
  double fY{COSMCtrlHelper::Latitude2TileY(centerPosition.m_fLatitude, nZoom)};
  double fInt{0};
  const double fFractionalZoom{modf(fZoom, &fInt)};
  const auto fOSMTileWidth{static_cast<Gdiplus::REAL>(OSMTileWidth + (fFractionalZoom * OSMTileWidth))};
  const auto fOSMTileHeight{static_cast<Gdiplus::REAL>(OSMTileHeight + (fFractionalZoom * OSMTileHeight))};
  fX += ((point.X - rClient.Width()/2.0) / fOSMTileWidth);
  fY += ((point.Y - rClient.Height()/2.0) / fOSMTileHeight);
  const double fMaxTile{pow(2.0, nZoom)};

  //Wrap the longitude if necessary
  while (fX < 0)
    fX += fMaxTile;
  while (fX >= fMaxTile)
    fX -= fMaxTile;

  if ((fX >= 0) && (fX <= fMaxTile) && (fY >= 0) && (fY <= fMaxTile))
  {
    bSuccess = true;
    position.m_fLatitude = COSMCtrlHelper::TileY2Latitude(fY, nZoom);
    position.m_fLongitude = COSMCtrlHelper::TileX2Longitude(fX, nZoom);
  }

  return bSuccess;
}
#else
_Return_type_success_(return != 0) bool COSMCtrl::ClientToPosition(_In_ const CD2DPointF& point, _In_ const COSMCtrlPosition& centerPosition, _In_ double fZoom, _In_ double fBearingOfTopOfMap, _In_ const CRect& rClient, _Out_ COSMCtrlPosition& position) noexcept
{
  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

  //Do the calculations
  const int nZoom{static_cast<int>(fZoom)};
  double fX{COSMCtrlHelper::Longitude2TileX(centerPosition.m_fLongitude, nZoom)};
  double fY{COSMCtrlHelper::Latitude2TileY(centerPosition.m_fLatitude, nZoom)};
  double fInt{0};
  const double fFractionalZoom{modf(fZoom, &fInt)};
  const double fOSMTileWidth{OSMTileWidth + (fFractionalZoom * OSMTileWidth)};
  const double fOSMTileHeight{OSMTileHeight + (fFractionalZoom * OSMTileHeight)};
  const double fDeltaX{(static_cast<double>(point.x) - rClient.Width()/2)};
  const double fDeltaY{(static_cast<double>(point.y) - rClient.Height()/2)};

  //Rotate the position around the center position
  const double fRadBearing{fBearingOfTopOfMap * 0.017453292519943295769236907684886};
  const double fCosBearing{cos(fRadBearing)};
  const double fSinBearing{sin(fRadBearing)};
  const double fRotateX{(fDeltaX*fCosBearing) - (fDeltaY*fSinBearing)};
  const double fRotateY{(fDeltaX*fSinBearing) + (fDeltaY*fCosBearing)};

  fX += (fRotateX / fOSMTileWidth);
  fY += (fRotateY / fOSMTileHeight);
  const double fMaxTile{pow(2.0, nZoom)};

  //Wrap the longitude if necessary
  while (fX < 0)
    fX += fMaxTile;
  while (fX >= fMaxTile)
    fX -= fMaxTile;

  if ((fX >= 0) && (fX <= fMaxTile) && (fY >= 0) && (fY <= fMaxTile))
  {
    bSuccess = true;
    position.m_fLatitude = COSMCtrlHelper::TileY2Latitude(fY, nZoom);
    position.m_fLongitude = COSMCtrlHelper::TileX2Longitude(fX, nZoom);
  }

  return bSuccess;
}
#endif //#ifdef COSMCTRL_NOD2D

#pragma warning(suppress: 26440)
void COSMCtrl::HandleLButtonDblClickMarker(_In_ UINT /*nFlags*/, _In_ const CPoint& /*point*/, _In_ MapItem /*item*/, _In_ INT_PTR /*nItem*/, _In_ INT_PTR /*nSubItem*/)
{
  //Nothing to do
}

#pragma warning(suppress: 26440)
void COSMCtrl::HandleLButtonDblClickCircle(_In_ UINT /*nFlags*/, _In_ const CPoint& /*point*/, _In_  MapItem /*item*/, _In_ INT_PTR /*nItem*/, _In_ INT_PTR /*nSubItem*/)
{
  //Nothing to do
}

void COSMCtrl::HandleLButtonDblClickPolyline(_In_ UINT /*nFlags*/, _In_ const CPoint& /*point*/, _In_ MapItem /*item*/, _In_ INT_PTR nItem, _In_ INT_PTR /*nSubItem*/)
{
  //We support decimation on double clicks for editable polylines
#pragma warning(suppress: 26446)
  auto& polyline{m_Polylines[nItem]};
  if (polyline->m_bEditable)
  {
    //Get the client rect
    CRect rClient;
    GetClientRect(&rClient);

    //Perform the decimation
    if (DecimatePolyline(*polyline, 1, rClient))
    {
      //Select all the nodes in the new polyline
      polyline->Select();

      //Force an update
      Invalidate(FALSE);
    }
  }
}

void COSMCtrl::HandleLButtonDblClickPolygon(_In_ UINT /*nFlags*/, _In_ const CPoint& /*point*/, _In_ MapItem /*item*/, _In_ INT_PTR nItem, _In_ INT_PTR /*nSubItem*/)
{
  //We support decimation on double clicks for editable polygons
#pragma warning(suppress: 26446)
  auto& polygon{m_Polygons[nItem]};
  if (polygon->m_bEditable)
  {
    //Get the client rect
    CRect rClient;
    GetClientRect(&rClient);

    //Perform the decimation
    if (DecimatePolygon(*polygon, 1, rClient))
    {
      //Select all the nodes in the new polygon
      polygon->Select();

      //Force an update
      Invalidate(TRUE);
    }
  }
}

void COSMCtrl::HandleLButtonDblClickMap(_In_ UINT /*nFlags*/, _In_ const CPoint& point, _In_ MapItem /*item*/, _In_ INT_PTR /*nItem*/, _In_ INT_PTR /*nSubItem*/)
{
  if (m_bAllowDoubleClickZoom)
  {
    //Get the client rect
    CRect rClient;
    GetClientRect(&rClient);

    //Get the coordinates of the double click position
    COSMCtrlPosition position;
  #ifdef COSMCTRL_NOD2D
    const Gdiplus::PointF tempPoint{static_cast<Gdiplus::REAL>(point.x), static_cast<Gdiplus::REAL>(point.y)};
  #else
    const CD2DPointF tempPoint{static_cast<FLOAT>(point.x), static_cast<FLOAT>(point.y)};
  #endif //#ifdef COSMCTRL_NOD2D
    if (ClientToPosition(tempPoint, rClient, position))
    {
      //Work out the new zoom level
      double fZoom{static_cast<double>(static_cast<int>(m_fZoom))}; //Try to keep the zoom at an integer level
      ++fZoom;
      if (fZoom > OSMMaxZoom)
        fZoom = OSMMaxZoom;

      //Set the new position and zoom level
      SetCenterAndZoom(position, fZoom, m_fDefaultAnimationDuration, true);
    }
  }
}

#pragma warning(suppress: 26434)
void COSMCtrl::OnLButtonDblClk(UINT nFlags, CPoint point)
{
  //Call the event handler if possible
  if (m_pEventHandler != nullptr)
  {
    if (m_pEventHandler->OnOSMCtrlLButtonDblClk(nFlags, point))
      return;
  }

  //Get the client rect
  CRect rClient;
  GetClientRect(&rClient);

  //Determine what we have double clicked on
  INT_PTR nItem{-1};
  INT_PTR nSubItem{-1};
  const MapItem item{HitTest(point, nItem, nSubItem, rClient)};
  switch (item)
  {
    case MapItem::Circle: //deliberate fallthrough
    case MapItem::CircleCircumference:
    {
      HandleLButtonDblClickCircle(nFlags, point, item, nItem, nSubItem);
      break;
    }
    case MapItem::Marker:
    {
      HandleLButtonDblClickMarker(nFlags, point, item, nItem, nSubItem);
      break;
    }
    case MapItem::Polyline:
    {
      HandleLButtonDblClickPolyline(nFlags, point, item, nItem, nSubItem);
      break;
    }
    case MapItem::Polygon:
    {
      HandleLButtonDblClickPolygon(nFlags, point, item, nItem, nSubItem);
      break;
    }
    case MapItem::Map:
    {
      HandleLButtonDblClickMap(nFlags, point, item, nItem, nSubItem);
      break;
    }
    default:
    {
      break;
    }
  }

  //Let the base class do its thing
  __super::OnLButtonDblClk(nFlags, point);
}

void COSMCtrl::StartDrag(_In_ const CPoint& point, _In_ MapItem item, _In_ INT_PTR nItem, _In_ INT_PTR nSubItem)
{
  //Capture the mouse and remember the initial drag position
  SetCapture();
  m_bMouseCapturedForDrag = true;

  //Remember the details of what we are dragging
  m_MouseDragItem = item;
  m_nMouseDragItemIndex = nItem;
  m_nMouseDragSubItemIndex = nSubItem;

  //Remember the current center position before we start dragging
  m_pointMouseCapture = point;
}

void COSMCtrl::SelectMarker(_In_ INT_PTR nIndex)
{
#pragma warning(suppress: 26446)
  auto& marker{m_Markers[nIndex]};

  //Mark the marker as selected
  marker->m_bSelected = true;

  //Force an update
  Invalidate(FALSE);
}

void COSMCtrl::SelectPolyline(_In_ INT_PTR nIndex)
{
#pragma warning(suppress: 26446)
  auto& polyline{m_Polylines[nIndex]};
  polyline->Select();

  //Force an update
  Invalidate(FALSE);
}

void COSMCtrl::SelectPolylineNode(_In_ INT_PTR nIndex, _In_ INT_PTR nSubItemIndex)
{
#pragma warning(suppress: 26446)
  auto& polyline{m_Polylines[nIndex]};
#pragma warning(suppress: 26446 26489)
  polyline->m_Nodes[nSubItemIndex].m_bSelected = true;

  //Force a update
#pragma warning(suppress: 26477)
  InvalidateRect(FALSE);
}

void COSMCtrl::SelectPolygon(_In_ INT_PTR nIndex)
{
  //Select the polygon
#pragma warning(suppress: 26446)
  auto& polygon{m_Polygons[nIndex]};
  polygon->Select();

  //Force an update
  Invalidate(FALSE);
}

void COSMCtrl::SelectCircle(_In_ INT_PTR nIndex)
{
  //Select the circle
#pragma warning(suppress: 26446)
  auto& circle{m_Circles[nIndex]};
  circle->m_bSelected = true;

  //Force an update
  Invalidate(FALSE);
}

void COSMCtrl::SelectPolygonNode(_In_ INT_PTR nIndex, _In_ INT_PTR nSubItemIndex)
{
#pragma warning(suppress: 26446)
  auto& polygon{m_Polygons[nIndex]};
#pragma warning(suppress: 26446 26489)
  polygon->m_Nodes[nSubItemIndex].m_bSelected = true;

  //Force an update
  Invalidate(FALSE);
}

#pragma warning(suppress: 26440)
int COSMCtrl::GetFirstSelectedPolyline() const
{
  for (std::vector<COSMCtrlPolyline>::size_type i{0}; i<m_Polylines.size(); i++)
  {
#pragma warning(suppress: 26446)
    const auto& polyline{m_Polylines[i]};
    if (polyline->FullySelected())
#pragma warning(suppress: 26472)
      return static_cast<int>(i);
  }

  //Nothing found
  return -1;
}

#pragma warning(suppress: 26440)
int COSMCtrl::GetFirstSelectedPolygon() const
{
  for (std::vector<COSMCtrlPolygon>::size_type i{0}; i<m_Polygons.size(); i++)
  {
#pragma warning(suppress: 26446)
    const auto& polygon{m_Polygons[i]};
    if (polygon->FullySelected())
#pragma warning(suppress: 26472)
      return static_cast<int>(i);
  }

  //Nothing found
  return -1;
}

#pragma warning(suppress: 26440)
int COSMCtrl::GetFirstSelectedMarker() const
{
  for (std::vector<COSMCtrlMarker>::size_type i{0}; i<m_Markers.size(); i++)
  {
#pragma warning(suppress: 26446)
    const auto& marker{m_Markers[i]};
    if (marker->m_bSelected)
#pragma warning(suppress: 26472)
      return static_cast<int>(i);
  }

  //Nothing found
  return -1;
}

#pragma warning(suppress: 26440)
int COSMCtrl::GetFirstSelectedCircle() const
{
  for (std::vector<COSMCtrlCircle>::size_type i{0}; i<m_Circles.size(); i++)
  {
#pragma warning(suppress: 26446)
    const auto& circle{m_Circles[i]};
    if (circle->m_bSelected)
#pragma warning(suppress: 26472)
      return static_cast<int>(i);
  }

  //Nothing found
  return -1;
}

size_t COSMCtrl::DeleteSelectedItems()
{
  //What will be the return value from this function
  size_t nItemsDeleted{0};

  //Delete any selected markers
  auto nOldSize{m_Markers.size()};
#pragma warning(suppress: 26489)
  m_Markers.erase(std::remove_if(m_Markers.begin(), m_Markers.end(), [](const std::shared_ptr<COSMCtrlMarker>& marker) noexcept -> bool { return marker->m_bSelected; }), m_Markers.end());
  nItemsDeleted += (nOldSize - m_Markers.size());

  //Delete any fully selected polylines
  nOldSize = m_Polylines.size();
#pragma warning(suppress: 26489)
  m_Polylines.erase(std::remove_if(m_Polylines.begin(), m_Polylines.end(), [](const std::shared_ptr<COSMCtrlPolyline>& polyline) noexcept -> bool { return polyline->FullySelected(); }), m_Polylines.end());
  nItemsDeleted += (nOldSize - m_Polylines.size());

  //Delete any selected nodes for each polyline
  for (auto& polyline : m_Polylines)
  {
#pragma warning(suppress: 26489)
    nOldSize = polyline->m_Nodes.size();
#pragma warning(suppress: 26489)
    polyline->m_Nodes.erase(std::remove_if(polyline->m_Nodes.begin(), polyline->m_Nodes.end(), [](const COSMCtrlNode& node) noexcept -> bool { return node.m_bSelected; }), polyline->m_Nodes.end());
    nItemsDeleted += (nOldSize - polyline->m_Nodes.size());
  }

  //Delete any polylines which now have less than 2 nodes
  nOldSize = m_Polylines.size();
#pragma warning(suppress: 26489)
  m_Polylines.erase(std::remove_if(m_Polylines.begin(), m_Polylines.end(), [](const std::shared_ptr<COSMCtrlPolyline>& polyline) noexcept -> bool { return polyline->m_Nodes.size() < 2; }), m_Polylines.end());
  nItemsDeleted += (nOldSize - m_Polylines.size());

  //Delete any fully selected polygons
  nOldSize = m_Polygons.size();
#pragma warning(suppress: 26489)
  m_Polygons.erase(std::remove_if(m_Polygons.begin(), m_Polygons.end(), [](const std::shared_ptr<COSMCtrlPolygon>& polygon) noexcept -> bool { return polygon->FullySelected(); }), m_Polygons.end());
  nItemsDeleted += (nOldSize - m_Polygons.size());

  //Delete any selected nodes for each polygon
  for (auto& polygon : m_Polygons)
  {
#pragma warning(suppress: 26489)
    nOldSize = polygon->m_Nodes.size();
#pragma warning(suppress: 26489)
    polygon->m_Nodes.erase(std::remove_if(polygon->m_Nodes.begin(), polygon->m_Nodes.end(), [](const COSMCtrlNode& node) noexcept -> bool { return node.m_bSelected; }), polygon->m_Nodes.end());
    nItemsDeleted += (nOldSize - polygon->m_Nodes.size());
  }

  //Delete any polygons which now have less than 3 nodes
  nOldSize = m_Polygons.size();
#pragma warning(suppress: 26489)
  m_Polygons.erase(std::remove_if(m_Polygons.begin(), m_Polygons.end(), [](const std::shared_ptr<COSMCtrlPolygon>& polygon) noexcept -> bool { return polygon->m_Nodes.size() < 3; }), m_Polygons.end());
  nItemsDeleted += (nOldSize - m_Polygons.size());

  //Delete any selected circles
  nOldSize = m_Circles.size();
#pragma warning(suppress: 26489)
  m_Circles.erase(std::remove_if(m_Circles.begin(), m_Circles.end(), [](const std::shared_ptr<COSMCtrlCircle>& circle) noexcept -> bool { return circle->m_bSelected; }), m_Circles.end());
  nItemsDeleted += (nOldSize - m_Circles.size());

  if (nItemsDeleted)
    Invalidate(FALSE);

  return nItemsDeleted;
}

void COSMCtrl::DeselectAllItems()
{
  //Get the client rect
  CRect rClient;
  GetClientRect(&rClient);

  //Deselect any selected markers
  for (auto& marker : m_Markers)
    marker->m_bSelected = false;

  //Deselect any selected polylines
  for (auto& polyline : m_Polylines)
  {
    if (polyline->AnySelected())
      polyline->Deselect();
  }

  //Deselect any selected polygons
  for (auto& polygon : m_Polygons)
  {
    if (polygon->AnySelected())
      polygon->Deselect();
  }

  //Deselect any selected circles
  for (auto& circle : m_Circles)
    circle->m_bSelected = false;

  //Force an update
  Invalidate(FALSE);
}

void COSMCtrl::SelectAllItems()
{
  //Get the client rect
  CRect rClient;
  GetClientRect(&rClient);

  //Select all markers
  for (auto& marker : m_Markers)
    marker->m_bSelected = true;

  //Select all polylines
  for (auto& polyline : m_Polylines)
    polyline->Select();

  //Select all polygons
  for (auto& polygon : m_Polygons)
    polygon->Select();

  //Select all circles
  for (auto& circle : m_Circles)
    circle->m_bSelected = true;

  Invalidate(FALSE);
}

void COSMCtrl::HandleLButtonDownStandard(_In_ UINT /*nFlags*/, _In_ const CPoint& point, _In_ MapItem item, _In_ INT_PTR /*nItem*/, _In_ INT_PTR /*nSubItem*/)
{
  switch (item)
  {
    case MapItem::ZoomBar:
    {
      //Capture the mouse
      SetCapture();
      m_bMouseCapturedForZoom = true;

      //Work out the zoom level and set it
      const double fZoom{min(max(((static_cast<double>(m_rZoomBar.bottom) - 8 - point.y) / 11.0) + OSMMinZoom, OSMMinZoom), OSMMaxZoom)};
      SetZoom(fZoom, 0, false);

      break;
    }
    case MapItem::ZoomIn:
    {
      //Increase the zoom
      double fZoom{GetZoom()};
      ++fZoom;
      if (fZoom > OSMMaxZoom)
        fZoom = OSMMaxZoom;
      SetZoom(fZoom, m_fDefaultAnimationDuration, true);
      break;
    }
    case MapItem::ZoomOut:
    {
      //Decrease the zoom
      double fZoom{GetZoom()};
      --fZoom;
      if (fZoom < OSMMinZoom)
        fZoom = OSMMinZoom;
      SetZoom(fZoom, m_fDefaultAnimationDuration, true);
      break;
    }
    case MapItem::ScrollerUp:
    {
      ScrollTileUp();
      break;
    }
    case MapItem::ScrollerDown:
    {
      ScrollTileDown();
      break;
    }
    case MapItem::ScrollerLeft:
    {
      ScrollTileLeft();
      break;
    }
    case MapItem::ScrollerRight:
    {
      ScrollTileRight();
      break;
    }
    default:
    {
      break;
    }
  }
}

void COSMCtrl::HandleLButtonDownPolylineCreation(_In_ UINT /*nFlags*/, _In_ const CPoint& point, _In_ MapItem /*item*/, _In_ INT_PTR /*nItem*/, _In_ INT_PTR /*nSubItem*/)
{
  //Get the client rect
  CRect rClient;
  GetClientRect(&rClient);

  //Create a new polyline and add it to the array
  COSMCtrlPosition newPosition;
#ifdef COSMCTRL_NOD2D
  const Gdiplus::PointF tempPoint{static_cast<Gdiplus::REAL>(point.x), static_cast<Gdiplus::REAL>(point.y)};
#else
  const CD2DPointF tempPoint{static_cast<FLOAT>(point.x), static_cast<FLOAT>(point.y)};
#endif //#ifdef COSMCTRL_NOD2D
  if (ClientToPosition(tempPoint, rClient, newPosition))
  {
    auto newPolyline{std::make_shared<COSMCtrlPolyline>()};
    const COSMCtrlNode node{newPosition.m_fLongitude, newPosition.m_fLatitude};
#pragma warning(suppress: 26489)
    newPolyline->m_Nodes.push_back(node);
    newPolyline->m_Nodes.push_back(node);
#pragma warning(suppress: 26486)
    InitializeNewPolyline(*newPolyline);

    //Deselect all other items before we continue
    DeselectAllItems();

    //Start off the new polyline as selected
#pragma warning(suppress: 26486)
    newPolyline->Select();

    //Add to the array
    m_Polylines.push_back(newPolyline);
    const auto nNewItem{m_Polylines.size() - 1};

    //Force an update
    Invalidate(FALSE);

    StartDrag(point, MapItem::Map, nNewItem, 1);
  }
}

void COSMCtrl::HandleLButtonDownPolylineCreationWithNode(_In_ UINT nFlags, _In_ const CPoint& point, _In_ MapItem item, _In_ INT_PTR nItem, _In_ INT_PTR nSubItem)
{
  //Get the client rect
  CRect rClient;
  GetClientRect(&rClient);

  //If we have selected the last node in the polyline, then continue the creation process
  COSMCtrlPosition newPosition;
#ifdef COSMCTRL_NOD2D
  const Gdiplus::PointF tempPoint{static_cast<Gdiplus::REAL>(point.x), static_cast<Gdiplus::REAL>(point.y)};
#else
  const CD2DPointF tempPoint{static_cast<FLOAT>(point.x), static_cast<FLOAT>(point.y)};
#endif //#ifdef COSMCTRL_NOD2D
  if (ClientToPosition(tempPoint, rClient, newPosition))
  {
    //Pull out the polyline we will be working on
#pragma warning(suppress: 26446)
    auto& polyline{m_Polylines[nItem]};

    //If we have selected the first node, then reverse the order of the nodes in the array
#pragma warning(suppress: 26489)
    const auto nExistingNodes{polyline->m_Nodes.size()};
    if (nSubItem == 0)
    {
      std::vector<COSMCtrlNode> tempNodes;
#pragma warning(suppress: 26489)
      tempNodes = polyline->m_Nodes;
      for (std::vector<COSMCtrlNode>::size_type i{0}; i<nExistingNodes; i++)
#pragma warning(suppress: 26446 26489)
        polyline->m_Nodes[i] = tempNodes[nExistingNodes - 1 - i];
      nSubItem = nExistingNodes - 1;
    }

    //Only continue with the creation if we have selected the last node in the polyline
#pragma warning(suppress: 26472)
    if (static_cast<std::vector<COSMCtrlNode>::size_type>(nSubItem) == (nExistingNodes - 1))
    {
      //Deselect all other items before we continue
      DeselectAllItems();

      //Add a new node to the polyline
      const COSMCtrlNode node0{newPosition.m_fLongitude, newPosition.m_fLatitude};
#pragma warning(suppress: 26489)
      polyline->m_Nodes.push_back(node0);
      const auto nNewItem{polyline->m_Nodes.size() - 1};

      //select all the nodes in the polyline
      SelectPolyline(nItem);

      StartDrag(point, MapItem::PolylineNode, nItem, nNewItem);
    }
    else
    {
      //Just start a new polyline at the node position
      HandleLButtonDownPolylineCreation(nFlags, point, item, nItem, nSubItem);
    }
  }
}

#ifndef COSMCTRL_NOD2D
void COSMCtrl::HandleLButtonDownRotate(_In_ UINT /*nFlags*/, _In_ const CPoint& point, _In_ MapItem /*item*/, _In_ INT_PTR /*nItem*/, _In_ INT_PTR /*nSubItem*/)
{
  //Get the client rect
  CRect rClient;
  GetClientRect(&rClient);

  //Get the position of where we have dragged to
  const COSMCtrlPosition newPosition;
#ifdef COSMCTRL_NOD2D
  const Gdiplus::PointF tempPoint{static_cast<Gdiplus::REAL>(point.x), static_cast<Gdiplus::REAL>(point.y)};
#else
  const CD2DPointF tempPoint{static_cast<FLOAT>(point.x), static_cast<FLOAT>(point.y)};
#endif //#ifdef COSMCTRL_NOD2D

  //Get the initial bearings and then start dragging
  m_fBearingAtTopOfMapMouseCapture = m_fBearingOfTopOfMap;
  StartDrag(point, MapItem::Map, -1, 1);
}
#endif //#ifndef COSMCTRL_NOD2D

void COSMCtrl::HandleLButtonDownSelection(_In_ UINT /*nFlags*/, _In_ const CPoint& point, _In_ MapItem /*item*/, _In_ INT_PTR /*nItem*/, _In_ INT_PTR /*nSubItem*/)
{
  //Get the client rect
  CRect rClient;
  GetClientRect(&rClient);

  //Create a new rectangular polygon in "m_SelectionPolygon"
  COSMCtrlPosition newPosition;
#ifdef COSMCTRL_NOD2D
  const Gdiplus::PointF tempPoint{static_cast<Gdiplus::REAL>(point.x), static_cast<Gdiplus::REAL>(point.y)};
#else
  const CD2DPointF tempPoint{static_cast<FLOAT>(point.x), static_cast<FLOAT>(point.y)};
#endif //#ifdef COSMCTRL_NOD2D
  if (ClientToPosition(tempPoint, rClient, newPosition))
  {
    //Deselect anything first (if the ctrl key was not held down)
    if ((GetKeyState(VK_CONTROL) & 0x8000) == 0)
      DeselectAllItems();

    m_SelectionPolygon->m_Nodes.clear();
    m_SelectionPolygon->m_Nodes.reserve(4);
    m_SelectionPolygon->m_Nodes.emplace_back(newPosition.m_fLongitude, newPosition.m_fLatitude);
    m_SelectionPolygon->m_Nodes.emplace_back(newPosition.m_fLongitude, newPosition.m_fLatitude);
    m_SelectionPolygon->m_Nodes.emplace_back(newPosition.m_fLongitude, newPosition.m_fLatitude);
    m_SelectionPolygon->m_Nodes.emplace_back(newPosition.m_fLongitude, newPosition.m_fLatitude);
    InitializeNewSelectionPolygon(*m_SelectionPolygon);

    //Force an update
    Invalidate(FALSE);

    StartDrag(point, MapItem::Map, -1, 1);
  }
}

void COSMCtrl::HandleLButtonDownPolygonCreation(_In_ UINT /*nFlags*/, _In_ const CPoint& point, _In_ MapItem /*item*/, _In_ INT_PTR /*nItem*/, _In_ INT_PTR /*nSubItem*/)
{
  //Get the client rect
  CRect rClient;
  GetClientRect(&rClient);

  //Create a new rectangular polygon and add it to the array
  COSMCtrlPosition newPosition;
#ifdef COSMCTRL_NOD2D
  const Gdiplus::PointF tempPoint{static_cast<Gdiplus::REAL>(point.x), static_cast<Gdiplus::REAL>(point.y)};
#else
  const CD2DPointF tempPoint{static_cast<FLOAT>(point.x), static_cast<FLOAT>(point.y)};
#endif //#ifdef COSMCTRL_NOD2D
  if (ClientToPosition(tempPoint, rClient, newPosition))
  {
    auto newPolygon{std::make_shared<COSMCtrlPolygon>()};
    newPolygon->m_Nodes.clear();
    newPolygon->m_Nodes.reserve(4);
    newPolygon->m_Nodes.emplace_back(newPosition.m_fLongitude, newPosition.m_fLatitude);
    newPolygon->m_Nodes.emplace_back(newPosition.m_fLongitude, newPosition.m_fLatitude);
    newPolygon->m_Nodes.emplace_back(newPosition.m_fLongitude, newPosition.m_fLatitude);
    newPolygon->m_Nodes.emplace_back(newPosition.m_fLongitude, newPosition.m_fLatitude);
#pragma warning(suppress: 26486)
    InitializeNewPolygon(*newPolygon);

    //Deselect all other items before we continue
    DeselectAllItems();

    //Start off the new polygon as selected
#pragma warning(suppress: 26486)
    newPolygon->Select();

    //Add to the array
    m_Polygons.push_back(newPolygon);
    const auto nNewItem{m_Polygons.size() - 1};

    //Force an update
    Invalidate(FALSE);

    StartDrag(point, MapItem::Map, nNewItem, 1);
  }
}

void COSMCtrl::HandleLButtonDownMarkerCreation(_In_ UINT /*nFlags*/, _In_ const CPoint& point, _In_ MapItem /*item*/, _In_ INT_PTR /*nItem*/, _In_ INT_PTR /*nSubItem*/)
{
  //Get the client rect
  CRect rClient;
  GetClientRect(&rClient);

  //Create a new marker and add it to the array
  COSMCtrlPosition newPosition;
#ifdef COSMCTRL_NOD2D
  const Gdiplus::PointF tempPoint{static_cast<Gdiplus::REAL>(point.x), static_cast<Gdiplus::REAL>(point.y)};
#else
  const CD2DPointF tempPoint{static_cast<FLOAT>(point.x), static_cast<FLOAT>(point.y)};
#endif //#ifdef COSMCTRL_NOD2D
  if (ClientToPosition(tempPoint, rClient, newPosition))
  {
    auto newMarker{std::make_shared<COSMCtrlMarker>()};
    newMarker->m_Position = newPosition;
#pragma warning(suppress: 26486)
    InitializeNewMarker(*newMarker);

    //Add to the array
    m_Markers.push_back(newMarker);
    const auto nNewItem{m_Markers.size() - 1};

    //Force an update
    Invalidate(FALSE);

    StartDrag(point, MapItem::Marker, nNewItem, -1);
  }
}

void COSMCtrl::HandleLButtonDownCircleCreation(_In_ UINT /*nFlags*/, _In_ const CPoint& point, _In_ MapItem /*item*/, _In_ INT_PTR /*nItem*/, _In_ INT_PTR /*nSubItem*/)
{
  //Get the client rect
  CRect rClient;
  GetClientRect(&rClient);

  //Create a new circle and add it to the array
  COSMCtrlPosition newPosition;
#ifdef COSMCTRL_NOD2D
  const Gdiplus::PointF tempPoint{static_cast<Gdiplus::REAL>(point.x), static_cast<Gdiplus::REAL>(point.y)};
#else
  const CD2DPointF tempPoint{static_cast<FLOAT>(point.x), static_cast<FLOAT>(point.y)};
#endif //#ifdef COSMCTRL_NOD2D
  if (ClientToPosition(tempPoint, rClient, newPosition))
  {
    auto newCircle{std::make_shared<COSMCtrlCircle>()};
    newCircle->m_Position = newPosition;
#pragma warning(suppress: 26486)
    InitializeNewCircle(*newCircle);

    //Add to the array
    m_Circles.push_back(newCircle);
    const auto nNewItem{m_Circles.size() - 1};

    //Force an update
    Invalidate(FALSE);

    StartDrag(point, MapItem::Circle, nNewItem, -1);
  }
}

void COSMCtrl::HandleLButtonDownPolygonCreationWithNode(_In_ UINT /*nFlags*/, _In_ const CPoint& point, _In_ MapItem /*item*/, _In_ INT_PTR nItem, _In_ INT_PTR nSubItem)
{
  //Get the client rect
  CRect rClient;
  GetClientRect(&rClient);

  //Insert a new node before the currently selected node
  COSMCtrlPosition newPosition;
#ifdef COSMCTRL_NOD2D
  const Gdiplus::PointF tempPoint{static_cast<Gdiplus::REAL>(point.x), static_cast<Gdiplus::REAL>(point.y)};
#else
  const CD2DPointF tempPoint{static_cast<FLOAT>(point.x), static_cast<FLOAT>(point.y)};
#endif //#ifdef COSMCTRL_NOD2D
  if (ClientToPosition(tempPoint, rClient, newPosition))
  {
    //Pull out the polygon we will be working on
#pragma warning(suppress: 26446)
    auto& polygon{m_Polygons[nItem]};

    //Deselect all other items before we continue
    DeselectAllItems();

    //Insert the new node
    const COSMCtrlNode node{newPosition.m_fLongitude, newPosition.m_fLatitude};
#pragma warning(suppress: 26489)
    const auto iter{polygon->m_Nodes.begin() + nSubItem};
    polygon->m_Nodes.insert(iter, node);

    //select all the nodes in the polygon
    SelectPolygon(nItem);

    StartDrag(point, MapItem::PolygonNode, nItem, nSubItem);
  }
}

void COSMCtrl::HandleLButtonDownPolygonNode(_In_ UINT /*nFlags*/, _In_ const CPoint& point, _In_ MapItem /*item*/, _In_ INT_PTR nItem, _In_ INT_PTR nSubItem)
{
  //Deselect anything first (if the ctrl key was not held down)
  if ((GetKeyState(VK_CONTROL) & 0x8000) == 0)
    DeselectAllItems();

  //select the node
  SelectPolygonNode(nItem, nSubItem);

  //If the polygon is editable, then drag
#pragma warning(suppress: 26446)
  const auto& polygon{m_Polygons[nItem]};
  if (polygon->m_bEditable)
    StartDrag(point, MapItem::PolygonNode, nItem, nSubItem);
}

void COSMCtrl::HandleLButtonDownPolygon(_In_ UINT /*nFlags*/, _In_ const CPoint& point, _In_ MapItem /*item*/, _In_ INT_PTR nItem, _In_ INT_PTR nSubItem)
{
  //Deselect anything first (if the ctrl key was not held down)
  if ((GetKeyState(VK_CONTROL) & 0x8000) == 0)
    DeselectAllItems();

  //Pull out the polygon we will be working on
#pragma warning(suppress: 26446)
  auto& polygon{m_Polygons[nItem]};

  //If the polygon we have hit is draggable, then drag it otherwise just drag the map
  if (polygon->m_bDraggable)
  {
    //The item is selected while it is dragged
    SelectPolygon(nItem);

    //Get the client rect
    CRect rClient;
    GetClientRect(&rClient);

  #ifdef COSMCTRL_NOD2D
    Gdiplus::PointF ptAnchor;
  #else
    CD2DPointF ptAnchor;
  #endif //#ifdef COSMCTRL_NOD2D
#pragma warning(suppress: 26446 26489)
    if (PositionToClient(polygon->m_Nodes[0].m_Position, rClient, ptAnchor))
    {
      //Remember the drag offset from the start drag position to the polyline's first position in pixels
    #ifdef COSMCTRL_NOD2D
      m_DragOffset = CSize{static_cast<int>(point.x - ptAnchor.X), static_cast<int>(point.y - ptAnchor.Y)};
    #else
      m_DragOffset = CSize{static_cast<int>(point.x - ptAnchor.x), static_cast<int>(point.y - ptAnchor.y)};
    #endif //#ifdef COSMCTRL_NOD2D

      //Also remember the original drag positions
#pragma warning(suppress: 26489)
      m_OriginalDragNodes = polygon->m_Nodes;

      //Finally start the dragging
      StartDrag(point, MapItem::Polygon, nItem, nSubItem);
    }
  }
  else if (polygon->m_bEditable)
  {
    //select the item
    SelectPolygon(nItem);
  }
  else if (m_bAllowDrag)
    StartDrag(point, MapItem::Map, -1, -1);
}

void COSMCtrl::HandleLButtonDownCircleCircumference(_In_ UINT /*nFlags*/, _In_ const CPoint& point, _In_ MapItem /*item*/, _In_ INT_PTR nItem, _In_ INT_PTR /*nSubItem*/)
{
  //Deselect anything first (if the ctrl key was not held down)
  if ((GetKeyState(VK_CONTROL) & 0x8000) == 0)
    DeselectAllItems();

  //Pull out the circle we will be working on
#pragma warning(suppress: 26446)
  const auto& circle{m_Circles[nItem]};

  //If the circle we have hit is editable, then drag it otherwise just drag the map
  if (circle->m_bEditable)
  {
    //The item is selected while it is dragged
    SelectCircle(nItem);

    //Finally start the dragging
    m_bModifyingCircleRadius = true;
    StartDrag(point, MapItem::Circle, nItem, -1);
  }
  else if (m_bAllowDrag)
    StartDrag(point, MapItem::Map, -1, -1);
}

void COSMCtrl::HandleLButtonDownCircle(_In_ UINT /*nFlags*/, _In_ const CPoint& point, _In_ MapItem /*item*/, _In_ INT_PTR nItem, _In_ INT_PTR nSubItem)
{
  //Deselect anything first (if the ctrl key was not held down)
  if ((GetKeyState(VK_CONTROL) & 0x8000) == 0)
    DeselectAllItems();

  //Pull out the circle we will be working on
#pragma warning(suppress: 26446)
  const auto& circle{m_Circles[nItem]};

  //If the circle we have hit is draggable, then drag it otherwise just drag the map
  if (circle->m_bDraggable)
  {
    //The item is selected while it is dragged
    SelectCircle(nItem);

    //Get the client rect
    CRect rClient;
    GetClientRect(&rClient);

  #ifdef COSMCTRL_NOD2D
    Gdiplus::PointF ptAnchor;
  #else
    CD2DPointF ptAnchor;
  #endif //#ifdef COSMCTRL_NOD2D
    if (PositionToClient(circle->m_Position, rClient, ptAnchor))
    {
      //Remember the drag offset from the start drag position to the polyline's first position in pixels
    #ifdef COSMCTRL_NOD2D
      m_DragOffset = CSize{static_cast<int>(point.x - ptAnchor.X), static_cast<int>(point.y - ptAnchor.Y)};
    #else
      m_DragOffset = CSize{static_cast<int>(point.x - ptAnchor.x), static_cast<int>(point.y - ptAnchor.y)};
    #endif //#ifdef COSMCTRL_NOD2D

      //Finally start the dragging
      m_bModifyingCircleRadius = false;
      StartDrag(point, MapItem::Circle, nItem, nSubItem);
    }
  }
  else if (circle->m_bEditable)
  {
    //select the item
    SelectCircle(nItem);
  }
  else if (m_bAllowDrag)
    StartDrag(point, MapItem::Map, -1, -1);
}

void COSMCtrl::HandleLButtonDownPolylineNode(_In_ UINT /*nFlags*/, _In_ const CPoint& point, _In_ MapItem /*item*/, _In_ INT_PTR nItem, _In_ INT_PTR nSubItem)
{
  //Deselect anything first (if the ctrl key was not held down)
  if ((GetKeyState(VK_CONTROL) & 0x8000) == 0)
    DeselectAllItems();

  //select the node
  SelectPolylineNode(nItem, nSubItem);

  //If the polygon is editable, then drag
#pragma warning(suppress: 26446)
  const auto& polyline{m_Polylines[nItem]};
  if (polyline->m_bEditable)
    StartDrag(point, MapItem::PolylineNode, nItem, nSubItem);
}

void COSMCtrl::HandleLButtonDownPolyline(_In_ UINT /*nFlags*/, _In_ const CPoint& point, _In_ MapItem /*item*/, _In_ INT_PTR nItem, _In_ INT_PTR nSubItem)
{
  //Deselect anything first (if the ctrl key was not held down)
  if ((GetKeyState(VK_CONTROL) & 0x8000) == 0)
    DeselectAllItems();

  //Pull out the polyline we will be working on
#pragma warning(suppress: 26446)
  auto& polyline{m_Polylines[nItem]};

  //If the polyline we have hit is draggable, then drag it otherwise just drag the map
  if (polyline->m_bDraggable)
  {
    //select the item
    SelectPolyline(nItem);

    //Get the client rect
    CRect rClient;
    GetClientRect(&rClient);

  #ifdef COSMCTRL_NOD2D
    Gdiplus::PointF ptAnchor;
  #else
    CD2DPointF ptAnchor;
  #endif //#ifdef COSMCTRL_NOD2D
#pragma warning(suppress: 26446 26489)
    if (PositionToClient(polyline->m_Nodes[0].m_Position, rClient, ptAnchor))
    {
      //Remember the drag offset from the start drag position to the polyline's first position in pixels
    #ifdef COSMCTRL_NOD2D
      m_DragOffset = CSize{static_cast<int>(point.x - ptAnchor.X), static_cast<int>(point.y - ptAnchor.Y)};
    #else
      m_DragOffset = CSize{static_cast<int>(point.x - ptAnchor.x), static_cast<int>(point.y - ptAnchor.y)};
    #endif //#ifdef COSMCTRL_NOD2D

      //Also remember the original drag nodes
#pragma warning(suppress: 26489)
      m_OriginalDragNodes = polyline->m_Nodes;

      //Finally start the dragging
      StartDrag(point, MapItem::Polyline, nItem, nSubItem);
    }
  }
  else if (polyline->m_bEditable)
  {
    //select the item
    SelectPolyline(nItem);
  }
  else if (m_bAllowDrag)
    StartDrag(point, MapItem::Map, -1, -1);
}

void COSMCtrl::HandleLButtonDownMarker(_In_ UINT /*nFlags*/, _In_ const CPoint& point, _In_ MapItem /*item*/, _In_ INT_PTR nItem, _In_ INT_PTR nSubItem)
{
  //Deselect anything first (if the ctrl key was not held down)
  if ((GetKeyState(VK_CONTROL) & 0x8000) == 0)
    DeselectAllItems();

  //If the marker we have hit is draggable, then drag it otherwise just drag the map
#pragma warning(suppress: 26446)
  const auto& marker{m_Markers[nItem]};

  //If the marker is draggable, then drag it
  if (marker->m_bDraggable)
  {
    //Select the item
    SelectMarker(nItem);

    //Get the client rect
    CRect rClient;
    GetClientRect(&rClient);

  #ifdef COSMCTRL_NOD2D
    Gdiplus::PointF ptAnchor;
  #else
    CD2DPointF ptAnchor;
  #endif //#ifdef COSMCTRL_NOD2D
    if (PositionToClient(marker->m_Position, rClient, ptAnchor))
    {
      //Remember the drag offset from the start drag position to the marker's anchor position in pixels
    #ifdef COSMCTRL_NOD2D
      m_DragOffset = CSize{static_cast<int>(point.x - ptAnchor.X), static_cast<int>(point.y - ptAnchor.Y)};
    #else
      m_DragOffset = CSize{static_cast<int>(point.x - ptAnchor.x), static_cast<int>(point.y - ptAnchor.y)};
    #endif //#ifdef COSMCTRL_NOD2D

      //Finally start the dragging
      StartDrag(point, MapItem::Marker, nItem, nSubItem);
    }
  }
  else if (m_bAllowDrag)
  {
    //Finally start the drag
    StartDrag(point, MapItem::Map, -1, -1);
  }
}

void COSMCtrl::HandleLButtonDownMap(_In_ UINT /*nFlags*/, _In_ const CPoint& point, _In_ MapItem /*item*/, _In_ INT_PTR /*nItem*/, _In_ INT_PTR /*nSubItem*/)
{
  if (m_bAllowDrag)
  {
    //Deselect all items before we drag
    DeselectAllItems();

    StartDrag(point, MapItem::Map, -1, -1);
  }
}

#pragma warning(suppress: 26434)
void COSMCtrl::OnLButtonDown(UINT nFlags, CPoint point)
{
  //Call the event handler if possible
  if (m_pEventHandler != nullptr)
  {
    if (m_pEventHandler->OnOSMCtrlLButtonDown(nFlags, point))
      return;
  }

  //Get the client rect
  CRect rClient;
  GetClientRect(&rClient);

  //Determine what we have clicked on
  INT_PTR nItem{-1};
  INT_PTR nSubItem{-1};
  const MapItem item{HitTest(point, nItem, nSubItem, rClient)};
  if (m_MapMode == MapMode::Selection)
  {
    switch (item)
    {
      case MapItem::PolylineNode: //deliberate fallthrough
      case MapItem::Polygon: //deliberate fallthrough
      case MapItem::PolygonNode: //deliberate fallthrough
      case MapItem::Polyline: //deliberate fallthrough
      case MapItem::Marker: //deliberate fallthrough
      case MapItem::Circle: //deliberate fallthrough
      case MapItem::CircleCircumference: //deliberate fallthrough
      case MapItem::Map:
      {
        HandleLButtonDownSelection(nFlags, point, item, nItem, nSubItem);
        break;
      }
      default:
      {
        HandleLButtonDownStandard(nFlags, point, item, nItem, nSubItem);
        break;
      }
    }
  }
  else if (m_MapMode == MapMode::Rotate)
  {
  #ifndef COSMCTRL_NOD2D //Rotate support is only available in the D2D code path
    HandleLButtonDownRotate(nFlags, point, item, nItem, nSubItem);
  #endif //#ifndef COSMCTRL_NOD2D
  }
  else if (m_MapMode == MapMode::MarkerCreation)
  {
    switch (item)
    {
      case MapItem::PolylineNode: //deliberate fallthrough
      case MapItem::Polygon: //deliberate fallthrough
      case MapItem::PolygonNode: //deliberate fallthrough
      case MapItem::Polyline: //deliberate fallthrough
      case MapItem::Marker: //deliberate fallthrough
      case MapItem::Circle: //deliberate fallthrough
      case MapItem::CircleCircumference: //deliberate fallthrough
      case MapItem::Map:
      {
        HandleLButtonDownMarkerCreation(nFlags, point, item, nItem, nSubItem);
        break;
      }
      default:
      {
        HandleLButtonDownStandard(nFlags, point, item, nItem, nSubItem);
        break;
      }
    }
  }
  else if (m_MapMode == MapMode::CircleCreation)
  {
    switch (item)
    {
      case MapItem::PolylineNode: //deliberate fallthrough
      case MapItem::Polygon: //deliberate fallthrough
      case MapItem::PolygonNode: //deliberate fallthrough
      case MapItem::Polyline: //deliberate fallthrough
      case MapItem::Marker: //deliberate fallthrough
      case MapItem::Circle: //deliberate fallthrough
      case MapItem::CircleCircumference: //deliberate fallthrough
      case MapItem::Map:
      {
        HandleLButtonDownCircleCreation(nFlags, point, item, nItem, nSubItem);
        break;
      }
      default:
      {
        HandleLButtonDownStandard(nFlags, point, item, nItem, nSubItem);
        break;
      }
    }
  }
  else if (m_MapMode == MapMode::PolylineCreation)
  {
    switch (item)
    {
      case MapItem::PolylineNode:
      {
        HandleLButtonDownPolylineCreationWithNode(nFlags, point, item, nItem, nSubItem);
        break;
      }
      case MapItem::Polygon: //deliberate fallthrough
      case MapItem::PolygonNode: //deliberate fallthrough
      case MapItem::Polyline: //deliberate fallthrough
      case MapItem::Marker: //deliberate fallthrough
      case MapItem::Circle: //deliberate fallthrough
      case MapItem::CircleCircumference: //deliberate fallthrough
      case MapItem::Map:
      {
        HandleLButtonDownPolylineCreation(nFlags, point, item, nItem, nSubItem);
        break;
      }
      default:
      {
        HandleLButtonDownStandard(nFlags, point, item, nItem, nSubItem);
        break;
      }
    }
  }
  else if (m_MapMode == MapMode::PolygonCreation)
  {
    switch (item)
    {
      case MapItem::PolygonNode:
      {
        HandleLButtonDownPolygonCreationWithNode(nFlags, point, item, nItem, nSubItem);
        break;
      }
      case MapItem::Polygon: //deliberate fallthrough
      case MapItem::Polyline: //deliberate fallthrough
      case MapItem::PolylineNode: //deliberate fallthrough
      case MapItem::Marker: //deliberate fallthrough
      case MapItem::Circle: //deliberate fallthrough
      case MapItem::CircleCircumference: //deliberate fallthrough
      case MapItem::Map:
      {
        HandleLButtonDownPolygonCreation(nFlags, point, item, nItem, nSubItem);
        break;
      }
      default:
      {
        HandleLButtonDownStandard(nFlags, point, item, nItem, nSubItem);
        break;
      }
    }
  }
  else
  {
    switch (item)
    {
      case MapItem::PolygonNode:
      {
        HandleLButtonDownPolygonNode(nFlags, point, item, nItem, nSubItem);
        break;
      }
      case MapItem::Polygon:
      {
        HandleLButtonDownPolygon(nFlags, point, item, nItem, nSubItem);
        break;
      }
      case MapItem::PolylineNode:
      {
        HandleLButtonDownPolylineNode(nFlags, point, item, nItem, nSubItem);
        break;
      }
      case MapItem::Polyline:
      {
        HandleLButtonDownPolyline(nFlags, point, item, nItem, nSubItem);
        break;
      }
      case MapItem::Marker:
      {
        HandleLButtonDownMarker(nFlags, point, item, nItem, nSubItem);
        break;
      }
      case MapItem::Circle:
      {
        HandleLButtonDownCircle(nFlags, point, item, nItem, nSubItem);
        break;
      }
      case MapItem::CircleCircumference:
      {
        HandleLButtonDownCircleCircumference(nFlags, point, item, nItem, nSubItem);
        break;
      }
      case MapItem::Map:
      {
        HandleLButtonDownMap(nFlags, point, item, nItem, nSubItem);
        break;
      }
      default:
      {
        HandleLButtonDownStandard(nFlags, point, item, nItem, nSubItem);
        break;
      }
    }
  }

  //let the base class do its thing
  __super::OnLButtonDown(nFlags, point);
}

void COSMCtrl::HandleMouseMovePolylineCreation(_In_ UINT /*nFlags*/, _In_ const CPoint& point)
{
  //Get the client rect
  CRect rClient;
  GetClientRect(&rClient);

  //Get the position of where we have dragged to
  COSMCtrlPosition newPosition;
#ifdef COSMCTRL_NOD2D
  const Gdiplus::PointF tempPoint{static_cast<Gdiplus::REAL>(point.x), static_cast<Gdiplus::REAL>(point.y)};
#else
  const CD2DPointF tempPoint{static_cast<FLOAT>(point.x), static_cast<FLOAT>(point.y)};
#endif //#ifdef COSMCTRL_NOD2D
  if (ClientToPosition(tempPoint, rClient, newPosition))
  {
#pragma warning(suppress: 26446)
    auto& polyline{m_Polylines[m_nMouseDragItemIndex]};

    //Update the position of the node of the polyline bearing dragged
#pragma warning(suppress: 26446 26489)
    polyline->m_Nodes[m_nMouseDragSubItemIndex].m_Position = newPosition;

    //Force an update
    Invalidate(FALSE);
  }
}

void COSMCtrl::HandleMouseMovePolygonCreation(_In_ UINT /*nFlags*/, _In_ const CPoint& point)
{
  //Get the client rect
  CRect rClient;
  GetClientRect(&rClient);

  //Get the position of where we have dragged to
  COSMCtrlPosition newPosition;
#ifdef COSMCTRL_NOD2D
  const Gdiplus::PointF tempPoint{static_cast<Gdiplus::REAL>(point.x), static_cast<Gdiplus::REAL>(point.y)};
#else
  const CD2DPointF tempPoint{static_cast<FLOAT>(point.x), static_cast<FLOAT>(point.y)};
#endif //#ifdef COSMCTRL_NOD2D
  if (ClientToPosition(tempPoint, rClient, newPosition))
  {
    //Pull out the polygon we are operating on
#pragma warning(suppress: 26446)
    auto& polygon{m_Polygons[m_nMouseDragItemIndex]};

#pragma warning(suppress: 26489)
    if (polygon->m_Nodes.size() == 4)
    {
      //Update the position of the second, third and fourth nodes of the polygon to support rectangle selection
#pragma warning(suppress: 26446 26489)
      polygon->m_Nodes[1].m_Position.m_fLongitude = newPosition.m_fLongitude;
#pragma warning(suppress: 26446 26489)
      polygon->m_Nodes[2].m_Position = newPosition;
#pragma warning(suppress: 26446 26489)
      polygon->m_Nodes[3].m_Position.m_fLatitude = newPosition.m_fLatitude;

      //Force an update
      Invalidate(FALSE);
    }
  }
}

void COSMCtrl::HandleMouseMoveCircleCreation(_In_ UINT /*nFlags*/, _In_ const CPoint& point)
{
  //Get the client rect
  CRect rClient;
  GetClientRect(&rClient);

  //Work out the new position of where we have dragged to
  COSMCtrlPosition newPosition;
#ifdef COSMCTRL_NOD2D
  const Gdiplus::PointF ptDrag{static_cast<Gdiplus::REAL>(point.x), static_cast<Gdiplus::REAL>(point.y)};
#else
  const CD2DPointF ptDrag{static_cast<FLOAT>(point.x), static_cast<FLOAT>(point.y)};
#endif //#ifdef COSMCTRL_NOD2D
  if (ClientToPosition(ptDrag, rClient, newPosition))
  {
    //Pull out the circle we are operating on
#pragma warning(suppress: 26446)
    auto& circle{m_Circles[m_nMouseDragItemIndex]};

    //Update the radius of the circle
    circle->m_fRadius = COSMCtrlHelper::DistanceBetweenPoints(circle->m_Position, newPosition, nullptr, nullptr);

    //Force an update
    Invalidate(FALSE);
  }
}

void COSMCtrl::HandleMouseMovePolygonCreationWithNode(_In_ UINT /*nFlags*/, _In_ const CPoint& point)
{
  //Get the client rect
  CRect rClient;
  GetClientRect(&rClient);

  //Get the position of where we have dragged to
  COSMCtrlPosition newPosition;
#ifdef COSMCTRL_NOD2D
  const Gdiplus::PointF tempPoint{static_cast<Gdiplus::REAL>(point.x), static_cast<Gdiplus::REAL>(point.y)};
#else
  const CD2DPointF tempPoint{static_cast<FLOAT>(point.x), static_cast<FLOAT>(point.y)};
#endif //#ifdef COSMCTRL_NOD2D
  if (ClientToPosition(tempPoint, rClient, newPosition))
  {
    //Pull out the polygon we are operating on
#pragma warning(suppress: 26446)
    auto& polygon{m_Polygons[m_nMouseDragItemIndex]};

    //Update the position of the node of the polygon being dragged
#pragma warning(suppress: 26446 26489)
    polygon->m_Nodes[m_nMouseDragSubItemIndex].m_Position = newPosition;

    //Force an update
    Invalidate(FALSE);
  }
}

bool COSMCtrl::ContainsPosition(_In_ const COSMCtrlPosition& topLeft, _In_ const COSMCtrlPosition& bottomRight, _In_ const COSMCtrlPosition& point) noexcept
{
    if ((point.m_fLatitude < topLeft.m_fLatitude) && (point.m_fLatitude > bottomRight.m_fLatitude) && (point.m_fLongitude > topLeft.m_fLongitude) && (point.m_fLongitude < bottomRight.m_fLongitude))
      return true;
    else
      return false;
}

void COSMCtrl::HandleMouseMoveSelection(_In_ UINT /*nFlags*/, _In_ const CPoint& point)
{
  //Get the client rect
  CRect rClient;
  GetClientRect(&rClient);

  //Get the position of where we have dragged to
  COSMCtrlPosition newPosition;
#ifdef COSMCTRL_NOD2D
  const Gdiplus::PointF tempPoint{static_cast<Gdiplus::REAL>(point.x), static_cast<Gdiplus::REAL>(point.y)};
#else
  const CD2DPointF tempPoint{static_cast<FLOAT>(point.x), static_cast<FLOAT>(point.y)};
#endif //#ifdef COSMCTRL_NOD2D
  if (ClientToPosition(tempPoint, rClient, newPosition))
  {
    //Update the position of the second, third and fourth nodes of the polygon to support rectangle selection
#pragma warning(suppress: 26446)
    m_SelectionPolygon->m_Nodes[1].m_Position.m_fLongitude = newPosition.m_fLongitude;
#pragma warning(suppress: 26446)
    m_SelectionPolygon->m_Nodes[2].m_Position = newPosition;
#pragma warning(suppress: 26446)
    m_SelectionPolygon->m_Nodes[3].m_Position.m_fLatitude = newPosition.m_fLatitude;

    //Get the new bounding rect of the selection
    //CRect rBounds;
    COSMCtrlPosition topLeftSelection;
    COSMCtrlPosition bottomRightSelection;
    if (m_SelectionPolygon->GetBoundingRect(topLeftSelection, bottomRightSelection))
    {
      //First do the hit testing of markers
      for (auto& marker : m_Markers)
      {
        if (ContainsPosition(topLeftSelection, bottomRightSelection, marker->m_Position))
          marker->m_bSelected = true;
      }

      //Next do the hit testing of polyline nodes
      for (auto& polyline : m_Polylines)
      {
#pragma warning(suppress: 26489)
        for (auto& node : polyline->m_Nodes)
        {
#pragma warning(suppress: 26489)
          if (ContainsPosition(topLeftSelection, bottomRightSelection, node.m_Position))
#pragma warning(suppress: 26489)
            node.m_bSelected = true;
        }
      }

      //next do the hit testing of polygons
      for (auto& polygon : m_Polygons)
      {
#pragma warning(suppress: 26489)
        for (auto& node : polygon->m_Nodes)
        {
#pragma warning(suppress: 26489)
          if (ContainsPosition(topLeftSelection, bottomRightSelection, node.m_Position))
#pragma warning(suppress: 26489)
            node.m_bSelected = true;
        }
      }

      //First do the hit testing of circles
      for (auto& circle : m_Circles)
      {
        if (ContainsPosition(topLeftSelection, bottomRightSelection, circle->m_Position))
          circle->m_bSelected = true;
      }

      //Force an update
      Invalidate(FALSE);
    }
  }
}

#ifndef COSMCTRL_NOD2D
void COSMCtrl::HandleMouseMoveRotate(_In_ UINT /*nFlags*/, _In_ const CPoint& point)
{
  //Get the client rect
  CRect rClient;
  GetClientRect(&rClient);

  //Rotate around the map center
  const CPoint ptCenter{rClient.left + rClient.Width() / 2, rClient.top + rClient.Height() / 2};
  const CPoint ptOldDelta{m_pointMouseCapture - ptCenter};
  const CPoint ptNewDelta{point - ptCenter};
  const double fOldBearing{atan2(ptOldDelta.y, ptOldDelta.x)};
  double fNewBearing{atan2(ptNewDelta.y, ptNewDelta.x) - fOldBearing};
  fNewBearing = m_fBearingAtTopOfMapMouseCapture - 90.0 * fNewBearing / M_PI_2;
  SetBearingOfTopOfMap(fNewBearing, 0, false);
}
#endif //#ifndef COSMCTRL_NOD2D

void COSMCtrl::HandleMouseMoveMap(_In_ UINT /*nFlags*/, _In_ const CPoint& point)
{
  //Work out the size of a tile at the current zoom level
  double fInt{0};
  const double fFractionalZoom{modf(m_fZoom, &fInt)};
  const double fOSMTileWidth{static_cast<Gdiplus::REAL>(OSMTileWidth + (fFractionalZoom * OSMTileWidth))};
  const double fOSMTileHeight{static_cast<Gdiplus::REAL>(OSMTileHeight + (fFractionalZoom * OSMTileHeight))};

  //Work out the longitude and latitude of the position where we have dragged to
  const int nZoom{static_cast<int>(m_fZoom)};
  const double fDeltaX{(static_cast<double>(point.x) - m_pointMouseCapture.x)/fOSMTileWidth};
  const double fDeltaY{(static_cast<double>(point.y) - m_pointMouseCapture.y)/fOSMTileHeight};
  const double fRadBearing{m_fBearingOfTopOfMap * 0.017453292519943295769236907684886};
  const double fCosBearing{cos(fRadBearing)};
  const double fSinBearing{sin(fRadBearing)};
  const double fRotateX{fDeltaX*fCosBearing - fDeltaY*fSinBearing};
  const double fRotateY{fDeltaX*fSinBearing + fDeltaY*fCosBearing};
  const double fX{COSMCtrlHelper::Longitude2TileX(m_CenterPosition.m_fLongitude, nZoom) - fRotateX};
  const double fNewLongitude{COSMCtrlHelper::TileX2Longitude(fX, nZoom)};
  const double fY{COSMCtrlHelper::Latitude2TileY(m_CenterPosition.m_fLatitude, nZoom) - fRotateY};
  const double fNewLatitude{COSMCtrlHelper::TileY2Latitude(fY, nZoom)};

  //Hive away the mouse position for the next mouse move
  m_pointMouseCapture = point;

  //Move to the new position (without forcing the download thread to restart if necessary)
  SetCenter(COSMCtrlPosition{fNewLongitude, fNewLatitude}, 0, false);

  //Force a update at this point
  UpdateWindow();
}

void COSMCtrl::HandleMouseMoveMarker(_In_ UINT /*nFlags*/, _In_ const CPoint& point)
{
  //Get the client rect
  CRect rClient;
  GetClientRect(&rClient);

  //Work out the new position of the marker
  COSMCtrlPosition newPosition;
#ifdef COSMCTRL_NOD2D
  const Gdiplus::PointF ptDrag{static_cast<Gdiplus::REAL>(point.x - m_DragOffset.cx), static_cast<Gdiplus::REAL>(point.y - m_DragOffset.cy)};
#else
  const CD2DPointF ptDrag{static_cast<FLOAT>(point.x - m_DragOffset.cx), static_cast<FLOAT>(point.y - m_DragOffset.cy)};
#endif //#ifdef COSMCTRL_NOD2D
  if (ClientToPosition(ptDrag, rClient, newPosition))
  {
    //Pull out the marker we are going to operate on
#pragma warning(suppress: 26446)
    auto& marker{m_Markers[m_nMouseDragItemIndex]};

    //Now update the marker's position
    marker->m_Position = newPosition;

    //Force an update
    Invalidate(FALSE);
  }
}

void COSMCtrl::HandleMouseMovePolylineNode(_In_ UINT /*nFlags*/, _In_ const CPoint& point)
{
  //Get the client rect
  CRect rClient;
  GetClientRect(&rClient);

  //Work out the new position of the node
  COSMCtrlPosition newPosition;
#ifdef COSMCTRL_NOD2D
  const Gdiplus::PointF tempPoint{static_cast<Gdiplus::REAL>(point.x), static_cast<Gdiplus::REAL>(point.y)};
#else
  const CD2DPointF tempPoint{static_cast<FLOAT>(point.x), static_cast<FLOAT>(point.y)};
#endif //#ifdef COSMCTRL_NOD2D
  if (ClientToPosition(tempPoint, rClient, newPosition))
  {
    //Pull out the polyline we are going to operate on
#pragma warning(suppress: 26446)
    auto& polyline{m_Polylines[m_nMouseDragItemIndex]};

    //move the node of the polyline
#pragma warning(suppress: 26446 26489)
    polyline->m_Nodes[m_nMouseDragSubItemIndex].m_Position = newPosition;

    //Force an update
    Invalidate(FALSE);
  }
}

void COSMCtrl::HandleMouseMovePolyline(_In_ UINT /*nFlags*/, _In_ const CPoint& point)
{
  //Get the client rect
  CRect rClient;
  GetClientRect(&rClient);

  //Work out the new position of the first point in the polyline
  COSMCtrlPosition newPosition;
#ifdef COSMCTRL_NOD2D
  const Gdiplus::PointF ptDrag{static_cast<Gdiplus::REAL>(point.x - m_DragOffset.cx), static_cast<Gdiplus::REAL>(point.y - m_DragOffset.cy)};
#else
  const CD2DPointF ptDrag{static_cast<FLOAT>(point.x - m_DragOffset.cx), static_cast<FLOAT>(point.y - m_DragOffset.cy)};
#endif //#ifdef COSMCTRL_NOD2D
  if (ClientToPosition(ptDrag, rClient, newPosition))
  {
    //Pull out the polyline we are going to operate on
#pragma warning(suppress: 26446)
    auto& polyline{m_Polylines[m_nMouseDragItemIndex]};

    //Update all the polyline's positions by the specified amount
#pragma warning(suppress: 26446)
    const COSMCtrlPosition oldNode0Position{m_OriginalDragNodes[0].m_Position};
    std::vector<COSMCtrlNode> newNodes;
#pragma warning(suppress: 26489)
    newNodes.reserve(polyline->m_Nodes.size());
    bool bAllowMove{true};
#pragma warning(suppress: 26489)
    for (std::vector<COSMCtrlNode>::size_type i{0}; (i<polyline->m_Nodes.size()) && bAllowMove; i++)
    {
#pragma warning(suppress: 26446)
      const COSMCtrlNode& oldNode{m_OriginalDragNodes[i]};
      COSMCtrlNode newNode{oldNode};
      newNode.m_Position.m_fLongitude += (newPosition.m_fLongitude - oldNode0Position.m_fLongitude);
      newNode.m_Position.m_fLatitude += (newPosition.m_fLatitude - oldNode0Position.m_fLatitude);

      //Fix up moves across the international date line
      newNode.m_Position.NormalizeLongitude();

    #ifdef COSMCTRL_NOD2D
      Gdiplus::PointF tempPoint;
    #else
      CD2DPointF tempPoint;
    #endif //#ifdef COSMCTRL_NOD2D
      bAllowMove = bAllowMove && PositionToClient(newNode.m_Position, rClient, tempPoint);
      newNodes.push_back(newNode);
    }
    if (bAllowMove)
    {
      //move the polyline
#pragma warning(suppress: 26489)
      polyline->m_Nodes = newNodes;

      //Force an update
      Invalidate(FALSE);
    }
  }
}

void COSMCtrl::HandleMouseMovePolygonNode(_In_ UINT /*nFlags*/, _In_ const CPoint& point)
{
  //Get the client rect
  CRect rClient;
  GetClientRect(&rClient);

  //Work out the new position of the node
  COSMCtrlPosition newPosition;
#ifdef COSMCTRL_NOD2D
  const Gdiplus::PointF tempPoint{static_cast<Gdiplus::REAL>(point.x), static_cast<Gdiplus::REAL>(point.y)};
#else
  const CD2DPointF tempPoint{static_cast<FLOAT>(point.x), static_cast<FLOAT>(point.y)};
#endif //#ifdef COSMCTRL_NOD2D
  if (ClientToPosition(tempPoint, rClient, newPosition))
  {
    //Pull out the polygon we are going to operate on
#pragma warning(suppress: 26446)
    auto& polygon{m_Polygons[m_nMouseDragItemIndex]};

    //move the node of the polygon
#pragma warning(suppress: 26446 26489)
    polygon->m_Nodes[m_nMouseDragSubItemIndex].m_Position = newPosition;

    //Force an update
    Invalidate(FALSE);
  }
}

void COSMCtrl::HandleMouseMovePolygon(_In_ UINT /*nFlags*/, _In_ const CPoint& point)
{
  //Get the client rect
  CRect rClient;
  GetClientRect(&rClient);

  //Work out the new position of the first point in the polygon
  COSMCtrlPosition newPosition;
#ifdef COSMCTRL_NOD2D
  const Gdiplus::PointF ptDrag{static_cast<Gdiplus::REAL>(point.x - m_DragOffset.cx), static_cast<Gdiplus::REAL>(point.y - m_DragOffset.cy)};
#else
  const CD2DPointF ptDrag{static_cast<FLOAT>(point.x - m_DragOffset.cx), static_cast<FLOAT>(point.y - m_DragOffset.cy)};
#endif //#ifdef COSMCTRL_NOD2D
  if (ClientToPosition(ptDrag, rClient, newPosition))
  {
    //Pull out the polygon we are going to operate on
#pragma warning(suppress: 26446)
    auto& polygon{m_Polygons[m_nMouseDragItemIndex]};

    //Update all the polygon's positions by the specified amount
#pragma warning(suppress: 26446)
    const COSMCtrlPosition oldNode0Position{m_OriginalDragNodes[0].m_Position};
    std::vector<COSMCtrlNode> newNodes;
#pragma warning(suppress: 26489)
    newNodes.reserve(polygon->m_Nodes.size());
    bool bAllowMove{true};
#pragma warning(suppress: 26489)
    for (std::vector<COSMCtrlNode>::size_type i{0}; (i<polygon->m_Nodes.size()) && bAllowMove; i++)
    {
#pragma warning(suppress: 26446)
      const COSMCtrlNode& oldNode{m_OriginalDragNodes[i]};
      COSMCtrlNode newNode{oldNode};
      newNode.m_Position.m_fLongitude += (newPosition.m_fLongitude - oldNode0Position.m_fLongitude);
      newNode.m_Position.m_fLatitude += (newPosition.m_fLatitude - oldNode0Position.m_fLatitude);

      //Fix up moves across the international date line
      newNode.m_Position.NormalizeLongitude();

    #ifdef COSMCTRL_NOD2D
      Gdiplus::PointF tempPoint;
    #else
      CD2DPointF tempPoint;
    #endif //#ifdef COSMCTRL_NOD2D
      bAllowMove = bAllowMove && PositionToClient(newNode.m_Position, rClient, tempPoint);
      newNodes.push_back(newNode);
    }
    if (bAllowMove)
    {
      //Move the polygon
#pragma warning(suppress: 26489)
      polygon->m_Nodes = newNodes;

      //Force an update
      Invalidate(FALSE);
    }
  }
}

void COSMCtrl::HandleMouseMoveCircle(_In_ UINT /*nFlags*/, _In_ const CPoint& point)
{
  //Get the client rect
  CRect rClient;
  GetClientRect(&rClient);

  //Modify the radius of the circle if that's the mode we are in
  if (m_bModifyingCircleRadius)
  {
    //Work out the new position of the marker
    COSMCtrlPosition newPosition;
  #ifdef COSMCTRL_NOD2D
    const Gdiplus::PointF ptDrag{static_cast<Gdiplus::REAL>(point.x), static_cast<Gdiplus::REAL>(point.y)};
  #else
    const CD2DPointF ptDrag{static_cast<FLOAT>(point.x), static_cast<FLOAT>(point.y)};
  #endif //#ifdef COSMCTRL_NOD2D
    if (ClientToPosition(ptDrag, rClient, newPosition))
    {
      //Pull out the circle we are going to operate on
#pragma warning(suppress: 26446)
      auto& circle{m_Circles[m_nMouseDragItemIndex]};

      //Work out the new radius of the circle
      const double fNewRadius{COSMCtrlHelper::DistanceBetweenPoints(circle->m_Position, newPosition, nullptr, nullptr)};

      //Work out the new radius in pixels of the circle
      const COSMCtrlPosition position90{COSMCtrlHelper::GetPosition(circle->m_Position, 90, fNewRadius, nullptr)};
      const COSMCtrlPosition position270{COSMCtrlHelper::GetPosition(circle->m_Position, 270, fNewRadius, nullptr)};
    #ifdef COSMCTRL_NOD2D
      Gdiplus::PointF pt90;
      Gdiplus::PointF pt270;
      Gdiplus::PointF ptCenter;
      Gdiplus::REAL fDiameter = 0;
    #else
      CD2DPointF pt90;
      CD2DPointF pt270;
      CD2DPointF ptCenter;
      FLOAT fDiameter = 0;
    #endif //#ifdef COSMCTRL_NOD2D
      if (PositionToClient(position90, rClient, pt90) && PositionToClient(position270, rClient, pt270) && PositionToClient(circle->m_Position, rClient, ptCenter))
      {
      #ifdef COSMCTRL_NOD2D
        fDiameter = pt90.X - pt270.X;
        const Gdiplus::PointF pt0{ptCenter.X, ptCenter.Y - fDiameter/2};
        const Gdiplus::PointF pt180{ptCenter.X, ptCenter.Y + fDiameter/2};
      #else
        fDiameter = pt90.x - pt270.x;
        const CD2DPointF pt0{ptCenter.x, ptCenter.y - (fDiameter/2)};
        const CD2DPointF pt180{ptCenter.x, ptCenter.y + (fDiameter/2)};
      #endif //#ifdef COSMCTRL_NOD2D

        COSMCtrlPosition position0;
        COSMCtrlPosition position180;
        if (ClientToPosition(pt0, rClient, position0) && ClientToPosition(pt180, rClient, position180))
        {
          //Update the radius of the circle
          circle->m_fRadius = COSMCtrlHelper::DistanceBetweenPoints(circle->m_Position, newPosition, nullptr, nullptr);

          //Force an update
          Invalidate(FALSE);
        }
      }
    }
  }
  else
  {
    //Work out the new position of the circle
    COSMCtrlPosition newPosition;
  #ifdef COSMCTRL_NOD2D
    const Gdiplus::PointF ptDrag{static_cast<Gdiplus::REAL>(point.x - m_DragOffset.cx), static_cast<Gdiplus::REAL>(point.y - m_DragOffset.cy)};
  #else
    const CD2DPointF ptDrag{static_cast<FLOAT>(point.x - m_DragOffset.cx), static_cast<FLOAT>(point.y - m_DragOffset.cy)};
  #endif //#ifdef COSMCTRL_NOD2D
    if (ClientToPosition(ptDrag, rClient, newPosition))
    {
      //Pull out the circle we are going to operate on
#pragma warning(suppress: 26446)
      auto& circle{m_Circles[m_nMouseDragItemIndex]};

      COSMCtrlCircle tempCircle{*circle};
      tempCircle.m_Position = newPosition;

      //Work out the new radius in pixels of the circle
      const COSMCtrlPosition position90{COSMCtrlHelper::GetPosition(tempCircle.m_Position, 90, tempCircle.m_fRadius, nullptr)};
      const COSMCtrlPosition position270{COSMCtrlHelper::GetPosition(tempCircle.m_Position, 270, tempCircle.m_fRadius, nullptr)};
    #ifdef COSMCTRL_NOD2D
      Gdiplus::PointF pt90;
      Gdiplus::PointF pt270;
      Gdiplus::PointF ptCenter;
      Gdiplus::REAL fDiameter{0};
    #else
      CD2DPointF pt90;
      CD2DPointF pt270;
      CD2DPointF ptCenter;
      FLOAT fDiameter{0};
    #endif //#ifdef COSMCTRL_NOD2D
      if (PositionToClient(position90, rClient, pt90) && PositionToClient(position270, rClient, pt270) && PositionToClient(tempCircle.m_Position, rClient, ptCenter))
      {
      #ifdef COSMCTRL_NOD2D
        fDiameter = pt90.X - pt270.X;
        const Gdiplus::PointF pt0{ptCenter.X, ptCenter.Y - (fDiameter/2)};
        const Gdiplus::PointF pt180{ptCenter.X, ptCenter.Y + (fDiameter/2)};
      #else
        fDiameter = pt90.x - pt270.x;
        const CD2DPointF pt0{ptCenter.x, ptCenter.y - (fDiameter/2)};
        const CD2DPointF pt180{ptCenter.x, ptCenter.y + (fDiameter/2)};
      #endif //#ifdef COSMCTRL_NOD2D

        COSMCtrlPosition position0;
        COSMCtrlPosition position180;
        if (ClientToPosition(pt0, rClient, position0) && ClientToPosition(pt180, rClient, position180))
        {
          //Now update the circle's position
          circle->m_Position = newPosition;

          //Force an update
          Invalidate(FALSE);
        }
      }
    }
  }
#pragma warning(suppress: 26812)
}

#pragma warning(suppress: 26434)
void COSMCtrl::OnMouseMove(UINT nFlags, CPoint point)
{
  //Call the event handler if possible
  if (m_pEventHandler != nullptr)
  {
    if (m_pEventHandler->OnOSMCtrlMouseMove(nFlags, point))
      return;
  }

  if (m_bMouseCapturedForDrag)
  {
    switch (m_MouseDragItem)
    {
      case MapItem::Map:
      {
        if (m_MapMode == MapMode::Selection)
          HandleMouseMoveSelection(nFlags, point);
        else if (m_MapMode == MapMode::PolylineCreation)
          HandleMouseMovePolylineCreation(nFlags, point);
        else if (m_MapMode == MapMode::PolygonCreation)
          HandleMouseMovePolygonCreation(nFlags, point);
        else if (m_MapMode == MapMode::CircleCreation)
          HandleMouseMoveCircleCreation(nFlags, point);
      #ifndef COSMCTRL_NOD2D
        else if (m_MapMode == MapMode::Rotate)
          HandleMouseMoveRotate(nFlags, point);
      #endif //#ifndef COSMCTRL_NOD2D
        else
          HandleMouseMoveMap(nFlags, point);
        break;
      }
      case MapItem::Marker:
      {
        HandleMouseMoveMarker(nFlags, point);
        break;
      }
      case MapItem::Circle: //Deliberate fallthrough
      case MapItem::CircleCircumference:
      {
        if (m_MapMode == MapMode::CircleCreation)
          HandleMouseMoveCircleCreation(nFlags, point);
        else
          HandleMouseMoveCircle(nFlags, point);
        break;
      }
      case MapItem::PolylineNode:
      {
        if (m_MapMode == MapMode::PolylineCreation)
          HandleMouseMovePolylineCreation(nFlags, point);
        else
          HandleMouseMovePolylineNode(nFlags, point);
        break;
      }
      case MapItem::Polyline:
      {
        HandleMouseMovePolyline(nFlags, point);
        break;
      }
      case MapItem::PolygonNode:
      {
        if (m_MapMode == MapMode::PolygonCreation)
          HandleMouseMovePolygonCreationWithNode(nFlags, point);
        else
          HandleMouseMovePolygonNode(nFlags, point);
        break;
      }
      case MapItem::Polygon:
      {
        HandleMouseMovePolygon(nFlags, point);
        break;
      }
      default:
      {
        ASSERT(false);
        break;
      }
    }
  }
  else if (m_bMouseCapturedForZoom)
  {
    //Work out the zoom level and set it
    const double fZoom{min(max(((static_cast<double>(m_rZoomBar.bottom) - 8 - point.y) / 11.0) + OSMMinZoom, OSMMinZoom), OSMMaxZoom)};
    SetZoom(fZoom, 0, false);
  }
  else
    __super::OnMouseMove(nFlags, point); //Let the base class do its thing
}

#pragma warning(suppress: 26434)
void COSMCtrl::OnLButtonUp(UINT nFlags, CPoint point)
{
  //Call the event handler if possible
  if (m_pEventHandler != nullptr)
  {
    if (m_pEventHandler->OnOSMCtrlLButtonUp(nFlags, point))
      return;
  }

  //Release mouse capture if we had captured it
  if (m_bMouseCapturedForDrag)
  {
    m_bMouseCapturedForDrag = false;
    ReleaseCapture();

    switch (m_MouseDragItem)
    {
      case MapItem::Map:
      {
        //Restart downloading of tiles if necessary
        if (m_bDownloadTiles && m_sCacheDirectory.GetLength())
        {
          StopAsyncDownload(false);
          StartAsyncDownload(true, false);
        }
        break;
      }
      default:
      {
        break;
      }
    }
  }
  else if (m_bMouseCapturedForZoom)
  {
    //Work out the zoom level and set it
    const double fZoom{min(max(((static_cast<double>(m_rZoomBar.bottom) - 8 - point.y) / 11.0) + OSMMinZoom, OSMMinZoom), OSMMaxZoom)};
    SetZoom(fZoom, 0, false);

    //Release the mouse
    m_bMouseCapturedForZoom = false;
    ReleaseCapture();

    //Restart downloading of tiles if necessary
    if (m_bDownloadTiles && m_sCacheDirectory.GetLength())
    {
      StopAsyncDownload(false);
      StartAsyncDownload(true, false);
    }
  }

  //let the base class do its thing
  __super::OnLButtonUp(nFlags, point);
}

#pragma warning(suppress: 26434)
void COSMCtrl::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags)
{
  //Call the event handler if possible
  if (m_pEventHandler != nullptr)
  {
    if (m_pEventHandler->OnOSMCtrlKeyDown(nChar, nRepCnt, nFlags))
      return;
  }

  if (m_bAllowKeyboard)
  {
    const bool bControl{(GetKeyState(VK_CONTROL) & 0x8000) != 0};
    if ((nChar == VK_BACK) || ((nChar == VK_LEFT) && bControl))
      ScrollTileLeft();
    else if (nChar == VK_LEFT)
      ScrollLeft(m_nScrollPixels);
    else if ((nChar == VK_END) || ((nChar == VK_RIGHT) && bControl))
      ScrollTileRight();
    else if (nChar == VK_RIGHT)
      ScrollRight(m_nScrollPixels);
    else if ((nChar == VK_PRIOR) || ((nChar == VK_UP) && bControl))
      ScrollTileUp();
    else if (nChar == VK_UP)
      ScrollUp(m_nScrollPixels);
    else if ((nChar == VK_NEXT) || ((nChar == VK_DOWN) && bControl))
      ScrollTileDown();
    else if (nChar == VK_DOWN)
      ScrollDown(m_nScrollPixels);
    else if (nChar == VK_F5)
      Refresh();
    else if (nChar == VK_DELETE)
      DeleteSelectedItems();
    else if ((nChar == 'A') && bControl)
      SelectAllItems();
  }

  //Let the base class do its thing
  __super::OnKeyDown(nChar, nRepCnt, nFlags);
}

#pragma warning(suppress: 26434)
void COSMCtrl::OnChar(UINT nChar, UINT nRepCnt, UINT nFlags)
{
  if (m_bAllowKeyboard)
  {
    if (nChar == '+')
    {
      //Increase the zoom
      double fZoom{GetZoom()};
      ++fZoom;
      if (fZoom > OSMMaxZoom)
        fZoom = OSMMaxZoom;
      SetZoom(fZoom, m_fDefaultAnimationDuration, true);
    }
    else if (nChar == '-')
    {
      //Decrease the zoom
      double fZoom{GetZoom()};
      --fZoom;
      if (fZoom < OSMMinZoom)
        fZoom = OSMMinZoom;
      SetZoom(fZoom, m_fDefaultAnimationDuration, true);
    }
  }
  
  //Let the base class do its thing
  __super::OnChar(nChar, nRepCnt, nFlags);
}

#pragma warning(suppress: 26434)
BOOL COSMCtrl::OnMouseWheel(UINT nFlags, short zDelta, CPoint pt)
{
  if (m_bAllowMouseZoom)
  {
    //Work out the new zoom level
    double fZoom{-1};
    if (zDelta > 0)
    {
      //Increase the zoom
      fZoom = GetZoom();
      ++fZoom;
      if (fZoom > OSMMaxZoom)
        fZoom = OSMMaxZoom;
    }
    else if (zDelta < 0)
    {
      //Decrease the zoom
      fZoom = GetZoom();
      --fZoom;
      if (fZoom < OSMMinZoom)
        fZoom = OSMMinZoom;
    }

    if (fZoom != -1)
    {
      //Get the client rect
      CRect rClient;
      GetClientRect(&rClient);

      //First get the position underneath the cursor
      COSMCtrlPosition position;
      CPoint ptClient{pt};
      ScreenToClient(&ptClient);

      if (rClient.PtInRect(ptClient))
      {
      #ifdef COSMCTRL_NOD2D
        const Gdiplus::PointF tempPoint{static_cast<Gdiplus::REAL>(ptClient.x), static_cast<Gdiplus::REAL>(ptClient.y)};
      #else
        const CD2DPointF tempPoint{static_cast<FLOAT>(ptClient.x), static_cast<FLOAT>(ptClient.y)};
      #endif //#ifdef COSMCTRL_NOD2D
        if (ClientToPosition(tempPoint, rClient, position))
        {
          const int nZoom{static_cast<int>(fZoom)};

          //Work out the size of a tile at the current zoom level
          double fInt{0};
          const double fFractionalZoom{modf(m_fZoom, &fInt)};
        #ifdef COSMCTRL_NOD2D
          const auto fOSMTileWidth{static_cast<Gdiplus::REAL>(OSMTileWidth + (fFractionalZoom * OSMTileWidth))};
          const auto fOSMTileHeight{static_cast<Gdiplus::REAL>(OSMTileHeight + (fFractionalZoom * OSMTileHeight))};
        #else
          const auto fOSMTileWidth{static_cast<FLOAT>(OSMTileWidth + (fFractionalZoom * OSMTileWidth))};
          const auto fOSMTileHeight{static_cast<FLOAT>(OSMTileHeight + (fFractionalZoom * OSMTileHeight))};
        #endif //#ifdef COSMCTRL_NOD2D

          //Calculate the new center of the map at the new zoom level which will keep the position underneath the cursor at the same location
          double fX{COSMCtrlHelper::Longitude2TileX(position.m_fLongitude, nZoom)};
          double fY{COSMCtrlHelper::Latitude2TileY(position.m_fLatitude, nZoom)};
          fX -= ((ptClient.x - (rClient.Width()/2.0) + rClient.left) / fOSMTileWidth);
          fY -= ((ptClient.y - (rClient.Height()/2.0) + rClient.top) / fOSMTileHeight);
          const double fMaxTile{pow(2.0, nZoom)};

          //Wrap the longitude if necessary
          while (fX < 0)
            fX += fMaxTile;
          while (fX >= fMaxTile)
            fX -= fMaxTile;

          if ((fX >= 0) && (fX <= fMaxTile) && (fY >= 0) && (fY <= fMaxTile))
          {
            COSMCtrlPosition newCenterPosition;
            newCenterPosition.m_fLatitude = COSMCtrlHelper::TileY2Latitude(fY, nZoom);
            newCenterPosition.m_fLongitude = COSMCtrlHelper::TileX2Longitude(fX, nZoom);

            SetCenterAndZoom(newCenterPosition, fZoom, m_fDefaultAnimationDuration, true);
          }
        }
      }
      else
        SetZoom(fZoom, m_fDefaultAnimationDuration, true);
    }
  }

  //let the base class do its thing
  return __super::OnMouseWheel(nFlags, zDelta, pt);
}

COSMCtrl::MapItem COSMCtrl::HitTest(_In_ const CPoint& point, _Out_ INT_PTR& nItem, _Out_ INT_PTR& nSubItem, _In_ const CRect& rClient) const
{
  //Set the out parameters to default values
  nItem = -1;
  nSubItem = -1;

  //First do the hit testing of markers
  for (std::vector<COSMCtrlMarker>::size_type i{0}; i<m_Markers.size(); i++)
  {
    //Pull out the marker we are working on
#pragma warning(suppress: 26446)
    const auto& marker{m_Markers[i]};

    if (marker->m_bHitTest)
    {
      CRect rMarker;
      if (GetBoundingRect(*marker, rMarker, rClient) && rMarker.PtInRect(point))
      {
        nItem = i;
        return MapItem::Marker;
      }
    }
  }

  //Now do the hit testing on all the controls
  if (m_rNorthScrollRose.PtInRect(point))
    return MapItem::ScrollerUp;
  else if (m_rSouthScrollRose.PtInRect(point))
    return MapItem::ScrollerDown;
  else if (m_rEastScrollRose.PtInRect(point))
    return MapItem::ScrollerRight;
  else if (m_rWestScrollRose.PtInRect(point))
    return MapItem::ScrollerLeft;
  else if (m_rZoomInZoomBar.PtInRect(point))
    return MapItem::ZoomIn;
  else if (m_rZoomOutZoomBar.PtInRect(point))
    return MapItem::ZoomOut;
  else if (m_rZoomBar.PtInRect(point))
    return MapItem::ZoomBar;
  else
  {
    //Next do the hit testing of polylines
    for (std::vector<COSMCtrlPolyline>::size_type i{0}; i<m_Polylines.size(); i++)
    {
      //Pull out the polyline we are working on
#pragma warning(suppress: 26446)
      const auto& polyline{m_Polylines[i]};

      if (polyline->m_bHitTest)
      {
        //First do hit testing on the polyline nodes
#pragma warning(suppress: 26489)
        for (std::vector<COSMCtrlNode>::size_type j{0}; j<polyline->m_Nodes.size(); j++)
        {
          CRect rNode;
          if (GetBoundingNodeRect(*polyline, j, rNode, rClient) && rNode.PtInRect(point))
          {
            nItem = i;
            nSubItem = j;
            return MapItem::PolylineNode;
          }
        }

        if (HitTest(point, *polyline, rClient))
        {
          nItem = i;
          return MapItem::Polyline;
        }
      }
    }

    //next do the hit testing of polygons
    for (std::vector<COSMCtrlPolygon>::size_type i{0}; i<m_Polygons.size(); i++)
    {
      //Pull out the polygon we are working on
#pragma warning(suppress: 26446)
      const auto& polygon{m_Polygons[i]};

      if (polygon->m_bHitTest)
      {
        //First do hit testing on the polygon nodes
#pragma warning(suppress: 26489)
        for (std::vector<COSMCtrlNode>::size_type j{0}; j<polygon->m_Nodes.size(); j++)
        {
          CRect rNode;
          if (GetBoundingNodeRect(*polygon, j, rNode, rClient) && rNode.PtInRect(point))
          {
            nItem = i;
            nSubItem = j;
            return MapItem::PolygonNode;
          }
        }

        if (HitTest(point, *polygon, rClient))
        {
          nItem = i;
          return MapItem::Polygon;
        }
      }
    }

    //next do the hit testing of circles
    for (std::vector<COSMCtrlCircle>::size_type i{0}; i<m_Circles.size(); i++)
    {
      //Pull out the circle we are working on
#pragma warning(suppress: 26446)
      const auto& circle{m_Circles[i]};

      if (circle->m_bHitTest)
      {
        //draw the item if the point is inside the bounding rect
        const MapItem tempItem{HitTest(point, *circle, rClient)};
        if (tempItem != MapItem::None)
        {
          nItem = i;
          return tempItem;
        }
      }
    }
    
    //Finally if the point is in the client area it must be the map, otherwise it is nothing
    if (rClient.PtInRect(point))
      return MapItem::Map;
    else
      return MapItem::None;
  }
}

bool COSMCtrl::ScrollUp(_In_ int nPixels)
{
  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

  //Get the client rect
  CRect rClient;
  GetClientRect(&rClient);

#ifdef COSMCTRL_NOD2D
  Gdiplus::PointF ptCenter;
#else
  CD2DPointF ptCenter;
#endif //#ifdef COSMCTRL_NOD2D
  if (PositionToClient(m_CenterPosition, rClient, ptCenter))
  {
  #ifdef COSMCTRL_NOD2D
    ptCenter.Y -= nPixels;
  #else
    ptCenter.y -= nPixels;
  #endif //#ifdef COSMCTRL_NOD2D
    COSMCtrlPosition newCenterPosition;
    if (ClientToPosition(ptCenter, rClient, newCenterPosition))
      bSuccess = SetCenter(newCenterPosition, 0, true);
  }

  return bSuccess;
}

bool COSMCtrl::ScrollDown(_In_ int nPixels)
{
  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

  //Get the client rect
  CRect rClient;
  GetClientRect(&rClient);

#ifdef COSMCTRL_NOD2D
  Gdiplus::PointF ptCenter;
#else
  CD2DPointF ptCenter;
#endif //#ifdef COSMCTRL_NOD2D
  if (PositionToClient(m_CenterPosition, rClient, ptCenter))
  {
  #ifdef COSMCTRL_NOD2D
    ptCenter.Y += nPixels;
  #else
    ptCenter.y += nPixels;
  #endif //#ifdef COSMCTRL_NOD2D
    COSMCtrlPosition newCenterPosition;
    if (ClientToPosition(ptCenter, rClient, newCenterPosition))
      bSuccess = SetCenter(newCenterPosition, 0, true);
  }

  return bSuccess;
}

bool COSMCtrl::ScrollLeft(_In_ int nPixels)
{
  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

  //Get the client rect
  CRect rClient;
  GetClientRect(&rClient);

#ifdef COSMCTRL_NOD2D
  Gdiplus::PointF ptCenter;
#else
  CD2DPointF ptCenter;
#endif //#ifdef COSMCTRL_NOD2D
  if (PositionToClient(m_CenterPosition, rClient, ptCenter))
  {
  #ifdef COSMCTRL_NOD2D
    ptCenter.X -= nPixels;
  #else
    ptCenter.x -= nPixels;
  #endif //#ifdef COSMCTRL_NOD2D
    COSMCtrlPosition newCenterPosition;
    if (ClientToPosition(ptCenter, rClient, newCenterPosition))
      bSuccess = SetCenter(newCenterPosition, 0, true);
  }

  return bSuccess;
}

bool COSMCtrl::ScrollRight(_In_ int nPixels)
{
  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

  //Get the client rect
  CRect rClient;
  GetClientRect(&rClient);

#ifdef COSMCTRL_NOD2D
  Gdiplus::PointF ptCenter;
#else
  CD2DPointF ptCenter;
#endif //#ifdef COSMCTRL_NOD2D
  if (PositionToClient(m_CenterPosition, rClient, ptCenter))
  {
  #ifdef COSMCTRL_NOD2D
    ptCenter.X += nPixels;
  #else
    ptCenter.x += nPixels;
  #endif //#ifdef COSMCTRL_NOD2D
    COSMCtrlPosition newCenterPosition;
    if (ClientToPosition(ptCenter, rClient, newCenterPosition))
      bSuccess = SetCenter(newCenterPosition, 0, true);
  }

  return bSuccess;
}

void COSMCtrl::ScrollTileUp()
{
  double fInt{0};
  const double fFractionalZoom{modf(m_fZoom, &fInt)};
  const double fOSMTileHeight{OSMTileHeight + (fFractionalZoom * OSMTileHeight)};

  ScrollUp(static_cast<int>(fOSMTileHeight));
}

void COSMCtrl::ScrollTileDown()
{
  double fInt{0};
  const double fFractionalZoom{modf(m_fZoom, &fInt)};
  const double fOSMTileHeight{OSMTileHeight + (fFractionalZoom * OSMTileHeight)};

  ScrollDown(static_cast<int>(fOSMTileHeight));
}

void COSMCtrl::ScrollTileLeft()
{
  double fInt{0};
  const double fFractionalZoom{modf(m_fZoom, &fInt)};
  const double fOSMTileWidth{OSMTileWidth + (fFractionalZoom * OSMTileWidth)};

  ScrollLeft(static_cast<int>(fOSMTileWidth));
}

void COSMCtrl::ScrollTileRight()
{
  double fInt{0};
  const double fFractionalZoom{modf(m_fZoom, &fInt)};
  const double fOSMTileWidth{OSMTileWidth + (fFractionalZoom * OSMTileWidth)};

  ScrollRight(static_cast<int>(fOSMTileWidth));
}

INT_PTR COSMCtrl::OnToolHitTest(CPoint point, TOOLINFO* pTI) const
{
  //Get the client rect
  CRect rClient;
  GetClientRect(&rClient);

  //Determine if the cursor is over a marker and if so setup the tooltip
  INT_PTR nItem{-1};
  INT_PTR nSubItem{-1};
  const MapItem item{HitTest(point, nItem, nSubItem, rClient)};
  CRect rMarker;
#pragma warning(suppress: 26446)
  if ((item == MapItem::Marker) && GetBoundingRect(*m_Markers[nItem], rMarker, rClient))
  {
    //Fill in the TOOLINFO struct
#pragma warning(suppress: 26496)
    AFXASSUME(pTI != nullptr);
    pTI->hwnd = GetSafeHwnd();
    pTI->uFlags = 0;
    pTI->uId = nItem;
    pTI->lpszText = LPSTR_TEXTCALLBACK;
    pTI->lParam = static_cast<int>(item);
    pTI->rect = rMarker;
    return pTI->uId;
  }
  else if ((item == MapItem::Circle) || (item == MapItem::Polygon) || (item == MapItem::Polyline) || (item == MapItem::PolygonNode) || (item == MapItem::PolylineNode))
  {
    //Fill in the TOOLINFO struct
#pragma warning(suppress: 26496)
    AFXASSUME(pTI != nullptr);
    pTI->hwnd = GetSafeHwnd();
    pTI->uFlags = 0;
    pTI->uId = nItem;
    pTI->lpszText = LPSTR_TEXTCALLBACK;
    pTI->lParam = static_cast<int>(item);
    pTI->rect = rClient;
    return pTI->uId;
  }

  return -1;
}

#pragma warning(suppress: 26440)
BOOL COSMCtrl::OnToolTipText(UINT /*nID*/, NMHDR* pNMHDR, LRESULT* pResult)
{
  //Validate our parameters
  ASSERT(pNMHDR != nullptr);
#pragma warning(suppress: 26496)
  AFXASSUME(pResult != nullptr);

#pragma warning(suppress: 26490)
  auto pTTT{reinterpret_cast<TOOLTIPTEXT*>(pNMHDR)};
#pragma warning(suppress: 26496)
  AFXASSUME(pTTT != nullptr);

  switch (static_cast<MapItem>(pTTT->lParam))
  {
    case MapItem::Marker:
    {
#pragma warning(suppress: 26446)
      const auto& marker{m_Markers[pTTT->hdr.idFrom]};
      m_sCurrentToolTipText = marker->m_sToolTipText;
      break;
    }
    case MapItem::Circle:
    {
#pragma warning(suppress: 26446)
      const auto& circle{m_Circles[pTTT->hdr.idFrom]};
      m_sCurrentToolTipText = circle->m_sToolTipText;
      break;
    }
    case MapItem::Polygon: //deliberate fallthrough
    case MapItem::PolygonNode:
    {
#pragma warning(suppress: 26446)
      const auto& polygon{m_Polygons[pTTT->hdr.idFrom]};
#pragma warning(suppress: 26485)
      m_sCurrentToolTipText = polygon->m_sToolTipText;
      break;
    }
    case MapItem::Polyline: //deliberate fallthrough
    case MapItem::PolylineNode:
    {
#pragma warning(suppress: 26446)
      const auto& polyline{m_Polylines[pTTT->hdr.idFrom]};
#pragma warning(suppress: 26485)
      m_sCurrentToolTipText = polyline->m_sToolTipText;
      break;
    }
    default:
    {
      ASSERT(false);
      break;
    }
  }

#pragma warning(suppress: 26465 26485 26492)
  pTTT->lpszText = const_cast<LPTSTR>(m_sCurrentToolTipText.GetString());
  *pResult = 0;

  return TRUE;
}

#pragma warning(suppress: 26434)
void COSMCtrl::OnVScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar)
{
  //Let the base class do its thing
  __super::OnVScroll(nSBCode, nPos, pScrollBar);

  //Handle the notification if it is from our control
#pragma warning(suppress: 26496)
  AFXASSUME(pScrollBar != nullptr);
  if (pScrollBar->GetSafeHwnd() == m_ctrlZoomBar.GetSafeHwnd())
  {
    switch (nSBCode)
    {
      case TB_BOTTOM:
      {
        SetZoom(OSMMinZoom, 0, true);
        break;
      }
      case TB_TOP:
      {
        SetZoom(OSMMaxZoom, 0, true);
        break;
      }
      case TB_LINEUP:
      {
        //Increase the zoom
        double fZoom{static_cast<double>(static_cast<int>(GetZoom()))}; //Try to keep the zoom at an integer level
        ++fZoom;
        SetZoom(fZoom, 0, true);
        break;
      }
      case TB_LINEDOWN:
      {
        //decrease the zoom
        double fZoom{static_cast<double>(static_cast<int>(GetZoom()))}; //Try to keep the zoom at an integer level
        --fZoom;
        SetZoom(fZoom, 0, true);
        break;
      }
      case TB_PAGEDOWN: //deliberate fallthrough
      case TB_PAGEUP:
      {
        SetZoom(OSMMaxZoom - (m_ctrlZoomBar.GetPos() / 10.0), m_fDefaultAnimationDuration, true);
        break;
      }
      case TB_THUMBTRACK: //deliberate fallthrough
      case TB_THUMBPOSITION:
      {
        SetZoom(OSMMaxZoom - (nPos / 10.0), 0, true);
        break;
      }
      default:
      {
        //Nothing to do
        break;
      }
    }
  }
}

_Return_type_success_(return != 0) bool COSMCtrl::GetBoundingNodeRect(_In_ const COSMCtrlPolyline& polyline, _In_ INT_PTR nIndex, _Out_ CRect& rBounds, _In_ const CRect& rClient) const
{
  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

#ifdef COSMCTRL_NOD2D
  Gdiplus::PointF ptPosition;
#else
  CD2DPointF ptPosition;
#endif //#ifdef COSMCTRL_NOD2D
#pragma warning(suppress: 26446)
  if (PositionToClient(polyline.m_Nodes[nIndex].m_Position, rClient, ptPosition))
  {
    bSuccess = true;
  #ifdef COSMCTRL_NOD2D
    rBounds = CRect{static_cast<int>(ptPosition.X - (polyline.m_fNodeWidth/2)), static_cast<int>(ptPosition.Y - (polyline.m_fNodeWidth/2)), static_cast<int>(ptPosition.X + (polyline.m_fNodeWidth/2)), static_cast<int>(ptPosition.Y + (polyline.m_fNodeWidth/2))};
  #else
    rBounds = CRect{static_cast<int>(ptPosition.x - (polyline.m_fNodeWidth/2)), static_cast<int>(ptPosition.y - (polyline.m_fNodeWidth/2)), static_cast<int>(ptPosition.x + (polyline.m_fNodeWidth/2)), static_cast<int>(ptPosition.y + (polyline.m_fNodeWidth/2))};
  #endif //#ifdef COSMCTRL_NOD2D
  }

  return bSuccess;
}

_Return_type_success_(return != 0) bool COSMCtrl::GetBoundingNodeRect(_In_ const COSMCtrlPolygon& polygon, _In_ INT_PTR nIndex, _Out_ CRect& rBounds, _In_ const CRect& rClient) const
{
  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

#ifdef COSMCTRL_NOD2D
  Gdiplus::PointF ptPosition;
#else
  CD2DPointF ptPosition;
#endif //#ifdef COSMCTRL_NOD2D
#pragma warning(suppress: 26446)
  if (PositionToClient(polygon.m_Nodes[nIndex].m_Position, rClient, ptPosition))
  {
    bSuccess = true;
  #ifdef COSMCTRL_NOD2D
    rBounds = CRect{static_cast<int>(ptPosition.X - (polygon.m_fNodeWidth/2)), static_cast<int>(ptPosition.Y - (polygon.m_fNodeWidth/2)), static_cast<int>(ptPosition.X + (polygon.m_fNodeWidth/2)), static_cast<int>(ptPosition.Y + (polygon.m_fNodeWidth/2))};
  #else
    rBounds = CRect{static_cast<int>(ptPosition.x - (polygon.m_fNodeWidth/2)), static_cast<int>(ptPosition.y - (polygon.m_fNodeWidth/2)), static_cast<int>(ptPosition.x + (polygon.m_fNodeWidth/2)), static_cast<int>(ptPosition.y + (polygon.m_fNodeWidth/2))};
  #endif //#ifdef COSMCTRL_NOD2D
  }

  return bSuccess;
}

_Return_type_success_(return != 0) bool COSMCtrl::GetBoundingRect(_In_ const COSMCtrlMarker& marker, _Out_ CRect& rBounds, _In_ const CRect& rClient) const
{
  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

#ifdef COSMCTRL_NOD2D
  Gdiplus::PointF ptMarker;
#else
  CD2DPointF ptMarker;
#endif //#ifdef COSMCTRL_NOD2D
  if (PositionToClient(marker.m_Position, rClient, ptMarker))
  {
#pragma warning(suppress: 26446)
    const COSMCtrlIcon* pIcon{m_Icons[marker.m_nIconIndex]};
#pragma warning(suppress: 26496)
    AFXASSUME(pIcon != nullptr);

  #ifdef COSMCTRL_NOD2D
#pragma warning(suppress: 26489)
    if (pIcon->m_Image != nullptr)
  #else
    if (pIcon->m_pImage != nullptr)
  #endif //#ifdef COSMCTRL_NOD2D
    {
      bSuccess = true;
    #ifdef COSMCTRL_NOD2D
#pragma warning(suppress: 26489)
      rBounds = CRect{static_cast<int>(ptMarker.X - pIcon->m_ptAnchor.x), static_cast<int>(ptMarker.Y - pIcon->m_ptAnchor.y), static_cast<int>(ptMarker.X - pIcon->m_ptAnchor.x + pIcon->m_Image->GetWidth()), static_cast<int>(ptMarker.Y - pIcon->m_ptAnchor.y + pIcon->m_Image->GetHeight())};
    #else
#pragma warning(suppress: 26489)
      const CD2DSizeU sizeIcon{pIcon->m_pImage->GetPixelSize()};
      rBounds = CRect{static_cast<int>(ptMarker.x - pIcon->m_ptAnchor.x), static_cast<int>(ptMarker.y - pIcon->m_ptAnchor.y), static_cast<int>(ptMarker.x - pIcon->m_ptAnchor.x + sizeIcon.width), static_cast<int>(ptMarker.y - pIcon->m_ptAnchor.y + sizeIcon.height)};
    #endif //#ifdef COSMCTRL_NOD2D
    }
  }

  return bSuccess;
}

INT_PTR COSMCtrl::ShowOperationsDialog(_In_ COSMCtrlMapOperationsDlg& dialog)
{
  //What will be the return value from this function (assume the worst)
  INT_PTR nReturn{-1};

  //Show the dialog if we have a valid selection rectangle
  if (m_SelectionPolygon->m_Nodes.size() == 4)
  {
    //Let the dialog know about the map
    dialog.m_pOSMCtrl = this;

    //Show the dialog
    nReturn = dialog.DoModal();
  }
  else
  {
    //Tell the user that they need a selection to use the map operation dialog
    AfxMessageBox(IDS_OSMCTRL_NO_VALID_SELECTION, MB_OK | MB_ICONEXCLAMATION);
  }

  return nReturn;
}

bool COSMCtrl::DecimatePolyline(_Inout_ COSMCtrlPolyline& polyline, _In_ int nNewNodesPerWay, _In_ const CRect& rClient)
{
  //Validate our parameters
  ASSERT(nNewNodesPerWay >= 1);

  //What will become the new nodes for this polyline
  std::vector<COSMCtrlNode> newNodes;

  //Accumulate the distance between all of the nodes in this polyline
  const auto nNodes{polyline.m_Nodes.size()};
  for (std::vector<COSMCtrlNode>::size_type i{1}; i<nNodes; i++)
  {
    //Get the distance and bearing between the current two nodes
#pragma warning(suppress: 26446)
    const COSMCtrlNode& prevNode{polyline.m_Nodes[i - 1]};
    double fBearing{0};
#pragma warning(suppress: 26446)
    const double fDistance{COSMCtrlHelper::DistanceBetweenPoints(prevNode.m_Position, polyline.m_Nodes[i].m_Position, &fBearing, nullptr)};

    //Add the previous node
    newNodes.push_back(prevNode);

    //Add the new interpolated nodes
    COSMCtrlPosition tempNode(prevNode.m_Position);
    for (int j{0}; j<nNewNodesPerWay; j++)
    {
      //Work out the new position
      const COSMCtrlPosition newPosition{COSMCtrlHelper::GetPosition(tempNode, fBearing, fDistance / (static_cast<double>(nNewNodesPerWay) + 1), nullptr)};

      //Make sure the new position is valid on the map
    #ifdef COSMCTRL_NOD2D
      Gdiplus::PointF ptPosition;
    #else
      CD2DPointF ptPosition;
    #endif //#ifdef COSMCTRL_NOD2D
      if (PositionToClient(newPosition, rClient, ptPosition))
      {
        const COSMCtrlNode newNode{newPosition.m_fLongitude, newPosition.m_fLatitude};

        //Add to the array
        newNodes.push_back(newNode);
      }
      else
        return false;

      //Prepare for the next loop around
      tempNode = newPosition;
    }
  }

  //Don't forget to add the last node
#pragma warning(suppress: 26446)
  newNodes.push_back(polyline.m_Nodes[nNodes - 1]);

  //Finally swap in the new array
  polyline.m_Nodes = newNodes;

  return true;
}

bool COSMCtrl::DecimatePolygon(_Inout_ COSMCtrlPolygon& polygon, _In_ int nNewNodesPerWay, _In_ const CRect& rClient)
{
  //Validate our parameters
  ASSERT(nNewNodesPerWay >= 1);

  //What will become the new nodes for this polyline
  std::vector<COSMCtrlNode> newNodes;

  //Accumulate the distance between all of the nodes in this polyline
  const auto nNodes{polygon.m_Nodes.size()};
  for (std::vector<COSMCtrlNode>::size_type i{1}; i<nNodes; i++)
  {
    //Get the distance and bearing between the current two nodes
#pragma warning(suppress: 26446)
    const COSMCtrlNode& prevNode{polygon.m_Nodes[i - 1]};
    double fBearing{0};
#pragma warning(suppress: 26446)
    const double fDistance{COSMCtrlHelper::DistanceBetweenPoints(prevNode.m_Position, polygon.m_Nodes[i].m_Position, &fBearing, nullptr)};

    //Add the previous node
    newNodes.push_back(prevNode);

    //Add the new interpolated nodes
    COSMCtrlPosition tempNode{prevNode.m_Position};
    for (int j{0}; j<nNewNodesPerWay; j++)
    {
      //Work out the new position
      const COSMCtrlPosition newPosition{COSMCtrlHelper::GetPosition(tempNode, fBearing, fDistance / (static_cast<double>(nNewNodesPerWay) + 1), nullptr)};

      //Make sure the new position is valid on the map
    #ifdef COSMCTRL_NOD2D
      Gdiplus::PointF ptPosition;
    #else
      CD2DPointF ptPosition;
    #endif //#ifdef COSMCTRL_NOD2D
      if (PositionToClient(newPosition, rClient, ptPosition))
      {
        const COSMCtrlNode newNode{newPosition.m_fLongitude, newPosition.m_fLatitude};

        //Add to the array
        newNodes.push_back(newNode);
      }
      else
        return false;

      //Prepare for the next loop around
      tempNode = newPosition;
    }
  }

  //Add the new nodes between the last node and the first node
#pragma warning(suppress: 26446)
  const COSMCtrlNode& prevNode{polygon.m_Nodes[nNodes - 1]};
  double fBearing{0};
#pragma warning(suppress: 26446)
  const double fDistance{COSMCtrlHelper::DistanceBetweenPoints(prevNode.m_Position, polygon.m_Nodes[0].m_Position, &fBearing, nullptr)};

  //Add the previous node
  newNodes.push_back(prevNode);

  //Add the new interpolated nodes
  COSMCtrlPosition tempNode{prevNode.m_Position};
  for (int j{0}; j<nNewNodesPerWay; j++)
  {
    //Work out the new position
    const COSMCtrlPosition newPosition{COSMCtrlHelper::GetPosition(tempNode, fBearing, fDistance / (static_cast<double>(nNewNodesPerWay) + 1), nullptr)};

    //Make sure the new position is valid on the map
  #ifdef COSMCTRL_NOD2D
    Gdiplus::PointF ptPosition;
  #else
    CD2DPointF ptPosition;
  #endif //#ifdef COSMCTRL_NOD2D
    if (PositionToClient(newPosition, rClient, ptPosition))
    {
      const COSMCtrlNode newNode{newPosition.m_fLongitude, newPosition.m_fLatitude};

      //Add to the array
      newNodes.push_back(newNode);
    }
    else
      return false;

    //Prepare for the next loop around
    tempNode = newPosition;
  }

  //Finally swap in the new array
  polygon.m_Nodes = newNodes;

  return true;
}

#pragma warning(suppress: 26429 26440 26461)
void COSMCtrl::OnNMClickCopyright(NMHDR* pNMHDR, LRESULT* pResult)
{
  //ShellExecute the URL
#pragma warning(suppress: 26490)
  const auto pNMLink{reinterpret_cast<const NMLINK*>(pNMHDR)};
#pragma warning(suppress: 26485)
  ShellExecuteW(nullptr, L"open", pNMLink->item.szUrl, nullptr, nullptr, SW_SHOWNORMAL);
  *pResult = 0;
}

#pragma warning(suppress: 26434)
void COSMCtrl::OnSize(UINT nType, int cx, int cy)
{
  //Get the client rect
  CRect rClient;
  GetClientRect(rClient);

  //Move any controls which are attached to the map

  //Fix up the copyright control position if necessary
  if (m_ctrlCopyright.GetSafeHwnd() != nullptr)
  {
    CPoint ptCopyright{GetControlPosition(m_CopyrightAnchorPosition, m_ptOffsetCopyright, rClient)};
    CRect rCtrl;
    m_ctrlCopyright.GetWindowRect(rCtrl);
    if ((m_CopyrightAnchorPosition == ControlAnchorPosition::TopRight) || (m_CopyrightAnchorPosition == ControlAnchorPosition::BottomRight))
      ptCopyright.x -= rCtrl.Width();
    if ((m_CopyrightAnchorPosition == ControlAnchorPosition::BottomLeft) || (m_CopyrightAnchorPosition == ControlAnchorPosition::BottomRight))
      ptCopyright.y -= rCtrl.Height();
    m_ctrlCopyright.MoveWindow(ptCopyright.x, ptCopyright.y, rCtrl.Width(), rCtrl.Height());
  }

  //Fix up the zoom bar control position if necessary
  if (m_ctrlZoomBar.GetSafeHwnd() != nullptr)
  {
    CPoint ptZoomBar{GetControlPosition(m_ZoomBarAnchorPosition, m_ptOffsetZoomBar, rClient)};
    CRect rCtrl;
    m_ctrlZoomBar.GetWindowRect(rCtrl);
    if ((m_ZoomBarAnchorPosition == ControlAnchorPosition::TopRight) || (m_ZoomBarAnchorPosition == ControlAnchorPosition::BottomRight))
      ptZoomBar.x -= rCtrl.Width();
    if ((m_ZoomBarAnchorPosition == ControlAnchorPosition::BottomLeft) || (m_ZoomBarAnchorPosition == ControlAnchorPosition::BottomRight))
      ptZoomBar.y -= rCtrl.Height();
    m_ctrlZoomBar.MoveWindow(ptZoomBar.x, ptZoomBar.y, rCtrl.Width(), rCtrl.Height());
  }

  //Restart downloading of tiles if necessary
  if (m_bDownloadTiles && m_sCacheDirectory.GetLength())
  {
    StopAsyncDownload(false);
    StartAsyncDownload(true, false);
  }

  //Let the base class do its thing
  __super::OnSize(nType, cx, cy);
}

#ifndef COSMCTRL_NOD2D
LRESULT COSMCtrl::OnDraw2D(WPARAM /*wParam*/, LPARAM lParam)
{
  //Pull out the render target
#pragma warning(suppress: 26490)
  auto pRenderTarget{reinterpret_cast<CHwndRenderTarget*>(lParam)};
#pragma warning(suppress: 26486)
  ASSERT_VALID(pRenderTarget);

  //Get the client rect
  CRect rClient;
  GetClientRect(rClient);

  //Call the helper function to do the heavy lifting
#pragma warning(suppress: 26486)
  Draw(pRenderTarget, rClient, nullptr, GetDrawScrollRose(), GetDrawZoomBar(), GetDrawScaleBar());

  return TRUE;
}

LRESULT COSMCtrl::OnRecreatedResources(WPARAM /*wParam*/, LPARAM /*lParam*/)
{
  DestroyD2DResources();
  return 0L;
}

#pragma warning(suppress: 26440)
void COSMCtrl::DestroyD2DResources()
{
  m_pEmptyTile = nullptr;
  for (auto pIcon : m_Icons)
  {
#pragma warning(suppress: 26496)
    AFXASSUME(pIcon != nullptr);
    pIcon->m_pImage = nullptr;
  }
  m_CachedTiles.clear();
}
#endif //#ifndef COSMCTRL_NOD2D

LRESULT COSMCtrl::OnDownloadComplete(WPARAM /*wParam*/, LPARAM lParam)
{
  //Pull out the success indicator of the download
#pragma warning(suppress: 26472)
  const auto hr{static_cast<HRESULT>(lParam)};

  const bool bNewTilesPending{m_tilesPendingToDownload.size() != 0};

  //Invalidate the tile just downloaded if successful
  if (hr == S_OK)
  {
    //Invalidate the specific area of the client which we have just download a tile for
  #ifdef COSMCTRL_NOD2D
    CRect rTile{static_cast<int>(m_CurrentlyDownloadingTile.m_rTile.X), static_cast<int>(m_CurrentlyDownloadingTile.m_rTile.Y), static_cast<int>(m_CurrentlyDownloadingTile.m_rTile.X + m_CurrentlyDownloadingTile.m_rTile.Width + 1),
                static_cast<int>(m_CurrentlyDownloadingTile.m_rTile.Y + m_CurrentlyDownloadingTile.m_rTile.Height + 1)};
  #else
    CRect rTile{static_cast<int>(m_CurrentlyDownloadingTile.m_rTile.left), static_cast<int>(m_CurrentlyDownloadingTile.m_rTile.top), static_cast<int>(m_CurrentlyDownloadingTile.m_rTile.right),
                static_cast<int>(m_CurrentlyDownloadingTile.m_rTile.bottom)};
  #endif //#ifdef COSMCTRL_NOD2D
    InvalidateRect(rTile, FALSE);
  }
  else
  {
  #ifdef _DEBUG
    //Just TRACE the error code for downloads which failed because of anything other than being cancelled
    if (HRESULT_CODE(hr) != ERROR_WINHTTP_OPERATION_CANCELLED)
    {
      TRACE(_T("COSMCtrl::OnDownloadComplete, failed to download tile %s, Error:%08X\n"), m_CurrentlyDownloadingTile.m_sFile.GetString(), hr);
    }
  #endif //#ifdef _DEBUG
    m_WinHTTPRequest.Close();
    if (m_bInitializedWinHTTPRequest)
      m_WinHTTPRequest.ReleaseResources();
    m_WinHTTPConnection.Close();
  }

  //Start using the new array if available
  if (bNewTilesPending)
  {
    m_tilesToDownload.clear();
    m_tilesToDownload.insert(m_tilesToDownload.end(), m_tilesPendingToDownload.begin(), m_tilesPendingToDownload.end());
    m_tilesPendingToDownload.clear();
  }

  //At this point we are not logically downloading
  m_bDownloading = false;

  //Start downloading of the next tile if necessary
  StartAsyncDownload(false, false);

  return 0L;
}
#pragma warning(suppress: 26812)

#pragma warning(suppress: 26434)
void COSMCtrl::OnTimer(UINT_PTR nIDEvent)
{
  if (nIDEvent == m_nAnimationTimer)
  {
    const ULONGLONG nCurrentTickCount{GetTickCount64()};

    double fFractionAnimation{((nCurrentTickCount - m_nStartAnimationTickCount) / 1000.0) / m_fCurrentAnimationDuration};
    if (fFractionAnimation > 1)
      fFractionAnimation = 1;

    double fZoom{-1};
    if (m_bDoingZoomAnimation)
      fZoom = m_fStartZoom + ((m_fFinalZoom - m_fStartZoom) * fFractionAnimation);

    //Get the current value of the longitude animation variable if necessary and update it
    double fLongitude{-181};
    if (m_bDoingPositionAnimation)
      fLongitude = m_StartCenterPosition.m_fLongitude + (m_FinalCenterPosition.m_fLongitude - m_StartCenterPosition.m_fLongitude) * fFractionAnimation;

    //Get the current value of the latitude animation variable if necessary and update it
    double fLatitude{-181};
    if (m_bDoingPositionAnimation)
      fLatitude = m_StartCenterPosition.m_fLatitude + (m_FinalCenterPosition.m_fLatitude - m_StartCenterPosition.m_fLatitude) * fFractionAnimation;

    //Get the current value of the bearing animation variable if necessary and update it
    double fBearing{-1};
    if (m_bDoingBearingAnimation)
      fBearing = m_fStartBearingAtTopOfMap + ((m_fFinalBearingAtTopOfMap - m_fStartBearingAtTopOfMap) * fFractionAnimation);

    //Hive away the new settings
    if (fZoom != -1)
      m_fZoom = fZoom;
    if (fLongitude != -181)
    {
      ASSERT(fLatitude != -181);
      m_CenterPosition = COSMCtrlPosition{fLongitude, fLatitude};
    }
    if (fBearing != -1)
      m_fBearingOfTopOfMap = fBearing;

    //Finally force a redraw
    Invalidate(TRUE);

    //Update the slider control with the new zoom value
    if (m_ctrlZoomBar.GetSafeHwnd() != nullptr)
      m_ctrlZoomBar.SetPos(static_cast<int>((OSMMaxZoom - GetZoom() - OSMMinZoom) * 10));

    //Restart downloading of tiles if we are at the end of the animation
    const bool bRestartDownload{fFractionAnimation >= 1};
    if (bRestartDownload && m_bDownloadTiles && m_sCacheDirectory.GetLength())
    {
      KillTimer(m_nAnimationTimer);
      m_nAnimationTimer = 0;
      StopAsyncDownload(false);
      StartAsyncDownload(true, false);
    }
  }
  else
    __super::OnTimer(nIDEvent);
}
